/*
$Id: bitree.c,v 1.17 2008/08/09 09:21:37 alex Exp $
	bitree.c--- binary tree
*/

#ifndef lint
static char * rcsid __attribute__((unused)) = "$Id: bitree.c,v 1.17 2008/08/09 09:21:37 alex Exp $";
#endif

#include	<stdio.h>
#include	<sys/types.h>
#include	<ctype.h>
#include	<string.h>

#define	FALSE	0
#define	TRUE	(!FALSE)

#include	"edb/bitree.h"

#if defined(__LIBEDB__)
extern int	_edb_delayed_free_mode;
extern int	edb_delayed_free_recycle_count;
extern void	*_edb_delayed_free_bitree;
#define	EdbDelayedFreePush(list, obj)	{	*(void **)(obj) = (list);	(list) = (obj);	}
#define	EdbDelayedFreePop(list, obj)	{	(obj) = (list);	(list) = *(void **)(obj);	}
#endif

#define	COMPARE(p, n, cmp)	(((p)->hint<(n)->hint) ? (-1) : (((p)->hint>(n)->hint) ? (1) : ((cmp) ? (*cmp)((p)->obj, (n)->obj) : 0)))

#define	UNBALANCE_ALLOWANCE	1

#define	BAL(t)	((t)->dn - (t)->dp)
#define	MAX(x,y)	((x)>(y)?(x):(y))
#define	MIN(x,y)	((x)<(y)?(x):(y))

#define	CL(t)	(((t)->p = (t)->n = NULL),  ((t)->dp = (t)->dn = 0), (t))
#define	PSET(t, tp)	{	\
			if(((t)->p = (tp))!=NULL) {	\
				(t)->dp = MAX((tp)->dp, (tp)->dn)+1;	\
			} else { (t)->dp = 0;	}	\
		}
#define	NSET(t, tn)	{	\
			if(((t)->n = (tn))!=NULL) {	\
				(t)->dn = MAX((tn)->dp, (tn)->dn)+1;	\
			 } else { (t)->dn = 0; }	\
		}
#define	SET(t, tp, tn)	{ PSET((t), (tp)); NSET((t), (tn)); }

static	char	dummy_object[] = "bitree dummy object";
BiTree_obj_t	bitree_dummy_object = &dummy_object;

/*#bitree_num--- number of object in bitree */
int
bitree_num(
	register BiTree t)
{
	register int	n = 0;

	if(t==NULL) return 0;
	if(t->p) n += bitree_num(t->p);
	if(t->n) n += bitree_num(t->n);
	return (n + 1);
}

/*#bitree_depth--- depth in bitree */
int
bitree_depth(
	register BiTree	tree)
{
	register int	dp, dn;

	if(tree==NULL) return 0;
	dp = bitree_depth(tree->p);
	dn = bitree_depth(tree->n);
	if(dp>dn) {
		return dp + 1;
	} else {
		return dn + 1;
	}
}

static BiTree	bitree_put_minimum	__P((BiTree, BiTree));
static BiTree	bitree_put_maximum	__P((BiTree, BiTree));
static BiTree	bitree_get_minimum	__P((BiTree, BiTree *));
static BiTree	bitree_get_maximum	__P((BiTree, BiTree *));

/*#bitree_rearrange--- rearrange objects */
BiTree
bitree_rearrange(
	register BiTree t)
{
	register	BiTree	p, n;
	BiTree	m;
	register int	bal;

	if(t==NULL) return NULL;
	bal = bitree_num(t->n) - bitree_num(t->p);
	while(bal<-1) {
		p = bitree_get_maximum(t->p, &m);
		n = t->n;
		n = bitree_put_minimum(n, CL(t));
		SET(m, p, n);
		t = m;
		bal += 2;
	}
	while(bal>1) {
		n = bitree_get_minimum(t->n, &m);
		p = t->p;
		p = bitree_put_maximum(p, CL(t));
		SET(m, p, n);
		t = m;
		bal -= 2;
	}
	p = bitree_rearrange(t->p);
	n = bitree_rearrange(t->n);
	SET(t, p, n);
	return t;
}

static
BiTree
bitree_rearrange0(
	register BiTree t)
{
	register	BiTree	p, n;

	while(t->p && (t->dn + UNBALANCE_ALLOWANCE < t->dp)) {
		p = t->p;
		PSET(t, p->n);
		NSET(p, t);
		t = p;
	}
	while(t->n && (t->dp + UNBALANCE_ALLOWANCE < t->dn)) {
		n = t->n;
		NSET(t, n->p);
		PSET(n, t);
		t = n;
	}
	return t;
}

/*#bitree_put_minimum--- put minimum */
static
BiTree
bitree_put_minimum(
	register BiTree	t,
	register BiTree	min)
{
	register BiTree	p;

	if(t==NULL) return min;
	p = (t->p) ? bitree_put_minimum(t->p, min) : min ;
	PSET(t, p);
	return bitree_rearrange0(t);
}

/*#bitree_put_maximum--- put maximum */
static
BiTree
bitree_put_maximum(
	register BiTree	t,
	register BiTree	max)
{
	register BiTree	n;

	if(t==NULL) return max;
	n = (t->n) ? bitree_put_maximum(t->n, max) : max;
	NSET(t, n);
	return bitree_rearrange0(t);
}

/*#bitree_get_minimum--- get minimum */
static
BiTree
bitree_get_minimum(
	register BiTree	t,
	register BiTree	*min)
{
	register BiTree	p, n;

	if(t==NULL) {
		*min = NULL;
		return NULL;
	}
	if(!t->p) {	/* I am minimum. */
		n = t->n;
		*min = CL(t);
		return n;
	}
	p = bitree_get_minimum(t->p, min);
	PSET(t, p);
	return bitree_rearrange0(t);
}

/*#bitree_get_maximum--- get maximum */
static
BiTree
bitree_get_maximum(
	register BiTree	t,
	register BiTree	*max)
{
	register BiTree	p, n;

	if(t==NULL) {
		*max = NULL;
		return NULL;
	}
	if(!t->n) {	/* I am maximum. */
		p = t->p;
		*max = CL(t);
		return p;
	}
	n = bitree_get_maximum(t->n, max);
	NSET(t, n);
	return bitree_rearrange0(t);
}

/*#bitree_add_obj--- add object to tree */
static
BiTree
bitree_add_obj(
	register BiTree			t,
	register BiTree			new,
	register BiTree_cmp_t	obj_compare,
	register int			(*err_callback) __P((BiTree_obj_t, BiTree_obj_t, void *)),
	register void			*data)
{
	register BiTree	p, n;
	register int	st;

	if(t==NULL) return new;
	st = COMPARE(new, t, obj_compare);
	if(st<0) {	/* prev side */
		p = bitree_add_obj(t->p, new, obj_compare, err_callback, data);
		PSET(t, p);
	} else if(st>0) {	/* next side */
		n = bitree_add_obj(t->n, new, obj_compare, err_callback, data);
		NSET(t, n);
	} else {
		/* equality, it means error  */
		if(err_callback) (*err_callback)(new->obj, t->obj, data);
#if defined(__LIBEDB__)
		if(_edb_delayed_free_mode>0) {
			EdbDelayedFreePush(_edb_delayed_free_bitree, new);
		} else {
			free(new);
		}
#else
		free(new);
#endif
	}
	if(t) t = bitree_rearrange0(t);
	return t;
}

/*#bitree_add--- add object to tree */
BiTree
bitree_add(
	register BiTree			tree,
	register BiTree_obj_t		obj,
	register BiTree_hint_t	hint,
	register BiTree_cmp_t	obj_compare,
	register int			(*err_callback) __P((BiTree_obj_t, BiTree_obj_t, void *)),
	register void			*data)
{
	register BiTree	new;

#if defined(__LIBEDB__)
	if(_edb_delayed_free_bitree) {
		++edb_delayed_free_recycle_count;
		EdbDelayedFreePop(_edb_delayed_free_bitree, new);
	} else {
		new = (BiTree)malloc(sizeof(struct BITREE));
	}
#else
	new = (BiTree)malloc(sizeof(struct BITREE));
#endif
	new->obj = obj;
	new->hint = hint;
	return bitree_add_obj(tree, CL(new), obj_compare, err_callback, data);
}

/*#bitree_remove_obj--- remove object from tree */
static
BiTree
bitree_remove_obj(
	register BiTree	t,
	register BiTree	target,
	register int	(*obj_compare) __P((BiTree_obj_t, BiTree_obj_t)),
	register BiTree	*rt)
{
	register BiTree	p, n;
	BiTree	m, dummy;
	register int	st;

	if(t==NULL) {
		*rt = NULL;
		return NULL;
	}
	st = COMPARE(target, t, obj_compare);
	if(st<0) {	/* prev side */
		p = bitree_remove_obj(t->p, target, obj_compare, rt);
		PSET(t, p);
	} else if(st>0) {	/* next side */
		n = bitree_remove_obj(t->n, target, obj_compare, rt);
		NSET(t, n);
	} else {
		/* equality, found! */
		(*rt) = t;
		if(!t->p && !t->n) {
			t = NULL;
		} else if(!t->n) {
			t = t->p;
		} else if(!t->p) {
			t = t->n;
		} else {
			if(BAL(t)<=0) {
				p = bitree_get_maximum(t->p, &m);
				SET(m, p, t->n);
				t = m;
			} else {
				n = bitree_get_minimum(t->n, &m);
				SET(m, t->p, n);
				t = m;
			}
		}
		dummy = CL(*rt);
	}
	if(t) t = bitree_rearrange0(t);
	return t;
}

/*#bitree_remove--- remove object from tree */
BiTree
bitree_remove(
	register BiTree			tree,
	register BiTree_obj_t		obj,
	register BiTree_hint_t	hint,
	register BiTree_cmp_t	obj_compare,
	register BiTree_obj_t		*robj)
{
	struct BITREE	target;
	BiTree	rtree;

	target.obj = obj;
	target.hint = hint;
	tree =  bitree_remove_obj(tree, CL(&target), obj_compare, &rtree);
	if(rtree) {
		*robj = rtree->obj;
#if defined(__LIBEDB__)
		if(_edb_delayed_free_mode>0) {
			EdbDelayedFreePush(_edb_delayed_free_bitree, rtree);
		} else {
			free(rtree);
		}
#else
		free(rtree);
#endif
	} else {
		*robj = NULL;
	}
	return tree;
}

/*#bitree_search--- search object */
BiTree_obj_t
bitree_search(
	register BiTree			t,
	register BiTree_obj_t		obj,
	register BiTree_hint_t	hint,
	register BiTree_cmp_t	obj_compare)
{
	register int	st;

	while(1) {
		if(t == NULL) return NULL;
		if(hint<t->hint) {
			st = -1;
		} else if(hint>t->hint) {
			st = 1;
		} else if(obj_compare) {
			st = (*obj_compare)(obj, t->obj);
		} else {
			st = 0;
		}
		if(st<0) {
			t = t->p;
		} else if(st>0) {
			t = t->n;
		} else {
			return t->obj;
		}
	}
}

/*#bitree_search_node--- search object */
BiTree
bitree_search_node(
	register BiTree			t,
	register BiTree_obj_t		obj,
	register BiTree_hint_t	hint,
	register BiTree_cmp_t	obj_compare)
{
	register int	st;

	while(1) {
		if(t == NULL) return NULL;
		if(hint<t->hint) {
			st = -1;
		} else if(hint>t->hint) {
			st = 1;
		} else if(obj_compare) {
			st = (*obj_compare)(obj, t->obj);
		} else {
			st = 0;
		}
		if(st<0) {
			t = t->p;
		} else if(st>0) {
			t = t->n;
		} else {
			return t;
		}
	}
}

/*#bitree_free--- free tree */
void
bitree_free(
	register BiTree t,
	register void (*obj_free) __P((BiTree_obj_t)))
{

	if(t == NULL) return;
	bitree_free(t->p, obj_free);
	bitree_free(t->n, obj_free);
	if(obj_free) (*obj_free)(t->obj);
#if defined(__LIBEDB__)
	if(_edb_delayed_free_mode>0) {
		EdbDelayedFreePush(_edb_delayed_free_bitree, t);
	} else {
		free(t);
	}
#else
	free(t);
#endif
}

/*#bitree_free2--- free tree with context */
void
bitree_free2(
	register BiTree	t,
	register void (*obj_free) __P((void *, BiTree_obj_t)),
	register void	*context)
{

	if(t == NULL) return;
	bitree_free2(t->p, obj_free, context);
	bitree_free2(t->n, obj_free, context);
	if(obj_free) (*obj_free)(context, t->obj);
#if defined(__LIBEDB__)
	if(_edb_delayed_free_mode>0) {
		EdbDelayedFreePush(_edb_delayed_free_bitree, t);
	} else {
		free(t);
	}
#else
	free(t);
#endif
}

/*#bitree_callback--- callback */
int
bitree_callback(
	register BiTree	t,
	register int	(*obj_callback) __P((BiTree_obj_t, void *)),
	register void	*data)
{
	register int cont = TRUE;

	if(t == NULL) return TRUE;

	cont = bitree_callback(t->p, obj_callback, data);
	if(!cont) return FALSE;

	cont = (*obj_callback)(t->obj, data);
	if(!cont) return FALSE;

	cont = bitree_callback(t->n, obj_callback, data);
	return cont;
}

/*#bitree_callback_hint--- callback */
int
bitree_callback_hint(
	register BiTree	t,
	register int	(*obj_callback) __P((BiTree_obj_t, BiTree_hint_t hint, void *)),
	register void	*data)
{
	register int cont = TRUE;

	if(t == NULL) return TRUE;

	cont = bitree_callback_hint(t->p, obj_callback, data);
	if(!cont) return FALSE;

	cont = (*obj_callback)(t->obj, t->hint, data);
	if(!cont) return FALSE;

	cont = bitree_callback_hint(t->n, obj_callback, data);
	return cont;
}

/*#bitree_callback_node--- callback */
int
bitree_callback_node(
	register BiTree	t,
	register int	(*obj_callback) __P((BiTree, void *)),
	register void	*data)
{
	register int cont = TRUE;

	if(t == NULL) return TRUE;

	cont = bitree_callback_node(t->p, obj_callback, data);
	if(!cont) return FALSE;

	cont = (*obj_callback)(t, data);
	if(!cont) return FALSE;

	cont = bitree_callback_node(t->n, obj_callback, data);
	return cont;
}

/*#bitree_hint_ucs4str--- make hint of ucs4string */
BiTree_hint_t
bitree_hint_ucs4str(
	register ucs4_t	*u)
{
	register BiTree_hint_t	h = 0;

	if(*u) h |= *u++;
#if 0
	h <<= 32; if(*u) h |= *u++;
	h <<= 32; if(*u) h |= *u++;
	h <<= 32; if(*u) h |= *u++;
#endif
	return h;
}

/*#bitree_hint_ucs4strcase--- make hint of ucs4string */
BiTree_hint_t
bitree_hint_ucs4strcase(
	register ucs4_t	*u)
{
	register BiTree_hint_t	h = 0;
	register ucs4_t	uc;

	if(*u) { uc = *u++; h |= ucs4_isupper(uc) ? ucs4_tolower(uc) : uc;	}
#if 0
	h <<= 32; if(*u) { uc = *u++; h |= ucs4_isupper(uc) ? ucs4_tolower(uc) : uc;	}
	h <<= 32; if(*u) { uc = *u++; h |= ucs4_isupper(uc) ? ucs4_tolower(uc) : uc;	}
	h <<= 32; if(*u) { uc = *u++; h |= ucs4_isupper(uc) ? ucs4_tolower(uc) : uc;	}
#endif
	return h;
}

/*#bitree_hint_str--- make hint of string */
BiTree_hint_t
bitree_hint_str(
	register char	*s)
{
	register BiTree_hint_t	h = 0;

	if(*s) h |= (unsigned char)*s++;
	h <<= 8; if(*s) h |= (unsigned char)*s++;
	h <<= 8; if(*s) h |= (unsigned char)*s++;
	h <<= 8; if(*s) h |= (unsigned char)*s++;
#if 0
	h <<= 8; if(*s) h |= (unsigned char)*s++;
	h <<= 8; if(*s) h |= (unsigned char)*s++;
	h <<= 8; if(*s) h |= (unsigned char)*s++;
	h <<= 8; if(*s) h |= (unsigned char)*s++;
#endif
	return h;
}

/*#bitree_hint_strcase--- make hint of string */
BiTree_hint_t
bitree_hint_strcase(
	register char	*s)
{
	register BiTree_hint_t	h = 0;
	register char	c;

	if(*s) { c= *s++; h |= (unsigned char)(isupper(c) ? tolower(c) : c);	}
	h <<= 8; if(*s) { c= *s++; h |= (unsigned char)(isupper(c) ? tolower(c) : c);	}
	h <<= 8; if(*s) { c= *s++; h |= (unsigned char)(isupper(c) ? tolower(c) : c);	}
	h <<= 8; if(*s) { c= *s++; h |= (unsigned char)(isupper(c) ? tolower(c) : c);	}
#if 0
	h <<= 8; if(*s) { c= *s++; h |= (unsigned char)(isupper(c) ? tolower(c) : c);	}
	h <<= 8; if(*s) { c= *s++; h |= (unsigned char)(isupper(c) ? tolower(c) : c);	}
	h <<= 8; if(*s) { c= *s++; h |= (unsigned char)(isupper(c) ? tolower(c) : c);	}
	h <<= 8; if(*s) { c= *s++; h |= (unsigned char)(isupper(c) ? tolower(c) : c);	}
#endif
	return h;
}

int
bitree_compare_str(register BiTree_obj_t obj1, register BiTree_obj_t obj2)
{

	return strcmp((char *)obj1, (char *)obj2);
}

int
bitree_compare_strcase(register BiTree_obj_t obj1, register BiTree_obj_t obj2)
{

	return strcasecmp((char *)obj1, (char *)obj2);
}

int
bitree_compare_ucs4str(register BiTree_obj_t obj1, register BiTree_obj_t obj2)
{

	return ucs4_strcmp((ucs4_t *)obj1, (ucs4_t *)obj2);
}

int
bitree_compare_ucs4strcase(register BiTree_obj_t obj1, register BiTree_obj_t obj2)
{

	return ucs4_strcasecmp((ucs4_t *)obj1, (ucs4_t *)obj2);
}


