/*
	$Id: catalogue.c,v 1.16 2011/04/18 09:45:28 alex Exp $
	catalogue.c--- catalogue
*/

#ifndef	lint
static char *rcsid __attribute__((unused)) = "$Id: catalogue.c,v 1.16 2011/04/18 09:45:28 alex Exp $";
#endif

#include	<stdio.h>
#include	<stdarg.h>
#include	"edb/edb.h"

EdbCatalogue
edb_catalogue_new()
{
	register EdbCatalogue	ca;

	ca = (EdbCatalogue)malloc(sizeof(struct EDB_CATALOGUE));
	if(ca==NULL) return NULL;
	bzero(ca, sizeof(struct EDB_CATALOGUE));
	return ca;
}

void
edb_catalogue_free(EdbCatalogue ca)
{

	if(ca==NULL) return;
	bitree_free(ca->bt, NULL);
	link_free(ca->link);
	free(ca);
}

void
edb_catalogue_free2(EdbCatalogue ca, void (*func)(), void *context)
{

	if(ca==NULL) return;
	bitree_free2(ca->bt, func, context);
	link_free(ca->link);
	free(ca);
}

void
edb_catalogue_clean(EdbCatalogue ca)
{

	if(ca==NULL) return;
	bitree_free(ca->bt, NULL);
	link_free(ca->link);
	bzero(ca, sizeof(struct EDB_CATALOGUE));
}

void
edb_catalogue_clean2(EdbCatalogue ca, void (*func)(), void *context)
{

	if(ca==NULL) return;
	bitree_free2(ca->bt, func, context);
	link_free(ca->link);
	bzero(ca, sizeof(struct EDB_CATALOGUE));
}

int
edb_catalogue_append(EdbCatalogue ca, eid_t eid, void *obj)
{
	register void	*robj;
	register Link	lobj;

	robj = bitree_search(ca->bt, obj, eid, NULL);
	if(robj) return FALSE;	/* already included */

	BiTreeAdd(ca->bt, obj, eid, NULL, NULL, NULL);
	lobj = link_atom_new(obj, NULL);
	if(ca->link && ca->tail_hint) {
		link_append(ca->tail_hint, lobj);
	} else {
		LinkAppend(ca->link, lobj);
	}
	ca->tail_hint = lobj;
	return TRUE;
}

void *
edb_catalogue_lookup(EdbCatalogue ca, eid_t eid)
{
	register void	*robj;

	robj = bitree_search(ca->bt, NULL, eid, NULL);
	return robj;
}

int
edb_catalogue_size(EdbCatalogue ca)
{

	if(!ca) return 0;
	return bitree_num(ca->bt);
}

int
edb_catalogue_callback(EdbCatalogue ca, int (*cb) __P((BiTree_obj_t, void *)), void *ud)
{

	return bitree_callback(ca->bt, cb, ud);
}

struct CATA_CBDATA {
	EdbCatalogue	ca;
};

static
int
edb_catalogue_append_cb(BiTree_obj_t obj, BiTree_hint_t hint, void *data)
{
	struct CATA_CBDATA *cbdata = data;

	edb_catalogue_append(cbdata->ca, hint, obj);
	return TRUE;
}

EdbCatalogue
edb_catalogue_duplicate(EdbCatalogue ca)
{
	struct CATA_CBDATA	cbdata;

	cbdata.ca = edb_catalogue_new();
	bitree_callback_hint(ca->bt, edb_catalogue_append_cb, &cbdata);

	return cbdata.ca;
}

EdbCatalogue
edb_catalogue_merge(EdbCatalogue ca1, EdbCatalogue ca2)
{
	struct CATA_CBDATA	cbdata;

	cbdata.ca = edb_catalogue_duplicate(ca1);
	bitree_callback_hint(ca2->bt, edb_catalogue_append_cb, &cbdata);
	return cbdata.ca;
}


