/*
$Id: check.c,v 1.19 2008/08/09 09:21:37 alex Exp $
	check.c--- check data base information
	(C) Copyright 1994 by Alex
*/

#ifndef	lint
static char * rcsid __attribute__((unused)) = "$Id: check.c,v 1.19 2008/08/09 09:21:37 alex Exp $";
#endif

#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<strings.h>
#include	<unistd.h>
#include	<ctype.h>
#include	<sys/types.h>
#include	<sys/socket.h>
#include	<netinet/in.h>
#include	<arpa/inet.h>
#include	"edb/edb.h"
#include	"edb/base64.h"
#include	"edb/pki.h"

/*
	edb_text_isAscii--- ascii text
	edb_text_isEnglish--- printable text (ascii)
	edb_text_isEID--- EID (positive integer)
	edb_text_isPronounce--- kana (hiragana|katakana|space)
	edb_text_isPhone--- phone number
	edb_text_isInteger--- integer number
	edb_text_isEmail--- integer number
	edb_text_isURL--- URL
*/
int
edb_text_isAscii(EdbContext ec, utf8_t *s)
{
	register utf8_t	c;

	while((c=*s)) {
		if(!isascii(c)) return FALSE;
		++s;
	}
	return TRUE;
}

int
edb_text_isEnglish(
	register EdbContext	ec,
	register utf8_t		*s)
{
	register ucs4_t	c, *u;

	u = UTF8toUCS4(s);
	while((c = *u++)) {
		if(!ucs4_isEnglish(c)) return FALSE;
	}
	return TRUE;
}

int
edb_text_isJapanese(
	register EdbContext	ec,
	register utf8_t		*s)
{
	register ucs4_t	c, *u;

	u = UTF8toUCS4(s);
	while((c = *u++)) {
		if(!ucs4_isJapanese(c)) return FALSE;
	}
	return TRUE;
}

int
edb_text_isPronounce(
	register EdbContext	ec,
	register utf8_t		*s)
{
	register ucs4_t	c, *u;

	u = UTF8toUCS4(s);
	while((c = *u++)) {
		if(!ucs4_isPronounce(c)) return FALSE;
	}
	return TRUE;
}

int
edb_text_isName(EdbContext ec, utf8_t *s)
{
	utf8_t	*en, *jp, *pr;
	eid_t	mapto;
	register int	st = TRUE;

	edb_mtext_split(ec, &mapto, &en, &jp, &pr, s);
	if(st && !edb_text_isEnglish(ec, en)) st = FALSE;
	if(st && !edb_text_isJapanese(ec, jp)) st = FALSE;
	if(st && !edb_text_isPronounce(ec, pr)) st = FALSE;
	FREE(en);
	FREE(jp);
	FREE(pr);
	return st;
}

int
edb_text_isSentence(EdbContext ec, utf8_t *s)
{
	utf8_t	*en, *jp, *pr;
	eid_t	mapto;
	register int	st = TRUE;

	edb_mtext_split(ec, &mapto, &en, &jp, &pr, s);
	if(st && !edb_text_isEnglish(ec, en)) st = FALSE;
	if(st && !edb_text_isJapanese(ec, jp)) st = FALSE;
	FREE(en);
	FREE(jp);
	FREE(pr);
	return st;
}

int
edb_text_isParagraph(EdbContext ec, utf8_t *s)
{
	utf8_t	*en, *jp, *pr;
	eid_t	mapto;
	register int	st = TRUE;

	edb_mtext_split(ec, &mapto, &en, &jp, &pr, s);
	if(st && !edb_text_isEnglish(ec, en)) st = FALSE;
	if(st && !edb_text_isJapanese(ec, jp)) st = FALSE;
	FREE(en);
	FREE(jp);
	FREE(pr);
	return st;
}

int
edb_text_isEID(EdbContext ec, utf8_t *s)
{
	register utf8_t	c;

	while((c=*s)) {
		if(!isdigit(c)) return FALSE;
		++s;
	}
	return TRUE;
}

int
edb_text_isBool(EdbContext ec, utf8_t *s)
{

	if(!strcmp((char *)s, "true") || !strcmp((char *)s, "false")) return TRUE;
	return FALSE;
}

int
edb_text_isInteger(EdbContext ec, utf8_t *s)
{
	register utf8_t	c;

	if(*s=='-') ++s;
	while((c=*s)) {
		if(!isdigit(c)) return FALSE;
		++s;
	}
	return TRUE;
}

int
edb_text_isReal(EdbContext ec, utf8_t *s)
{
	double	f;

	return (sscanf((char *)s, "%lf", &f) == 1);
}

int
edb_text_isPhone(EdbContext ec, utf8_t *s)
{
	register utf8_t	c;

	while((c=*s)) {
		if(!(isdigit(c) || c=='+' || c=='-')) return FALSE;
		++s;
	}
	return TRUE;
}

int
edb_text_isDate(EdbContext ec, utf8_t *s)
{

	if(strlen((char *)s)!=8 || !str_isdigit((char *)s)) return FALSE;
	return TRUE;
}

int
edb_text_isValidDate(EdbContext ec, utf8_t *s)
{
	int	year, month, day;
	utf8_t	str[BUFSIZ];

	if(!edb_text_isDate(ec, s)) return FALSE;

	strncpy((char *)str, (char *)s, 4);		str[4] = '\0';	year = ATOI((char *)str);
	strncpy((char *)str, (char *)&s[4], 2);	str[2] = '\0';	month = ATOI((char *)str);
	strncpy((char *)str, (char *)&s[6], 2);	str[2] = '\0';	day = ATOI((char *)str);
	return edb_date_is_valid(year, month, day);
}

static
int
check_username(utf8_t *s)
{

	if(strlen((char *)s)==0) return FALSE;
	while(*s) {
		if(!(isalnum(*s) || *s == '-' || *s == '_' || *s == '.' || *s == '+')) return FALSE;
		++s;
	}
	return TRUE;
}

static
int
check_hostname(utf8_t *s)
{

	while(*s) {
#if 0
		if(!(isalnum(*s) || *s == '-' || *s == '_' )) return FALSE;
#else
		if(!(isalnum(*s) || *s == '-')) return FALSE;
#endif
		++s;
	}
	return TRUE;
}

int
edb_text_isEmail(EdbContext ec, utf8_t *s)
{
	Link addr_l, dom_l;
	Link l;

	if(rindex((char *)s, ' ')) return FALSE;
	if(rindex((char *)s, '\t')) return FALSE;
	if(!rindex((char *)s, '@')) return FALSE;

	addr_l = str_split((char *)s, '@');
	if(link_atoms(addr_l) != 2) {
		link_free(addr_l);
		return FALSE;
	}
	if(!check_username(addr_l->d)) {
		link_free(addr_l);
		return FALSE;
	}

	dom_l = str_split((char *)addr_l->n->d, '.');
	for(l=dom_l; l; l = l->n) {
		if(strlen(l->d)==0 || !check_hostname(l->d)) {
			link_free(addr_l);
			link_free(dom_l);
			return FALSE;
		}
	}

	link_free(addr_l);
	link_free(dom_l);
	return TRUE;
}

static
int
check_dns_hostname(ucs4_t *s)
{
	ucs4_t	*buf = UCS4_STRDUP(s);

	ucs4_str_capitalize(buf, s);
	s = buf;
	while(*s) {
#if 0
		if(!(ucs4_isalnum(*s) || *s == '-' || *s == '_' || *s == '.')) return FALSE;
#else
		if(!(ucs4_isalnum(*s) || *s == '-' || *s == '.')) return FALSE;
#endif
		if(s[0]=='.' && s[1] == '.') return FALSE;
		++s;
	}

	return TRUE;
}

int
edb_text_isURL(EdbContext ec, register utf8_t *s)
{
	ucs4_t	*u;
	Link l, lurl;
	utf8_t proto[BUFSIZ];
	ucs4_t *parent_dir = UTF8toUCS4((utf8_t *)"..");

	u = UTF8toUCS4(s);
	if(u == NULL || *u == UCS4_NULL) return FALSE;

	if(ucs4_index(u, '"')) return FALSE;
	if(ucs4_index(u, UCS4_SPACE)) return FALSE;

	lurl = ucs4_str_split(u, '/');
	if(lurl == NULL) return FALSE;

	if(link_atoms(lurl)<3) {
		link_free(lurl);
		return FALSE;
	}

	if(ucs4_strlen((ucs4_t *)lurl->n->d)!=0) {
		link_free(lurl);
		return FALSE;
	}

	l = lurl;
	/* check protocol */
	convUCS4toUTF8((ucs4_t *)l->d, proto);
	if(strcmp((char *)proto, "http:")
		&& strcmp((char *)proto, "https:")
		&& strcmp((char *)proto, "gopher:")
		&& strcmp((char *)proto, "ftp:")) {
		link_free(lurl);
		return FALSE;
	}

	l = l->n;
	/* // */
	if(ucs4_strlen((ucs4_t *)l->d) != 0) {
		link_free(lurl);
		return FALSE;
	}

	l = l->n;
	/* host:port */
	if(l->d == NULL) {
		link_free(lurl);
		return FALSE;
	} else {
		Link	lhost;
		int	n;

		lhost = ucs4_str_split((ucs4_t *)l->d, ':');
		n = link_atoms(lhost);
		if((n>=1 && !check_dns_hostname((ucs4_t *)lhost->d))
			|| (n>=2 && !ucs4_str_isdigit((ucs4_t *)lhost->n->d))) {
			link_free(lhost);
			link_free(lurl);
			return FALSE;
		}
		link_free(lhost);
	}

	/* path */
	while((l=l->n)) {
		if(ucs4_strlen((ucs4_t *)l->d)==0 || !ucs4_strcmp((ucs4_t *)l->d, parent_dir)) {
			link_free(lurl);
			return FALSE;
		}
	}

	link_free(lurl);

	return TRUE;
}

int
edb_text_isIPv4CIDR(EdbContext ec, register utf8_t *s)
{
	register Link	lstr;
	register int	status = FALSE;

	if(!STRisVALID(s)) return FALSE;
	lstr = str_split((char *)s, '/');
	if(link_atoms(lstr) == 2) {
		if(edb_text_isIPv4Address(ec, lstr->d)) {
			register utf8_t *v = lstr->n->d;

			if(STRisVALID(v) && str_isdigit((char *)v)) {
				register int	m = ATOI((char *)v);

				if(EDB_DNS_MINIMUM_SUBNETMASK<=m && m<32) status = TRUE;
			}
		}
	}
	link_free(lstr);
	return status;
}

int
edb_text_isIPv4Address(EdbContext ec, register utf8_t *s)
{
	register Link	l, lstr;
	register int	status;

	if(!STRisVALID(s)) return FALSE;
	lstr = str_split((char *)s, '.');
	if(link_atoms(lstr) != 4) {
		link_free(lstr);
		return FALSE;
	}
	status = TRUE;
	ForLink(l, lstr) {
		register utf8_t	*v = l->d;
		register int	d;

		if(!STRisVALID(v) || !str_isdigit((char *)v)) {
			status = FALSE;
		} else {
			d = ATOI((char *)v);
			if(d<0 || 255<d) status = FALSE;
		}
	}
	link_free(lstr);
	return status;
}

int
edb_text_isInetName(EdbContext ec, register utf8_t *s)
{

	if(!STRisVALID(s)) return FALSE;
	while(*s) {
		if(!(ucs4_isalnum(*s) || *s == '-')) return FALSE;
		++s;
	}
	return TRUE;
}

int
edb_text_isInetFQDN(EdbContext ec, register utf8_t *s)
{
	register Link	l, lstr;
	register int	len;

	if(!STRisVALID(s)) return FALSE;
	if(s[0]=='.') return FALSE;
	len = strlen((char *)s);
	if(s[len-1]=='.') return FALSE;

	lstr = str_split((char *)s, '.');
	ForLink(l, lstr) {
		if(!edb_text_isInetName(ec, l->d)) {
			link_free(lstr);
			return FALSE;
		}
	}
	link_free(lstr);
	return TRUE;
}

#define	PEM_X509_BEGIN_CERTIFICATE "-----BEGIN CERTIFICATE-----"
#define	PEM_X509_END_CERTIFICATE "-----END CERTIFICATE-----"

int
edb_text_isCertificate(EdbContext ec, register utf8_t *s)
{

	return edb_pki_check_certificate(ec, s);
}

int
edb_date_is_valid(y, m, d)
{
	static int	mdays[12] = {31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};

	if(y<=0) return FALSE;
	if(m<=0 || m>12) return FALSE;
	if(d<=0 || d>mdays[m-1]) return FALSE;
	/* leap year check */
	if(m==2 && d==29) {
		if((y%4)!=0) return FALSE;
		if((y%400)==0) return TRUE;
		if((y%100)!=0) return TRUE;
		return FALSE;
	}
	return TRUE;
}

int
edb_date_lastday(y, m)
{
	static int	mdays[12] = {31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};

	/* leap year check */
	if(m<=0 || 12<m) return 31;
	if(m==2) {
		if((y%4)!=0) return 28;
		if((y%400)==0) return 29;
		if((y%100)!=0) return 29;
		return 28;
	} else {
		return mdays[m-1];
	}
}

int
edb_date_kind(y, m, d)
{
	struct tm	tm;
	time_t	tloc;

	tm.tm_year = y - 1900;
	tm.tm_mon = m - 1;
	tm.tm_mday = d;
	tm.tm_hour = 0;
	tm.tm_min = 0;
	tm.tm_sec = 0;
	tm.tm_isdst = 0;
	tm.tm_zone = "JST";
	tm.tm_gmtoff = 9*60*60;
	tloc = mktime(&tm);
	bcopy(localtime(&tloc), &tm, sizeof(struct tm));
	return tm.tm_wday;
}

int
check_addr_in_network(
	utf8_t			*addr,		/* xx.xx.xx.xx */
	utf8_t			*network)		/* xx.xx.xx.xx/mm.mm.mm.mm */
{
	Link	l_net;
	unsigned int ip, net, mask;
	utf8_t	*s;

	if(addr==NULL) return FALSE;

	ip = ntohl(inet_addr((char *)addr));

	l_net = str_split((char *)network, '/');
	if(l_net) {
		s = l_net->d;
		net = ntohl(inet_addr((char *)s));
	} else {
		net = ~0;
	}

	if(l_net->n) {
		s = l_net->n->d;
		mask = ntohl(inet_addr((char *)s));
	} else {
		mask = ~0;
	}
	link_free(l_net);

	return (ip&mask)==(net&mask);
}

