/*
	$Id: classify.c,v 1.14 2008/08/09 09:21:37 alex Exp $
	classify.c--- classify
*/

#ifndef	lint
static char *rcsid __attribute__((unused)) = "$Id: classify.c,v 1.14 2008/08/09 09:21:37 alex Exp $";
#endif

#include	<stdio.h>
#include	<stdarg.h>
#include	<string.h>
#include	"edb/edb.h"
#include	"edb/classify.h"
#include	<netinet/in.h>
#include	<arpa/inet.h>

static
BiTree
edb_classify_append(BiTree bt, int id)
{
	struct EDB_CLASSIFY_OBJ	*robj;

	robj = (struct EDB_CLASSIFY_OBJ *)bitree_search(bt, NULL, id, NULL);
	if(robj) {
		++robj->count;
	} else {
		robj = (struct EDB_CLASSIFY_OBJ *)malloc(sizeof(struct EDB_CLASSIFY_OBJ));
		bzero(robj, sizeof(struct EDB_CLASSIFY_OBJ));
		robj->id = id;
		robj->count = 1;
		BiTreeAdd(bt, robj, id, NULL, NULL, NULL);
	}
	return bt;
}

int
edb_classify_count(EdbContext ec, EdbClassify bt, int id)
{
	struct EDB_CLASSIFY_OBJ	*robj;

	robj = (struct EDB_CLASSIFY_OBJ *)bitree_search(bt, NULL, id, NULL);
	if(robj) return robj->count;
	return 0;
}

int
edb_classify_obj_id(EdbContext ec, EdbClassifyObj obj)
{

	return obj->id;
}

int
edb_classify_obj_count(EdbContext ec, EdbClassifyObj obj)
{

	return obj->count;
}

static
void
edb_classify_free_func(void *obj)
{

	FREE(obj);
}

void
edb_classify_free(EdbContext ec, EdbClassify class)
{

	if(class==NULL) return;
	bitree_free(class, edb_classify_free_func);
}

int
edb_classify_num(EdbContext ec, EdbClassify class)
{

	return bitree_num(class);
}

void
edb_classify_callback(EdbContext ec, EdbClassify class, int (*cb) __P((EdbClassifyObj, void *)), void *data)
{

	bitree_callback(class, (int (*) __P((BiTree_obj_t, void *)))cb, data);
}

static
EdbClassify
edb_classify_table_column_by_EID_func(EdbContext ec, utf8_t *query)
{
	int	n;
	Link	lf, l_field;
	BiTree	class = NULL;

	l_field = edb_pgsql_query_one_field(ec, query, &n);
	FREE(query);

	ForLink(lf, l_field) {
		Link	l, l_eid;
		int	none = FALSE;

		if(STRisVALID((char *)lf->d)) {
			l_eid = edb_array_split(lf->d);
			if(l_eid) {
				str_sort(l_eid);
				l_eid = str_uniq(l_eid);
				ForLink(l, l_eid) {
					class = edb_classify_append(class, ATOI((char *)l->d));
				}
				EdbLinkFree(l_eid);
			} else {
				none = TRUE;
			}
		} else {
			none = TRUE;
		}
		if(none) class = edb_classify_append(class, 0);
	}
	EdbLinkFree(l_field);

	return class;
}

EdbClassify
edb_classify_table_column_by_EID(EdbContext ec, EdbTableInfo ti, EdbColumnInfo ci, char *fmt, ...)
{
	utf8_t	*query, *target;
	va_list	ap;
	BiTree	class = NULL;

	target = (utf8_t *)strdup("");
	if(fmt) {
		char	*s;

		va_start(ap, fmt);
		vasprintf(&s, fmt, ap);
		va_end(ap);
		if(STRisVALID(s)) {
			FREE(target);
			asprintf((char **)&target, " WHERE %s", s);
		}
		FREE(s);
	}
	asprintf((char **)&query, "SELECT %s FROM %s%s", ColumnSN_EID(ci), TABLE_SN(ti), target);
	class = edb_classify_table_column_by_EID_func(ec, query);
	FREE(target);
	return class;
}

EdbClassify
edb_classify_tableAS_column_by_EID(EdbContext ec, EdbTableInfo ti, utf8_t *as, EdbColumnInfo ci, char *fmt, ...)
{
	utf8_t	*query, *target;
	va_list	ap;
	BiTree	class = NULL;

	target = (utf8_t *)strdup("");
	if(fmt) {
		char	*s;

		va_start(ap, fmt);
		vasprintf(&s, fmt, ap);
		va_end(ap);
		if(STRisVALID(s)) {
			FREE(target);
			asprintf((char **)&target, " WHERE %s", s);
		}
		FREE(s);
	}
	asprintf((char **)&query, "SELECT %s FROM %s as %s%s", ColumnSN_EID(ci), TABLE_SN(ti), as, target);
	class = edb_classify_table_column_by_EID_func(ec, query);
	FREE(target);
	return class;
}

static
EdbClassify
edb_classify_table_column_by_YEAR_func(EdbContext ec, EdbTableInfo ti, EdbColumnInfo ci, int fiscal, utf8_t *query)
{
	int	n;
	int	elements;
	Link	lf, l_field;
	BiTree	class = NULL;

	l_field = edb_pgsql_query_one_field(ec, query, &n);

	elements = edb_table_column_reldb_elements(ec, ti, NULL, ci);
	switch(ci->typeInfo->typeID) {
	case EDB_TYPE_YEAR:
		if(elements==1) {
			ForLink(lf, l_field) {
				if(STRisVALID((char *)lf->d)) {
					register int y = ATOI((char *)lf->d);

					class = edb_classify_append(class, y);
				} else {
					class = edb_classify_append(class, 0);
				}
			}
		} else {
			ForLink(lf, l_field) {
				Link	l, l_value;
				int	none = FALSE;
				int	py = 0;

				if(STRisVALID((char *)lf->d)) {
					l_value = edb_array_split(lf->d);
					if(l_value) {
						str_sort(l_value);
						l_value = str_uniq(l_value);
						ForLink(l, l_value) {
							register int y = ATOI((char *)l->d);

							if(py==y) continue;
							py = y;
							class = edb_classify_append(class, y);
						}
						EdbLinkFree(l_value);
					} else {
						none = TRUE;
					}
				} else {
					none = TRUE;
				}
				if(none) class = edb_classify_append(class, 0);
			}
		}
		break;
	case EDB_TYPE_DATE:
		if(elements==1) {
			ForLink(lf, l_field) {
				if(STRisVALID((char *)lf->d)) {
					register int d = ATOI((char *)lf->d);

					class = edb_classify_append(class, fiscal ? DATE_FiscalY(d) : DATE_Y(d));
				} else {
					class = edb_classify_append(class, 0);
				}
			}
		} else {
			ForLink(lf, l_field) {
				Link	l, l_value;
				int	none = FALSE;
				int	y = 0;

				if(STRisVALID((char *)lf->d)) {
					l_value = edb_array_split(lf->d);
					if(l_value) {
						str_sort(l_value);
						l_value = str_uniq(l_value);
						ForLink(l, l_value) {
							register int d = ATOI((char *)l->d);
							int yy = fiscal ? DATE_FiscalY(d) : DATE_Y(d);

							if(y==yy) continue;
							y = yy;
							class = edb_classify_append(class, y);
						}
						EdbLinkFree(l_value);
					} else {
						none = TRUE;
					}
				} else {
					none = TRUE;
				}
				if(none) class = edb_classify_append(class, 0);
			}
		}
		break;
	case EDB_TYPE_DATE2:
		if(elements==1) {
			ForLink(lf, l_field) {
				if(STRisVALID((char *)lf->d)) {
					register quad_t dd = ATOQ((char *)lf->d);
					register int	d1, d2, y, y1, y2;

					d1 = (dd/100000000)%100000000;
					y1 = fiscal ? DATE_FiscalY(d1) : DATE_Y(d1);
					d2 = dd%100000000;
					y2 = fiscal ? DATE_FiscalY(d2) : DATE_Y(d2);
					for(y=y1; y<=y2; ++y) {
						class = edb_classify_append(class, y);
					}
				} else {
					class = edb_classify_append(class, 0);
				}
			}
		} else {
			ForLink(lf, l_field) {
				Link	l, l_value;
				int	none = FALSE;

				if(STRisVALID((char *)lf->d)) {
					l_value = edb_array_split(lf->d);
					if(l_value) {
						str_sort(l_value);
						l_value = str_uniq(l_value);
						ForLink(l, l_value) {
							register quad_t dd = ATOQ((char *)lf->d);
							register int	d1, d2, y, y1, y2;

							d1 = (dd/100000000)%100000000;
							y1 = fiscal ? DATE_FiscalY(d1) : DATE_Y(d1);
							d2 = dd%100000000;
							y2 = fiscal ? DATE_FiscalY(d2) : DATE_Y(d2);
							for(y=y1; y<=y2; ++y) {
								class = edb_classify_append(class, y);
							}
						}
						EdbLinkFree(l_value);
					} else {
						none = TRUE;
					}
				} else {
					none = TRUE;
				}
				if(none) class = edb_classify_append(class, 0);
			}
		}
		break;
	}
	EdbLinkFree(l_field);

	return class;
}

EdbClassify
edb_classify_table_column_by_YEAR(EdbContext ec, EdbTableInfo ti, EdbColumnInfo ci, int fiscal, char *fmt, ...)
{
	utf8_t	*query, *target;
	va_list	ap;
	BiTree	class = NULL;

	target = (utf8_t *)strdup("");
	if(fmt) {
		char	*s;

		va_start(ap, fmt);
		vasprintf(&s, fmt, ap);
		va_end(ap);
		if(STRisVALID((char *)s)) {
			FREE(target);
			asprintf((char **)&target, " WHERE %s", s);
		}
		FREE(s);
	}
	asprintf((char **)&query, "SELECT %s FROM %s%s", ColumnSN(ci), TABLE_SN(ti), target);
	FREE(target);
	class = edb_classify_table_column_by_YEAR_func(ec, ti, ci, fiscal, query);
	FREE(query);
	return class;
}

EdbClassify
edb_classify_tableAS_column_by_YEAR(EdbContext ec, EdbTableInfo ti, utf8_t *as, EdbColumnInfo ci, int fiscal, char *fmt, ...)
{
	utf8_t	*query, *target;
	va_list	ap;
	BiTree	class = NULL;

	target = (utf8_t *)strdup("");
	if(fmt) {
		char	*s;

		va_start(ap, fmt);
		vasprintf(&s, fmt, ap);
		va_end(ap);
		if(STRisVALID((char *)s)) {
			FREE(target);
			asprintf((char **)&target, " WHERE %s", s);
		}
		FREE(s);
	}
	asprintf((char **)&query, "SELECT %s FROM %s as %s%s", ColumnSN(ci), TABLE_SN(ti), as, target);
	FREE(target);
	class = edb_classify_table_column_by_YEAR_func(ec, ti, ci, fiscal, query);
	FREE(query);
	return class;
}

static
int
cidr24_addr(
	char			*addr)		/* xx.xx.xx.xx */
{
	unsigned int ip;

	if(addr==NULL) return 0;
	ip = ntohl(inet_addr(addr));
	return (ip>>8);
}

static
EdbClassify
edb_classify_table_column_by_CIDR24_func(EdbContext ec, EdbTableInfo ti, EdbColumnInfo ci, utf8_t *query)
{
	int	n;
	int	elements;
	Link	lf, l_field;
	BiTree	class = NULL;

	l_field = edb_pgsql_query_one_field(ec, query, &n);

	elements = edb_table_column_reldb_elements(ec, ti, NULL, ci);
	switch(ci->typeInfo->typeID) {
	case EDB_TYPE_IPv4ADDR:
		if(elements==1) {
			ForLink(lf, l_field) {
				if(STRisVALID((char *)lf->d)) {
					register int c24 = cidr24_addr((char *)lf->d);

					class = edb_classify_append(class, c24);
				} else {
					class = edb_classify_append(class, 0);
				}
			}
		} else {
			ForLink(lf, l_field) {
				Link	l, l_value;
				int	none = FALSE;
				int	pc = 0;

				if(STRisVALID((char *)lf->d)) {
					l_value = edb_array_split(lf->d);
					if(l_value) {
						str_sort(l_value);
						l_value = str_uniq(l_value);
						ForLink(l, l_value) {
							register int c24 = cidr24_addr((char *)l->d);

							if(pc==c24) continue;
							pc = c24;
							class = edb_classify_append(class, c24);
						}
						EdbLinkFree(l_value);
					} else {
						none = TRUE;
					}
				} else {
					none = TRUE;
				}
				if(none) class = edb_classify_append(class, 0);
			}
		}
		break;
	}
	EdbLinkFree(l_field);

	return class;
}

EdbClassify
edb_classify_table_column_by_CIDR24(EdbContext ec, EdbTableInfo ti, EdbColumnInfo ci, char *fmt, ...)
{
	utf8_t	*query, *target;
	va_list	ap;
	BiTree	class = NULL;

	target = (utf8_t *)strdup("");
	if(fmt) {
		char	*s;

		va_start(ap, fmt);
		vasprintf(&s, fmt, ap);
		va_end(ap);
		if(STRisVALID((char *)s)) {
			FREE(target);
			asprintf((char **)&target, " WHERE %s", s);
		}
		FREE(s);
	}
	asprintf((char **)&query, "SELECT %s FROM %s%s", ColumnSN(ci), TABLE_SN(ti), target);
	FREE(target);
	class = edb_classify_table_column_by_CIDR24_func(ec, ti, ci, query);
	FREE(query);
	return class;
}

EdbClassify
edb_classify_tableAS_column_by_CIDR24(EdbContext ec, EdbTableInfo ti, utf8_t *as, EdbColumnInfo ci, char *fmt, ...)
{
	utf8_t	*query, *target;
	va_list	ap;
	BiTree	class = NULL;

	target = (utf8_t *)strdup("");
	if(fmt) {
		char	*s;

		va_start(ap, fmt);
		vasprintf(&s, fmt, ap);
		va_end(ap);
		if(STRisVALID((char *)s)) {
			FREE(target);
			asprintf((char **)&target, " WHERE %s", s);
		}
		FREE(s);
	}
	asprintf((char **)&query, "SELECT %s FROM %s as %s%s", ColumnSN(ci), TABLE_SN(ti), as, target);
	FREE(target);
	class = edb_classify_table_column_by_CIDR24_func(ec, ti, ci, query);
	FREE(query);
	return class;
}


