/*
	$Id: condition.c,v 1.20 2012/04/03 03:51:13 alex Exp alex $
	condition.c--- make condition
*/

#ifndef	lint
static char *rcsid __attribute__((unused)) = "$Id";
#endif

#include	<stdio.h>
#include	<stdarg.h>
#include	<string.h>
#include	"edb/edb.h"

utf8_t *
edb_condition_make(EdbContext ec, EdbTableInfo ti, int logic, ...)
{
	va_list	ap;
	int		not = FALSE;
	char		*as = "";
	char		*column, *s, *s2;
	utf8_t	*condition;
	Link		l_condition = NULL;
	int		op, skip;
	EdbColumnInfo	super_ci = NULL;

	if(ti==NULL) {
		edb_trace(ec, "edb_condition_make: ti==NULL!\n");
		return NULL;
	}
	va_start(ap, logic);
	skip = FALSE;
	while((op = va_arg(ap, int)) != ConditionTerminate) {
		char	*prefix;

		s = NULL;
		if(op==COND_NOP) {	/* no operation */
			continue;
		}
		if(op==COND_SKIP_NEXT) {	/* no operation */
			skip = !skip;
			continue;
		}
		if(op==COND_PREFIX_NOT) {	/* not */
			if(!skip) not = !not;
			skip = FALSE;
			continue;
		}
		if(op==COND_PREFIX_TNAME) {
			s2 = va_arg(ap, char *);
			if(!skip) {
				asprintf(&s, "%s.", s2);
				as = STRDUP(s);
				FREE(s);
			}
			continue;
		}
		prefix = not ? "NOT ": "";
		if(op==COND_AVAILABLE) {	/* available */
			if(!skip) {
				asprintf(&s, "%s%s%s", prefix, as, EDB_BASE_AVAILABLE);
				LinkAppend(l_condition, link_atom_new(s, free));
			}
		} else if(op==COND_CONDITION) {
			s2 = va_arg(ap, char *);
			if(!skip) {
				asprintf(&s, "%s(%s)", prefix, s2);
				LinkAppend(l_condition, link_atom_new(s, free));
			}
		} else if(op==COND_COLUMN) {
			s2 = va_arg(ap, char *);
			if(!skip) super_ci = edb_table_seek_column(ec, ti, (utf8_t *)s2);
		} else {
			eid_t	eid;
			Link	l, lstr;
			EdbCatalogue	ca;
			int	v_int, v_int1, v_int2;
			double	v_double;
			utf8_t	*v_str;
			EdbColumnInfo	ci = NULL;
			char	*special_csn = NULL;

			column = va_arg(ap, char *);
			if(!strcmp(column, "REF") || !strcmp(column, "@.REF")) {
				asprintf(&special_csn, "%s*", EDB_COLUMN_L_EID);
			} else if(!strcmp(column, "PERM") || !strcmp(column, "@.PERM")) {
				asprintf(&special_csn, "%s*", EDB_COLUMN_L_PEID);
			} else if(!strcmp(column, "OWN") || !strcmp(column, "@.OWN")) {
				asprintf(&special_csn, "%s", EDB_BASE_OWNER);
			} else if(!strcmp(column, "MAP") || !strcmp(column, "@.MAP")) {
				asprintf(&special_csn, "%s", EDB_BASE_MAPTO_EID);
			} else if(!strcmp(column, "EID") || !strcmp(column, "@.EID")) {
				asprintf(&special_csn, "%s", EDB_BASE_EID);
			} else if(super_ci && column[0]=='@') {
				char	*xn;

				asprintf(&xn, "%s%s", ColumnXN(super_ci), &column[1]);
				ci = edb_table_seek_column(ec, ti, (utf8_t *)xn);
				FREE(xn);
			} else {
				ci = edb_table_seek_column(ec, ti, (utf8_t *)column);
			}
			if(!STRisVALID(special_csn) && ci==NULL) {
				edb_trace(ec, "edb_condition_make: ci==NULL!\n");
				return NULL;
			}
			s = NULL;
			switch(op) {
			case COND_EID_MATCH_NO_EXPAND:
				eid = va_arg(ap, eid_t);
				if(!skip) {
					if(STRisVALID(special_csn)) {
						asprintf(&s, "%s%s%s=%d", prefix, as, special_csn, eid);
					} else {
						asprintf(&s, "%s%s%s*=%d", prefix, as, ColumnSN_EID(ci), eid);
					}
				}
				break;
			case COND_EID_MATCH:
				eid = va_arg(ap, eid_t);
				if(!skip) {
					ca = edb_tuple_expand(ec, edb_tuple_get(ec, eid, 0));
					/* Caution: ca is not necessary to be free! */
					if(ca) {
						lstr = NULL;
						ForCatalogueLink(l, ca) {
							EdbTupleInfo	tuple = CatalogueObject(l);

							if(STRisVALID(special_csn)) {
								asprintf(&s2, "%s%s=%d", as, special_csn, TupleEID(tuple));
							} else {
								asprintf(&s2, "%s%s*=%d", as, ColumnSN_EID(ci), TupleEID(tuple));
							}
							LinkAppend(lstr, link_atom_new(s2, free));
						}
						s2 = str_concatenate_str(lstr, " OR ");
						EdbLinkFree(lstr);
						asprintf(&s, "%s(%s)", prefix, s2);
						FREE(s2);
					}
				}
				break;
			case COND_CATALOGUE_MATCH_NO_EXPAND:
				ca = va_arg(ap, EdbCatalogue);
				if(!skip) {
					lstr = NULL;
					ForCatalogueLink(l, ca) {
						EdbTupleInfo	tuple = CatalogueObject(l);

						if(STRisVALID(special_csn)) {
							asprintf(&s2, "%s%s=%d", as, special_csn, TupleEID(tuple));
						} else {
							asprintf(&s2, "%s%s*=%d", as, ColumnSN_EID(ci), TupleEID(tuple));
						}
						LinkAppend(lstr, link_atom_new(s2, free));
					}
					s2 = str_concatenate_str(lstr, " OR ");
					EdbLinkFree(lstr);
					asprintf(&s, "%s(%s)", prefix, s2);
					FREE(s2);
				}
				break;
			case COND_CATALOGUE_MATCH:
				ca = va_arg(ap, EdbCatalogue);
				if(!skip) {
					lstr = NULL;
					ForCatalogueLink(l, ca) {
						Link			l2;
						EdbCatalogue	ca2;
						EdbTupleInfo	tuple = CatalogueObject(l);

						ca2 = edb_tuple_expand(ec, tuple);
						/* Caution: ca2 is not necessary to be free! */
						ForCatalogueLink(l2, ca2) {
							EdbTupleInfo	tuple2 = CatalogueObject(l2);

							if(STRisVALID(special_csn)) {
								asprintf(&s2, "%s%s=%d", as, special_csn, TupleEID(tuple2));
							} else {
								asprintf(&s2, "%s%s*=%d", as, ColumnSN_EID(ci), TupleEID(tuple2));
							}
							LinkAppend(lstr, link_atom_new(s2, free));
						}
					}
					s2 = str_concatenate_str(lstr, " OR ");
					EdbLinkFree(lstr);
					asprintf(&s, "%s(%s)", prefix, s2);
					FREE(s2);
				}
				break;
			case COND_INT_EQ:
				v_int = va_arg(ap, int);
				if(!skip) {
					if(ci) {
						asprintf(&s, "%s%s%s%s=%d", prefix, as, ColumnSN(ci), edb_table_column_reldb_elements(ec, ti, super_ci, ci)==1 ? "" : "*", v_int);
					} else {
						edb_trace(ec, "edb_condition_make: invalid column!\n");
					}
				}
				break;
			case COND_INT_LE:
				v_int = va_arg(ap, int);
				if(!skip) {
					if(ci) {
						asprintf(&s, "%s%s%s%s<=%d", prefix, as, ColumnSN(ci), edb_table_column_reldb_elements(ec, ti, super_ci, ci)==1 ? "" : "*", v_int);
					} else {
						edb_trace(ec, "edb_condition_make: invalid column!\n");
					}
				}
				break;
			case COND_INT_LT:
				v_int = va_arg(ap, int);
				if(!skip) {
					if(ci) {
						asprintf(&s, "%s%s%s%s<%d", prefix, as, ColumnSN(ci), edb_table_column_reldb_elements(ec, ti, super_ci, ci)==1 ? "" : "*", v_int);
					} else {
						edb_trace(ec, "edb_condition_make: invalid column!\n");
					}
				}
				break;
			case COND_INT_GE:
				v_int = va_arg(ap, int);
				if(!skip) {
					if(ci) {
						asprintf(&s, "%s%s%s%s>=%d", prefix, as, ColumnSN(ci), edb_table_column_reldb_elements(ec, ti, super_ci, ci)==1 ? "" : "*", v_int);
					} else {
						edb_trace(ec, "edb_condition_make: invalid column!\n");
					}
				}
				break;
			case COND_INT_GT:
				v_int = va_arg(ap, int);
				if(!skip) {
					if(ci) {
						asprintf(&s, "%s%s%s%s>%d", prefix, as, ColumnSN(ci), edb_table_column_reldb_elements(ec, ti, super_ci, ci)==1 ? "" : "*", v_int);
					} else {
						edb_trace(ec, "edb_condition_make: invalid column!\n");
					}
				}
				break;
			case COND_DATE2_OVERLAP:
				v_int1 = va_arg(ap, int);
				v_int2 = va_arg(ap, int);
				if(!skip) {
					if(ci) {
						asprintf(&s, "%s(NOT (%d<int4(%s%s/int8(100000000)) OR int4(%s%s%%int8(100000000))<%d))",
							prefix, v_int2, as, ColumnSN(ci), as, ColumnSN(ci), v_int1);
					} else {
						edb_trace(ec, "edb_condition_make: invalid column!\n");
					}
				}
				break;
			case COND_REAL_EQ:
				v_double = va_arg(ap, double);
				if(!skip) {
					if(ci) {
						asprintf(&s, "%s%s%s%s=%f", prefix, as, ColumnSN(ci), edb_table_column_reldb_elements(ec, ti, super_ci, ci)==1 ? "" : "*", v_double);
					} else {
						edb_trace(ec, "edb_condition_make: invalid column!\n");
					}
				}
				break;
			case COND_REAL_LE:
				v_double = va_arg(ap, double);
				if(!skip) {
					if(ci) {
						asprintf(&s, "%s%s%s%s<=%f", prefix, as, ColumnSN(ci), edb_table_column_reldb_elements(ec, ti, super_ci, ci)==1 ? "" : "*", v_double);
					} else {
						edb_trace(ec, "edb_condition_make: invalid column!\n");
					}
				}
				break;
			case COND_REAL_LT:
				v_double = va_arg(ap, double);
				if(!skip) {
					if(ci) {
						asprintf(&s, "%s%s%s%s<%f", prefix, as, ColumnSN(ci), edb_table_column_reldb_elements(ec, ti, super_ci, ci)==1 ? "" : "*", v_double);
					} else {
						edb_trace(ec, "edb_condition_make: invalid column!\n");
					}
				}
				break;
			case COND_REAL_GE:
				v_double = va_arg(ap, double);
				if(!skip) {
					if(ci) {
						asprintf(&s, "%s%s%s%s>=%f", prefix, as, ColumnSN(ci), edb_table_column_reldb_elements(ec, ti, super_ci, ci)==1 ? "" : "*", v_double);
					} else {
						edb_trace(ec, "edb_condition_make: invalid column!\n");
					}
				}
				break;
			case COND_REAL_GT:
				v_double = va_arg(ap, double);
				if(!skip) {
					if(ci) {
						asprintf(&s, "%s%s%s%s>%f", prefix, as, ColumnSN(ci), edb_table_column_reldb_elements(ec, ti, super_ci, ci)==1 ? "" : "*", v_double);
					} else {
						edb_trace(ec, "edb_condition_make: invalid column!\n");
					}
				}
				break;
			case COND_TEXT_EQ:
#if 0
				v_str = va_arg(ap, utf8_t *);
				s = edb_text_for_pgsql(ec, v_str);
				v_str = STRDUP(s);
				FREE(s);
				if(!skip) {
					if(ci) {
						if(edb_table_column_reldb_elements(ec, ti, super_ci, ci)==1 && ci->typeInfo->typeID == EDB_TYPE_NAME) {
							asprintf(&s, "%s (english(%s%s)='%s' OR japanese(%s%s)='%s')", prefix, as, ColumnSN(ci), v_str, as, ColumnSN(ci), v_str);
						} else {
							asprintf(&s, "%s%s%s%s='%s'", prefix, as, ColumnSN(ci), edb_table_column_reldb_elements(ec, ti, super_ci, ci)==1 ? "" : "*", v_str);
						}
					} else {
						edb_trace(ec, "edb_condition_make: invalid column!\n");
					}
				}
#else
				v_str = va_arg(ap, utf8_t *);
				s = (char *)edb_text_for_pgsql_literal(ec, v_str);
				v_str = (utf8_t *)STRDUP(s);
				FREE(s);
				if(!skip) {
					if(ci) {
						if(edb_table_column_reldb_elements(ec, ti, super_ci, ci)==1 && ci->typeInfo->typeID == EDB_TYPE_NAME) {
							asprintf(&s, "%s (english(%s%s)=%s OR japanese(%s%s)=%s)", prefix, as, ColumnSN(ci), v_str, as, ColumnSN(ci), v_str);
						} else {
							asprintf(&s, "%s%s%s%s=%s", prefix, as, ColumnSN(ci), edb_table_column_reldb_elements(ec, ti, super_ci, ci)==1 ? "" : "*", v_str);
						}
					} else {
						edb_trace(ec, "edb_condition_make: invalid column!\n");
					}
				}
#endif
				break;
			case COND_REGEX_TEXT_EQ:
#if	defined(TEXT_VARIANT_CONVERSION)
#if 0
				v_str = va_arg(ap, utf8_t *);
				s = edb_text_for_pgsql_variant(ec, v_str);
				v_str = STRDUP(s);
				FREE(s);
				if(!skip) {
					if(ci) {
						if(edb_table_column_reldb_elements(ec, ti, super_ci, ci)==1 && ci->typeInfo->typeID == EDB_TYPE_NAME) {
							asprintf(&s, "%s (english(%s%s)~'%s' OR japanese(%s%s)~'%s')", prefix, as, ColumnSN_VARIANT(ci), v_str, as, ColumnSN_VARIANT(ci), v_str);
						} else {
							asprintf(&s, "%s%s%s%s~'%s'", prefix, as, ColumnSN_VARIANT(ci), edb_table_column_reldb_elements(ec, ti, super_ci, ci)==1 ? "" : "*", v_str);
						}
					} else {
						edb_trace(ec, "edb_condition_make: invalid column!\n");
					}
				}
#else
				if(ci &&
					(ci->typeInfo->typeID == EDB_TYPE_IPv4CIDR
					|| ci->typeInfo->typeID == EDB_TYPE_IPv4ADDR)) {
					v_str = va_arg(ap, utf8_t *);
					s = (char *)edb_text_for_pgsql_literal(ec, v_str);
					v_str = (utf8_t *)STRDUP(s);
					FREE(s);
					if(!skip) {
						if(ci) {
							if(edb_table_column_reldb_elements(ec, ti, super_ci, ci)==1 && ci->typeInfo->typeID == EDB_TYPE_NAME) {
								asprintf(&s, "%s (english(%s%s)~%s OR japanese(%s%s)~%s)", prefix, as, ColumnSN(ci), v_str, as, ColumnSN(ci), v_str);
							} else {
								asprintf(&s, "%s%s%s%s~%s", prefix, as, ColumnSN(ci), edb_table_column_reldb_elements(ec, ti, super_ci, ci)==1 ? "" : "*", v_str);
							}
						} else {
							edb_trace(ec, "edb_condition_make: invalid column!\n");
						}
					}
				} else {
					v_str = va_arg(ap, utf8_t *);
					s = (char *)edb_text_for_pgsql_literal_variant(ec, v_str);
					v_str = (utf8_t *)STRDUP(s);
					FREE(s);
					if(!skip) {
						if(ci) {
							if(edb_table_column_reldb_elements(ec, ti, super_ci, ci)==1 && ci->typeInfo->typeID == EDB_TYPE_NAME) {
								asprintf(&s, "%s (english(%s%s)~%s OR japanese(%s%s)~%s)", prefix, as, ColumnSN_VARIANT(ci), v_str, as, ColumnSN_VARIANT(ci), v_str);
							} else {
								asprintf(&s, "%s%s%s%s~%s", prefix, as, ColumnSN_VARIANT(ci), edb_table_column_reldb_elements(ec, ti, super_ci, ci)==1 ? "" : "*", v_str);
							}
						} else {
							edb_trace(ec, "edb_condition_make: invalid column!\n");
						}
					}
				}
#endif
#else
#if 0
				v_str = va_arg(ap, utf8_t *);
				s = edb_text_for_pgsql(ec, v_str);
				v_str = STRDUP(s);
				FREE(s);
				if(!skip) {
					if(ci) {
						if(edb_table_column_reldb_elements(ec, ti, super_ci, ci)==1 && ci->typeInfo->typeID == EDB_TYPE_NAME) {
							asprintf(&s, "%s (english(%s%s)~'%s' OR japanese(%s%s)~'%s')", prefix, as, ColumnSN(ci), v_str, as, ColumnSN(ci), v_str);
						} else {
							asprintf(&s, "%s%s%s%s~'%s'", prefix, as, ColumnSN(ci), edb_table_column_reldb_elements(ec, ti, super_ci, ci)==1 ? "" : "*", v_str);
						}
					} else {
						edb_trace(ec, "edb_condition_make: invalid column!\n");
					}
				}
#else
				v_str = va_arg(ap, utf8_t *);
				s = edb_text_for_pgsql_literal(ec, v_str);
				v_str = STRDUP(s);
				FREE(s);
				if(!skip) {
					if(ci) {
						if(edb_table_column_reldb_elements(ec, ti, super_ci, ci)==1 && ci->typeInfo->typeID == EDB_TYPE_NAME) {
							asprintf(&s, "%s (english(%s%s)~%s OR japanese(%s%s)~%s)", prefix, as, ColumnSN(ci), v_str, as, ColumnSN(ci), v_str);
						} else {
							asprintf(&s, "%s%s%s%s~%s", prefix, as, ColumnSN(ci), edb_table_column_reldb_elements(ec, ti, super_ci, ci)==1 ? "" : "*", v_str);
						}
					} else {
						edb_trace(ec, "edb_condition_make: invalid column!\n");
					}
				}
#endif
#endif	/* TEXT_VARIANT_CONVERSION */
				break;
			default:
				edb_trace(ec, "edb_condition_make: unresolved condition!\n");
				return NULL;
				break;
			}
			if(!skip) {
				if(STRisVALID(s)) LinkAppend(l_condition, link_atom_new(s, free));
			} else {
				FREE(s);
			}
			FREE(special_csn);
		}
		not = FALSE;
		skip = FALSE;
	}
	va_end(ap);

	if(logic==CONDITION_LOGIC_OR) {
		condition = (utf8_t *)str_concatenate_str(l_condition, " OR ");
	} else {
		condition = (utf8_t *)str_concatenate_str(l_condition, " AND ");
	}
	EdbLinkFree(l_condition);

	return condition;
}


