/*
	$Id: conf.c,v 1.43 2012/04/03 03:51:13 alex Exp alex $
	conf.c--- configuration
*/

#ifndef	lint
static char *rcsid __attribute__((unused)) = "$Id: conf.c,v 1.43 2012/04/03 03:51:13 alex Exp alex $";
#endif

#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<unistd.h>
#include	<netdb.h>
#include	<time.h>
#include	<sys/types.h>
#include	<sys/socket.h>
#include	<netinet/in.h>
#include	"edb/edb.h"
#include	"edb/xml.h"

#define	DEFAULT_HOST	"localhost"
#define	DEFAULT_PORT	"5432"

#if	defined(EDB_DEBUGGER)
static
void
edb_configure_debugger(EdbContext ec)
{
	static eid_t	debugger[] = EDB_DEBUGGER;
	int	i;

	ec->ca_debugger = edb_catalogue_new();
	for(i=0; i<Ns(debugger) && EIDisValid(debugger[i]); ++i) {
		edb_catalogue_append(ec->ca_debugger, debugger[i], "");
	}
}
#endif

static
EdbDBPath
edb_dbpath_parse(char *dbpath)
{
	EdbDBPath	db;
	int	n;
	Link	lstr;

	db = (EdbDBPath)malloc(sizeof(struct _EDB_DBPATH));
	bzero(db, sizeof(struct _EDB_DBPATH));

	if(!STRisValid(dbpath)) dbpath = EDB_DBPATH_DEFAULT;

	db->path = strdup((char *)dbpath);
	lstr = str_split((char *)dbpath, ':');
	n = link_atoms(lstr);
	if(n==1) {
		db->host = strdup(DEFAULT_HOST);
		db->port = strdup(DEFAULT_PORT);
		db->dbname = strdup(lstr->d);
	} else if(n==2) {
		db->host = strdup(DEFAULT_HOST);
		db->port = strdup(lstr->d);
		db->dbname = strdup(lstr->n->d);
	} else if(n>=3) {
		db->host = strdup(lstr->d);
		db->port = strdup(lstr->n->d);
		db->dbname = strdup(lstr->n->n->d);
	}
	if(STRisValid(db->dbname)) {
		char	*p;
		int	replica = 0;

		p = strchr(db->dbname, '.');
		if(p) {
			*p++ = '\0';
			replica = ATOI(p);
		}
		db->replica = replica;
	}
	link_free(lstr);
	return db;
}

static
int
whois_get(EdbContext ec, char *host, char *query)
{
	register FILE *sfi, *sfo;
	int	s;
	struct sockaddr_in sin;
	struct hostent *hp;
	struct servent *sp;

	hp = gethostbyname(host);
	if(hp == NULL) {
		fprintf(stderr, "whois_get: %s: not exist.\n", host);
		return FALSE;
	}

	sp = getservbyname("nicname", "tcp");
	if(sp==NULL) {
		fprintf(stderr, "whois_get: %s: not exist.\n", "whois/tcp");
		return FALSE;
	}
	host = hp->h_name;
	s = socket(hp->h_addrtype, SOCK_STREAM, 0);
	if(s < 0) {
		perror("whois_get: socket");
		return FALSE;
	}
	bzero(&sin, sizeof (sin));
	sin.sin_family = hp->h_addrtype;
	if(bind(s, (struct sockaddr *)&sin, sizeof (sin)) < 0) {
		perror("whois_get: bind");
		return FALSE;
	}
	bcopy(hp->h_addr, &sin.sin_addr, hp->h_length);
	sin.sin_port = sp->s_port;
	if(connect(s, (struct sockaddr *)&sin, sizeof (sin)) < 0) {
		perror("whois_get: connect");
		return FALSE;
	}
	sfi = fdopen(s, "r");
	sfo = fdopen(s, "w");
	if(sfi == NULL || sfo == NULL) {
		perror("fdopen");
		close(s);
		return FALSE;
	}
	fprintf(sfo, "%s\r\n", query);
	(void)fflush(sfo);
	edb_dbpath_reset(ec);
	while(!feof(sfi)) {
		char	*path;

		path = file_getline(sfi);
		if(path==NULL) break;
		edb_dbpath_set(ec, path, FALSE);
		FREE(path);
	}
	fclose(sfo);
	fclose(sfi);
	return TRUE;
}

/*#edb_configure--- main function */
EdbContext
edb_configure(char *path)
{
	char	*s;
	EdbContext	ec;
	Link	l;
	EdbDBPath	db;
	char	*whois_serv = NULL, *whois_query = NULL;

	srandomdev();

	ec = (EdbContext)malloc(sizeof(struct EDB_CONTEXT));
	if(ec == NULL) return NULL;
	bzero(ec, sizeof(struct EDB_CONTEXT));

#if 0
	ec->host = DEFAULT_HOST;
	ec->port = DEFAULT_PORT;
	ec->dbName = EDB_DB;
#else
	db = edb_dbpath_parse(EDB_DBPATH_DEFAULT);
	ec->host = db->host;
	ec->port = db->port;
	ec->dbName = db->dbname;
	ec->number_of_replica = db->replica;
#endif
	ec->options = NULL;
	ec->tty = NULL;
	ec->pgsqlUser = NULL;
	ec->pgsqlPassword = NULL;
	ec->transaction = FALSE;
	ec->locked = FALSE;
	ec->japanese = TRUE;
	ec->masterUID = EDB_MASTER_UID;
	ec->master = FALSE;
	ec->operator = 0;
	ec->browser_url = EDB_BROWSER_URL_DEFAULT;
	ec->l_mynetwork = NULL;
	ec->max_mapto_recurse = MAX_MAPTO_RECURSE;
	ec->max_hierarchy_recurse = MAX_HIERARCHY_RECURSE;
	ec->dont_use_reldb = FALSE;

	/* read configuration file */
	if(path) {
		Link	lstr;

		lstr = file_load_as_linked_str(path, NULL);
		if(lstr==NULL) return NULL;
		ForLink(l, lstr) {
			char *p, tag[BUFSIZ], value[BUFSIZ];
			int	m;

			if((p = strchr(l->d, '#'))) *p = '\0';
			if(strlen(l->d)==0) continue;
			str_oneline(l->d);
			if((m=sscanf(l->d, "%s %s", tag, value)) == 0) continue;
			if(!strcasecmp(tag, "pgsqlHost")) {
				ec->host = (m>1) ? strdup(value) : DEFAULT_HOST ;
			} else if(!strcasecmp(tag, "pgsqlPort")) {
				ec->port = (m>1) ? strdup(value) : DEFAULT_PORT ;
			} else if(!strcasecmp(tag, "pgsqlOptions")){
				ec->options = (m>1) ? strdup(value) : NULL ;
			} else if(!strcasecmp(tag, "pgsqlTty")) {
				ec->tty = (m>1) ? strdup(value) : NULL ;
			} else if(!strcasecmp(tag, "pgsqlDBName")) {
				ec->dbName = (m>1) ? strdup(value) : NULL ;
			} else if(!strcasecmp(tag, "pgsqlDBPath")) {
				int	n;

				db = edb_dbpath_parse(value);
				ec->host = db->host;
				ec->port = db->port;
				ec->dbName = db->dbname;
				ec->number_of_replica = db->replica;
#if 0
				if((n=link_atoms(ec->l_dbpath))==0) {
					LinkAppend(ec->l_dbpath, link_atom_new(db, NULL));
				} else if(random()%(n+1)==0) {
					LinkAppendTop(link_atom_new(db, NULL), ec->l_dbpath);
				} else {
					LinkAppend(ec->l_dbpath, link_atom_new(db, NULL));
				}
#else
				if((n=link_atoms(ec->l_dbpath))==0) {
					ec->l_dbpath = link_atom_new(db, NULL);
				} else {
					ec->l_dbpath = link_insert(ec->l_dbpath, (int)(random()%(n+1)), link_atom_new(db, NULL));
				}
#endif
			} else if(!strcasecmp(tag, "WhoisServer")) {
				whois_serv = strdup(value);
			} else if(!strcasecmp(tag, "WhoisQuery")) {
				whois_query = strdup(value);
			} else if(!strcasecmp(tag, "pgsqlUser")) {
				ec->pgsqlUser = (m>1) ? strdup(value) : NULL ;
			} else if(!strcasecmp(tag, "pgsqlPassword")) {
				ec->pgsqlPassword = (m>1) ? strdup(value) : NULL ;
			} else if(!strcasecmp(tag, "EDBMasterUID")) {
				ec->masterUID = (m>1) ? atoi(value) : 0 ;
			} else if(!strcasecmp(tag, "MyNetwork")) {
				LinkAppend(ec->l_mynetwork, link_atom_new_sdup(value));
			} else if(!strcasecmp(tag, "CaptionOmit")) {
				m = ATOI(value);
				if(EIDisVALID(m)) edb_caption_omit_append(ec, m);
			} else if(!strcasecmp(tag, "RelationalDB")) {
				ec->dont_use_reldb = !strcasecmp(value, "none");
			} else {
#if 0
				fprintf(stderr, "Configuration Error\n");
#endif
			}
		}
		link_free(lstr);
	}
	if(STRisValid(whois_serv) && STRisValid(whois_query)) {
		whois_get(ec, whois_serv, whois_query);
	}

	/* check master */
	if(ec->masterUID == getuid()) ec->master = TRUE;

	/* remote addr */
	ec->remote_addr = NULL;
	ec->university = FALSE;
	s = getenv("REMOTE_ADDR");
	if(s) {
		ec->remote_addr = strdup(s);
		ForLink(l, ec->l_mynetwork) {
			if(check_addr_in_network((utf8_t *)ec->remote_addr, l->d)) {
				ec->university = TRUE;
				break;
			}
		}
	}
#ifdef EDB_READER
	ec->readonly = TRUE;
#else /* !EDB_READER */
#endif /*EDB_READER */

	/* init number of query / command */
	ec->number_of_query = 0;
	ec->number_of_command = 0;
	ec->number_of_list_hit = 0;

	/* for XML */
	xmlInitParser();
	xmlSubstituteEntitiesDefault(1);

	edb_trace_set(ec, stderr);
	gettimeofday(&ec->configured_time, NULL);

	if(ec->l_dbpath) ec->l_dbpath = link_randomize(ec->l_dbpath);
#if	defined(EDB_DEBUGGER)
	edb_configure_debugger(ec);
#endif

	return ec;
}

void
edb_dbpath_reset(EdbContext ec)
{

	link_free(ec->l_dbpath);
	ec->l_dbpath = NULL;
}

void
edb_dbpath_set(EdbContext ec, char *path, int exclusive)
{
	EdbDBPath	db;

	if(exclusive) edb_dbpath_reset(ec);
	db = edb_dbpath_parse(path);
	ec->host = db->host;
	ec->port = db->port;
	ec->dbName = db->dbname;
	ec->number_of_replica = db->replica;
	LinkAppend(ec->l_dbpath, link_atom_new(db, NULL));
}

void
edb_configure_propagate_authority(EdbContext dst, EdbContext src)
{

	dst->operator = src->operator;
	dst->staff = src->staff;
	dst->Odin = src->Odin;
}

void
edb_configure_master_authority(EdbContext ec)
{

	ec->operator = EDB_MASTER_UID;
	ec->staff = TRUE;
	ec->Odin = TRUE;
}

void
edb_configure_user_authority(EdbContext ec, eid_t user)
{

	ec->operator = user;
	ec->staff = FALSE;
	ec->Odin = FALSE;
}

