/*
	$Id: datum.c,v 1.59 2012/04/03 03:51:13 alex Exp alex $
	datum.c--- datum info
*/

#ifndef	lint
static char *rcsid __attribute__((unused)) = "$Id: datum.c,v 1.59 2012/04/03 03:51:13 alex Exp alex $";
#endif

#include	<stdio.h>
#include	<stdarg.h>
#include	<string.h>
#include	<ctype.h>
#define	EDB_TYPE_DEFINITION
#include	"edb/edb.h"
#include	"edb/pki.h"

#define	EDB_MAX_PASSPHRASE_LEN	128

static	utf8_t	*edb_datum_text_null_string = (utf8_t *)"";	/* null string for EdbDatum */

#define	DatumTextFree(ec, s) if((s) && (s)!=edb_datum_text_null_string) { free((s)); (s) = edb_datum_text_null_string;	}

static	BiTree	bt_shared_text = NULL;
struct	SHARED_TEXT {
	int		refs;
	utf8_t	text[0];
};
typedef	struct SHARED_TEXT *SharedText;

static
int
edb_datum_text_compare_for_search(register void *o1, register void *o2)
{
	register utf8_t *t1 = (utf8_t *)o1;
	register utf8_t *t2 = ((SharedText)o2)->text;

	return strcmp((char *)t1, (char *)t2);
}

static
int
edb_datum_text_compare(register void *o1, register void *o2)
{
	register utf8_t *t1 = ((SharedText)o1)->text;
	register utf8_t *t2 = ((SharedText)o2)->text;

	return strcmp((char *)t1, (char *)t2);
}

static
utf8_t *
edb_datum_text_use(register EdbContext ec, register utf8_t *text)
{
	register SharedText	st;

	if(!STRisValid(text)) return edb_datum_text_null_string;
	st = bitree_search(bt_shared_text, text, bitree_hint_str((char *)text), edb_datum_text_compare_for_search);
	if(st==NULL) {
		register int	len;

		len = sizeof(struct SHARED_TEXT) + sizeof(utf8_t)*(strlen((char *)text)+1);
		st = (SharedText)Edb_malloc(len);
		bzero(st, len);
		strcpy((char *)st->text, (char *)text);
		BiTreeAdd(bt_shared_text, st, bitree_hint_str((char *)st->text), edb_datum_text_compare, NULL, NULL);
	}
	++st->refs;
	return st->text;
}

static
void
edb_datum_text_unuse(register EdbContext ec, register utf8_t *text)
{
	register SharedText	st;

	if(!STRisValid(text)) return;
	st = bitree_search(bt_shared_text, text, bitree_hint_str((char *)text), edb_datum_text_compare_for_search);
	if(st) {
		if(st->refs>0) {
			--st->refs;
		} else {
			edb_trace(ec, "edb_datum_text_unuse: reference counter underflow. (text: '%s').\n", text);
		}
		if(st->refs==0) {
			SharedText	robj;

			BiTreeRemove(bt_shared_text, text, bitree_hint_str((char *)text), edb_datum_text_compare_for_search, (BiTree_obj_t)&robj);
			if(robj==NULL) edb_trace(ec, "edb_datum_text_unuse: remove-object is not found. (text: '%s').\n", text);
			Edb_free(robj);
		}
	} else {
		edb_trace(ec, "edb_datum_text_unuse: can't find text '%s'.\n", text);
	}
}

EdbTypeInfo
edb_datum_type_to(EdbContext ec, char *name)
{
	int	i;

	for(i=0; i<Ns(edb_datum_type_table); ++i) {
		if(!strcasecmp(name, edb_datum_type_table[i].name)) return &edb_datum_type_table[i];
	}
	return &edb_datum_type_table[0];
}

EdbCTypeInfo
edb_datum_ctype_new(EdbContext ec, EdbColumnInfo ci, char *name)
{
	EdbCTypeInfo	ctype;

	ctype = (EdbCTypeInfo)malloc(sizeof(struct EDB_CTYPE_INFO));
	bzero(ctype, sizeof(struct EDB_CTYPE_INFO));
	ctype->name = strdup(name);
	ctype->ci = ci;
	ctype->info = edb_datum_type_to(ec, name);
	return ctype;
}

void
edb_datum_ctype_free(EdbContext ec, EdbCTypeInfo ctype)
{

	if(ctype==NULL) return;
	FREE(ctype->name);
	FREE(ctype->v_minimum);
	FREE(ctype->v_maximum);
	FREE(ctype->v_default);
	FREE(ctype->cryptography);
	FREE(ctype);
}

EdbTypeInfo
edb_datum_typeID_to(EdbContext ec, int typeID)
{
	int	i;

	for(i=0; i<Ns(edb_datum_type_table); ++i) {
		if(typeID == edb_datum_type_table[i].typeID) return &edb_datum_type_table[i];
	}
	return &edb_datum_type_table[0];
}

void
edb_datum_linkto_type_description(EdbContext ec, EdbTypeInfo typeinfo, Doc doc)
{

	doc_printf(doc, "<a href=\"/edb-manual/edb/type-definition.html#TYPE-%s\">%s</a>\n", typeinfo->name, typeinfo->desc);
}

/*
	edb_datum_new--- new datum
	edb_datum_free--- new datum
*/
EdbDatum
edb_datum_new(EdbContext ec, void *super)
{
	register EdbDatum	datum;

	if(_edb_delayed_free_datum) {
		++edb_delayed_free_recycle_count;
		EdbDelayedFreePop(_edb_delayed_free_datum, datum);
	} else {
		datum = Edb_malloc(sizeof(struct EDB_DATUM));
	}
	bzero(datum, sizeof(struct EDB_DATUM));
	datum->super = super;
	DatumRead(datum) = EDB_AUTH_INHERIT;
	DatumDateFrom(datum) = EDB_DATE_MIN;
	DatumDateTo(datum) = EDB_DATE_MAX;
	datum->d_english = edb_datum_text_null_string;
	datum->d_japanese = edb_datum_text_null_string;
	datum->d_pronounce = edb_datum_text_null_string;
	return datum;
}

void
edb_datum_free(EdbContext ec, EdbDatum datum)
{

	if(datum==NULL) return;
	if(datum->english_shared) {
		edb_datum_text_unuse(ec, datum->d_english);
	} else {
		DatumTextFree(ec, datum->d_english);
	}
	if(datum->japanese_shared) {
		edb_datum_text_unuse(ec, datum->d_japanese);
	} else {
		DatumTextFree(ec, datum->d_japanese);
	}
	if(datum->pronounce_shared) {
		edb_datum_text_unuse(ec, datum->d_pronounce);
	} else {
		DatumTextFree(ec, datum->d_pronounce);
	}
	EdbLinkFree2(ec, datum->l_child);
	if(_edb_delayed_free_mode>0) {
		EdbDelayedFreePush(_edb_delayed_free_datum, datum);
	} else {
		Edb_free(datum);
	}
}

static
int
edb_datum_modify_TEXT(EdbContext ec, EdbCTypeInfo ctype, EdbDatum datum)
{
	register ucs4_t	*usrc, *udst, *u;
	utf8_t	*s;

	if(STRisValid(datum->d_english)) {
		usrc = UTF8toUCS4(datum->d_english);
		udst = edb_text_conv(ec, usrc);
		for(u=udst; *u; ++u) {
			if(!ucs4_isEnglish(*u)) {
				edb_trace(ec, "Modify: character (0x%x) -> SPACE.\n", *u);
				*u = UCS4_SPACE;
			}
		}
		s = UCS4toUTF8(udst);
		FREE(udst);
		str_oneline((char *)s);
		DatumTextFree(ec, datum->d_english);
		datum->d_english = STRisValid(s) ? (utf8_t *)strdup((char *)s) : edb_datum_text_null_string;
	}
	return TRUE;
}

static
int
edb_datum_modify_TEXT2(EdbContext ec, EdbCTypeInfo ctype, EdbDatum datum)
{

	register ucs4_t	*usrc, *udst, *u, *ud;
	utf8_t	*s;

	if(STRisValid(datum->d_english)) {
		usrc = UTF8toUCS4(datum->d_english);
		udst = edb_text_conv(ec, usrc);
		for(u=ud=udst; *u; ++u) {
			if(!ucs4_isEnglish(*u)) {
				edb_trace(ec, "Modify: character (0x%x) is deleted.\n", *u);
			} else {
				*ud++ = *u;
			}
		}
		*ud = UCS4_NULL;
		s = UCS4toUTF8(udst);
		FREE(udst);
		str_oneline((char *)s);
		DatumTextFree(ec, datum->d_english);
		datum->d_english = STRisValid(s) ? (utf8_t *)strdup((char *)s) : edb_datum_text_null_string;
	}
	if(STRisValid(datum->d_english)) {
		if(!strchr((char *)datum->d_english, ' ')) {
			edb_trace(ec, "Modify: duplicate string (%s).\n", datum->d_english);
			asprintf((char **)&s, "%s %s", datum->d_english, datum->d_english);
			DatumTextFree(ec, datum->d_english);
			datum->d_english = s;
		}
	}
	return TRUE;
}

static
int
edb_datum_modify_NAME(EdbContext ec, EdbCTypeInfo ctype, EdbDatum datum)
{
	register ucs4_t	*usrc, *udst, *u;
	utf8_t	*s;

	if(STRisValid(datum->d_english)) {
		usrc = UTF8toUCS4(datum->d_english);
		udst = edb_text_conv(ec, usrc);
		for(u=udst; *u; ++u) {
			if(!ucs4_isEnglish(*u)) {
				edb_trace(ec, "Modify: character (0x%x) -> SPACE.\n", *u);
				*u = UCS4_SPACE;
			}
		}
		s = UCS4toUTF8(udst);
		FREE(udst);
		str_oneline((char *)s);
		DatumTextFree(ec, datum->d_english);
		datum->d_english = STRisValid(s) ? (utf8_t *)strdup((char *)s) : edb_datum_text_null_string;
	}

	if(STRisValid(datum->d_japanese)) {
		usrc = UTF8toUCS4(datum->d_japanese);
		udst = edb_text_conv(ec, usrc);
		for(u=udst; *u; ++u) {
			if(!ucs4_isJapanese(*u)) {
				edb_trace(ec, "Modify: character (0x%x) -> SPACE.\n", *u);
				*u = UCS4_SPACE;
			}
		}
		s = UCS4toUTF8(udst);
		FREE(udst);
		str_oneline((char *)s);
		DatumTextFree(ec, datum->d_japanese);
		datum->d_japanese = STRisValid(s) ? (utf8_t *)strdup((char *)s) : edb_datum_text_null_string;
	}

	if(STRisValid(datum->d_pronounce)) {
		usrc = UTF8toUCS4(datum->d_pronounce);
		udst = edb_text_conv(ec, usrc);
		for(u=udst; *u; ++u) {
			if(!ucs4_isPronounce(*u)) {
				edb_trace(ec, "Modify: character (0x%x) -> SPACE.\n", *u);
				*u = UCS4_SPACE;
			}
		}
		s = UCS4toUTF8(udst);
		FREE(udst);
		str_oneline((char *)s);
		DatumTextFree(ec, datum->d_pronounce);
		datum->d_pronounce = STRisValid(s) ? (utf8_t *)strdup((char *)s) : edb_datum_text_null_string;
	}

	return TRUE;
}

static
int
edb_datum_modify_SENTENCE(EdbContext ec, EdbCTypeInfo ctype, EdbDatum datum)
{

	return edb_datum_modify_NAME(ec, ctype, datum);
}

static
int
edb_datum_modify_INTEGER(EdbContext ec, EdbCTypeInfo ctype, EdbDatum datum)
{

	if(edb_datum_modify_TEXT(ec, ctype, datum)) {
		if(STRisValid(datum->d_english)) {
			double	d = ATOF((char *)datum->d_english);
			int		v, v0;

			v = d;
			if(d<INT_MIN) {
				v = INT_MIN;
				edb_trace(ec, "Modify: integer (%f) -> (%d).\n", d, v);
			}
			if(d>INT_MAX) {
				v = INT_MAX;
				edb_trace(ec, "Modify: integer (%f) -> (%d).\n", d, v);
			}
			if(STRisValid(ctype->v_minimum) && (d<(v0=ATOI((char *)ctype->v_minimum)))) {
				v = v0;
				edb_trace(ec, "Modify: integer (%f) -> (%d).\n", d, v);
			}
			if(STRisValid(ctype->v_maximum) && (d>(v0=ATOI((char *)ctype->v_maximum)))) {
				v = v0;
				edb_trace(ec, "Modify: integer (%f) -> (%d).\n", d, v);
			}
			DatumTextFree(ec, datum->d_english);
			asprintf((char **)&datum->d_english, "%d", v);
		}
	}
	return TRUE;
}

static
int
edb_datum_modify_REAL(EdbContext ec, EdbCTypeInfo ctype, EdbDatum datum)
{

	if(edb_datum_modify_TEXT(ec, ctype, datum)) {
		if(STRisValid(datum->d_english)) {
			double	d = ATOF((char *)datum->d_english), v;

			if(STRisValid(ctype->v_minimum) && (d<(v=ATOF((char *)ctype->v_minimum)))) {
				edb_trace(ec, "Modify: real (%f) -> (%f).\n", d, v);
				d = v;
			}
			if(STRisValid(ctype->v_maximum) && (d>(v=ATOF((char *)ctype->v_maximum)))) {
				edb_trace(ec, "Modify: real (%f) -> (%f).\n", d, v);
				d = v;
			}
			DatumTextFree(ec, datum->d_english);
			asprintf((char **)&datum->d_english, "%f", d);
		}
	}
	return TRUE;
}

static
int
edb_datum_modify_DATE(EdbContext ec, EdbCTypeInfo ctype, EdbDatum datum)
{

	if(edb_datum_modify_TEXT(ec, ctype, datum)) {
		if(STRisValid(datum->d_english)) {
			int	d = ATOI((char *)datum->d_english);

			if(d<10000) {	/* 年の指定なし */
				DatumTextFree(ec, datum->d_english);
				datum->d_english = edb_datum_text_null_string;
			} else if(strlen((char *)datum->d_english)!=8) {
				DatumTextFree(ec, datum->d_english);
				asprintf((char **)&datum->d_english, "%08d", d);
			}
		}
	}
	return TRUE;
}

static
int
edb_datum_modify_DATE2(EdbContext ec, EdbCTypeInfo ctype, EdbDatum datum)
{

	if(edb_datum_modify_TEXT(ec, ctype, datum)) {
		if(STRisValid(datum->d_english)) {
			int	d1, d2;

			sscanf((char *)datum->d_english, "%d %d", &d1, &d2);
			if(d1<10000 && d2<10000) {
				DatumTextFree(ec, datum->d_english);
				datum->d_english = edb_datum_text_null_string;
			} else if(d2<d1) {
				d2 = d1;
				DatumTextFree(ec, datum->d_english);
				asprintf((char **)&datum->d_english, "%08d %08d", d1, d2);
			} else if(strlen((char *)datum->d_english)!=8*2+1) {
				DatumTextFree(ec, datum->d_english);
				asprintf((char **)&datum->d_english, "%08d %08d", d1, d2);
			}
		}
	}
	return TRUE;
}

static
int
edb_datum_modify_PICTURE(EdbContext ec, EdbCTypeInfo ctype, EdbDatum datum)
{
	register char	*s;

	if(STRisValid(datum->d_english)) {
		s = STRDUP((char *)datum->d_english);
		str_oneline(s);
		if(strcmp((char *)datum->d_english, s)) {
			DatumTextFree(ec, datum->d_english);
			datum->d_english = (utf8_t *)strdup(s);
		}
	}
	return TRUE;
}

static
int
edb_datum_modify_IPV4CIDR(EdbContext ec, EdbCTypeInfo ctype, EdbDatum datum)
{

	if(STRisValid(datum->d_english)) {
		int	ip1, ip2, ip3, ip4, m;

		if(sscanf((char *)datum->d_english, "%d.%d.%d.%d/%d", &ip1, &ip2, &ip3, &ip4, &m)==5) {
			char	*s;

			asprintf((char **)&s, "%d.%d.%d.%d/%d", ip1, ip2, ip3, ip4, m);
			if(strcmp((char *)datum->d_english, s)) {
				DatumTextFree(ec, datum->d_english);
				datum->d_english = (utf8_t *)s;
			} else {
				FREE(s);
			}
		}
	}
	return TRUE;
}

static
int
edb_datum_modify_IPV4ADDRESS(EdbContext ec, EdbCTypeInfo ctype, EdbDatum datum)
{

	if(STRisValid(datum->d_english)) {
		int	ip1, ip2, ip3, ip4;

		if(sscanf((char *)datum->d_english, "%d.%d.%d.%d", &ip1, &ip2, &ip3, &ip4)==4) {
			char	*s;

			asprintf((char **)&s, "%d.%d.%d.%d", ip1, ip2, ip3, ip4);
			if(strcmp((char *)datum->d_english, s)) {
				DatumTextFree(ec, datum->d_english);
				datum->d_english = (utf8_t *)s;
			} else {
				FREE(s);
			}
		}
	}
	return TRUE;
}

static
int
edb_datum_modify_INETNAME(EdbContext ec, EdbCTypeInfo ctype, EdbDatum datum)
{

	if(STRisValid(datum->d_english)) {
		register char	*s = STRDUP((char *)datum->d_english);

		str_onephrase(s);
		str_capitalize1(s);
		if(strcmp((char *)datum->d_english, s)) {
			DatumTextFree(ec, datum->d_english);
			datum->d_english = STRisValid(s) ? (utf8_t *)strdup(s) : edb_datum_text_null_string;
		}
	}
	return TRUE;
}

static
int
edb_datum_modify_INETFQDN(EdbContext ec, EdbCTypeInfo ctype, EdbDatum datum)
{

	if(STRisValid(datum->d_english)) {
		register char	*s = STRDUP((char *)datum->d_english);

		str_onephrase(s);
		str_capitalize1(s);
		if(strcmp((char *)datum->d_english, s)) {
			DatumTextFree(ec, datum->d_english);
			datum->d_english = STRisValid(s) ? (utf8_t *)strdup(s) : edb_datum_text_null_string;
		}
	}
	return TRUE;
}

static
int
edb_datum_modify_CERTIFICATE(EdbContext ec, EdbCTypeInfo ctype, EdbDatum datum)
{

	if(STRisValid(datum->d_english)) {
		register char	*s = STRDUP((char *)datum->d_english);

		str_oneline(s);
		if(strcmp((char *)datum->d_english, s)) {
			DatumTextFree(ec, datum->d_english);
			datum->d_english = STRisValid(s) ? (utf8_t *)strdup(s) : edb_datum_text_null_string;
		}
	}
	return TRUE;
}

static
utf8_t *
edb_datum_encrypt(EdbContext ec, EdbCTypeInfo ctype, utf8_t *str)
{

	if(!ctype) return NULL;
	if(!STRisValid(ctype->cryptography)) return NULL;
	if(!strcmp(ctype->cryptography, "RSA")) {
		utf8_t	*enc;
		char		*cert;

		asprintf((char **)&cert, "%s.crt", ColumnXN(ctype->ci));
		enc = edb_pki_rsa_encrypt(ec, str, strlen((char *)str), cert);
		if(STRisValid(enc)) str_oneline((char *)enc);
		FREE(cert);
		return enc;
	}
	return NULL;
}

static
int
edb_datum_modify_PASSPHRASE(EdbContext ec, EdbCTypeInfo ctype, EdbDatum datum)
{
	int	len, minl, maxl;

	edb_datum_modify_TEXT(ec, ctype, datum);
	if(ctype==NULL || !STRisValid(datum->d_english)) return TRUE;
	if(!STRisValid(ctype->cryptography)) return TRUE;

	len = strlen(ctype->cryptography);
	if(datum->d_english[0] =='{'	/* } */
		&& !strncmp((char *)&datum->d_english[1], ctype->cryptography, len)
		&& !isalnum(datum->d_english[len+1]))
		return TRUE;
	minl = 1;
	maxl = EDB_MAX_PASSPHRASE_LEN;
	if(STRisValid(ctype->v_minimum)) {
		len = ATOI((char *)ctype->v_minimum);
		minl = max(minl, len);
	}
	if(STRisValid(ctype->v_maximum)) {
		len = ATOI((char *)ctype->v_maximum);
		maxl = min(maxl, len);
	}
	len = strlen((char *)datum->d_english);
	if(minl<=len && len<=maxl) {
		utf8_t	*enc;

		enc = edb_datum_encrypt(ec, ctype, datum->d_english);
		len = strlen(ctype->cryptography);
		if(STRisValid(enc) && enc[0] =='{'	/* } */
			&& !strncmp((char *)&enc[1], ctype->cryptography, len)
			&& !isalnum(enc[len+1])) {
			DatumTextFree(ec, datum->d_english);
			datum->d_english = STRisValid(enc) ? (utf8_t *)strdup((char *)enc) : edb_datum_text_null_string;
		}
		FREE(enc);
	}
	return TRUE;
}

/*
	edb_dabum_dup
*/
EdbDatum
edb_datum_dup(EdbContext ec, void *super, EdbDatum src)
{
	register EdbDatum	new;

	new = edb_datum_new(ec, super);
	bcopy(src, new, sizeof(struct EDB_DATUM));
	new->super = super;
	if(!ec->edb_datum_text_sharable) {
		new->d_english = STRisValid(new->d_english) ? (utf8_t *)strdup((char *)new->d_english) : edb_datum_text_null_string;
		new->d_japanese = STRisValid(new->d_japanese) ? (utf8_t *)strdup((char *)new->d_japanese) : edb_datum_text_null_string;
		new->d_pronounce = STRisValid(new->d_pronounce) ? (utf8_t *)strdup((char *)new->d_pronounce) : edb_datum_text_null_string;
		new->english_shared = new->japanese_shared = new->pronounce_shared = FALSE;
	} else {
		new->d_english = edb_datum_text_use(ec, new->d_english);
		new->d_japanese = edb_datum_text_use(ec, new->d_japanese);
		new->d_pronounce = edb_datum_text_use(ec, new->d_pronounce);
		new->english_shared = new->japanese_shared = new->pronounce_shared = TRUE;
	}
	new->l_child = NULL;
	return new;
}

static
int
datum_string_compare(register utf8_t *s1, register utf8_t *s2)
{

	if(!s1 && !s2) return 0;
	if(!s1 && s2) return -1;
	if(s1 && !s2) return 1;
	return strcmp((char *)s1, (char *)s2);
}

int
edb_datum_diff(EdbContext ec, EdbDatum d1, EdbDatum d2)
{

	if(!d1 && !d2) return FALSE;
	if(!d1) return TRUE;
	if(!d2) return TRUE;
	if(datum_string_compare(d1->d_english, d2->d_english)) return TRUE;
	if(datum_string_compare(d1->d_japanese, d2->d_japanese)) return TRUE;
	if(datum_string_compare(d1->d_pronounce, d2->d_pronounce)) return TRUE;
	return FALSE;
}

static
int
edb_datum_check_TEXT2(register EdbContext ec, register EdbCTypeInfo ctype, register EdbDatum datum)
{
	register int	status = TRUE;

	if(STRisValid(datum->d_english)) {
		register int	n = 0;
		register utf8_t	*s = datum->d_english;

		while(s && (s=(utf8_t *)strchr((char *)s, ' '))) { ++n; ++s; }
		if(n==0) {
			edb_trace(ec, "datum_check: TEXT2 : there is no space.\n");
			status = FALSE;
		} else if(n>1) {
			edb_trace(ec, "datum_check: TEXT2 : there are too many spaces.\n");
			status = FALSE;
		}
	}
	return status;
}

static
int
edb_datum_check_YEAR(register EdbContext ec, register EdbCTypeInfo ctype, register EdbDatum datum)
{
	register int	status = TRUE;

	if(STRisValid(datum->d_english)) {
		if(!edb_text_isInteger(ec, datum->d_english)) {
			edb_trace(ec, "datum_check: YEAR : value is not integer.\n");
			status = FALSE;
		} else {
			register int	v = ATOI((char *)datum->d_english);

			if(v<1 || 9999<v) {
				edb_trace(ec, "datum_check: YEAR : value is out of range.\n");
				status = FALSE;
			}
		}
	}
	return status;
}

static
int
edb_datum_check_MINUTE(register EdbContext ec, register EdbCTypeInfo ctype, register EdbDatum datum)
{
	register int	status = TRUE;

	if(STRisValid(datum->d_english)) {
		if(!edb_text_isInteger(ec, datum->d_english)) {
			edb_trace(ec, "datum_check: MINUTE : value is not integer.\n");
			status = FALSE;
		} else {
			register int	v = ATOI((char *)datum->d_english);

			if(v<0) {
				edb_trace(ec, "datum_check: MINUTE : value is out of range.\n");
				status = FALSE;
			}
		}
	}
	return status;
}

static
int
edb_datum_check_EMAIL(register EdbContext ec, register EdbCTypeInfo ctype, register EdbDatum datum)
{
	register int	status = TRUE;

	if(STRisValid(datum->d_english) && !edb_text_isEmail(ec, datum->d_english)) {
		edb_trace(ec, "datum_check: EMAIL : value is not e-mail address.\n");
		status = FALSE;
	}
	return status;
}

static
int
edb_datum_check_BOOL(register EdbContext ec, register EdbCTypeInfo ctype, register EdbDatum datum)
{
	register int	status = TRUE;

	if(STRisValid(datum->d_english) && !edb_text_isBool(ec, datum->d_english)) {
		edb_trace(ec, "datum_check: BOOL : value is not boolean value.\n");
		status = FALSE;
	}
	return status;
}

static
int
edb_datum_check_INT(register EdbContext ec, register EdbCTypeInfo ctype, register EdbDatum datum)
{
	register int	status = TRUE;

	if(STRisValid(datum->d_english) && !edb_text_isInteger(ec, datum->d_english)) {
		edb_trace(ec, "datum_check: INTEGER : value is not integer.\n");
		status = FALSE;
	}
	return status;
}

static
int
edb_datum_check_REAL(register EdbContext ec, register EdbCTypeInfo ctype, register EdbDatum datum)
{
	register int	status = TRUE;

	if(STRisValid(datum->d_english) && !edb_text_isReal(ec, datum->d_english)) {
		edb_trace(ec, "datum_check: REAL : value is not real.\n");
		status = FALSE;
	}
	return status;
}

static
int
edb_datum_check_URL(register EdbContext ec, register EdbCTypeInfo ctype, register EdbDatum datum)
{
	register int	status = TRUE;

	if(STRisValid(datum->d_english) && !edb_text_isURL(ec, datum->d_english)) {
		edb_trace(ec, "datum_check: URL : value is not URL.\n");
		status = FALSE;
	}
	return status;
}

static
int
edb_datum_check_PHONE(register EdbContext ec, register EdbCTypeInfo ctype, register EdbDatum datum)
{
	register int	status = TRUE;

	if(STRisValid(datum->d_english) && !edb_text_isPhone(ec, datum->d_english)) {
		edb_trace(ec, "datum_check: PHONE : value is not phone number.\n");
		status = FALSE;
	}
	return status;
}

static
int
edb_datum_check_DATE(register EdbContext ec, register EdbCTypeInfo ctype, register EdbDatum datum)
{
	register int	status = TRUE;

	if(STRisValid(datum->d_english) && !edb_text_isDate(ec, datum->d_english)) {
		edb_trace(ec, "datum_check: DATE : value is not date (YYYYMMDD).\n");
		status = FALSE;
	}
	return status;
}

static
int
edb_datum_check_DATE2(register EdbContext ec, register EdbCTypeInfo ctype, register EdbDatum datum)
{
	register int	status = TRUE;

	if(STRisValid(datum->d_english)) {
		utf8_t	*s = datum->d_english;

		if(strlen((char *)s)!=(8*2+1)) {
			edb_trace(ec, "datum_check: DATE2 : value is not period (YYYYMMDD YYYYMMDD).\n");
			status = FALSE;
		} else {
			char	str[BUFSIZ];

			strncpy(str, (char *)s, 8);
			str[8]='\0';
			if(!edb_text_isDate(ec, (utf8_t *)str) || s[8]!=' ' || !edb_text_isDate(ec, (utf8_t *)&s[9])) {
				edb_trace(ec, "datum_check: DATE2 : value is not period (YYYYMMDD YYYYMMDD).\n");
				status = FALSE;
			}
		}
	}
	if(STRisValid(datum->d_japanese)) status = FALSE;
	if(STRisValid(datum->d_pronounce)) status = FALSE;
	return status;
}

static
int
edb_datum_check_IPV4CIDR(register EdbContext ec, register EdbCTypeInfo ctype, register EdbDatum datum)
{
	register int	status = TRUE;

	if(STRisValid(datum->d_english) && !edb_text_isIPv4CIDR(ec, datum->d_english)) {
		edb_trace(ec, "datum_check: IPV4CIDR : value is not CIDR.\n");
		status = FALSE;
	}
	if(STRisValid(datum->d_japanese)) status = FALSE;
	if(STRisValid(datum->d_pronounce)) status = FALSE;
	return status;
}

static
int
edb_datum_check_IPV4ADDR(register EdbContext ec, register EdbCTypeInfo ctype, register EdbDatum datum)
{
	register int	status = TRUE;

	if(STRisValid(datum->d_english) && !edb_text_isIPv4Address(ec, datum->d_english)) {
		edb_trace(ec, "datum_check: IPV4ADDR : value is not IP-Address.\n");
		status = FALSE;
	}
	if(STRisValid(datum->d_japanese)) status = FALSE;
	if(STRisValid(datum->d_pronounce)) status = FALSE;
	return status;
}

static
int
edb_datum_check_INETNAME(register EdbContext ec, register EdbCTypeInfo ctype, register EdbDatum datum)
{
	register int	status = TRUE;

	if(STRisValid(datum->d_english) && !edb_text_isInetName(ec, datum->d_english)) {
		edb_trace(ec, "datum_check: INETNAME : value is not inet name.\n");
		status = FALSE;
	}
	if(STRisValid(datum->d_japanese)) status = FALSE;
	if(STRisValid(datum->d_pronounce)) status = FALSE;
	return status;
}

static
int
edb_datum_check_INETFQDN(register EdbContext ec, register EdbCTypeInfo ctype, register EdbDatum datum)
{
	register int	status = TRUE;

	if(STRisValid(datum->d_english) && !edb_text_isInetFQDN(ec, datum->d_english)) {
		edb_trace(ec, "datum_check: INETFQDN : value is not inet FQDN.\n");
		status = FALSE;
	}
	if(STRisValid(datum->d_japanese)) status = FALSE;
	if(STRisValid(datum->d_pronounce)) status = FALSE;
	return status;
}

static
int
edb_datum_check_CERTIFICATE(register EdbContext ec, register EdbCTypeInfo ctype, register EdbDatum datum)
{
	register int	status = TRUE;

	if(STRisValid(datum->d_english) && !edb_text_isCertificate(ec, datum->d_english)) {
		edb_trace(ec, "datum_check: CERTIFICATE : value is not certificate.\n");
		status = FALSE;
	}
	if(STRisValid(datum->d_japanese)) status = FALSE;
	if(STRisValid(datum->d_pronounce)) status = FALSE;
	return status;
}

static
int
edb_datum_check_PASSPHRASE(register EdbContext ec, register EdbCTypeInfo ctype, register EdbDatum datum)
{
	register int	status = TRUE;
	register int	len, minl, maxl;

	if(STRisValid(datum->d_japanese)) return FALSE;
	if(STRisValid(datum->d_pronounce)) return FALSE;

	if(!STRisValid(datum->d_english)) return TRUE;
	if(ctype==NULL) return TRUE;
	if(STRisValid(ctype->cryptography)) {
		len = strlen(ctype->cryptography);
		if(datum->d_english[0] =='{'	/* } */
			&& !strncmp((char *)&datum->d_english[1], ctype->cryptography, len)
			&& !isalnum(datum->d_english[len+1]))
			return TRUE;
	}
	minl = 1;
	maxl = EDB_MAX_PASSPHRASE_LEN;
	if(STRisValid(ctype->v_minimum)) {
		len = ATOI((char *)ctype->v_minimum);
		minl = max(minl, len);
	}
	if(STRisValid(ctype->v_maximum)) {
		len = ATOI((char *)ctype->v_maximum);
		maxl = min(maxl, len);
	}
	len = strlen((char *)datum->d_english);
	if(len<minl) {
		edb_trace(ec, "datum_check: PASSPHRASE : passphrase is too short.\n");
		status = FALSE;
	}
	if(maxl<len) {
		edb_trace(ec, "datum_check: PASSPHRASE : passphrase is too long.\n");
		status = FALSE;
	}
	return status;
}

int
edb_datum_check(EdbContext ec, EdbCTypeInfo ctype, EdbDatum datum)
{
	EdbTypeInfo	typeinfo = ctype->info;
	int	status = TRUE;

	if(datum->d_eid!=0 && !EIDisValid(datum->d_eid)) status = FALSE;
	if(typeinfo->multilingual) {
		if(STRisValid(datum->d_english) && !edb_text_isEnglish(ec, datum->d_english)) status = FALSE;
		if(STRisValid(datum->d_japanese) && !edb_text_isJapanese(ec, datum->d_japanese)) status = FALSE;
		if(STRisValid(datum->d_pronounce) && !edb_text_isPronounce(ec, datum->d_pronounce)) status = FALSE;
	} else {
		if(STRisValid(datum->d_english) && !edb_text_isEnglish(ec, datum->d_english)) status = FALSE;
		if(STRisValid(datum->d_japanese)) status = FALSE;
		if(STRisValid(datum->d_pronounce)) status = FALSE;
	}
	if(typeinfo->check_datum && !(*typeinfo->check_datum)(ec, ctype, datum)) status = FALSE;
	return status;
}

static
utf8_t	*
edb_datum_text_conv(EdbContext ec, register utf8_t *src)
{
	ucs4_t	*s, *d;
	char	*str;

	s = UTF8toUCS4(src);
	d = edb_text_conv(ec, s);
	str = (char *)UCS4toUTF8(d);
	FREE(d);
	str_oneline(str);
	return (utf8_t *)strdup(str);
}

int
edb_datum_set(
	register EdbContext ec,
	register EdbDatum datum,
	register eid_t	eid,
	register utf8_t *en,
	register utf8_t *ja,
	register utf8_t *pr)
{
	register utf8_t *en0 = datum->d_english;
	register utf8_t *ja0 = datum->d_japanese;
	register utf8_t *pr0 = datum->d_pronounce;

	if(!datum) return FALSE;
	if(!ec->edb_datum_text_sharable) {
		datum->d_eid = eid;
		datum->d_english = STRisValid(en) ? (utf8_t *)strdup((char *)en) : edb_datum_text_null_string;
		datum->d_japanese = STRisValid(ja) ? (utf8_t *)strdup((char *)ja) : edb_datum_text_null_string;
		datum->d_pronounce = STRisValid(pr) ? (utf8_t *)strdup((char *)pr) : edb_datum_text_null_string;
		if(datum->english_shared) {
			edb_datum_text_unuse(ec, en0);
		} else {
			DatumTextFree(ec, en0);
		}
		if(datum->japanese_shared) {
			edb_datum_text_unuse(ec, ja0);
		} else {
			DatumTextFree(ec, ja0);
		}
		if(datum->pronounce_shared) {
			edb_datum_text_unuse(ec, pr0);
		} else {
			DatumTextFree(ec, pr0);
		}
		datum->english_shared = datum->japanese_shared = datum->pronounce_shared = FALSE;
	} else {
		datum->d_eid = eid;
		datum->d_english = edb_datum_text_use(ec, en);
		datum->d_japanese = edb_datum_text_use(ec, ja);
		datum->d_pronounce = edb_datum_text_use(ec, pr);
		if(datum->english_shared) {
			edb_datum_text_unuse(ec, en0);
		} else {
			DatumTextFree(ec, en0);
		}
		if(datum->japanese_shared) {
			edb_datum_text_unuse(ec, ja0);
		} else {
			DatumTextFree(ec, ja0);
		}
		if(datum->pronounce_shared) {
			edb_datum_text_unuse(ec, pr0);
		} else {
			DatumTextFree(ec, pr0);
		}
		datum->english_shared = datum->japanese_shared = datum->pronounce_shared = TRUE;
	}
	return TRUE;
}

EdbDatum
edb_datum_parse_xml_node(EdbContext ec, void *super, EdbCTypeInfo ctype, xmlDocPtr doc, xmlNodePtr node)
{
	EdbTypeInfo	typeinfo = ctype->info;
	EdbDatum	datum;
	utf8_t	*s;

	datum = edb_datum_new(ec, super);
	DatumRead(datum) = xml_get_element_attr_permission(ec, doc, node, (xmlChar *)XML_AN_READ, EDB_AUTH_INHERIT, NULL);
	// for nothing
	if(super!=NULL && ((EdbTupleColumn)super)->ci->canbeNothing) {
		utf8_t *s_assert = xml_get_element_attr(ec, doc, node, (xmlChar *)XML_AN_ASSERT, NULL);
		if(STRisValid(s_assert)) {
			if(!strcmp((char *)s_assert, XML_AV_ASSERT_NOTHING)) ((EdbTupleColumn)super)->nothing = TRUE;
			FREE(s_assert);
		}
	}
	DatumDateFrom(datum) = xml_get_element_attr_integer(ec, doc, node, (xmlChar *)XML_AN_DATE_FROM, EDB_DATE_MIN, NULL);
	DatumDateTo(datum) = xml_get_element_attr_integer(ec, doc, node, (xmlChar *)XML_AN_DATE_TO, EDB_DATE_MAX, NULL);
	switch(typeinfo->typeID) {
	case	EDB_TYPE_NAME:
	case	EDB_TYPE_SENTENCE:
	case	EDB_TYPE_PARAGRAPH:
		datum->d_eid = xml_get_element_attr_integer(ec, doc, node, (xmlChar *)XML_AN_MAPTO, 0, NULL);
		datum->d_english = (utf8_t *)xml_get_element_text(ec, doc, node, node->name, (xmlChar *)XML_EN_ENGLISH, NULL);
		datum->d_japanese = (utf8_t *)xml_get_element_text(ec, doc, node, node->name, (xmlChar *)XML_EN_JAPANESE, NULL);
		datum->d_pronounce = (utf8_t *)xml_get_element_text(ec, doc, node, node->name, (xmlChar *)XML_EN_PRONOUNCE, NULL);
		break;
	case	EDB_TYPE_EID:
		datum->d_eid = xml_get_element_attr_integer(ec, doc, node, (xmlChar *)XML_AN_MAPTO, 0, NULL);
		if(EIDisValid(datum->d_eid)) {
			datum->d_english = xml_get_element_text(ec, doc, node, node->name, (xmlChar *)XML_EN_ENGLISH, NULL);
			datum->d_japanese = xml_get_element_text(ec, doc, node, node->name, (xmlChar *)XML_EN_JAPANESE, NULL);
			datum->d_pronounce = xml_get_element_text(ec, doc, node, node->name, (xmlChar *)XML_EN_PRONOUNCE, NULL);
		}
		break;
	case	EDB_TYPE_TEXT:
	case	EDB_TYPE_INT:
	case	EDB_TYPE_REAL:
	case	EDB_TYPE_PASSPHRASE:
	case	EDB_TYPE_EMAIL:
	case	EDB_TYPE_URL:
	case	EDB_TYPE_PHONE:
	case	EDB_TYPE_DATE:
	case EDB_TYPE_DATE2:
	case	EDB_TYPE_MONETARY:
	case	EDB_TYPE_BOOL:
	case	EDB_TYPE_PICTURE:
	default:
		datum->d_english = xml_get_element_text(ec, doc, node, node->name, XML_EN_ENGLISH, NULL);
		break;
	}

#ifdef EDB_READER
#else /* !EDB_READER */
	if(typeinfo->modify) {
		(*typeinfo->modify)(ec, ctype, datum);
	} else {
		if(STRisValid(datum->d_english)) {
			datum->d_english = edb_datum_text_conv(ec, s=datum->d_english);
			DatumTextFree(ec, s);
		}
		if(STRisValid(datum->d_japanese)) {
			datum->d_japanese = edb_datum_text_conv(ec, s=datum->d_japanese);
			DatumTextFree(ec, s);
		}
		if(STRisValid(datum->d_pronounce)) {
			datum->d_pronounce = edb_datum_text_conv(ec, s=datum->d_pronounce);
			DatumTextFree(ec, s);
		}
	}
#endif /* EDB_READER */

	DatumAvail(datum) = TRUE;
	if(ec->edb_datum_text_sharable) {
		edb_datum_set(ec, datum, DatumEID(datum), DatumEnglish(datum), DatumJapanese(datum), DatumPronounce(datum));
	}

	return datum;
}

int
edb_datum_get_DATE2(
	register EdbContext	ec,
	register EdbDatum	datum,
	register int32_t		*date1,
	register int32_t		*date2)
{
	int	d1, d2;

	if(!datum || !STRisValid(datum->d_english)) {
		if(date1) *date1 = 0;
		if(date2) *date2 = 0;
		return FALSE;
	}
	d1 = d2 = 0;
	if(sscanf((char *)datum->d_english, "%d %d", &d1, &d2)==1) d2 = d1;
	if(date1) *date1 = d1;
	if(date2) *date2 = d2;
	return TRUE;
}

utf8_t	*
edb_datum_make_pgsql_value(EdbContext ec, EdbCTypeInfo ctype, EdbDatum datum)
{
	EdbTypeInfo	typeinfo = ctype->info;
	utf8_t	*value = NULL;

	if(typeinfo->typeID==EDB_TYPE_DATE2) {
		int	d1 = 0, d2 = 0;

		if(STRisValid(datum->d_english) && sscanf((char *)datum->d_english, "%d %d", &d1, &d2)==1) d2 = d1;
		asprintf((char **)&value, "%08d%08d", d1, d2);
	} else if(typeinfo->typeID==EDB_TYPE_PICTURE) {
		/* NOTICE: if type is PICTURE, we don't store picture data on RDB. */
		value = (utf8_t *)strdup("PICTURE");
	} else if(typeinfo->typeID==EDB_TYPE_CERTIFICATE) {
		/* NOTICE: if type is CERTIFICATE, we don't store certificate data on RDB. */
		value = (utf8_t *)strdup("CERTIFICATE");
	} else if(typeinfo->typeID==EDB_TYPE_PASSPHRASE) {
		/* NOTICE: if type is PASSPHRASE, we don't store passphrase data on RDB. */
		value = (utf8_t *)strdup("PASSPHRASE");
	} else if(typeinfo->multilingual) {
		value = edb_mtext_concatenate(ec, datum->d_eid, datum->d_english, datum->d_japanese, datum->d_pronounce);
	} else {
		value = (utf8_t *)strdup_null((char *)datum->d_english);
	}
	return value;
}

#if	defined(TEXT_VARIANT_CONVERSION)
utf8_t	*
edb_datum_make_pgsql_variant_value(EdbContext ec, EdbCTypeInfo ctype, EdbDatum datum)
{
	EdbTypeInfo	typeinfo = ctype->info;
	utf8_t	*value = NULL;

	if(typeinfo->typeID==EDB_TYPE_DATE2) {
		int	d1 = 0, d2 = 0;

		if(STRisValid(datum->d_english) && sscanf((char *)datum->d_english, "%d %d", &d1, &d2)==1) d2 = d1;
		asprintf((char **)&value, "%08d%08d", d1, d2);
	} else if(typeinfo->typeID==EDB_TYPE_PICTURE) {
		/* NOTICE: if type is PICTURE, we don't store picture data on RDB. */
		value = (utf8_t *)strdup("PICTURE");
	} else if(typeinfo->typeID==EDB_TYPE_CERTIFICATE) {
		/* NOTICE: if type is CERTIFICATE, we don't store certificate data on RDB. */
		value = (utf8_t *)strdup("CERTIFICATE");
	} else if(typeinfo->typeID==EDB_TYPE_PASSPHRASE) {
		/* NOTICE: if type is PASSPHRASE, we don't store passphrase data on RDB. */
		value = (utf8_t *)strdup("PASSPHRASE");
	} else if(typeinfo->multilingual) {
		value = edb_mtext_concatenate_variant(ec, datum->d_eid, datum->d_english, datum->d_japanese, datum->d_pronounce);
	} else {
		value = edb_text_variant_conversion(datum->d_english, TRUE);
	}
	return value;
}
#endif	/* TEXT_VARIANT_CONVERSION */

utf8_t	*
edb_datum_make_xml(EdbContext ec, EdbCTypeInfo ctype, EdbDatum datum, utf8_t **m_attr, utf8_t **p_attr, int mode)
{
	EdbTypeInfo	typeinfo = ctype->info;
	utf8_t	*xml, *english, *japanese, *pronounce, *s;
	Link	l_attr = NULL;

	*m_attr = NULL;
	*p_attr = NULL;
	xml = NULL;
	if(datum==NULL) return NULL;

	/* attribute (mapto, read) */
	if(EIDisValid(datum->d_eid)) {
		asprintf((char **)m_attr, "%s=\"%d\"", XML_AN_MAPTO, datum->d_eid);
	}
	if(EDB_DATE_MIN < DatumDateFrom(datum)) {
		asprintf((char **)&s, "%s=\"%08d\"", XML_AN_DATE_FROM, DatumDateFrom(datum));
		LinkAppend(l_attr, link_atom_new(s, free));
	}
	if(DatumDateTo(datum) < EDB_DATE_MAX) {
		asprintf((char **)&s, "%s=\"%08d\"", XML_AN_DATE_TO, DatumDateTo(datum));
		LinkAppend(l_attr, link_atom_new(s, free));
	}
	if(DatumRead(datum)!=EDB_AUTH_INHERIT) {
		asprintf((char **)&s, "%s=\"%s\"", XML_AN_READ, edb_auth_xmlstring(ec, DatumRead(datum), FALSE));
		LinkAppend(l_attr, link_atom_new(s, free));
	}
	if(l_attr) {
		*p_attr = (utf8_t *)str_concatenate(l_attr, ' ');
		EdbLinkFree(l_attr);
	}
	if(EIDisValid(datum->d_eid) && !mode) return NULL;

	/* text */
	english = (utf8_t *)strdup("");
	japanese = (utf8_t *)strdup("");
	pronounce = (utf8_t *)strdup("");
	if(STRisValid(datum->d_english)) {
		FREE(english);
		english = xml_make_element_text(ec, (xmlChar *)XML_EN_ENGLISH, datum->d_english);
	}

	if(typeinfo->multilingual) {
		if(STRisValid(datum->d_japanese)) {
			FREE(japanese);
			japanese = xml_make_element_text(ec, (xmlChar *)XML_EN_JAPANESE, datum->d_japanese);
		}
		if(STRisValid(datum->d_pronounce)) {
			FREE(pronounce);
			pronounce = xml_make_element_text(ec, (xmlChar *)XML_EN_PRONOUNCE, datum->d_pronounce);
		}
	}
	asprintf((char **)&xml, "%s%s%s", english, japanese, pronounce);
	FREE(english);
	FREE(japanese);
	FREE(pronounce);
	if(xml[0]=='\0') {
		FREE(xml);
		return NULL;
	}
	return xml;
}

struct _PARSE_TEXT {
	EdbContext	ec;
	Link			l_eid;
};

static	int	edb_datum_get_reference_list_parse_text		__P((struct _PARSE_TEXT *, utf8_t *));

static
int
edb_datum_get_reference_list_parse_text_callback(
	register void *context,
	register utf8_t c,
	register utf8_t *s)
{
	register struct _PARSE_TEXT	*pt = context;
//	register EdbContext	ec = pt->ec;
	register eid_t	eid;

	switch(c) {
	case 'P':
		eid = ATOI((char *)s);
		if(EIDisValid(eid)) LinkAppend(pt->l_eid, link_atom_new_eid2s(eid));
		break;
	case '_':
	case '^':
	case 'A':
	case 'M':
	case 'N':
	case 'D':
	case 'I':
	default:
		if(STRisValid(s) && strchr((char *)s, '\\')) edb_datum_get_reference_list_parse_text(pt, s);
		break;
	}
	return TRUE;
}

static
int
edb_datum_get_reference_list_parse_text(
	register struct _PARSE_TEXT *pt,
	utf8_t *s)
{
	register int	status = TRUE;

	while((s=(utf8_t *)strchr((char *)s, '\\'))) {
		++s;
		if(!edb_text_parse(&s, edb_datum_get_reference_list_parse_text_callback, pt)) status = FALSE;
	}
	return status;
}

Link
edb_datum_get_reference_list(EdbContext ec, EdbDatum datum)
{
	register Link	l_eid = NULL;
	register eid_t	eid;

	eid = datum->d_eid;
	if(EIDisValid(eid)) {
		LinkAppend(l_eid, link_atom_new_eid2s(eid));
	} else {
		struct _PARSE_TEXT	pt;

		pt.ec = ec;
		pt.l_eid = NULL;
		if(STRisValid(datum->d_english) && strchr((char *)datum->d_english, '\\')) {
			edb_datum_get_reference_list_parse_text(&pt, datum->d_english);
		}
		if(STRisValid(datum->d_japanese) && strchr((char *)datum->d_japanese, '\\')) {
			edb_datum_get_reference_list_parse_text(&pt, datum->d_japanese);
		}
		if(pt.l_eid) LinkAppend(l_eid, pt.l_eid);
	}
	return l_eid;
}

int
edb_datum_check_reference(EdbContext ec, EdbDatum datum, eid_t eid)
{

	if(!EIDisValid(eid)) return FALSE;
	if(datum==NULL) return FALSE;
	if(datum->d_eid == eid) {
		return TRUE;
	} else {
		register Link	l_eid;
		Link	l;

		l_eid = edb_datum_get_reference_list(ec, datum);
		ForLink(l, l_eid) {
			if(eid==ATOI((char *)l->d)) break;
		}
		EdbLinkFree(l_eid);
		return (l!=NULL);
	}
}

int
edb_datum_check_text(EdbContext ec, EdbDatum datum)
{
	register eid_t	eid;
	register int	status = TRUE;

	eid = datum->d_eid;
	if(EIDisValid(eid)) {
		return TRUE;
	} else {
		struct _PARSE_TEXT	pt;

		pt.ec = ec;
		pt.l_eid = NULL;
		if(STRisValid(datum->d_english) && strchr((char *)datum->d_english, '\\')) {
			if(!edb_datum_get_reference_list_parse_text(&pt, datum->d_english)) status = FALSE;
		}
		if(STRisValid(datum->d_japanese) && strchr((char *)datum->d_japanese, '\\')) {
			if(!edb_datum_get_reference_list_parse_text(&pt, datum->d_japanese)) status = FALSE;
		}
		EdbLinkFree(pt.l_eid);
	}
	return status;
}

#define	DATE_INNER(d, f, t)		((f)<=(d) && (d)<=(t))
#define	DATE_FUTURE(d, f, t)	((d)<(f) && (d)<(t))
#define	DATE_PAST(d, f, t)		((f)<(d) && (t)<(d))
#define	PERIOD_OVERLAP(f1, t1, f2, t2)	(((f1) <= (t2)) && ((f2) <= (t1)))

static
int
edb_datum_date_in_period(
	register EdbContext	ec,
	register EdbDatum	datum,
	register int		d)
{
	register EdbTupleInfo	tuple = NULL;

	if(datum==NULL) return FALSE;
	while(datum) {
		register EdbTupleColumn	tc;

		if(!DATE_INNER(d, DatumDateFrom(datum), DatumDateTo(datum))) return FALSE;
		if((tc = DatumSuper(datum))) {
			if(tc->ci->super) {
				datum =  tc->super;
				tuple = NULL;
			} else {
				datum = NULL;
				tuple =  tc->super;
			}
		} else {
			datum = NULL;
		}
	}
	if(tuple && !DATE_INNER(d, TupleDateFrom(tuple), TupleDateTo(tuple))) return FALSE;
	return TRUE;
}

int
edb_datum_period_overlapped(EdbContext ec, EdbDatum datum1, EdbDatum datum2)
{

	if(datum1==NULL || datum2==NULL) return FALSE;

	if(DatumDateFrom(datum1) > DatumDateTo(datum1)) return FALSE;
	if(DatumDateFrom(datum2) > DatumDateTo(datum2)) return FALSE;
	if(PERIOD_OVERLAP(DatumDateFrom(datum1), DatumDateTo(datum1), DatumDateFrom(datum2), DatumDateTo(datum2) ) ) return TRUE;
	return FALSE;
}

int
edb_datum_period_compare_for_sort_ascend(LinkObj d1, LinkObj d2)
{
	register EdbDatum	datum1 = d1->d;
	register EdbDatum	datum2 = d2->d;

	if(DatumDateFrom(datum1)!=DatumDateFrom(datum2)) {
		return DatumDateFrom(datum1) - DatumDateFrom(datum2);
	} else if(DatumDateTo(datum1)!=DatumDateTo(datum2)) {
		return DatumDateTo(datum1) - DatumDateTo(datum2);
	} else {
		return d1->idx - d2->idx;
	}
}

int
edb_datum_period_compare_for_sort_descend(LinkObj d1, LinkObj d2)
{
	register EdbDatum	datum1 = d1->d;
	register EdbDatum	datum2 = d2->d;

	if(DatumDateFrom(datum1)!=DatumDateFrom(datum2)) {
		return DatumDateFrom(datum2) - DatumDateFrom(datum1);
	} else if(DatumDateTo(datum1)!=DatumDateTo(datum2)) {
		return DatumDateTo(datum2) - DatumDateTo(datum1);
	} else {
		return d1->idx - d2->idx;
	}
}

int
edb_datum_is_empty(EdbContext ec, EdbDatum datum)
{

	if(datum==NULL) return TRUE;
	if(EIDisValid(DatumEID(datum))) return FALSE;
	if(STRisValid(datum->d_english)) return FALSE;
	if(STRisValid(datum->d_japanese)) return FALSE;
	if(STRisValid(datum->d_pronounce)) return FALSE;
	return TRUE;
}

int
edb_datum_at_present(EdbContext ec, EdbDatum datum)
{

	return edb_datum_date_in_period(ec, datum, ec->present_date);
}

static
int
edb_datum_date_in_future(
	register EdbContext	ec,
	register EdbDatum	datum,
	register int		d)
{
	register EdbTupleInfo	tuple = NULL;

	if(datum==NULL) return FALSE;
	while(datum) {
		register EdbTupleColumn	tc;

		if(DATE_PAST(d, DatumDateFrom(datum), DatumDateTo(datum))) return FALSE;
		if((tc = DatumSuper(datum))) {
			if(tc->ci->super) {
				datum =  tc->super;
				tuple = NULL;
			} else {
				datum = NULL;
				tuple =  tc->super;
			}
		} else {
			datum = NULL;
		}
	}
	if(tuple && DATE_PAST(d, TupleDateFrom(tuple), TupleDateTo(tuple))) return FALSE;
	return TRUE;
}

int
edb_datum_at_future(EdbContext ec, EdbDatum datum)
{

	if(edb_datum_at_present(ec, datum)) return FALSE;
	return edb_datum_date_in_future(ec, datum, ec->present_date);
}

Caption
edb_datum_make_caption(EdbContext ec, EdbDatum datum, int mode)
{
	EdbTupleColumn tc = (EdbTupleColumn)DatumSuper(datum);

	if(DatumEIDisValid(datum)) {
		EdbTupleInfo	rtuple;

		rtuple = edb_tuple_get(ec, DatumEID(datum), 0);
		if(rtuple) {
			Caption	acap;
			Link	lm, l_mci = NULL;

			ForLink(lm, tc->ci->l_maplookup) {
				EdbMapLookup	map = lm->d;

				if(TABLE_EID(TupleTable(rtuple))==edb_maplookup_table_eid(ec, map)) {
					l_mci = map->l_caption_ci;
					break;
				}
			}

			return edb_tuple_make_caption(ec, rtuple, mode);
		}
	}

	Caption cap = edb_caption_new(ec);
	if(cap==NULL) return NULL;

	if(tc && tc->ci->typeInfo->typeID == EDB_TYPE_PICTURE) {
		asprintf((char **)&cap->c_main, "\\P{%d}", DatumEID(datum));
	} else if(tc && tc->ci->typeInfo->typeID == EDB_TYPE_TEXT) {
		char	*capstr = "", *s;

		if(DatumEnglishIsValid(datum)) {
			if(ec->japanese) {
				asprintf(&s, "%s%s%s",
					STRisVALID(tc->ci->prefix.japanese) ? tc->ci->prefix.japanese:(utf8_t *)"",
					DatumEnglish(datum),
					STRisVALID(tc->ci->suffix.japanese) ? tc->ci->suffix.japanese:(utf8_t *)"");
				capstr = STRDUP(s);
				FREE(s);
			} else {
				asprintf(&s, "%s%s%s",
					STRisVALID(tc->ci->prefix.english) ? tc->ci->prefix.english:(utf8_t *)"",
					DatumEnglish(datum),
					STRisVALID(tc->ci->suffix.english) ? tc->ci->suffix.english:(utf8_t *)"");
				capstr = STRDUP(s);
				FREE(s);
			}
		}
		cap->c_main = (utf8_t *)strdup_null(capstr);
		cap->c_main_english = (utf8_t *)strdup_null((char *)DatumEnglish(datum));
		cap->c_main_japanese = (utf8_t *)strdup_null((char *)DatumEnglish(datum));
		cap->c_main_pronounce = (utf8_t *)strdup_null("");
	} else {
		char	*capstr = "", *s;

		if(ec->japanese) {
			if(DatumJapaneseIsValid(datum)) {
				asprintf(&s, "%s%s%s",
					tc && STRisVALID(tc->ci->prefix.japanese) ? tc->ci->prefix.japanese:(utf8_t *)"",
					DatumJapanese(datum),
					tc && STRisVALID(tc->ci->suffix.japanese) ? tc->ci->suffix.japanese:(utf8_t *)"");
				capstr = STRDUP(s);
				FREE(s);
			} else if(DatumEnglishIsValid(datum)) {
				asprintf(&s, "%s%s%s",
					tc && STRisVALID(tc->ci->prefix.english) ? tc->ci->prefix.english:(utf8_t *)"",
					DatumEnglish(datum),
					tc && STRisVALID(tc->ci->suffix.english) ? tc->ci->suffix.english:(utf8_t *)"");
				capstr = STRDUP(s);
				FREE(s);
			}
		} else {
			if(DatumEnglishIsValid(datum)) {
				asprintf(&s, "%s%s%s",
					tc && STRisVALID(tc->ci->prefix.english) ? tc->ci->prefix.english:(utf8_t *)"",
					DatumEnglish(datum),
					tc && STRisVALID(tc->ci->suffix.english) ? tc->ci->suffix.english:(utf8_t *)"");
				capstr = STRDUP(s);
				FREE(s);
			} else if(DatumJapaneseIsValid(datum)) {
				asprintf(&s, "%s%s%s",
					tc && STRisVALID(tc->ci->prefix.japanese) ? tc->ci->prefix.japanese:(utf8_t *)"",
					DatumJapanese(datum),
					tc && STRisVALID(tc->ci->suffix.japanese) ? tc->ci->suffix.japanese:(utf8_t *)"");
				capstr = STRDUP(s);
				FREE(s);
			}
		}
		cap->c_main = (utf8_t *)strdup_null(capstr);
		cap->c_main_english = (utf8_t *)strdup_null((char *)DatumEnglish(datum));
		cap->c_main_japanese = (utf8_t *)strdup_null((char *)DatumJapanese(datum));
		cap->c_main_pronounce = (utf8_t *)strdup_null((char *)DatumPronounce(datum));
	}
	return cap;
}

