/*
	$Id: doc.c,v 1.40 2011/04/18 09:45:28 alex Exp $
	doc.c--- document
*/

#ifndef	lint
static char *rcsid __attribute__((unused)) = "$Id: doc.c,v 1.40 2011/04/18 09:45:28 alex Exp $";
#endif

#include	<stdio.h>
#include	<stdlib.h>
#include	<stdarg.h>
#include	<unistd.h>
#include	<sys/types.h>
#include	<string.h>
#include	<signal.h>
#include	<time.h>
#include	"edb/edb.h"

static	int	doc_ucs4_putc		__P((Doc, ucs4_t));
static	int	doc_ucs4_puts		__P((Doc, ucs4_t *));
static	int	doc_text_ucs4_putc	__P((Doc, ucs4_t));
static	int	doc_text_ucs4_puts	__P((Doc, ucs4_t *));

#define	MATH_NONE	0
#define	MATH_IMPL	1
#define	MATH_EXPL	2

/*
	doc_open	- allocate document and fopen
	doc_sopen	- allocate document without file (buffered string)
	doc_close - free docuemnt and fclose
	doc_fpopen - allocate docuemtn with specified fp
	doc_tmpopen - allocate docuemtn with tmpopen()
	doc_fpclose - free document without fclose
*/
Doc
doc_open(int dml, char *path, char *mode)
{
	Doc	doc;

	doc = (Doc)malloc(sizeof(struct DOC));
	if(doc==NULL) return NULL;
	bzero(doc, sizeof(struct DOC));
	doc->dml = dml;
	doc->fp = fopen(path, mode);
	if(doc->fp == NULL) {
		free(doc);
		return NULL;
	}
	doc->inside = FALSE;
	doc->math = MATH_NONE;
	doc->lastc = '\0';
	return doc;
}

Doc
doc_sopen(int dml)
{
	Doc	doc;

	doc = (Doc)malloc(sizeof(struct DOC));
	if(doc==NULL) return NULL;
	bzero(doc, sizeof(struct DOC));
	doc->dml = dml;
	doc->buf_size = BUFSIZ;
	doc->buf = (ucs4_t *)malloc(sizeof(ucs4_t)*doc->buf_size);
	doc->buf_cp = 0;
	doc->inside = FALSE;
	doc->math = MATH_NONE;
	doc->lastc = '\0';
	return doc;
}

int
doc_close(Doc doc)
{
	int	status;

	if(doc==NULL) return EOF;
	if(doc->fp) {
		status = fclose(doc->fp);
	} else {
		status = 0;
	}
	FREE(doc->buf);
	free(doc);
	return status;
}

Doc
doc_fpopen(int dml, FILE *fp)
{
	Doc	doc;

	doc = (Doc)malloc(sizeof(struct DOC));
	if(doc==NULL) return NULL;
	bzero(doc, sizeof(struct DOC));
	doc->dml = dml;
	doc->fp = fp;
	doc->inside = FALSE;
	doc->math = MATH_NONE;
	doc->lastc = '\0';
	return doc;
}

Doc
doc_tmpopen(int dml)
{
	FILE	*fp;

	fp = tmpfile();
	if(fp==NULL) return NULL;
	return doc_fpopen(dml, fp);
}

void
doc_fpclose(Doc doc)
{

	FREE(doc->buf);
	FREE(doc);
}

int
doc_flush(Doc doc)
{

	if(doc->fp) {
		return fflush(doc->fp);
	} else {
		return 0;
	}
}

static
int
doc_ucs4_sputc(Doc doc, ucs4_t c)
{
	if(doc->buf==NULL) return EOF;

	if(!(doc->buf_cp+1<doc->buf_size)) {
		doc->buf_size += BUFSIZ;
		doc->buf = (ucs4_t *)realloc(doc->buf, sizeof(ucs4_t)*doc->buf_size);
	}
	doc->buf[doc->buf_cp++] = c;
	return 1;
}

ucs4_t *
doc_ucs4_sgets(Doc doc)
{
	ucs4_t	*u4;

	if(doc->buf==NULL) return NULL;

	u4 = (ucs4_t *)malloc(sizeof(ucs4_t)*(doc->buf_cp+1));
	if(u4==NULL) return NULL;
	bcopy(doc->buf, u4, sizeof(ucs4_t)*doc->buf_cp);
	u4[doc->buf_cp] = UCS4_NULL;
	doc->buf_cp = 0;
	return u4;
}

char *
doc_sgets(Doc doc)
{
	ucs4_t	*u4;
	utf8_t	*u8;

	if(doc->buf==NULL) return NULL;

	u4 = doc_ucs4_sgets(doc);
	if(u4==NULL) return NULL;

	u8 = UCS4toUTF8(u4);
	FREE(u4);
	return strdup((char *)u8);
}

int
doc_sflush(Doc doc)
{

	doc->buf_cp = 0;
	return 0;
}

static
void
doc_math_check(Doc doc)
{

	if(doc->math!=MATH_NONE) {
		switch(doc->dml) {
		case DocML_XML:
			break;
		case DocML_HTML:
#if 1
			if(doc->fp) {
				fputs("</var>", doc->fp);
			} else {
				char *s = "</var>";

				while(*s) doc_ucs4_sputc(doc, *s++);
			}
#else
			fputs("</var>", doc->fp);
#endif
			doc->lastc = '>';
			break;
		case DocML_TeX:
#if 1
			if(doc->fp) {
				fputs("$", doc->fp);
			} else {
				doc_ucs4_sputc(doc, '$');
			}
#else
			fputs("$", doc->fp);
#endif
			doc->lastc = '$';
		case DocML_PLAIN:
		default:
			break;
		}
		doc->math = MATH_NONE;
	}
}

int
doc_printf(Doc doc, char *fmt, ...)
{
	va_list	ap;
	int	n;
	char	*str;

	va_start(ap, fmt);
	n = vasprintf(&str, fmt, ap);
	va_end(ap);
	doc_puts(doc, str);
	FREE(str);
	return n;
}

static
int
doc_ucs4_puts(Doc doc, ucs4_t *s)
{
	ucs4_t	*u = s;

	while(*u) doc_ucs4_putc(doc, *u++);
	return 0;
}

int
doc_puts(Doc doc, char *s)
{
	ucs4_t	*u = UTF8toUCS4((utf8_t *)s);

	while(*u) doc_ucs4_putc(doc, *u++);
	return 0;
}

#if 0
int
doc_punctuation_puts(Doc doc, char *s)
{

	if(doc->dml!=DocML_TeX) doc_puts(doc, s);
	return 0;
}
#endif

int
doc_prefix_set(Doc doc, char *s)
{

	doc_prefix_reset(doc);
	doc->prefix = strdup(s);
	return 0;
}

int
doc_prefix_reset(Doc doc)
{

	FREE(doc->prefix);
	doc->prefix = NULL;
	return 0;
}

#define	UCS4_BITS(c)	(0x80 | ((c) & 0x3f))
#define	UCS4_PUT(c)	if(fputc((c), doc->fp) == EOF) return EOF
#define	UCS4_PUTBITS(c)	UCS4_PUT(UCS4_BITS(c))

static
int
doc_ucs4_putc(Doc doc, ucs4_t	c)
{

	if(STRisVALID(doc->prefix)) {
		register char *s = STRDUP(doc->prefix);

		doc_prefix_reset(doc);
		doc_puts(doc, s);
	}

	if(doc->dml==DocML_TeX) {
		if((doc->lastc=='$' && c=='$') 
			|| (doc->lastc=='?' && c=='`')
			|| (doc->lastc=='!' && c=='`')) {
			if(doc->fp) {
				fputc(' ', doc->fp);
			} else {
				doc_ucs4_sputc(doc, ' ');
			}
		}
	}
	doc->lastc = c;

	if(doc->fp==NULL) {
		return doc_ucs4_sputc(doc, c);
	}

	if(c < 0x80) {
		UCS4_PUT(c);
	} else if(c < 0x800) {
		UCS4_PUT(0xc0 | (c >>  6));
		UCS4_PUTBITS(c);
	} else if(c < 0x10000) {
		UCS4_PUT(0xe0 | (c >> 12));
		UCS4_PUTBITS(c >>  6);
		UCS4_PUTBITS(c);
	} else if(c < 0x200000) {
		UCS4_PUT(0xf0 | (c >> 18));
		UCS4_PUTBITS(c >> 12);
		UCS4_PUTBITS(c >>  6);
		UCS4_PUTBITS(c);
	} else if(c < 0x400000) {
		UCS4_PUT(0xf8 | (c >> 24));
		UCS4_PUTBITS(c >> 18);
		UCS4_PUTBITS(c >> 12);
		UCS4_PUTBITS(c >>  6);
		UCS4_PUTBITS(c);
	} else if(c < 0x80000000) {
		UCS4_PUT(0xfc | (c >> 30));
		UCS4_PUTBITS(c >> 24);
		UCS4_PUTBITS(c >> 18);
		UCS4_PUTBITS(c >> 12);
		UCS4_PUTBITS(c >>  6);
		UCS4_PUTBITS(c);
	} else {	/* Error, cannot convert */
		return 1;
	}

	return 0;
}

int
doc_putc(Doc doc, int c)
{

	return doc_ucs4_putc(doc, c);
}

int
doc_comment_printf(Doc doc, char *fmt, ...)
{
	va_list	ap;
	int	n;
	char	*str;

	va_start(ap, fmt);
	n = vasprintf(&str, fmt, ap);
	va_end(ap);
	switch(doc->dml) {
	case DocML_XML:
		doc_puts(doc, "! ");
		doc_puts(doc, str);
		doc_puts(doc, "\n");
		break;
	case DocML_HTML:
		doc_text_puts3(doc, "<!-- ", str, " -->\n");
		break;
	case DocML_TeX:
		doc_puts(doc, "% ");
		doc_puts(doc, str);
		doc_puts(doc, "\n");
		break;
	default:
		doc_puts(doc, "! ");
		doc_puts(doc, str);
		doc_puts(doc, "\n");
		break;
	}
	FREE(str);
	return n;
}

static
int
doc_text_ucs4_parse(void *context, ucs4_t sc, ucs4_t *s)
{
	Doc	doc = (Doc)context;
	ucs4_t	*d;

	if(sc=='{' || sc=='}' || sc=='\\') {
		doc_text_ucs4_putc(doc, sc);
		return TRUE;
	}
	if(sc=='L') {	/* line feed */
		switch(doc->dml) {
		case DocML_XML:
			doc_puts(doc, "\\L");
			break;
		case DocML_HTML:
			doc_puts(doc, "<br />");
			break;
		case DocML_TeX:
			doc_puts(doc, "\\\\");
			break;
		case DocML_PLAIN:
		default:
			doc_puts(doc, "\n");
			break;
		}
		return TRUE;
	}
	if(!UCS4_STRisVALID(s)) return FALSE;
	if(sc == 'A') {
		switch(doc->dml) {
		case DocML_XML:
			doc_puts(doc, "\\A{");
			doc_text_ucs4_puts(doc, s);
			doc_puts(doc, "}");
			break;
		case DocML_HTML:
			doc_puts(doc, "<span style=\"color:red\">");
			doc_text_ucs4_puts(doc, s);
			doc_puts(doc, "</span>");
			break;
		case DocML_TeX:
			doc_puts(doc, "\\EdbAttention{");
			doc_text_ucs4_puts(doc, s);
			doc_puts(doc, "}");
			break;
		case DocML_PLAIN:
		default:
			doc_text_ucs4_puts(doc, s);
			break;
		}
	} else if(sc == 'C') {
		switch(doc->dml) {
		case DocML_XML:
			doc_puts(doc, "\\C{");
			d = text_ucs4_codename_to_text(s);
			doc_ucs4_puts(doc, d);
			doc_puts(doc, "}");
			break;
		case DocML_HTML:
			d = text_ucs4_codename_to_html(s);
			doc_ucs4_puts(doc, d);
			break;
		case DocML_TeX:
			d = text_ucs4_codename_to_tex(s);
			if(d[0]=='$') {
				if(doc->math==MATH_NONE) {
					doc_ucs4_puts(doc, d);
					doc_puts(doc, "{}");
					doc->math = MATH_IMPL;
				} else {
					doc_ucs4_puts(doc, &d[1]);
					doc_puts(doc, "{}");
				}
			} else if(d[0]=='\\') {
				if(doc->math==MATH_NONE) {
					doc_ucs4_puts(doc, d);
					doc_puts(doc, "{}");
				} else if(doc->math==MATH_IMPL) {
					doc_puts(doc, "$");
					doc->math = MATH_NONE;
					doc_ucs4_puts(doc, d);
					doc_puts(doc, "{}");
				} else {
					doc_puts(doc, "$");
					doc->math = MATH_NONE;
					doc_ucs4_puts(doc, d);
					doc_puts(doc, "{}");
					doc_puts(doc, "$");
					doc->math = MATH_EXPL;
				}
			} else {
				doc_ucs4_puts(doc, d);
				doc_puts(doc, "{}");
			}
			break;
		case DocML_PLAIN:
		default:
			d = text_ucs4_codename_to_text(s);
			doc_ucs4_puts(doc, d);
			break;
		}
	} else if(sc == 'M') {
		switch(doc->dml) {
		case DocML_XML:
			doc_puts(doc, "\\M{");
			doc_text_ucs4_puts(doc, s);
			doc_puts(doc, "}");
			break;
		case DocML_HTML:
			if(doc->math==MATH_NONE) {
				doc->math = MATH_EXPL;
				doc_puts(doc, "<var>");
				doc_text_ucs4_puts(doc, s);
				doc_puts(doc, "</var>");
				doc->math = MATH_NONE;
			} else {
				int	math = doc->math;

				doc->math = MATH_NONE;
				doc_puts(doc, "</var>");
				doc_text_ucs4_puts(doc, s);
				doc_puts(doc, "<var>");
				doc->math = math;
			}
			break;
		case DocML_TeX:
			if(doc->math==MATH_NONE) {
				doc_ucs4_putc(doc, '$');
				doc->math = MATH_EXPL;
				doc_text_ucs4_puts(doc, s);
				doc->math = MATH_IMPL;
			} else if(doc->math==MATH_EXPL) {
				int	math = doc->math;

				doc_puts(doc, "\\mbox{");
				doc->math = MATH_NONE;
				doc_text_ucs4_puts(doc, s);
				doc_math_check(doc);
				doc_puts(doc, "}");
				doc->math = math;
			} else {	/* doc->math==MATH_IMPL */
				int	math = doc->math;

				doc->math = MATH_EXPL;
				doc_text_ucs4_puts(doc, s);
				doc->math = math;
			}
			break;
		case DocML_PLAIN:
		default:
			doc_text_ucs4_puts(doc, s);
			break;
		}
	} else if(sc == '_') {
		switch(doc->dml) {
		case DocML_XML:
			doc_puts(doc, "\\_{");
			doc_text_ucs4_puts(doc, s);
			doc_puts(doc, "}");
			break;
		case DocML_HTML:
			doc_puts(doc, "<sub>");
			doc_text_ucs4_puts(doc, s);
			doc_puts(doc, "</sub>");
			break;
		case DocML_TeX:
			if(doc->math==MATH_NONE) {
				doc_puts(doc, "$_{");
				doc->math = MATH_EXPL;
				doc_text_ucs4_puts(doc, s);
				doc_puts(doc, "}");
				doc->math = MATH_IMPL;
			} else {
				int	math = doc->math;

				doc->math = MATH_EXPL;
				doc_puts(doc, "_{");
				doc_text_ucs4_puts(doc, s);
				doc_puts(doc, "}");
				doc->math = math;
			}
			break;
		case DocML_PLAIN:
		default:
			doc_text_ucs4_puts(doc, s);
			break;
		}
	} else if(sc == '^') {
		switch(doc->dml) {
		case DocML_XML:
			doc_puts(doc, "\\^{");
			doc_text_ucs4_puts(doc, s);
			doc_puts(doc, "}");
			break;
		case DocML_HTML:
			doc_puts(doc, "<sup>");
			doc_text_ucs4_puts(doc, s);
			doc_puts(doc, "</sup>");
			break;
		case DocML_TeX:
			if(doc->math==MATH_NONE) {
				doc_puts(doc, "$^{");
				doc->math = MATH_EXPL;
				doc_text_ucs4_puts(doc, s);
				doc_puts(doc, "}");
				doc->math = MATH_IMPL;
			} else {
				int	math = doc->math;

				doc->math = MATH_EXPL;
				doc_puts(doc, "^{");
				doc_text_ucs4_puts(doc, s);
				doc_puts(doc, "}");
				doc->math = math;
			}
			break;
		case DocML_PLAIN:
		default:
			doc_text_ucs4_puts(doc, s);
			break;
		}
	} else if(sc == 'P') {
		utf8_t	*u8 = UCS4toUTF8(s);

		if(doc->picture_func) {
			(*doc->picture_func)(doc, (char *)u8, doc->picture_udata);
		} else {
			switch(doc->dml) {
			case DocML_XML:
				doc_puts(doc, "\\P{");
				doc_text_ucs4_puts(doc, s);
				doc_puts(doc, "}");
				break;
			case DocML_HTML:
				doc_printf(doc, "<img src=\"/cgi-bin/edb_browse?ACT=PICTURE&EID=%s\" alt=\"\\P{%s}\" height=\"32\" align=\"top\" />", u8, u8);
				break;
			case DocML_TeX:
			case DocML_PLAIN:
			default:
				doc_text_ucs4_puts(doc, s);
				break;
			}
		}
	} else if(sc == 'N') {
		switch(doc->dml) {
		case DocML_XML:
			doc_puts(doc, "\\N{");
			doc_text_ucs4_puts(doc, s);
			doc_puts(doc, "}");
			break;
		case DocML_HTML:
			doc_puts(doc, "<i>");
			doc_text_ucs4_puts(doc, s);
			doc_puts(doc, "</i>");
			break;
		case DocML_TeX:
			doc_puts(doc, "{\\it ");
			doc_text_ucs4_puts(doc, s);
			doc_puts(doc, "}");
			break;
		case DocML_PLAIN:
		default:
			doc_text_ucs4_puts(doc, s);
			break;
		}
	} else if(sc == 'D') {
		switch(doc->dml) {
		case DocML_XML:
			doc_puts(doc, "\\D{");
			doc_text_ucs4_puts(doc, s);
			doc_puts(doc, "}");
			break;
		case DocML_HTML:
			doc_puts(doc, "<del>");
			doc_text_ucs4_puts(doc, s);
			doc_puts(doc, "</del>");
			break;
		case DocML_TeX:
			doc_puts(doc, "\\EdbDelete{");
			doc_text_ucs4_puts(doc, s);
			doc_puts(doc, "}");
			break;
		case DocML_PLAIN:
		default:
		/*
			doc_text_ucs4_puts(doc, s);
		*/
			break;
		}
	} else if(sc == 'I') {
		switch(doc->dml) {
		case DocML_XML:
			doc_puts(doc, "\\I{");
			doc_text_ucs4_puts(doc, s);
			doc_puts(doc, "}");
			break;
		case DocML_HTML:
			doc_puts(doc, "<ins>");
			doc_text_ucs4_puts(doc, s);
			doc_puts(doc, "</ins>");
			break;
		case DocML_TeX:
			doc_puts(doc, "\\EdbInsert{");
			doc_text_ucs4_puts(doc, s);
			doc_puts(doc, "}");
			break;
		case DocML_PLAIN:
		default:
			doc_text_ucs4_puts(doc, s);
			break;
		}
	} else {
		doc_text_ucs4_puts(doc, s);
	}
	return TRUE;
}

static
int
doc_text_ucs4_puts(Doc doc, ucs4_t *s)
{
	register ucs4_t	c;
	int	inside = doc->inside;

	if(s==NULL) return 0;
	doc->inside = TRUE;
	while((c=*s++)) {
		if(c=='\\') {
			edb_text_ucs4_parse(&s, doc_text_ucs4_parse, doc);
		} else {
			doc_text_ucs4_putc(doc, c);
		}
	}
	if(!inside) doc_math_check(doc);
	doc->inside = inside;
	return 0;
}

int
doc_text_puts(Doc doc, char *s)
{
	register ucs4_t	*u;

	if(s==NULL) return 0;
	u = UTF8toUCS4((utf8_t *)s);
	doc_text_ucs4_puts(doc, u);
	return 0;
}

static
int
doc_text_ucs4_putc(Doc doc, ucs4_t c)
{
	int	inside = doc->inside;
	ucs4_t	*u;

	doc->inside = TRUE;
	switch(doc->dml) {
	case DocML_XML:
		switch(c) {
		case	'<':	doc_puts(doc, "&lt;");	break;
		case	'>':	doc_puts(doc, "&gt;");	break;
		case	'&':	doc_puts(doc, "&amp;");	break;
		case	'\'':	doc_puts(doc, "&apos;");	break;
		case	'"':	doc_puts(doc, "&quot;");	break;
		default:	doc_ucs4_putc(doc, c);	break;
		}
		break;
	case DocML_HTML:
		switch(c) {
		case '&':	doc_puts(doc, "&amp;");	break;
		case '"':	doc_puts(doc, "&quot;");	break;
		case '<':	doc_puts(doc, "&lt;");	break;
		case '>':	doc_puts(doc, "&gt;");	break;
		case '\n':	doc_puts(doc, "<br />\n");	break;
		default:	doc_ucs4_putc(doc, c);	break;
		}
		break;
	case DocML_TeX:
		if(doc->math==MATH_IMPL) {
			doc_ucs4_putc(doc, '$');
			doc->math = MATH_NONE;
		}
		switch(c) {
		case ' ':
			if(doc->math==MATH_NONE && doc->lastc == '.') doc_ucs4_putc(doc, '\\');
			doc_ucs4_putc(doc, ' ');
			break;
		case '%':	doc_puts(doc, "\\%");	break;
		case '{':	doc_puts(doc, "\\{");		break;
		case '}':	doc_puts(doc, "\\}");		break;
		case '&':	doc_puts(doc, "\\&");		break;
		case '#':	doc_puts(doc, "\\#");		break;
		case '$':	doc_puts(doc, "\\$");		break;
		case '^':	doc_puts(doc, doc->math ? "\\hat{}" : "\\^{}");	break;
		case '_':	doc_puts(doc, "\\_");		break;
		case '<':	doc_puts(doc, doc->math ? "<" : "{$<$}");	break;
		case '>':	doc_puts(doc, doc->math ? ">" : "{$>$}");	break;
		case '~':	doc_puts(doc, doc->math ? "\\tilde{}" : "\\~{}");	break;
		case '\\':	doc_puts(doc, doc->math ? "\\backslash{}" : "{$\\backslash$}");	break;
		default:
			u = text_ucs4_unicode_to_tex(c);
			if(!STRisVALID(u)) {
				doc_ucs4_putc(doc, c);
			} else if(u[0]=='$') {
				if(doc->math==MATH_NONE) {
					doc_ucs4_puts(doc, u);
					doc_puts(doc, "{}");
					doc->math = MATH_IMPL;
				} else {
					doc_ucs4_puts(doc, &u[1]);
					doc_puts(doc, "{}");
				}
			} else if(u[0]=='\\') {
				if(doc->math==MATH_NONE) {
					doc_ucs4_puts(doc, u);
					doc_puts(doc, "{}");
				} else if(doc->math==MATH_IMPL) {
					doc_puts(doc, "$");
					doc->math = MATH_NONE;
					doc_ucs4_puts(doc, u);
					doc_puts(doc, "{}");
				} else {
					doc_puts(doc, "$");
					doc->math = MATH_NONE;
					doc_ucs4_puts(doc, u);
					doc_puts(doc, "{}");
					doc_puts(doc, "$");
					doc->math = MATH_EXPL;
				}
			} else {
				doc_ucs4_puts(doc, u);
			}
			break;
		}
		if(doc->math==MATH_NONE) {
			if((doc->split_mode==DOC_SPLIT_ALL)
				|| (doc->split_mode==DOC_SPLIT_SOME && ucs4_isascii(c) && strchr(",-", c))
				) {
				doc_puts(doc, "{\\EdbZSp}");
			}
		}
		break;
	case DocML_PLAIN:
	default:
		doc_ucs4_putc(doc, c);
		break;
	}
	if(!inside) doc_math_check(doc);
	doc->inside = inside;
	return 0;
}

int
doc_text_putc(Doc doc, int c)
{
	int	inside = doc->inside;

	doc->inside = TRUE;
	doc_text_ucs4_putc(doc, c);
	if(!inside) doc_math_check(doc);
	doc->inside = inside;
	return 0;
}

int
doc_text_puts3(Doc doc, char *h, char *s, char *t)
{

	if(h) doc_puts(doc, h);
	doc_text_puts(doc, s);
	if(t) doc_puts(doc, t);
	return 0;
}

int
doc_text_printf(Doc doc, char *fmt, ...)
{
	va_list	ap;
	int	n;
	char	*str;

	va_start(ap, fmt);
	n = vasprintf(&str, fmt, ap);
	va_end(ap);
	doc_text_puts(doc, str);
	FREE(str);
	return n;
}

void
doc_color_begin(Doc doc, char *fgc)
{

	if(doc==NULL) return;
	switch(doc->dml) {
	case DocML_HTML:
		doc_puts(doc, "<span style=\"color:");
		doc_puts(doc, fgc);
		doc_puts(doc, "\">");
		break;
	default:
		break;
	}
}

void
doc_color_end(Doc doc)
{

	if(doc==NULL) return;
	switch(doc->dml) {
	case DocML_HTML:
		doc_puts(doc, "</span>");
		break;
	default:
		break;
	}
}

void
doc_underline_begin(Doc doc)
{

	if(doc==NULL) return;
	switch(doc->dml) {
	case DocML_HTML:
		doc_puts(doc, "<u>");
		break;
	case DocML_TeX:
		doc_puts(doc, "\\underline{");
		break;
	default:
		break;
	}
}

void
doc_underline_end(Doc doc)
{

	if(doc==NULL) return;
	switch(doc->dml) {
	case DocML_HTML:
		doc_puts(doc, "</u>");
		break;
	case DocML_TeX:
		doc_puts(doc, "}");
		break;
	default:
		break;
	}
}

void
doc_put_datetime(Doc doc, time_t t)
{
	struct tm	tm;
	char	*mname[] = { "Jan.", "Feb.", "Mar.", "Apr.", "May", "Jun.", "Jul.", "Aug.", "Sep.", "Oct.", "Nov.", "Dec."};
	char	*kname_j[] = { "日", "月", "火", "水", "木", "金", "土"};
	char	*kname_e[] = { "Sun.", "Mon.", "Tue.", "Wed.", "Thu.", "Fri.", "Sat."};

	localtime_r(&t, &tm);
	if(doc->codeset != DocCodeset_Latin) {
		doc_printf(doc, "%d年%d月%d日(%s) %02d:%02d:%02d",
			tm.tm_year + 1900,
			tm.tm_mon + 1,
			tm.tm_mday,
			kname_j[tm.tm_wday],
			tm.tm_hour,
			tm.tm_min,
			tm.tm_sec);
	} else {
		doc_printf(doc, "%s %s %d %02d:%02d:%02d %d",
			kname_e[tm.tm_wday],
			mname[tm.tm_mon],
			tm.tm_mday,
			tm.tm_hour,
			tm.tm_min,
			tm.tm_sec,
			tm.tm_year + 1900);
	}
}
