/*
	$Id: edb_publish.c,v 1.11 2003/04/30 04:17:42 alex Exp alex $
	edb_publish---
*/

#ifndef	lint
static char *rcsid = "$Id: edb_publish.c,v 1.11 2003/04/30 04:17:42 alex Exp alex $";
#endif

#include	<stdio.h>
#include	<string.h>
#include	"edb/edb.h"

static
Link
edb_publish_get_eid_list(EdbContext ec, char *table, char *constraint)
{
	char	*query;
	char	*latest;
	int	n;
	Link	l_eid;

	latest = (char *)edb_create_latest_phrase(ec, (utf8_t *)table, (utf8_t *)"t1", (utf8_t *)"t2", NULL, 0);
	asprintf(&query,
		"SELECT %s FROM %s AS t1 WHERE %s %s %s ORDER BY %s",
		EDB_BASE_EID,
		table,
		constraint ? constraint : "",
		constraint ? "AND" : "",
		latest,
		EDB_BASE_EID);
	free(latest);
	l_eid = edb_pgsql_query_one_field(ec, (utf8_t *)query, &n);
	free(query);
	return l_eid;
}

static
void
edb_tuple_publish(EdbContext dst, EdbContext src, char *sqlname)
{
	Link	l, l_tuple_eid;

	l_tuple_eid = edb_publish_get_eid_list(src, sqlname, NULL);

	for(l=l_tuple_eid; l; l=l->n) {
		eid_t	eid = atoi((char *)l->d);
		EdbTupleInfo	tuple;
		Link	l_peid, l_eid;
		eoid_t	eoid;
		Link	lc;

		edb_begin(src, FALSE);
		tuple = edb_tuple_get(src, eid, 0);
		if(tuple == NULL) {
			fprintf(stderr, "ERROR: tuple read error: table=%s EID=%d\n", sqlname, eid);
			continue;
		}
		edb_end(src);

		if(!TupleAvail(tuple)) continue;

#if 0
printf("\tEID=%d\n", TupleEID(tuple); fflush(stdout);
#endif
		edb_begin(dst, TRUE);
#if 0
		eoid = edb_tuple_put_reldb(dst, tuple, FALSE);
#endif

		edb_tuple_get_ref_eid(dst, tuple, &l_eid, &l_peid, NULL);
#if 0
		edb_index_put(dst, TupleBase(tuple), EDB_EID_IS_TUPLE, TABLE_EID(TupleTable(tuple)), l_peid, l_eid);
#endif
		link_free(l_peid);
		link_free(l_eid);
		edb_end(dst);
	}

	link_free(l_tuple_eid);
}

static
void
edb_table_publish(EdbContext dst, EdbContext src)
{
	Link	l, l_table_eid;

	l_table_eid = edb_publish_get_eid_list(src, EDB_TABLE_TABLE, NULL);

	for(l=l_table_eid; l; l=l->n) {
		eid_t	eid = atoi((char *)l->d);
		EdbTableInfo	ti;
		int	c;
		char	command[EDB_BUFSIZ];
		eoid_t	eoid;

		edb_begin(src, FALSE);
		ti = edb_table_get(src, eid);
		if(ti == NULL) {
			fprintf(stderr, "ERROR: table read error: EID=%d\n", eid);
			continue;
		}
		edb_end(src);

		if(!TABLE_Avail(ti)) continue;

#if 0
printf("table=%s\n", ti->sqlname);
#endif
		edb_begin(dst, TRUE);
#if 0
		eoid = edb_table_put_reldb(dst, ti);
#endif
		sprintf(command, "CREATE TABLE %s (", ti->sqlname);
		for(c=0; c<ti->ncol; ++c) {
			EdbColumnInfo ci = ti->column[c];
			char	*sql_type;

#if 0
			char sqlname[BUFSIZ];

			ci->col = c+1;
			if(ci->sqlname) free(ci->sqlname);
			sprintf(sqlname, "c%d", ci->col);
			ci->sqlname = strdup(sqlname);
#endif
			sql_type = edb_datum_type_to_sql_type(src, ColumnTypeName(ci));
			sprintf(&command[strlen(command)], "%s %s%s%s",
				ci->sqlname,
				sql_type,
				ci->elements==1 ? "" : "[]",
				(c<ti->ncol-1) ? "," : "");
		}
		sprintf(&command[strlen(command)], ") INHERITS (edb_base)");
		if(!edb_pgsql_command_exec(dst, (utf8_t *)command)) {
			fprintf(stderr, "ERROR: cannot create table: EID=%d : %s\n", eid, command);
		}
#if 0
		edb_index_put(dst, TABLE_Base(ti), EDB_EID_IS_TABLE, 0, NULL, NULL);
#endif

#if 0
		edb_table_create_index(dst, ti, EDB_BASE_EID, FALSE);
		edb_table_create_index(dst, ti, EDB_BASE_EOID, TRUE);
#endif

		sprintf(command, "GRANT SELECT ON %s TO PUBLIC", ti->sqlname);
		if(!edb_pgsql_command_exec(dst, (utf8_t *)command)) {
			fprintf(stderr, "ERROR: cannot grant table: EID=%d : %s\n", eid, command);
		}
		edb_end(dst);

		edb_tuple_publish(dst, src, (char *)ti->sqlname);
	}

	link_free(l_table_eid);
}

/*#edb_publish--- publish */
static
void
edb_publish(EdbContext dst, EdbContext src)
{

	edb_table_publish(dst, src);
}

/*#main--- main function */
int
main(int ac, char *av[])
{
	EdbContext	edb = NULL;
	EdbContext	www = NULL;
	char	*srcdbname = NULL, *dstdbname = NULL;
	int	i;

	for(i=1; i<ac; ++i) {
		if(!strcmp(av[i], "-D") && i+1<ac) {
			srcdbname = av[i+1];
			++i;
			continue;
		} else {
			dstdbname = av[i];
		}
	}

	edb = edb_configure(NULL);
	if(srcdbname) {
		edb_dbpath_set(edb, srcdbname, TRUE);
		printf("SRCNAME=%s\n", srcdbname);
	}
	edb_open(edb);
	edb_trace_set(edb, stderr);
	edb->operator = EDB_MASTER_UID;

	www = edb_configure(PUBLISH_DBCONF);
	if(dstdbname) {
		edb_dbpath_set(www, dstdbname, TRUE);
		printf("DSTNAME=%s\n", dstdbname);
	}
	www->pgsqlUser = NULL;
	www->pgsqlPassword = NULL;
	edb_open(www);
	edb_trace_set(www, stderr);
	www->operator = EDB_MASTER_UID;

	edb_publish(www, edb);

	edb_close(edb);
	edb_close(www);
	exit(0);
}


