/*
	$Id: edb.c,v 1.49 2012/04/03 03:51:13 alex Exp alex $
	edb.c--- 
*/

#ifndef	lint
static char *rcsid __attribute__((unused)) = "$Id: edb.c,v 1.49 2012/04/03 03:51:13 alex Exp alex $";
#endif

#include	<stdio.h>
#include	<stdlib.h>
#include	<stdarg.h>
#include	<string.h>
#include	<time.h>
#include	"libpq-fe.h"
#include	"pg_config.h"
#include	"edb/edb.h"

#if 0
#define	DELAYED_FREE_SIZE				256
#define	DELAYED_FREE_SIZE_FOR_COMMAND	1024
#define	DELAYED_FREE_SIZE_FOR_QUERY		8192
#else
#define	DELAYED_FREE_SIZE				1
#define	DELAYED_FREE_SIZE_FOR_COMMAND	2
#define	DELAYED_FREE_SIZE_FOR_QUERY		4
#endif

static PGresult * edb_pgsql_get_result	__P((EdbContext));

int
edb_pgsql_pid(EdbContext ec)
{

	return PQbackendPID(ec->pg_conn->conn);
}

int
edb_pgsql_is_connected(EdbContext ec)
{

	return (ec->number_of_pg_connections>0 && ec->pg_conn!=NULL && ec->pg_conn->conn!=NULL);
}

static
int
edb_pgsql_cmd_exec(EdbContext ec, utf8_t *cmd)
{
	register PGresult	*res;

	if(!STRisValid(cmd)) return TRUE;
	if(ec->pg_conn->conn==NULL) return FALSE;
	if(PQisnonblocking(ec->pg_conn->conn)) {
		FREE(ec->pg_conn->last_command);
		ec->pg_conn->last_command = (utf8_t *)strdup((char *)cmd);
		if(!PQsendQuery(ec->pg_conn->conn, (char *)cmd)) {
			edb_trace(ec, "edb_pgsql_cmd_exec: %s\n %s", cmd, PQerrorMessage(ec->pg_conn->conn));
			edb_trace_to_syslog(ec, "edb_pgsql_cmd_exec: %s\n %s", cmd, PQerrorMessage(ec->pg_conn->conn));
			return FALSE;
		}
		while(PQflush(ec->pg_conn->conn)>0) edb_delayed_free_exec(ec, DELAYED_FREE_SIZE);
		return TRUE;
	} else {
		res = PQexec(ec->pg_conn->conn, (char *)cmd);
		if(PQresultStatus(res) != PGRES_COMMAND_OK) {
			edb_trace(ec, "edb_pgsql_cmd_exec: %s\n %s", cmd, PQerrorMessage(ec->pg_conn->conn));
			edb_trace_to_syslog(ec, "edb_pgsql_cmd_exec: %s\n %s", cmd, PQerrorMessage(ec->pg_conn->conn));
			PQclear(res);
			return FALSE;
		}
		PQclear(res);
		return TRUE;
	}
}

static
void
edb_pgsql_wait_for_ready(EdbContext ec, register int size)
{

	while(PQisBusy(ec->pg_conn->conn)) {
		if(edb_delayed_free_exec(ec, size)<size) break;
		PQconsumeInput(ec->pg_conn->conn);
	}
}

static
void
edb_pgsql_wait_for_ready2(EdbContext ec)
{

	PQconsumeInput(ec->pg_conn->conn);
	while(PQisBusy(ec->pg_conn->conn)) {
		edb_process_switch(ec, TRUE);
		PQconsumeInput(ec->pg_conn->conn);
	}
}

static
int
edb_pgsql_consume_result(EdbContext ec, int wait)
{

	if(ec->pg_conn->conn==NULL) return FALSE;
	if(PQisnonblocking(ec->pg_conn->conn)) {
		while(wait || PQconsumeInput(ec->pg_conn->conn)) {
			register PGresult   *res;

			if(!wait && PQisBusy(ec->pg_conn->conn)) return FALSE;
			if(wait) PQconsumeInput(ec->pg_conn->conn);
#if 0
			if(edb_process_number_of_active_processes(ec)==0) {
				edb_pgsql_wait_for_ready(ec, DELAYED_FREE_SIZE);
			} else {
				edb_pgsql_wait_for_ready2(ec);
			}
			res = PQgetResult(ec->pg_conn->conn);
#else
			res = edb_pgsql_get_result(ec);
#endif
			if(res == NULL) break;
			if(PQresultStatus(res) != PGRES_COMMAND_OK) {
				edb_trace(ec, "edb_pgsql_consume_result: There is non-PGRES_COMMAND_OK status.\n");
				if(STRisValid(ec->pg_conn->last_command)) {
					edb_trace(ec, " latest command: %s.\n", ec->pg_conn->last_command);
					edb_trace_to_syslog(ec, " latest command: %s.\n", ec->pg_conn->last_command);
				}
				edb_trace(ec, " This process may be wrong.\n");
				edb_trace(ec, " %s", PQerrorMessage(ec->pg_conn->conn));
			}
			PQclear(res);
		}
		FREE(ec->pg_conn->last_command);
	} else {
		edb_delayed_free_exec(ec, DELAYED_FREE_SIZE);
	}
	return TRUE;
}

static
int
edb_pgsql_flush_command_queue(EdbContext ec)
{
	register int	status = TRUE;

	if(ec->pg_conn->conn==NULL) return FALSE;
	if(ec->pg_conn->l_pgsql_command_queue) {
		if(PQisnonblocking(ec->pg_conn->conn)) {
			register utf8_t	*cmd;

			cmd = (utf8_t *)str_concatenate_str(ec->pg_conn->l_pgsql_command_queue, " ; ");
			status = edb_pgsql_cmd_exec(ec, cmd);
			FREE(cmd);
		} else {
			register Link	l;

			ForLink(l, ec->pg_conn->l_pgsql_command_queue) if(!edb_pgsql_cmd_exec(ec, l->d)) status = FALSE;
		}
		EdbLinkFree(ec->pg_conn->l_pgsql_command_queue);
		ec->pg_conn->l_pgsql_command_queue = NULL;
	}
	return status;
}

int
edb_pgsql_flush(EdbContext ec, int force)
{

#if 0
	edb_pgsql_consume_result(ec, TRUE);
	return edb_pgsql_flush_command_queue(ec);
#else
	if(edb_pgsql_consume_result(ec, force)) return edb_pgsql_flush_command_queue(ec);
	return TRUE;
#endif
}

int
edb_pgsql_setnonblocking(EdbContext ec, int arg)
{
	int	i, status;

	edb_pgsql_flush(ec, TRUE);
	edb_pgsql_consume_result(ec, TRUE);
	return PQsetnonblocking(ec->pg_conn->conn, arg);
}

static
int
edb_pgsql_command_exec_wait(EdbContext ec, utf8_t *cmd)
{
	register PGresult	*res;

	edb_pgsql_flush(ec, TRUE);
	edb_pgsql_consume_result(ec, TRUE);

	++ec->number_of_command;
#if 0
	res = PQexec(ec->pg_conn->conn, (char *)cmd);
#else
	if(ec->pg_conn->conn==NULL) return FALSE;
	if(PQisnonblocking(ec->pg_conn->conn)) {
		if(!PQsendQuery(ec->pg_conn->conn, (char *)cmd)) {
			edb_trace(ec, "edb_pgsql_command_exec_wait: %s\n %s", cmd, PQerrorMessage(ec->pg_conn->conn));
			edb_trace_to_syslog(ec, "edb_pgsql_command_exec_wait: %s\n %s", cmd, PQerrorMessage(ec->pg_conn->conn));
			return FALSE;
		}
		while(PQflush(ec->pg_conn->conn)>0) edb_delayed_free_exec(ec, DELAYED_FREE_SIZE);
#if 0
		if(edb_process_number_of_active_processes(ec)==0) {
			if(edb_delayed_free_exec(ec, DELAYED_FREE_SIZE_FOR_COMMAND)>=DELAYED_FREE_SIZE_FOR_COMMAND) {
				PQconsumeInput(ec->pg_conn->conn);
				edb_pgsql_wait_for_ready(ec, DELAYED_FREE_SIZE);
			}
		} else {
			edb_pgsql_wait_for_ready2(ec);
		}
		res = PQgetResult(ec->pg_conn->conn);
#else
		res = edb_pgsql_get_result(ec);
#endif
	} else {
		edb_delayed_free_exec(ec, DELAYED_FREE_SIZE_FOR_COMMAND);
		res = PQexec(ec->pg_conn->conn, (char *)cmd);
	}
#endif
	if(PQresultStatus(res) != PGRES_COMMAND_OK) {
		edb_trace(ec, "edb_pgsql_command_exec_wait: %s\n", cmd);
		edb_trace_to_syslog(ec, "edb_pgsql_command_exec_wait: %s\n", cmd);
		edb_trace(ec, "%s", PQerrorMessage(ec->pg_conn->conn));
		PQclear(res);
		return FALSE;
	}
	PQclear(res);
	return TRUE;
}

int
edb_pgsql_command_exec(EdbContext ec, utf8_t *cmd)
{
	register int	status;

	++ec->number_of_command;
	if(STRisValid(cmd)) LinkAppend(ec->pg_conn->l_pgsql_command_queue, link_atom_new_sdup((char *)cmd));
	if(link_atoms(ec->pg_conn->l_pgsql_command_queue)>ec->pgsql_revolver_cmds) {
		return edb_pgsql_flush(ec, TRUE);
	} else if(edb_pgsql_consume_result(ec, FALSE)) {
		return edb_pgsql_flush_command_queue(ec);
	} else {
		return TRUE;
	}
	return status;
}

static
PGresult *
edb_pgsql_query_exec(EdbContext ec, utf8_t *query)
{

	edb_pgsql_consume_result(ec, TRUE);
	edb_pgsql_flush_command_queue(ec);
	edb_pgsql_consume_result(ec, TRUE);
	++ec->number_of_query;
#if 0
	return PQexec(ec->pg_conn->conn, (char *)query);
#else
	if(PQisnonblocking(ec->pg_conn->conn)) {
		if(!PQsendQuery(ec->pg_conn->conn, (char *)query)) {
			edb_trace(ec, "edb_pgsql_query_exec: %s\n %s", query, PQerrorMessage(ec->pg_conn->conn));
			edb_trace_to_syslog(ec, "edb_pgsql_query_exec: %s\n %s", query, PQerrorMessage(ec->pg_conn->conn));
			return FALSE;
		}
		while(PQflush(ec->pg_conn->conn)>0) edb_delayed_free_exec(ec, DELAYED_FREE_SIZE);
#if 0
		if(edb_process_number_of_active_processes(ec)==0) {
			if(edb_delayed_free_exec(ec, DELAYED_FREE_SIZE_FOR_QUERY)>=DELAYED_FREE_SIZE_FOR_QUERY) {
				PQconsumeInput(ec->pg_conn->conn);
				edb_pgsql_wait_for_ready(ec, DELAYED_FREE_SIZE);
			}
		} else {
			edb_pgsql_wait_for_ready2(ec);
		}
		return PQgetResult(ec->pg_conn->conn);
#else
		return edb_pgsql_get_result(ec);
#endif
	} else {
		edb_delayed_free_exec(ec, DELAYED_FREE_SIZE_FOR_QUERY);
		return PQexec(ec->pg_conn->conn, (char *)query);
	}
#endif
}

static
int
edb_pg_stat_activity(EdbContext ec, utf8_t *dbname)
{
	register int	users;
	utf8_t		*query;

	if(dbname) {
		asprintf((char **)&query, "SELECT COUNT(*) FROM pg_stat_activity WHERE datname='%s'", dbname);
	} else {
		asprintf((char **)&query, "SELECT COUNT(*) FROM pg_stat_activity");
	}
	users = edb_pgsql_query_int(ec, query);
	FREE(query);
	return users;
}

static
int
edb_pg_server_version(EdbContext ec)
{
	utf8_t		*query;
	utf8_t		*server_version;
	int			version;
	int			major, minor, patch, m;

	asprintf((char **)&query, "SELECT setting FROM pg_settings WHERE name='server_version'");
	server_version = edb_pgsql_query_text(ec, query);
	FREE(query);
	m = sscanf((char *)server_version, "%d.%d.%d", &major, &minor, &patch);
	FREE(server_version);
#if 0
	version = 0;
	if(m>=1) version |= (major<<16);
	if(m>=2) version |= (minor<<8);
	if(m>=3) version |= (patch);
#else
	if(m<1) major = 0;
	if(m<2) minor = 0;
	if(m<3) patch = 0;
	version = PG_SV_mmp(major, minor, patch);
#endif
	return version;
}

#define	MINIMUM_ACTIVITY	1	/* 1 for dwarf */
#define	LOW_LA			1.0	/* load average is enough low */
#define	LA_BASE			0.3
#define	LA_DELTA			0.1

int
edb_open(EdbContext ec)
{
	int	users = 0;
	int	replica;
	char	*dbName;
	int	replica_offset = getpid();

	/* set randomize seed */

	if(ec->pg_conn && ec->pg_conn->conn) return TRUE;
	if(ec->pg_connections==NULL) {
		ec->pg_connections = (EdbPGConn)malloc(sizeof(struct _EDB_PG_CONN));
		ec->number_of_pg_connections = 1;
	}
	ec->pg_conn = &ec->pg_connections[0];
	bzero(ec->pg_connections, sizeof(struct _EDB_PG_CONN));

	/* make a connection to the database */
	if(ec->l_dbpath == NULL) {
		ec->pg_conn->conn = PQsetdbLogin(ec->host, ec->port, ec->options, ec->tty, ec->dbName, ec->pgsqlUser, ec->pgsqlPassword);
		++ec->open_trials;
		/* check to see that the backend connection was successfully made */
		if(PQstatus(ec->pg_conn->conn) == CONNECTION_BAD) {
			edb_trace(ec, "Connection to database '%s' failed.\n", ec->dbName);
			edb_trace(ec, "%s", PQerrorMessage(ec->pg_conn->conn));
			PQfinish(ec->pg_conn->conn);
			return FALSE;
		}
	} else {
		Link	l;
		double	loadaverage=0.0;

		ec->pg_conn->conn = NULL;
		ForLink(l, ec->l_dbpath) {
			EdbDBPath	db = (EdbDBPath)l->d;
			double	la;
			int		u;
			register PGconn	*conn;
			register PGresult	*res;
			char	*s;
			char	*dbn = NULL;

			if(db->replica>0) {
				asprintf(&dbn, "%s%d", db->dbname, (replica_offset)%db->replica);
			} else {
				dbn = strdup(db->dbname);
			}
			conn = PQsetdbLogin(db->host, db->port, ec->options, ec->tty, dbn, ec->pgsqlUser, ec->pgsqlPassword);
			++ec->open_trials;
			/* check to see that the backend connection was successfully made */
			if(PQstatus(conn) == CONNECTION_BAD) {
				PQfinish(conn);
				FREE(dbn);
				continue;
			}
			/* get load average */
			res = PQexec(conn, "SELECT getloadaverage()");
			if(PQresultStatus(res) != PGRES_TUPLES_OK) {
				PQclear(res);
				PQfinish(conn);
				FREE(dbn);
				continue;
			}
			s = PQgetvalue(res, 0, 0);
			PQclear(res);
			la = ATOF(s);
//			if(la<=LOW_LA) la = LOW_LA;

#if 0
			if(ec->readonly) {
				u = 1;
			} else {
				/* get number of users */
#if 1
				if(db->replica>0) {
					asprintf((char **)&command, "SELECT COUNT(*) FROM pg_stat_activity WHERE datname~'%s.'", db->dbname);
				} else {
					asprintf((char **)&command, "SELECT COUNT(*) FROM pg_stat_activity WHERE datname~'%s'", db->dbname);
				}
				res = PQexec(conn, (char *)command);
				FREE(command);
#else
				res = PQexec(conn, "SELECT COUNT(*) FROM pg_stat_activity");
#endif
				if(PQresultStatus(res) != PGRES_TUPLES_OK) {
					PQclear(res);
					PQfinish(conn);
					FREE(dbn);
					continue;
				}
				s = PQgetvalue(res, 0, 0);
				PQclear(res);
				u = ATOI(s);
			}
#else
			u = 1;
#endif

			if(ec->pg_conn->conn==NULL || la*u<loadaverage) {
				if(ec->pg_conn->conn) PQfinish(ec->pg_conn->conn);
				ec->pg_conn->conn = conn;
				ec->host = db->host;
				ec->port = db->port;
				ec->dbName = db->dbname;
				ec->number_of_replica = db->replica;
				loadaverage = la*u;
			} else {
				PQfinish(conn);
			}
			/*
			If connection is established, loadaverage is low and no other user,
			then don't try other database server.
			ec->number_of_replica for each edb_dwarf, 1 for own connection.
			*/
			if(ec->pg_conn->conn && (loadaverage<=LA_BASE+LA_DELTA*ec->open_trials) && (ec->number_of_replica==0 || u<=ec->number_of_replica+1)) break;
		}
	}

	if(ec->pg_conn->conn && !ec->readonly && ec->number_of_replica>0) {
		utf8_t	*dbn = NULL;

		dbName = strdup(ec->dbName);

		asprintf((char **)&dbn, "%s%d", dbName, replica_offset%ec->number_of_replica);
		users = edb_pg_stat_activity(ec, dbn);
		ec->dbName = (char *)dbn;

		if(users>MINIMUM_ACTIVITY) {
			for(replica=1; replica<ec->number_of_replica; ++replica) {
				register PGconn	*conn;
				int	u;

				asprintf((char **)&dbn, "%s%d", dbName, (replica + replica_offset)%ec->number_of_replica);
				conn = PQsetdbLogin(ec->host, ec->port, ec->options, ec->tty, (char *)dbn, ec->pgsqlUser, ec->pgsqlPassword);
				++ec->open_trials;
				/* check to see that the backend connection was successfully made */
				if(PQstatus(conn) == CONNECTION_BAD) {
					PQfinish(conn);
					FREE(dbn);
					continue;
				}
				u = edb_pg_stat_activity(ec, dbn);
				if(u<users) {
					PQfinish(ec->pg_conn->conn);
					ec->pg_conn->conn = conn;
					users = u;
					ec->dbName = (char *)dbn;
					if(u<=MINIMUM_ACTIVITY) break;
				} else {
					PQfinish(conn);
					FREE(dbn);
				}
			}
		}
	}
	if(ec->pg_conn->conn) {
		ec->pg_server_version = edb_pg_server_version(ec);
		edb_title_initialize(ec);
	}
	return (ec->pg_conn->conn!=NULL);
}

#if 0
int
edb_setup_processes(EdbContext ec, int num)
{

	edb_process_initialize(ec);
	if(!ec->readonly) return ec->number_of_pg_connections;
	if(num==0) {
		register PGresult	*res;

		res = PQexec(ec->pg_conn->conn, "SELECT getnumberofcpus()");
		if(PQresultStatus(res) == PGRES_TUPLES_OK) {
			char	*s = PQgetvalue(res, 0, 0);
//			num = ATOI(s)+1;
			num = ATOI(s)*2;
		}
		PQclear(res);
		if(num<1) num = 1;
	}
	if(&ec->pg_connections[0]!=ec->pg_conn) return ec->number_of_pg_connections;
	if(ec->number_of_pg_connections<num) {
		int	i;

		ec->pg_connections = (EdbPGConn)realloc(ec->pg_connections, sizeof(struct _EDB_PG_CONN)*num);
		ec->pg_conn = &ec->pg_connections[0];
		bzero(&ec->pg_connections[ec->number_of_pg_connections], sizeof(struct _EDB_PG_CONN)*(num-ec->number_of_pg_connections));
		for(i=ec->number_of_pg_connections; i<num; ++i) {
			EdbPGConn	pgc = &ec->pg_connections[ec->number_of_pg_connections];

			pgc->conn = PQsetdbLogin(ec->host, ec->port, ec->options, ec->tty, ec->dbName, ec->pgsqlUser, ec->pgsqlPassword);
			/* check to see that the backend connection was successfully made */
			if(PQstatus(pgc->conn) == CONNECTION_BAD) {
				edb_trace(ec, "Connection to database '%s' failed.\n", ec->dbName);
				edb_trace(ec, "%s", PQerrorMessage(pgc->conn));
				PQfinish(pgc->conn);
				pgc->conn = NULL;
				continue;
			}
			PQsetnonblocking(pgc->conn, 1);
			++ec->number_of_pg_connections;
		}
	}
	return ec->number_of_pg_connections;
}
#else
int
edb_setup_processes(EdbContext ec, int num)
{

	edb_process_initialize(ec);
	if(!ec->readonly) return ec->number_of_pg_connections;
	if(num==0) {
		register PGresult	*res;

		res = PQexec(ec->pg_conn->conn, "SELECT getnumberofcpus()");
		if(PQresultStatus(res) == PGRES_TUPLES_OK) {
			char	*s = PQgetvalue(res, 0, 0);
//			num = ATOI(s)+1;
			num = ATOI(s)*2;
		}
		PQclear(res);
		if(num<1) num = 1;
	}
	if(&ec->pg_connections[0]!=ec->pg_conn) return ec->number_of_pg_connections;
	if(ec->number_of_pg_connections<num) {
		int	i;

		ec->pg_connections = (EdbPGConn)realloc(ec->pg_connections, sizeof(struct _EDB_PG_CONN)*num);
		ec->pg_conn = &ec->pg_connections[0];
		bzero(&ec->pg_connections[ec->number_of_pg_connections], sizeof(struct _EDB_PG_CONN)*(num-ec->number_of_pg_connections));
		for(i=ec->number_of_pg_connections; i<num; ++i) {
			EdbPGConn	pgc = &ec->pg_connections[ec->number_of_pg_connections];

			if(link_atoms(ec->l_dbpath)>1) {
				int n = link_atoms(ec->l_dbpath);

//				EdbDBPath	db = (EdbDBPath)link_get_obj(ec->l_dbpath, (int)(((unsigned long int)random())%n));
				EdbDBPath	db = (EdbDBPath)link_get_obj(ec->l_dbpath, i%n);
				pgc->conn = PQsetdbLogin(db->host, db->port, ec->options, ec->tty, db->dbname, ec->pgsqlUser, ec->pgsqlPassword);
			} else {
				pgc->conn = PQsetdbLogin(ec->host, ec->port, ec->options, ec->tty, ec->dbName, ec->pgsqlUser, ec->pgsqlPassword);
			}
			/* check to see that the backend connection was successfully made */
			if(PQstatus(pgc->conn) == CONNECTION_BAD) {
				edb_trace(ec, "Connection to database '%s' failed.\n", ec->dbName);
				edb_trace(ec, "%s", PQerrorMessage(pgc->conn));
				PQfinish(pgc->conn);
				pgc->conn = NULL;
				continue;
			}
			PQsetnonblocking(pgc->conn, 1);
			++ec->number_of_pg_connections;
		}
	}
	return ec->number_of_pg_connections;
}
#endif

int
edb_close(EdbContext ec)
{
	int	i;

	/* close the connection to the database and cleanup */
	if(ec->delayed_free) edb_delayed_free_mode(ec, 0);
	for(i=0; i<ec->number_of_pg_connections; ++i) {
		EdbPGConn	pgc = &ec->pg_connections[i];
		if(pgc->conn) {
			PQfinish(pgc->conn);
			pgc->conn = NULL;
		}
	}
	ec->pg_conn = NULL;
	FREE(ec->pg_connections);
	ec->number_of_pg_connections = 0;
	return TRUE;
}

/*
	edb_pgsql_begin
	edb_pgsql_end
*/
int
edb_pgsql_begin(EdbContext ec, int exclusive)
{
	char	*command;
	struct tm	tm;

	/* begin transaction */
	if(!edb_pgsql_command_exec_wait(ec, (utf8_t *)"BEGIN")) return FALSE;
	ec->pg_conn->transaction = TRUE;

	if(ec->readonly || !exclusive) {
		/* set transaction isolation level */
		if(!edb_pgsql_command_exec_wait(ec, (utf8_t *)"SET TRANSACTION ISOLATION LEVEL SERIALIZABLE READ ONLY")) {
			return FALSE;
		}
	} else {
		/* lock database */
		asprintf((char **)&command, "LOCK TABLE edb_lock IN ACCESS %s MODE", exclusive ? "EXCLUSIVE" : "SHARE");
		if(!edb_pgsql_command_exec_wait(ec, (utf8_t *)command)) {
			FREE(command);
			return FALSE;
		}
		FREE(command);
		ec->locked = TRUE;
	}

	/* get current transaction time */
//	ec->now = edb_pgsql_query_int(ec, (utf8_t *)"SELECT int4(abstime(now()))");
	ec->now = edb_pgsql_query_int(ec, (utf8_t *)"SELECT int4(extract(epoch from now()))");
	ec->pg_conn->age = edb_getage(ec);
	if(ec->age<ec->pg_conn->age) ec->age = ec->pg_conn->age;

	localtime_r(&ec->now, &tm);
	ec->present_date = YMD_DATE(tm.tm_year + 1900, tm.tm_mon+1, tm.tm_mday);
	ec->present_wday = tm.tm_wday;
	ec->present_time = HMS_TIME(tm.tm_hour, tm.tm_min, tm.tm_sec);
	return TRUE;
}

int
edb_pgsql_end(EdbContext ec)
{

	/* end transaction */
	edb_pgsql_command_exec_wait(ec, (utf8_t *)"END");
	ec->pg_conn->age = ec->age;
	ec->locked = FALSE;
	ec->pg_conn->transaction = FALSE;
	return TRUE;
}

/*
	edb_begin
	edb_end
*/
int
edb_begin(EdbContext ec, int exclusive)
{

	if(! ec->transaction) {
		EdbPGConn	pg_save = ec->pg_conn;
		int i;

		for(i=ec->number_of_pg_connections-1; i>=0; --i) {
			ec->pg_conn = &ec->pg_connections[i];
			edb_pgsql_begin(ec, exclusive);
		}
		ec->pg_conn = pg_save;
		ec->transaction = TRUE;
	}
	return TRUE;
}

int
edb_end(EdbContext ec)
{

	if(ec->transaction) {
		EdbPGConn	pg_save = ec->pg_conn;
		int i;

		for(i=0; i<ec->number_of_pg_connections; ++i) {
			ec->pg_conn = &ec->pg_connections[i];
			edb_pgsql_end(ec);
		}
		ec->pg_conn = pg_save;
		ec->transaction = FALSE;
	}
	return TRUE;
}

int
edb_listen(EdbContext ec, utf8_t *table)
{
	register int	status;
	utf8_t	*command;

	asprintf((char **)&command, "LISTEN %s", table);
	status = edb_pgsql_command_exec_wait(ec, command);
	FREE(command);
	return status;
}

int
edb_notify(EdbContext ec, utf8_t *table)
{
	register int	status;
	utf8_t	*command;

	if(ec->transaction) {
		edb_trace(ec, "edb_notify: WARNING : NOTIFY is in transaction. It should be in outside of transaction.\n");
	}
	asprintf((char **)&command, "NOTIFY %s", table);
	status = edb_pgsql_command_exec_wait(ec, command);
	FREE(command);
	return status;
}

utf8_t *
edb_accept_notify(EdbContext ec)
{
	PGnotify	*notify;
	utf8_t	*table;

	PQconsumeInput(ec->pg_conn->conn);
	notify = PQnotifies(ec->pg_conn->conn);
	if(notify) {
		struct tm	tm;
		time_t	t;
		static char	*mname[] = { "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"};

		time(&t);
		localtime_r(&t, &tm);
		edb_trace(ec, "Received Notify on \"%s\" from pid(%d) at %s %d %02d:%02d:%02d %d.\n", notify->relname, notify->be_pid,
			mname[tm.tm_mon],
			tm.tm_mday,
			tm.tm_hour,
			tm.tm_min,
			tm.tm_sec,
			tm.tm_year + 1900);
		table = (utf8_t *)strdup(notify->relname);
		FREE(notify);
		return table;
	}
	return NULL;
}


/*#edb_getage--- get age of DB (maximum eoid in xmldb) */
eoid_t
edb_getage(EdbContext ec)
{

	return edb_pgsql_query_int(ec, (utf8_t *)"SELECT " EDB_XMLDB_COLUMN_EOID
		" FROM " EDB_XMLDB_TABLE
		" ORDER BY " EDB_XMLDB_COLUMN_EOID " DESC"
		" LIMIT 1 OFFSET 0");
}

/*#edb_reldb_is_usable--- if relational database is usable */
int
edb_reldb_is_usable(EdbContext ec)
{

	return (!ec->dont_use_reldb);
}

/*#edb_setdatabasespeaking--- set database speaking */
int
edb_setdatabasespeaking(EdbContext ec, char lang[])
{
	utf8_t	*query, *text;

	asprintf((char **)&query, "SELECT setdatabasespeaking('%s')", lang);
	text = edb_pgsql_query_one_text(ec, query);
	FREE(query);
	if(text && !strcasecmp((char *)text, "Japanese")) {
		ec->japanese = TRUE;
	} else if(text && !strcasecmp((char *)text, "English")) {
		ec->japanese = FALSE;
	}
	FREE(text);
	return TRUE;
}

/*#edb_process_setdatabasespeaking--- set database speaking */
int
edb_process_setdatabasespeaking(EdbContext ec, char lang[])
{
	utf8_t	*query, *text;

	asprintf((char **)&query, "SELECT setdatabasespeaking('%s')", lang);
	text = edb_pgsql_query_one_text(ec, query);
	FREE(query);
	FREE(text);
	return TRUE;
}

/*
	edb_pgsql_query--- query
	edb_pgsql_query_one_field--- query (must one fields)
	edb_pgsql_query_one_tuple--- query (must one tuple)
	edb_pgsql_query_one_text--- query (must one text)
*/
Link
edb_pgsql_query(EdbContext ec, utf8_t *query, int *ntuples, int *nfields)
{
	register PGresult	*res;
	int	t, f, nt, nf;
	Link	l_tuple;

	if(nfields) *nfields = 0;
	if(ntuples) *ntuples = 0;
	res = edb_pgsql_query_exec(ec, query);
	if(PQresultStatus(res) != PGRES_TUPLES_OK) {
		edb_trace(ec, "edb_pgsql_query: %s\n", query);
		edb_trace_to_syslog(ec, "edb_pgsql_query: %s\n", query);
		edb_trace(ec, "%s", PQerrorMessage(ec->pg_conn->conn));
		PQclear(res);
		return NULL;
	}
	nf = PQnfields(res);
	nt = PQntuples(res);

	l_tuple = NULL;
	for(t = 0; t < nt; ++t) {
		Link	l_field = NULL;

		for(f = 0; f < nf; ++f) {
			char	*s;

			s = PQgetisnull(res, t, f) ? "" : PQgetvalue(res, t, f);
			LinkAppendTop(link_atom_new_sdup(s), l_field);
		}
		LinkReverse(l_field);
		LinkAppendTop(link_atom_new(l_field, link_free), l_tuple);
	}
	LinkReverse(l_tuple);
	PQclear(res);

	if(ntuples) *ntuples = nt;
	if(nfields) *nfields = nf;
	return l_tuple;
}

int
edb_pgsql_query_callback(
	EdbContext ec,
	utf8_t *query,
	EdbPgsqlQuery_callback_func_t		cb_func,
	void							*cb_data)
{
	register PGresult	*res;
	int	t, f, nt, nf;
	int	status = TRUE;

	if(cb_func==NULL) return FALSE;
	res = edb_pgsql_query_exec(ec, query);
	if(PQresultStatus(res) != PGRES_TUPLES_OK) {
		edb_trace(ec, "edb_pgsql_query: %s\n", query);
		edb_trace_to_syslog(ec, "edb_pgsql_query: %s\n", query);
		edb_trace(ec, "%s", PQerrorMessage(ec->pg_conn->conn));
		PQclear(res);
		return FALSE;
	}
	nf = PQnfields(res);
	nt = PQntuples(res);

	for(t = 0; t < nt && status; ++t) {
		Link	l_field = NULL;

		for(f = 0; f < nf; ++f) {
			char	*s;

			s = PQgetisnull(res, t, f) ? "" : PQgetvalue(res, t, f);
			LinkAppendTop(link_atom_new_sdup(s), l_field);
		}
		LinkReverse(l_field);
		status = status && (*cb_func)(ec, l_field, cb_data);
		EdbLinkFree(l_field);
	}
	PQclear(res);
	return status;
}

Link
edb_pgsql_query_one_field(EdbContext ec, utf8_t *query, int *ntuples)
{
	int	nfields;
	Link	l_tuple, l, l_new;

	l_tuple = edb_pgsql_query(ec, query, ntuples, &nfields);
	if(l_tuple == NULL) return NULL;
	if(nfields != 1) {
		EdbLinkFree(l_tuple);
		return NULL;
	}

	l_new = NULL;
	ForLink(l, l_tuple) {
		LinkAppendTop(link_atom_new_sdup((char *)((Link)l->d)->d), l_new);
	}
	EdbLinkFree(l_tuple);
	return link_reverse(l_new);
}

Link
edb_pgsql_query_one_tuple(EdbContext ec, utf8_t *query, int *nfields)
{
	int	ntuples;
	Link	l_tuple, l_new;

	l_tuple = edb_pgsql_query(ec, query, &ntuples, nfields);
	if(l_tuple == NULL) return NULL;
	if(ntuples != 1) {
		EdbLinkFree(l_tuple);
		return NULL;
	}

	l_new = (Link)l_tuple->d;
	l_tuple->d = NULL;
	EdbLinkFree(l_tuple);
	return l_new;
}

utf8_t *
edb_pgsql_query_one_text(EdbContext ec, utf8_t *query)
{
	int	ntuples, nfields;
	register Link	l_tuple;
	register utf8_t	*text;

	l_tuple = edb_pgsql_query(ec, query, &ntuples, &nfields);
	if(l_tuple == NULL) return NULL;
	if(ntuples != 1 || nfields !=1) {
		EdbLinkFree(l_tuple);
		return NULL;
	}

	text = (utf8_t *)strdup((char *)((Link)l_tuple->d)->d);
	EdbLinkFree(l_tuple);
	return text;
}

/*
	edb_pgsql_query_int--- dump table
*/
int
edb_pgsql_query_int(EdbContext ec, utf8_t *query)
{
	register utf8_t	*text;
	register int	v;

	text = edb_pgsql_query_one_text(ec, query);
	if(text==NULL) return 0;
	v = atoi((char *)text);
	FREE(text);
	return v;
}

/*
	edb_pgsql_query_text--- dump table
*/
utf8_t *
edb_pgsql_query_text(EdbContext ec, utf8_t *query)
{

	return edb_pgsql_query_one_text(ec, query);
}

/*
	edb_pgsql_query_texts--- query texts
*/
int
edb_pgsql_query_texts(EdbContext ec, utf8_t ***array, int *nelems, utf8_t *query)
{
	register PGresult	*res;
	int	i;

	res = edb_pgsql_query_exec(ec, query);
	if (PQresultStatus(res) != PGRES_TUPLES_OK)
	{
		edb_trace(ec, "edb_pgsql_query_texts: %s\n", query);
		edb_trace_to_syslog(ec, "edb_pgsql_query_texts: %s\n", query);
		edb_trace(ec, "%s", PQerrorMessage(ec->pg_conn->conn));
		PQclear(res);
		return 0;
	}
	if(PQnfields(res)!=1) {
		PQclear(res);
		return 0;
	}
	*nelems = PQntuples(res);
	*array = (utf8_t **)malloc(sizeof(utf8_t *)*(*nelems));
	for(i=0; i < (*nelems); ++i) {
		(*array)[i] = (utf8_t *)strdup(PQgetvalue(res, i, 0));
	}
	PQclear(res);

	return (*nelems);
}

/*
	edb_pgsql_query_integers--- dump table
*/
int
edb_pgsql_query_integers(EdbContext ec, int **array, int *nelems, utf8_t *query)
{
	register PGresult	*res;
	int	i;

	res = edb_pgsql_query_exec(ec, query);
	if (PQresultStatus(res) != PGRES_TUPLES_OK)
	{
		edb_trace(ec, "edb_pgsql_query_int: %s\n", query);
		edb_trace_to_syslog(ec, "edb_pgsql_query_int: %s\n", query);
		edb_trace(ec, "%s", PQerrorMessage(ec->pg_conn->conn));
		PQclear(res);
		return FALSE;
	}
	if(PQnfields(res)!=1) {
		PQclear(res);
		return FALSE;
	}
	*nelems = PQntuples(res);
	*array = (int *)malloc(sizeof(int)*(*nelems));
	for(i=0; i < (*nelems); ++i) {
		(*array)[i] = atoi(PQgetvalue(res, i, 0));
	}
	PQclear(res);

	return (*nelems);
}

/*
	edb_pgsql_seek_eid--- seek a eid
*/
int
edb_pgsql_seek_eid(EdbContext ec, utf8_t *table, char *fmt, ...)
{
	register PGresult	*res;
	int	value;
	utf8_t	*query;
	utf8_t	*latest;
	char	*target;
	va_list	ap;

	latest = edb_create_latest_phrase(ec, table, (utf8_t *)"t1", (utf8_t *)"t2", NULL, 0);
	if(fmt) {
		va_start(ap, fmt);
		vasprintf((char **)&target, fmt, ap);
		va_end(ap);
		asprintf((char **)&query, "SELECT %s FROM %s AS t1 WHERE %s AND %s",
			EDB_BASE_EID,
			table,
			target,
			latest);
		FREE(target);
	} else {
		asprintf((char **)&query, "SELECT %s FROM %s AS t1 WHERE %s",
			EDB_BASE_EID,
			table,
			latest);
	}
	FREE(latest);
	res = edb_pgsql_query_exec(ec, query);
	if (PQresultStatus(res) != PGRES_TUPLES_OK) {
		edb_trace(ec, "edb_pgsql_query_int: %s\n", query);
		edb_trace_to_syslog(ec, "edb_pgsql_query_int: %s\n", query);
		edb_trace(ec, "%s", PQerrorMessage(ec->pg_conn->conn));
		PQclear(res);
		FREE(query);
		return FALSE;
	}
	if(PQnfields(res)!=1 || PQntuples(res)!=1) {
		PQclear(res);
		FREE(query);
		return FALSE;
	}
	value = atoi(PQgetvalue(res, 0, 0));
	PQclear(res);
	FREE(query);

	return value;
}

/*
	edb_EID_get--- get EID
*/
int
edb_EID_get(EdbContext ec, utf8_t *table, char *fmt, ...)
{
	int	eid;
	register PGresult	*res;
	utf8_t *query;
	char *target;
	char	*s;
	va_list	ap;

	va_start(ap, fmt);
	vasprintf((char **)&target, fmt, ap);
	va_end(ap);
	asprintf((char **)&query, "SELECT %s FROM %s WHERE %s ORDER BY %s DESC",
		EDB_BASE_EID, table, target, EDB_BASE_MTIME);
	FREE(target);
	res = edb_pgsql_query_exec(ec, query);
	if (PQresultStatus(res) != PGRES_TUPLES_OK || PQnfields(res) <1 || PQntuples(res)!=1) {
		edb_trace(ec, "edb_eid_get: %s", query);
		edb_trace_to_syslog(ec, "edb_eid_get: %s", query);
		edb_trace(ec, "%s", PQerrorMessage(ec->pg_conn->conn));
		PQclear(res);
		FREE(query);
		return 0;
	}

	s = PQgetvalue(res, 0, 0);
	eid = atoi(s);
	PQclear(res);
	FREE(query);

	return eid;
}

/*
	edb_EIDs_get--- get EID
*/
int
edb_EIDs_get(EdbContext ec, utf8_t *table, int **values, int *nvalue, char *fmt, ...)
{
	int	i;
	register PGresult	*res;
	utf8_t *query;
	char *target;
	utf8_t *latest;
	va_list	ap;

	*values = NULL;
	*nvalue = 0;
	latest = edb_create_latest_phrase(ec, table, (utf8_t *)"t1", (utf8_t *)"t2", NULL, 0);
	if(fmt) {
		va_start(ap, fmt);
		vasprintf((char **)&target, fmt, ap);
		va_end(ap);
		asprintf((char **)&query,
			"SELECT %s FROM %s AS t1 WHERE %s AND %s ORDER BY %s",
			EDB_BASE_EID,
			table,
			target,
			latest,
			EDB_BASE_EID);
		FREE(target);
	} else {
		asprintf((char **)&query,
			"SELECT %s FROM %s AS t1 WHERE %s ORDER BY %s",
			EDB_BASE_EID,
			table,
			latest,
			EDB_BASE_EID);
	}
	FREE(latest);
	res = edb_pgsql_query_exec(ec, query);
	if (PQresultStatus(res) != PGRES_TUPLES_OK || PQnfields(res) !=1) {
		edb_trace(ec, "edb_EIDs_get: %s\n", query);
		edb_trace_to_syslog(ec, "edb_EIDs_get: %s\n", query);
		edb_trace(ec, "edb_EIDs_get: %s", PQerrorMessage(ec->pg_conn->conn));
		PQclear(res);
		FREE(query);
		return 0;
	}
	*nvalue = PQntuples(res);
	*values = (int *)malloc(sizeof(int)*(*nvalue));
	FREE(query);

	for(i=0; i<*nvalue; ++i) {
		(*values)[i] = atoi(PQgetvalue(res, i, 0));
	}
	PQclear(res);

	return (*nvalue);
}


/*
	edb_get_eoids_from_eid--- get OID
*/
int
edb_get_eoids_from_eid(EdbContext ec, utf8_t *table, eid_t eid, int **values, int *nvalue)
{
	int	i;
	register PGresult	*res;
	utf8_t	*query;

	*values = NULL;
	*nvalue = 0;
	asprintf((char **)&query,
		"SELECT " EDB_XMLDB_COLUMN_EOID
		" FROM " EDB_XMLDB_TABLE
		" WHERE eid = %d"
		" ORDER BY " EDB_XMLDB_COLUMN_EOID " DESC",
		eid);
	res = edb_pgsql_query_exec(ec, query);
	if (PQresultStatus(res) != PGRES_TUPLES_OK || PQnfields(res) !=1) {
		edb_trace(ec, "query failed\n");
		edb_trace(ec, "%s", PQerrorMessage(ec->pg_conn->conn));
		PQclear(res);
		FREE(query);
		return 0;
	}
	*nvalue = PQntuples(res);
	*values = (int *)malloc(sizeof(int)*(*nvalue));
	FREE(query);

	for(i=0; i<*nvalue; ++i) {
		(*values)[i] = atoi(PQgetvalue(res, i, 0));
	}
	PQclear(res);

	return (*nvalue);
}

utf8_t *
edb_create_latest_phrase(EdbContext ec, utf8_t *tname, utf8_t *as1, utf8_t *as2, struct tm *tm, eid_t eid)
{
	utf8_t	*teid;
	utf8_t	*phrase;

	if(eid==0) {
		asprintf((char **)&teid, "%s.%s", as1, EDB_BASE_EID);
	} else {
		asprintf((char **)&teid, "%d", eid);
	}
	asprintf((char **)&phrase, "%s.%s = (SELECT MAX(%s.%s) FROM %s AS %s WHERE %s.%s = %s)",
		as1, EDB_BASE_EOID,
		as2, EDB_BASE_EOID,
		tname, as2,
		as2, EDB_BASE_EID,
		teid);
	FREE(teid);
#if 0
	if(tm) {	/* tm == NULL means current time */
		sprintf(&phrase[strlen(phrase)], 
			" AND %s.%s < '%04d-%02d-%02d %02d:%02d:%02d'::datetime",
			as2,	EDB_BASE_MTIME,
			tm->tm_year, tm->tm_mon, tm->tm_mday,
			tm->tm_hour, tm->tm_min, tm->tm_sec);
	}
#endif
	return phrase;
}

void
edb_vacuum(EdbContext ec)
{

	edb_pgsql_command_exec_wait(ec, (utf8_t *)"VACUUM FULL ANALYZE");
}

EdbCatalogue
edb_get_users(EdbContext ec)
{
	EdbTableInfo	ti;
	utf8_t	*query;
	utf8_t	*orderby;
	utf8_t	*order;
	int	ni;
	EdbCatalogue	ca_users;
	Link	l, l_xml, l_eid;

	ca_users = edb_catalogue_new();

	ti = edb_table_get_by_name(ec, (utf8_t *)EDB_TABLE_Person_XN);
	if(ti==NULL) return ca_users;

	if((order = edb_table_make_order(ec, ti, (utf8_t *)"t2"))!=NULL) {
		asprintf((char **)&orderby, " ORDER BY %s", order);
		FREE(order);
	} else {
		orderby = (utf8_t *)strdup("");
	}

	asprintf((char **)&query,
		"SELECT t1." EDB_XMLDB_COLUMN_XML " FROM " EDB_XMLDB_TABLE " AS t1, %s AS t2, %s AS t3"
		" WHERE t1." EDB_XMLDB_COLUMN_TEID "=%d"
			" AND t2." EDB_BASE_AVAILABLE
			" AND t3." EDB_BASE_AVAILABLE
			" AND t1." EDB_XMLDB_COLUMN_EOID "=t2." EDB_BASE_EOID
			" AND t1." EDB_XMLDB_COLUMN_EID "=t3." EDB_BASE_EID
		" %s",
		TABLE_SN(ti), EDB_AUTH_TABLE,
		TABLE_EID(ti),
		orderby);
	FREE(orderby);

	l_xml = edb_pgsql_query_one_field(ec, query, &ni);
	FREE(query);
	if(l_xml==NULL) return ca_users;

	l_eid = NULL;
	ForLink(l, l_xml) {
		EdbTupleInfo	tuple;

		tuple = edb_tuple_parse_xml(ec, ti, (utf8_t *)l->d);
		tuple = edb_eid_list_store(ec, TupleEID(tuple), EDB_EID_IS_TUPLE, tuple);
		LinkAppend(l_eid, link_atom_new_eid2s(TupleEID(tuple)));
	}
	EdbLinkFree(l_xml);

	ForLink(l, l_eid) {
		eid_t	eid = ATOI(l->d);
		EdbTupleInfo	tuple = edb_tuple_get(ec, eid, 0);

		if(tuple) edb_catalogue_append(ca_users, TupleEID(tuple), tuple);
	}
	EdbLinkFree(l_eid);

	return ca_users;
}

void
edb_sharable_text_mode(EdbContext ec, int mode)
{

	ec->edb_datum_text_sharable = mode;
}


int	_edb_delayed_free_mode = 0;
void	*_edb_delayed_free_bitree = NULL;
void	*_edb_delayed_free_link_atom = NULL;
void	*_edb_delayed_free_tuple_column = NULL;
void	*_edb_delayed_free_datum = NULL;
void	*_edb_delayed_free_index = NULL;
void	*_edb_delayed_free_xmldb = NULL;
int	edb_delayed_free_objects_count = 0;
int	edb_delayed_free_executed_count = 0;
int	edb_delayed_free_recycle_count = 0;

void
edb_delayed_free_mode(EdbContext ec, int mode)
{

	if(mode) {
		if(ec->delayed_free) return;
		ec->delayed_free = TRUE;
		++_edb_delayed_free_mode;
	} else {
		if(!ec->delayed_free) return;
		ec->delayed_free = FALSE;
		--_edb_delayed_free_mode;
		if(_edb_delayed_free_mode<=0) {
			_edb_delayed_free_mode = 0;
			edb_delayed_free_exec(ec, 0);
		} else {
			while(ec->delayed_free_tuple) edb_delayed_free_exec(ec, 1);
			while(ec->l_delayed_link_free2) edb_delayed_free_exec(ec, 1);
		}
	}
}

int
edb_delayed_free_exec(register EdbContext ec, register int num)
{
	register void	*p;
	register int	n = 0;

	++edb_delayed_free_executed_count;
	while(num==0 || n<num) {
		if(ec->delayed_free_tuple) {
			EdbDelayedFreePop(ec->delayed_free_tuple, p);
			edb_tuple_free_exec(ec, (EdbTupleInfo)p);
			++n;
		} else {
			break;
		}
	}
	while((num==0 || n<num) && ec->l_delayed_link_free2) {
		register Link	l = ec->l_delayed_link_free2;

		ec->l_delayed_link_free2 = ec->l_delayed_link_free2->n;
		if(ec->l_delayed_link_free2) ec->l_delayed_link_free2->p = NULL;
		l->n = NULL;
		link_atom_free2(ec, l);
		++n;
	}
	while(num==0 || n<num) {
		register int	st = FALSE;

		if(_edb_delayed_free_bitree) {
			EdbDelayedFreePop(_edb_delayed_free_bitree, p);
			free(p);
			++n;
			st = TRUE;
		}
		if(_edb_delayed_free_tuple_column) {
			EdbDelayedFreePop(_edb_delayed_free_tuple_column, p);
			Edb_free(p);
			++n;
			st = TRUE;
		}
		if(_edb_delayed_free_link_atom) {
			EdbDelayedFreePop(_edb_delayed_free_link_atom, p);
			free(p);
			++n;
			st = TRUE;
		}
		if(_edb_delayed_free_datum) {
			EdbDelayedFreePop(_edb_delayed_free_datum, p);
			Edb_free(p);
			++n;
			st = TRUE;
		}
		if(_edb_delayed_free_index) {
			EdbDelayedFreePop(_edb_delayed_free_index, p);
			Edb_free(p);
			++n;
			st = TRUE;
		}
		if(_edb_delayed_free_xmldb) {
			EdbDelayedFreePop(_edb_delayed_free_xmldb, p);
			Edb_free(p);
			++n;
			st = TRUE;
		}
		if(!st) break;
	}
#if 0
	if(num==0 && n>0) edb_trace(ec, "edb_delayed_free_exec: %d objects were freed.\n", n);
#endif
	edb_delayed_free_objects_count += n;
	return n;
}

static
void
edb_process_lock(EdbContext ec)
{

	if(!ec->process_initialized) return;
	pthread_mutex_lock(&ec->process_mutex);
}

static
void
edb_process_unlock(EdbContext ec)
{

	if(!ec->process_initialized) return;
	pthread_mutex_unlock(&ec->process_mutex);
}

void
edb_process_initialize(EdbContext ec)
{
	size_t	size;
	int		policy;

	if(!ec->process_initialized) {
		pthread_mutexattr_init(&ec->process_mutexattr);
		pthread_mutexattr_setprotocol(&ec->process_mutexattr, PTHREAD_PRIO_NONE);
		pthread_mutex_init(&ec->process_mutex, &ec->process_mutexattr);
		pthread_cond_init(&ec->process_cond, NULL);
		pthread_attr_init(&ec->process_attr);
		pthread_attr_setstacksize(&ec->process_attr, (size_t)0x1000000);

		pthread_attr_getschedpolicy(&ec->process_attr, &policy);
//		edb_trace(ec, "Policy: %d %d\n", policy, SCHED_FIFO);
		policy = SCHED_FIFO;
		pthread_attr_setschedpolicy(&ec->process_attr, policy);
//		pthread_attr_getschedpolicy(&ec->process_attr, &policy);
//		edb_trace(ec, "Policy: %d %d\n", policy, SCHED_FIFO);

		ec->process_initialized = TRUE;
		ec->pg_connections[0].thread = pthread_self();
//		edb_trace(ec, "Main thread: %08x\n", ec->pg_connections[0].thread);
//		edb_trace(ec, "Thread-Safe: %d\n", PQisthreadsafe());
		ec->pg_connections[0].state = EDB_PROCESS_EXECUTING;
		pthread_mutex_lock(&ec->process_mutex);
	}
//	pthread_attr_destroy(&ec->thread_attr);
//	pthread_cond_destroy(&ec->process_cond);
//	pthread_mutex_destroy(&ec->process_mutex);
}

struct PROCESS_CONTEXT {
	EdbContext		ec;
	EdbPGConn		pgc;
	EdbProcessFunc	process;
	void				*data;
};
typedef struct PROCESS_CONTEXT *ProcContext;

#if defined(__DARWIN__)
static void pthread_yield() {}
#endif

static
void *
edb_process_starter(void *ctxt)
{
	struct PROCESS_CONTEXT	*pctxt = ctxt;
	EdbContext	ec = pctxt->ec;
	EdbPGConn	pgc = pctxt->pgc;

	edb_process_lock(ec);
	++pgc->number_of_executed;
	ec->pg_conn = pgc;
	ec->pg_conn->number_of_switched = 0;
	(*pctxt->process)(pctxt->data);
	pgc->state = EDB_PROCESS_TERMINATED;
	pthread_cond_signal(&ec->process_cond);
	edb_process_unlock(ec);
	pthread_yield();
	FREE(pctxt);
	pthread_exit(NULL);
	return NULL;
}

void
edb_process_create(EdbContext ec, EdbProcessFunc process, void *data)
{
	int	i, executed;
	EdbPGConn pgc;

	if(edb_process_number_of_usable_processes(ec)<=0) {
		(*process)(data);
		return;
	}
	executed = FALSE;
	while(!executed) {
		edb_process_number_of_idle_processes(ec);
		for(i=1; i<ec->number_of_pg_connections; ++i) {
			ProcContext	pctxt;

			++ec->process_last_created;
			if((ec->process_last_created>=ec->number_of_pg_connections)
				|| (ec->process_last_created<1)) {
				ec->process_last_created = 1;
			}
			pgc = &ec->pg_connections[ec->process_last_created];
			if(pgc->state==EDB_PROCESS_EXECUTING) continue;
			if(pgc->state==EDB_PROCESS_TERMINATED) {
				pthread_join(pgc->thread, NULL);
				pgc->state = EDB_PROCESS_IDLE;
			}

			if(pgc->age!=ec->age) continue;

			pctxt = (ProcContext)malloc(sizeof(struct PROCESS_CONTEXT));
			bzero(pctxt, sizeof(struct PROCESS_CONTEXT));
			pctxt->ec = ec;
			pctxt->pgc = pgc;
			pgc->thread = NULL;
			pgc->state = EDB_PROCESS_EXECUTING;
			pctxt->process = process;
			pctxt->data = data;
			if(pthread_create(&pgc->thread, &ec->process_attr, edb_process_starter, pctxt)!=0) {
				edb_trace(ec, "edb_process_create: thread creation is failed.\n");
				pgc->state = EDB_PROCESS_IDLE;
				FREE(pctxt);
				continue;
			}
			executed = TRUE;
			break;
		}
		edb_process_switch(ec, edb_delayed_free_exec(ec, 1024)<=0);
	}
}

#if defined(PG_VERSION_NUM) && PG_VERSION_NUM>=80000
static
PGresult *
edb_pgsql_get_result(EdbContext ec)
{
	register PGresult   *res = NULL;

	if(edb_process_number_of_active_processes(ec)==0) {
		PQconsumeInput(ec->pg_conn->conn);
		edb_pgsql_wait_for_ready(ec, DELAYED_FREE_SIZE);
		res = PQgetResult(ec->pg_conn->conn);
	} else if(ec->process_initialized && PQisthreadsafe()) {
		EdbPGConn	pgc = ec->pg_conn;

		++pgc->number_of_switched;
		pthread_cond_signal(&ec->process_cond);
		edb_process_unlock(ec);
		res = PQgetResult(pgc->conn);
		edb_process_lock(ec);
		ec->pg_conn = pgc;
	} else {
		edb_pgsql_wait_for_ready2(ec);
		res = PQgetResult(ec->pg_conn->conn);
	}
	return res;
}
#else	/* for PostgreSQL 7.4 */
static
PGresult *
edb_pgsql_get_result(EdbContext ec)
{
	register PGresult   *res = NULL;

	if(edb_process_number_of_active_processes(ec)==0) {
		PQconsumeInput(ec->pg_conn->conn);
		edb_pgsql_wait_for_ready(ec, DELAYED_FREE_SIZE);
	} else {
		edb_pgsql_wait_for_ready2(ec);
	}
	return PQgetResult(ec->pg_conn->conn);
}
#endif

#define	PROCESS_IDLING_UTIME		5	// in micro-second

void
edb_process_switch(EdbContext ec, int waiting)
{
	EdbPGConn	pgc;

	if(!ec->process_initialized) return;
	if(ec->number_of_pg_connections<=1) return;
	++ec->process_number_of_switched;
	if(pthread_cond_signal(&ec->process_cond)!=0) edb_trace(ec, "pthread_cond_signal: error.\n");
	if(edb_process_number_of_active_processes(ec)>0) {
		pgc = ec->pg_conn;
		++pgc->number_of_switched;
		if(waiting) {
#if 1
			struct timeval	tv;
			struct timespec ts;

			gettimeofday(&tv, NULL);
			ts.tv_sec = tv.tv_sec;
			ts.tv_nsec = (tv.tv_usec+PROCESS_IDLING_UTIME)*1000L;
			if(ts.tv_nsec>=1000*1000*1000L) {
				++ts.tv_sec;
				ts.tv_nsec -= 1000*1000*1000L;
			}
			pthread_cond_timedwait(&ec->process_cond, &ec->process_mutex, &ts);
//			pthread_cond_wait(&ec->process_cond, &ec->process_mutex);
#else
			edb_process_unlock(ec);
			pthread_yield();
			edb_process_lock(ec);
#endif
		} else {
			edb_process_unlock(ec);
			pthread_yield();
			edb_process_lock(ec);
		}
		ec->pg_conn = pgc;
	}
}

void
edb_process_waitall(EdbContext ec)
{
	int	i;

	if(!ec->process_initialized) return;
	if(ec->number_of_pg_connections<=1) return;
	if(!pthread_equal(ec->pg_connections[0].thread, pthread_self())) return;
	for(i=1; i<ec->number_of_pg_connections; ++i) {
		EdbPGConn pgc = &ec->pg_connections[i];

		while(pgc->state!=EDB_PROCESS_IDLE) {
			if(pgc->state==EDB_PROCESS_TERMINATED) {
				pthread_join(pgc->thread, NULL);
				pgc->state = EDB_PROCESS_IDLE;
				break;
			}
			edb_process_switch(ec, TRUE);
		}
	}
}

int
edb_process_number_of_active_processes(EdbContext ec)
{
	int	i, num;

	if(!ec->process_initialized) return 0;
	num = 0;
	for(i=1; i<ec->number_of_pg_connections; ++i) {
		EdbPGConn pgc = &ec->pg_connections[i];

		switch(pgc->state) {
		case EDB_PROCESS_IDLE:
			break;
		case EDB_PROCESS_EXECUTING:
			++num;
			break;
		case EDB_PROCESS_TERMINATED:
			if(pthread_equal(ec->pg_connections[0].thread, pthread_self())) {
				pthread_join(pgc->thread, NULL);
				pgc->state = EDB_PROCESS_IDLE;
			} else {
				++num;
			}
			break;
		}
	}
	return num;
}

int
edb_process_number_of_usable_processes(EdbContext ec)
{
	int	i, num;

	if(!ec->process_initialized) return 0;
	num = 0;
	for(i=1; i<ec->number_of_pg_connections; ++i) {
		EdbPGConn pgc = &ec->pg_connections[i];

		if(pgc->age!=ec->age) continue;
		++num;
	}
	return num;
}

int
edb_process_number_of_idle_processes(EdbContext ec)
{
	int	i, num;

	if(!ec->process_initialized) return 0;
	num = 0;
	for(i=1; i<ec->number_of_pg_connections; ++i) {
		EdbPGConn pgc = &ec->pg_connections[i];

		if(pgc->age!=ec->age) continue;
		if(pgc->state==EDB_PROCESS_EXECUTING) continue;
		if(pgc->state==EDB_PROCESS_TERMINATED) {
			if(pthread_equal(ec->pg_connections[0].thread, pthread_self())) {
				pthread_join(pgc->thread, NULL);
				pgc->state = EDB_PROCESS_IDLE;
				++num;
			}
		} else {
			++num;
		}
	}
	return num;
}


