/*
	$Id: eid.c,v 1.38 2012/04/03 03:51:13 alex Exp alex $
	eid.c--- eid info
*/

#ifndef	lint
static char *rcsid __attribute__((unused)) = "$Id: eid.c,v 1.38 2012/04/03 03:51:13 alex Exp alex $";
#endif

#include	<stdio.h>
#include	<stdarg.h>
#include	<string.h>
#include	"edb/edb.h"

/*
	edb_eid_list_store
	edb_eid_list_restore
*/
struct	_EID_OBJ	{
	eid_t			eid;
	eid_t			teid;
	void			*obj;
	EdbIndexInfo	ei;
};

void *
edb_eid_list_store(EdbContext ec, eid_t eid, int type, void *obj)
{
	struct _EID_OBJ	*robj;

	robj = bitree_search(ec->bt_eid_list, NULL, eid, NULL);
	if(robj) {
		if(robj->obj && obj != robj->obj) {
			if(type==EDB_EID_IS_TABLE) {
				edb_table_free(ec, obj);
			} else {
				edb_tuple_free(ec, obj);
			}
		} else {
			robj->obj = obj;
			if(type==EDB_EID_IS_TABLE) {
				robj->teid = 0;
			} else {
				robj->teid = TABLE_EID(TupleTable((EdbTupleInfo)robj->obj));
				if(robj->ei==NULL) robj->ei = edb_tuple_make_eid_index(ec, (EdbTupleInfo)robj->obj);
			}
		}
		robj->eid = eid;
		return robj->obj;
	} else {
		robj = (struct _EID_OBJ *)malloc(sizeof(struct _EID_OBJ));
		bzero(robj, sizeof(struct _EID_OBJ));
		robj->eid = eid;
		robj->obj = obj;
		if(type==EDB_EID_IS_TABLE) {
			robj->teid = 0;
			if(robj->ei==NULL) {
				robj->ei = edb_index_new(ec);
				bcopy(TABLE_Base((EdbTableInfo)robj->obj), &robj->ei->base, sizeof(struct EDB_BASE_INFO));
			}
		} else {
			robj->teid = TABLE_EID(TupleTable((EdbTupleInfo)robj->obj));
			if(robj->ei==NULL) robj->ei = edb_tuple_make_eid_index(ec, (EdbTupleInfo)robj->obj);
		}
		BiTreeAdd(ec->bt_eid_list, robj, eid, NULL, NULL, NULL);
		return obj;
	}
}

void	*
edb_eid_list_restore(EdbContext ec, eid_t eid)
{
	struct _EID_OBJ	*robj;

	robj = bitree_search(ec->bt_eid_list, NULL, eid, NULL);
	if(robj) {
		if(robj->obj) ++ec->number_of_list_hit;
		return robj->obj;
	}
	return NULL;
}

static
void
edb_eid_list_free(EdbContext ec, struct _EID_OBJ *robj)
{
	if(robj==NULL) return;
	if(robj->obj) {
		if(robj->teid==0) {
#if 0
			edb_table_free(ec, (EdbTableInfo)robj->obj);
#else
			LinkAppend(ec->l_garbage2, link_atom_new(robj->obj, edb_table_free));
#endif
		} else {
			edb_tuple_free(ec, (EdbTupleInfo)robj->obj);
		}
	}
	if(robj->ei) edb_index_free(ec, robj->ei);
	FREE(robj);
}

eid_t
edb_eid_list_remove(EdbContext ec, eid_t eid)
{
	struct _EID_OBJ	*robj, *rrobj;

	robj = bitree_search(ec->bt_eid_list, NULL, eid, NULL);
	if(robj == NULL) return eid;
	BiTreeRemove(ec->bt_eid_list, robj, eid, NULL, (BiTree_obj_t *)&rrobj);
	robj = rrobj;
	if(robj) edb_eid_list_free(ec, robj);
	return eid;
}

EdbIndexInfo
edb_eid_list_store_index(EdbContext ec, eid_t eid, EdbIndexInfo ei)
{
	struct _EID_OBJ	*robj;

	robj = bitree_search(ec->bt_eid_list, NULL, eid, NULL);
	if(robj) {
		if(robj->ei && ei!=robj->ei) {
			edb_index_free(ec, ei);
			ei = robj->ei;
		} else {
			robj->ei = ei;
		}
	} else {
		robj = (struct _EID_OBJ *)malloc(sizeof(struct _EID_OBJ));
		bzero(robj, sizeof(struct _EID_OBJ));
		robj->eid = eid;
		robj->teid = ei->teid;
		robj->ei = ei;
		BiTreeAdd(ec->bt_eid_list, robj, eid, NULL, NULL, NULL);
	}
	return ei;
}

EdbIndexInfo
edb_eid_list_restore_index(EdbContext ec, eid_t eid)
{
	struct _EID_OBJ	*robj;

	robj = bitree_search(ec->bt_eid_list, NULL, eid, NULL);
	if(robj) {
		if(robj->ei) ++ec->number_of_list_hit;
		return robj->ei;
	}
	return NULL;
}

static
int
eid_list_purge_callback(register void *o1, register void *o2)
{
	struct _EID_OBJ *robj = o1;
	BiTree *bt_eid = o2;

	if(robj->teid==0) {
		struct _EID_OBJ *nobj = (struct _EID_OBJ *)malloc(sizeof(struct _EID_OBJ));
		bcopy(robj, nobj, sizeof(struct _EID_OBJ));
		BiTreeAdd(*bt_eid, nobj, nobj->eid, NULL, NULL, NULL);
		robj->obj = NULL;
		robj->ei = NULL;
	}
	return TRUE;
}

void
edb_eid_list_purge(EdbContext ec)
{
	BiTree	new_tree = NULL;

	bitree_callback(ec->bt_eid_list, eid_list_purge_callback, (void *)&new_tree);
	bitree_free2(ec->bt_eid_list, (void (*)())edb_eid_list_free, ec);
	ec->bt_eid_list = new_tree;
}

void
edb_eid_clear_auth_check(EdbContext ec)
{

}

