/*
	$Id: html.c,v 1.41 2011/04/18 09:45:28 alex Exp $
	html.c--- HTML misc command
*/

#ifndef	lint
static char *rcsid __attribute__((unused)) = "$Id: html.c,v 1.41 2011/04/18 09:45:28 alex Exp $";
#endif

#include	<stdio.h>
#include	<stdlib.h>
#include	<stdarg.h>
#include	<unistd.h>
#include	<sys/types.h>
#include	<ctype.h>
#include	<string.h>
#include	<signal.h>
#include	"edb/edb.h"
#include	"edb/html.h"

char *
html_make_icon_img_tag(EdbContext ec, char *icon, char *fmt, ...)
{
	va_list	ap;
	char	*s, *comment, *imgtag;

	if(fmt) {
		va_start(ap, fmt);
		vasprintf(&comment, fmt, ap);
		va_end(ap);
		for(s=comment; *s; ++s) {
			if(!(isalnum(*s) || *s == ' ' || ((*s)&0x80))) *s = '.';
		}
		asprintf(&imgtag, "<img src=\"%s\" border=\"0\" alt=\"%s\" />", icon, comment);
		FREE(comment);
	} else {
		asprintf(&imgtag, "<img src=\"%s\" border=\"0\" />", icon);
	}
	return imgtag;
}

char *
html_make_target(EdbContext ec, eid_t eid, char *fmt, ...)
{
	va_list	ap;
	char	*extension;
	char	*target;

	if(fmt) {
		va_start(ap, fmt);
		vasprintf(&extension, fmt, ap);
		va_end(ap);
#if 0
		asprintf(&target, "target=\"EDB:EID=%d:%s\"", eid, extension);
#else
		asprintf(&target, "target=\"EID%d%s\"", eid, extension);
#endif
		FREE(extension);
	} else {
#if 0
		asprintf(&target, "target=\"EDB:EID=%d\"", eid);
#else
		asprintf(&target, "target=\"EID%d\"", eid);
#endif
	}
	return target;
}

void
html_FORM_begin(EdbContext ec, Doc doc, char *url, eid_t eid, char *fmt, ...)
{
	va_list	ap;
	char	*target;
	char	*extension;

	if(fmt) {
		va_start(ap, fmt);
		vasprintf(&extension, fmt, ap);
		va_end(ap);
#if 0
		asprintf(&target, "target=\"EDB:EID=%d:%s\"", eid, extension);
#else
		asprintf(&target, "target=\"EID%d%s\"", eid, extension);
#endif
		FREE(extension);
	} else {
#if 0
		asprintf(&target, "target=\"EDB:EID=%d\"", eid);
#else
		target = strdup("");
#endif
	}
	doc_printf(doc, "<form method=\"POST\" action=\"%s\"", url);
	if(target && strlen(target)>0) doc_printf(doc, " %s", target);
	doc_printf(doc, ">\n");
	FREE(target);
}

void
html_FORM_encode_begin(EdbContext ec, Doc doc, char *url, eid_t eid, char *enctype)
{

	doc_printf(doc, "<form method=\"POST\" action=\"%s\" enctype=\"%s\">", url, enctype);
}

void
html_FORM_end(EdbContext ec, Doc doc)
{

	doc_printf(doc, "</form>\n");
}

#if 0
void
html_form_INPUT(EdbContext ec, Doc doc, char *type, char *name, char *fmt, ...)
{
	va_list	ap;
	char	*value;

	doc_printf(doc, "<input type=\"%s\"", type);
	if(name) doc_printf(doc, " name=\"%s\"", name);
	if(fmt) {
		va_start(ap, fmt);
		vasprintf(&value, fmt, ap);
		va_end(ap);
		doc_text_puts3(doc, " value=\"", value, "\"");
		FREE(value);
	}
	doc_printf(doc, " />\n");
}
#else
void
html_form_INPUT(EdbContext ec, Doc doc, char *type, char *name, char *fmt, ...)
{
	va_list	ap;
	char	*value, *s;

	doc_printf(doc, "<input type=\"%s\"", type);
	if(name) doc_printf(doc, " name=\"%s\"", name);
	if(fmt) {
		va_start(ap, fmt);
		vasprintf(&value, fmt, ap);
		va_end(ap);
		doc_puts(doc, " value=\"");
		for(s=value; *s; ++s) {
			char	c = *s;
			switch(c) {
			case '&':	doc_puts(doc, "&amp;");	break;
			case '"':	doc_puts(doc, "&quot;");	break;
			case '<':	doc_puts(doc, "&lt;");	break;
			case '>':	doc_puts(doc, "&gt;");	break;
			case '\t':	doc_puts(doc, " ");		break;
			case '\n':	doc_puts(doc, "\n");		break;
			default:	doc_putc(doc, c);		break;
			}
		}
		doc_puts(doc, "\"");
		FREE(value);
	}
	doc_printf(doc, " />\n");
}
#endif

void
html_form_input_TEXT(EdbContext ec, Doc doc, int size, char *name, char *fmt, ...)
{
	va_list	ap;
	char	*value, *s;

	doc_printf(doc, "<input type=\"TEXT\" name=\"%s\"", name);
	if(fmt) {
		va_start(ap, fmt);
		vasprintf(&value, fmt, ap);
		va_end(ap);
		doc_puts(doc, " value=\"");
		for(s=value; *s; ++s) {
			char	c = *s;
			switch(c) {
			case '&':	doc_puts(doc, "&amp;");	break;
			case '"':	doc_puts(doc, "&quot;");	break;
			case '<':	doc_puts(doc, "&lt;");	break;
			case '>':	doc_puts(doc, "&gt;");	break;
			case '\t':	doc_puts(doc, " ");		break;
			case '\n':	doc_puts(doc, "\n");		break;
			default:	doc_putc(doc, c);		break;
			}
		}
		doc_puts(doc, "\"");
		FREE(value);
	}
	if(size>0) doc_printf(doc, " size=\"%d\"", size);
	doc_printf(doc, " />\n");
}

void
html_form_readonly_TEXT(EdbContext ec, Doc doc, int size, char *name, char *fmt, ...)
{
	va_list	ap;
	char	*value, *s;

	doc_printf(doc, "<input type=\"TEXT\" name=\"%s\"", name);
	if(fmt) {
		va_start(ap, fmt);
		vasprintf(&value, fmt, ap);
		va_end(ap);
		doc_puts(doc, " value=\"");
		for(s=value; *s; ++s) {
			char	c = *s;
			switch(c) {
			case '&':	doc_puts(doc, "&amp;");	break;
			case '"':	doc_puts(doc, "&quot;");	break;
			case '<':	doc_puts(doc, "&lt;");	break;
			case '>':	doc_puts(doc, "&gt;");	break;
			case '\t':	doc_puts(doc, " ");		break;
			case '\n':	doc_puts(doc, "\n");		break;
			default:	doc_putc(doc, c);		break;
			}
		}
		doc_puts(doc, "\"");
		FREE(value);
	}
	if(size>0) doc_printf(doc, " size=\"%d\"", size);
	doc_printf(doc, " readonly=\"readonly\" />\n");
}

void
html_form_TEXTAREA(EdbContext ec, Doc doc, int row, int col, char *name, char *fmt, ...)
{
	va_list	ap;
	char	*value, *s;

	doc_printf(doc, "<textarea name=\"%s\"", name);
	if(row>0) doc_printf(doc, " rows=\"%d\"", row);
	if(col>0) doc_printf(doc, " cols=\"%d\"", col);
	doc_printf(doc, " wrap=\"virtual\"");
	doc_puts(doc, ">\n");
	if(fmt) {
		va_start(ap, fmt);
		vasprintf(&value, fmt, ap);
		va_end(ap);
		for(s=value; *s; ++s) {
			char	c = *s;
			switch(c) {
			case '&':	doc_puts(doc, "&amp;");	break;
			case '"':	doc_puts(doc, "&quot;");	break;
			case '<':	doc_puts(doc, "&lt;");	break;
			case '>':	doc_puts(doc, "&gt;");	break;
			case '\t':	doc_puts(doc, " ");		break;
			case '\n':	doc_puts(doc, "\n");		break;
			default:	doc_putc(doc, c);		break;
			}
		}
		FREE(value);
	}
	doc_printf(doc, "</textarea>\n");
}

void
html_form_readonly_TEXTAREA(EdbContext ec, Doc doc, int row, int col, char *name, char *fmt, ...)
{
	va_list	ap;
	char	*value, *s;

	doc_printf(doc, "<textarea name=\"%s\"", name);
	if(row>0) doc_printf(doc, " rows=\"%d\"", row);
	if(col>0) doc_printf(doc, " cols=\"%d\"", col);
	doc_printf(doc, " wrap=\"virtual\"");
	doc_puts(doc, " readonly=\"readonly\">\n");
	if(fmt) {
		va_start(ap, fmt);
		vasprintf(&value, fmt, ap);
		va_end(ap);
		for(s=value; *s; ++s) {
			char	c = *s;
			switch(c) {
			case '&':	doc_puts(doc, "&amp;");	break;
			case '"':	doc_puts(doc, "&quot;");	break;
			case '<':	doc_puts(doc, "&lt;");	break;
			case '>':	doc_puts(doc, "&gt;");	break;
			case '\t':	doc_puts(doc, " ");		break;
			case '\n':	doc_puts(doc, "\n");		break;
			default:	doc_putc(doc, c);		break;
			}
		}
		FREE(value);
	}
	doc_printf(doc, "</textarea>\n");
}


void
html_form_input_PASSPHRASE(EdbContext ec, Doc doc, char *name)
{

	doc_printf(doc, "<input type=\"PASSWORD\" name=\"%s\" size=\"48\" />\n", name);
}

void
html_form_input_RADIO(EdbContext ec, Doc doc, char *name, int selected, char *fmt, ...)
{
	va_list	ap;
	char	*value;

	doc_printf(doc, "<input type=\"radio\" name=\"%s\"", name);
	if(fmt) {
		va_start(ap, fmt);
		vasprintf(&value, fmt, ap);
		va_end(ap);
		doc_printf(doc, " value=\"%s\"", value);
		FREE(value);
	}
	if(selected) doc_printf(doc, " checked=\"checked\"");
	doc_printf(doc, " />\n");
}

void
html_form_SELECT_begin(EdbContext ec, Doc doc, char *name, int size, int multi)
{

	doc_printf(doc, "<select name=\"%s\"", name);
	if(size) doc_printf(doc, " size=\"%d\"", size);
	if(multi) doc_printf(doc, " multiple");
	doc_printf(doc, ">\n");
}

void
html_form_SELECT_end(EdbContext ec, Doc doc)
{

	doc_printf(doc, "</select>\n");
}

void
html_form_select_OPTION(EdbContext ec, Doc doc, int selected, int value, char *fmt, ...)
{
	va_list	ap;
	char	*caption;

	if(fmt) {
		va_start(ap, fmt);
		vasprintf(&caption, fmt, ap);
		va_end(ap);
	} else {
		caption = strdup("");
	}
	doc_printf(doc, "<option value=\"%d\"%s>", value, selected ? " selected=\"selected\"" : "");
	doc_text_puts(doc, caption);
	doc_printf(doc, "</option>\n");
	FREE(caption);
}

void
html_form_select_OPTION_string(EdbContext ec, Doc doc, int selected, char *value, char *fmt, ...)
{
	va_list	ap;
	char	*caption;

	if(fmt) {
		va_start(ap, fmt);
		vasprintf(&caption, fmt, ap);
		va_end(ap);
	} else {
		caption = strdup("");
	}
	doc_printf(doc, "<option value=\"%s\"%s>", value, selected ? " selected=\"selected\"" : "");
	doc_text_puts(doc, caption);
	doc_printf(doc, "</option>\n");
	FREE(caption);
}

void
html_form_select_ERA(EdbContext ec, Doc doc, char *name)
{

	html_form_SELECT_begin(ec, doc, name, 0, FALSE);
	html_form_select_OPTION(ec, doc, TRUE, 1, ec->japanese ? "西暦":"A.D.");
	html_form_select_OPTION(ec, doc, FALSE, 1867, ec->japanese ? "明治":"Meiji");
	html_form_select_OPTION(ec, doc, FALSE, 1912, ec->japanese ? "大正":"Taisho");
	html_form_select_OPTION(ec, doc, FALSE, 1926, ec->japanese ? "昭和":"Showa");
	html_form_select_OPTION(ec, doc, FALSE, 1989, ec->japanese ? "平成":"Heisei");
	html_form_select_OPTION(ec, doc, FALSE, 2019, ec->japanese ? "令和":"Reiwa");
	html_form_SELECT_end(ec, doc);
}

void
html_form_select_MONTH(EdbContext ec, Doc doc, char *name, int cvalue)
{
	int	m;
	static char	*mname_en[] = {
		"January", "February", "March", "April", "May", "June",
		"July", "August", "September", "October", "November", "December"
	};
	static char	*mname_jp[] __attribute__((unused)) = {
		"睦月", "如月", "弥生", "卯月", "皐月", "水無月",
		"文月", "葉月", "長月", "神無月", "霜月", "師走"
	};

	html_form_SELECT_begin(ec, doc, name, 0, FALSE);
	html_form_select_OPTION(ec, doc, cvalue==0, 0, "初月");
	for(m=1; m<=12; ++m) {
		if(ec->japanese) {
			html_form_select_OPTION(ec, doc, cvalue==m, m, "%d月", m);
		} else {
			html_form_select_OPTION(ec, doc, cvalue==m, m, "%s", mname_en[m-1]);
		}
	}
	html_form_select_OPTION(ec, doc, cvalue==99, 99, "末月");
	html_form_SELECT_end(ec, doc);
}

void
html_form_select_DAY(EdbContext ec, Doc doc, char *name, int cvalue)
{
	int	d;
	static char	*dname_en[] = {
		"1st", "2nd", "3rd", "4th", "5th", "6th", "7th", "8th", "9th", "10th",
		"11th", "12th", "13th", "14th", "15th", "16th", "17th", "18th", "19th", "20th",
		"21st", "22nd", "23rd", "24th", "25th", "26th", "27th", "28th", "29th", "30th",
		"31st"
	};

	html_form_SELECT_begin(ec, doc, name, 0, FALSE);
	html_form_select_OPTION(ec, doc, cvalue==0, 0, "初日");
	for(d=1; d<=31; ++d) {
		if(ec->japanese) {
			html_form_select_OPTION(ec, doc, cvalue==d, d, "%d日", d);
		} else {
			html_form_select_OPTION(ec, doc, cvalue==d, d, "%s", dname_en[d-1]);
		}
	}
	html_form_select_OPTION(ec, doc, cvalue==99, 99, "末日");
	html_form_SELECT_end(ec, doc);
}

void
html_form_input_FILE(EdbContext ec, Doc doc, int size, char *name, char *accept, char *fmt, ...)
{
	va_list	ap;
	char	*value;

	doc_printf(doc, "<input type=\"FILE\" name=\"%s\"", name);
	if(STRisValid(accept)) doc_printf(doc, " accept=\"%s\"", accept);
	if(fmt) {
		va_start(ap, fmt);
		vasprintf(&value, fmt, ap);
		va_end(ap);
		doc_printf(doc, " value=\"%s\"", value);
		FREE(value);
	}
	if(size>0) doc_printf(doc, " size=\"%d\"", size);
	doc_printf(doc, " />\n");
}

static
struct	PARSE_LIST {
	char	m;
	char	*bgn;
	char	*end;
} parse_list[] = {
	{	'_',	"<sub>",	"</sub>"	},
	{	'^',	"<sup>",	"</sup>"	},
	{	'M',	"<var>",	"</var>"	},
	{	'C',	"",		""		}
};

/*
	text_parse
	\?{text}
*/
static
void
html_text_parse(EdbContext ec, char **sp, FILE *fp)
{
	char	*s = *sp, *e, sc;
	utf8_t	*d;
	int	i, count;

	sc = s[0];
	for(i=0; i<Ns(parse_list); ++i) {
		if(sc == parse_list[i].m && s[1]=='{') {
			fputs(parse_list[i].bgn, fp);
			++*sp;
			++*sp;
			s = strdup(*sp);
			/* find corresponding {} */
			count = 0;
			for(e=s; *e; ++e, ++*sp) {
				if(*e=='{') {
					++count;
				} else if(*e=='}') {
					if(count==0) {
						*e++ = '\0';
						++*sp;
						break;
					}
					--count;
				}
			}
			if(sc == 'C') {
				d = edb_text_codename_to_html(ec, (utf8_t *)s);
				fputs((char *)d, fp);
			} else {
				html_text_puts(ec, s, fp);
			}
			FREE(s);
			fputs(parse_list[i].end, fp);
			return;
		}
	}
	fprintf(fp, "%c", sc);
	*sp = ++s;
}

/*
	\_{}	下付き
	\^{}	上付き
	\\	\
	\{	{
	\}	}
	\M	数式モード
	\C	文字モード
*/

int
html_text_puts(EdbContext ec, char *s, FILE *fp)
{
	register char	c;

	while((c=*s++)!='\0') {
		if(c=='\\') {
			html_text_parse(ec, &s, fp);
		} else {
			html_text_putc(ec, c, fp);
		}
	}
	return 0;
}

int
html_text_putc(EdbContext ec, char c, FILE *fp)
{
	switch(c) {
	case '&':	fputs("&amp;", fp);	break;
	case '"':	fputs("&quot;", fp);	break;
	case '<':	fputs("&lt;", fp);	break;
	case '>':	fputs("&gt;", fp);	break;
	case '\n':	fputs("<br />\n", fp);	break;
	default:	fputc(c, fp);		break;
	}
	return 0;
}

int
html_text_puts3(EdbContext ec, char *h, char *s, char *t, FILE *fp)
{

	if(h) fputs(h, fp);
	html_text_puts(ec, s, fp);
	if(t) fputs(t, fp);
	return 0;
}

int
html_message_put(EdbContext ec, char *msgs[], FILE *fp)
{
	char	**s;

	fprintf(fp, "<table border=\"5\" align=\"center\" width=\"100%%\"><tr><td style=\"font-size:smaller\">\n");
	for(s=msgs; *s; ++s) {
		if(strlen(*s)>0) {
			fprintf(fp, "<p>　");
			html_text_puts(ec, *s, fp);
			fputs("</p>\n", fp);
		}
	}
	fprintf(fp, "</td></tr></table>\n");
	return 0;
}

static
struct	MSG_CMD {
	char	*cmd;
	char	*value;

} msg_cmd[] = {
{	"WEB-PORTAL-URL",		WEB_PORTAL_URL		},
{	"ADMIN-MAILADDR",		EDB_ADMIN_MAILADDR	},
{	"NAME-OF-DB",			NAME_OF_DB			},
};

int
html_file_msg_put(EdbContext ec, char *name, FILE *fp)
{
	Link	lstr, l;
	char	*pathname;

	asprintf(&pathname, "%s/libdata/msg/%s", EDBDIR, name);
	lstr = file_load_as_linked_str(pathname, NULL);
	if(lstr) {
		fprintf(fp, "<table border=\"5\" align=\"center\" width=\"100%%\"><tr><td style=\"font-size:smaller\">\n");
		for(l=lstr; l; l=l->n) {
			char *s = (char *)l->d;

			if(strlen(s)>0) {
				char	*p;

				fprintf(fp, "<p>&nbsp;&nbsp;");
				s = STRDUP(s);
				while((p=strstr(s, "%%EDB:"))!=NULL) {
					int	i;

					*p = '\0';
					html_text_puts(ec, s, fp);
					p += 6;
					s = p;
					for(i=0; i<Ns(msg_cmd); ++i) {
						int	len;

						if(!strncmp(p, msg_cmd[i].cmd, len=strlen(msg_cmd[i].cmd))
							&& !strncmp(&p[len], "%%", 2)) {
							html_text_puts(ec, msg_cmd[i].value, fp);
							s = p + len + 2;
							break;
						}
					}
				}
				html_text_puts(ec, s, fp);
				fputs("</p>\n", fp);
			}
		}
		fprintf(fp, "</td></tr></table>\n");
		link_free(lstr);
	}
	FREE(pathname);
	return 0;
}

void
html_print_postscript(EdbContext ec, FILE *fp)
{
	char	buf[BUFSIZ];

	ctime_r(&ec->now, buf);
	str_oneline(buf);
	fprintf(fp, "<span style=\"font-size:smaller\">");
#if defined(EDB_ADMIN_MAILADDR_SHOW)
	fprintf(fp, "<i>%s &lt;%s&gt; at %s.</i>\n", EDB_WG_NAME, EDB_ADMIN_MAILADDR_SHOW, buf);
#else
	fprintf(fp, "<i>%s &lt;%s&gt; at %s.</i>\n", EDB_WG_NAME, EDB_ADMIN_MAILADDR, buf);
#endif
	fprintf(fp, "(%s:%s:%s v.%d.%d, o=%d/%d, c=%d, q=%d, h=%d)\n",
		ec->host, ec->port, ec->dbName, PG_SV_major(ec), PG_SV_minor(ec),
		ec->open_trials, link_atoms(ec->l_dbpath), ec->number_of_command, ec->number_of_query, ec->number_of_list_hit);
	fprintf(fp, "</span>");
}

int
edb_html_make_d_arg(EdbContext ec, char *name, int arg)
{

	sprintf(name, "D_ARG%d", arg++);
	return arg;
}

void
html_print_HEAD(EdbContext ec, char *title, char *refresh_url, int delay, int body, FILE *fp)
{
	char	*bgc = "white";

	fprintf(fp, "Content-Type: text/html\n\n");
	fprintf(fp, "<?xml version=\"1.0\" encoding=\"utf-8\" ?>\n");
	fprintf(fp, "<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Transitional//EN\" \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd\">\n");
	fprintf(fp, "<html xmlns=\"http://www.w3.org/1999/xhtml\" xml:lang=\"ja\" lang=\"ja\">\n");
	fprintf(fp, "	<head>\n");
	fprintf(fp, "		<meta http-equiv=\"Content-Type\" content=\"text/html; charset=utf-8\" />\n");
	fprintf(fp, "		<meta http-equiv=\"Content-Style-Type\" content=\"text/css; charset=utf-8\" />\n");
	fprintf(fp, "		<meta http-equiv=\"Pragma\" content=\"no-cache\" />\n");
	fprintf(fp, "		<meta http-equiv=\"Cache-Control\" content=\"no-cache\" />\n");
	fprintf(fp, "		<meta http-equiv=\"Expires\" content=\"Thu, 01 Dec 1994 16:00:00 GMT\" />\n");
	fprintf(fp, "		<meta http-equiv=\"X-UA-Compatible\" content=\"IE=10\" />\n");
#if defined(EDB_READER)
	fprintf(fp, "		<link rel=\"stylesheet\" type=\"text/css\" href=\"" WEB_PORTAL_URL "/edb-css/edb.css\" />\n");
#else
	fprintf(fp, "		<link rel=\"stylesheet\" type=\"text/css\" href=\"/edb-css/edb.css\" />\n");
#endif
	if(refresh_url) fprintf(fp, "<meta http-equiv=\"refresh\" content=\"%d; URL=%s\" />\n", delay, refresh_url);
	fprintf(fp, "		<title>%s</title>\n", title);
	fprintf(fp, "	</head>\n");

	if(body) {
		if(ec->operator!=0) {
			if(ec->Odin) {
				bgc = "#ff4040";	/* light red */
			} else if(ec->staff) {
				bgc = "#ffcccc";	/* pink */
			} else {
				bgc = "#ffffff";		/* white */
			}
		} else {
			if(ec->university) {
				bgc = "#ccffff";	/* skyblue (light) */
			} else {
				bgc = "#80ffff";	/* skyblue (dark) */
			}
		}
		fprintf(fp, "	<body style=\"color:black; background-color:%s; font-size:10pt;\">\n", bgc);
	}
}

void
html_print_FOOT(EdbContext ec, int body, FILE *fp)
{

	if(body) fprintf(fp, "	</body>\n");
	fprintf(fp, "</html>\n");
}

