/*
$Id: organization.c,v 1.8 2008/08/09 09:21:37 alex Exp $
	organization.c--- organization
*/

#ifndef lint
static char * rcsid __attribute__((unused)) = "$Id: organization.c,v 1.8 2008/08/09 09:21:37 alex Exp $";
#endif

#include	<stdio.h>
#include	<stdlib.h>
#include	<strings.h>
#include	"edb/edb.h"
#include	"edb/organization.h"

#define	DATE_INNER(d, f, t)	((f)<=(d) && (d)<=(t))

static
int
edb_organization_is_inside_func(EdbContext ec, EdbTupleInfo o, eid_t ancestor, int recurse)
{
	register EdbTupleColumn	tc;
	register EdbDatum		d;

	if(o==NULL || !TupleAvail(o)) return FALSE;
	if(!TupleIsOrganization(o)) return FALSE;
	if(!DATE_INNER(ec->present_date, TupleDateFrom(o), TupleDateTo(o))) return FALSE;
	if(TupleEID(o)==ancestor) return TRUE;
	if(ancestor==ORGAN_UNIV_EID) {
		if(o->tuple_location==TUPLE_LOCATION_INSIDE) return TRUE;
		if(o->tuple_location==TUPLE_LOCATION_OUTSIDE) return FALSE;
	}
	if(recurse<=0) return FALSE;

	tc = edb_tuple_seek_column_by_xmlname(ec, o, (utf8_t *)"@.upper");
	if(tc==NULL) return FALSE;
	ForTCDatum(d, tc) {
		register EdbTupleInfo		u;
		register EdbTupleColumn	mtc;
		register EdbDatum		md;

		if(!edb_datum_at_present(ec, d)) continue;
		if(!DatumEIDisVALID(d)) continue;

		u = edb_tuple_get(ec, DatumEID(d), 0);
		if(u==NULL) continue;
		if(!TupleIsOrganization(u)) continue;

		mtc = edb_tuple_seek_column_by_xmlname(ec, u, (utf8_t *)"@.member");
		if(mtc==NULL) continue;
		ForTCDatum(md, mtc) {
			if(!edb_datum_at_present(ec, md)) continue;
			if(!DatumEIDisVALID(md)) continue;
			if(DatumEID(md)==TupleEID(o)) {
				if(edb_organization_is_inside_func(ec, u, ancestor, recurse-1)) return TRUE;
				break;
			}
		}
	}
	return FALSE;
}

int
edb_organization_is_inside(EdbContext ec, EdbTupleInfo o, eid_t ancestor)
{

	if(o==NULL) return FALSE;
	if(!TupleIsOrganization(o)) return FALSE;
	if(ancestor==ORGAN_UNIV_EID) {
		register int	inside;

		switch(o->tuple_location) {
		case TUPLE_LOCATION_INSIDE: return TRUE;
		case TUPLE_LOCATION_OUTSIDE: return FALSE;
		default:
			inside = edb_organization_is_inside_func(ec, o, ancestor, ec->max_hierarchy_recurse);
			o->tuple_location = inside ? TUPLE_LOCATION_INSIDE : TUPLE_LOCATION_OUTSIDE;
			return inside;
		}
	} else {
		EdbTupleInfo	u;

		if(o->tuple_location==TUPLE_LOCATION_UNKNOWN) edb_organization_is_inside(ec, o, ORGAN_UNIV_EID);
		u = edb_tuple_get(ec, ORGAN_UNIV_EID, 0);
		if(u==NULL) return FALSE;
		if(o->tuple_location==TUPLE_LOCATION_INSIDE) {
			return edb_organization_is_inside_func(ec, u, ancestor, ec->max_hierarchy_recurse);
		} else if(o->tuple_location==TUPLE_LOCATION_OUTSIDE) {
			return !edb_organization_is_inside_func(ec, u, ancestor, ec->max_hierarchy_recurse);
		} else {
			return edb_organization_is_inside_func(ec, o, ancestor, ec->max_hierarchy_recurse);
		}
	}
}

#if 0
static
void
edb_organization_collect_persons_func(EdbContext ec, EdbCatalogue ca, EdbTupleInfo organ, int recurse)
{
	EdbTupleColumn	tc;
	EdbDatum	datum;

	if(!organ) return;
	if(--recurse<=0) return;
	tc = edb_tuple_seek_column_by_xmlname(ec, organ, (utf8_t *)"@.member");
	if(tc==NULL) return;
	ForTCDatum(datum, tc) {
		EdbTupleInfo		member;

		if(!DatumEIDisVALID(datum)) continue;
		if(!edb_datum_at_present(ec, datum)) continue;
		member = edb_tuple_get(ec, DatumEID(datum), 0);
		if(TupleIsOrganization(member)) {
			edb_organization_collect_persons_func(ec, ca, member, recurse);
		} else if(TupleIsPerson(member)) {
			edb_catalogue_append(ca, TupleEID(member), member);
		}
	}
	return;
}
#endif

static
void
edb_organization_collect_titled_persons_func(EdbContext ec, EdbCatalogue ca, EdbTupleInfo organ, EdbCatalogue ca_title,int recurse)
{
	EdbTupleColumn	tc;
	EdbDatum	datum;

	if(!organ) return;
	if(--recurse<=0) return;
	tc = edb_tuple_seek_column_by_xmlname(ec, organ, (utf8_t *)"@.member");
	if(tc==NULL) return;
	ForTCDatum(datum, tc) {
		EdbTupleInfo		member;

		if(!DatumEIDisVALID(datum)) continue;
		if(!edb_datum_at_present(ec, datum)) continue;
		member = edb_tuple_get(ec, DatumEID(datum), 0);
		if(TupleIsOrganization(member)) {
			edb_organization_collect_titled_persons_func(ec, ca, member, ca_title, recurse);
		} else if(TupleIsPerson(member)) {
			if(ca_title && edb_catalogue_size(ca_title)>0) {
				EdbTupleColumn	ttc;
				EdbDatum	td;

				ttc = edb_tuple_seek_TC_by_xmlname(ec, organ, tc, datum, (utf8_t *)"@.title");
				ForTCDatum(td, ttc) {
					if(!DatumEIDisValid(td)) continue;
					if(!edb_datum_at_present(ec, td)) continue;
					if(edb_catalogue_lookup(ca_title, DatumEID(td))) {
						edb_catalogue_append(ca, TupleEID(member), member);
						break;
					}
				}
			} else {
				edb_catalogue_append(ca, TupleEID(member), member);
			}
		}
	}
	return;
}

EdbCatalogue
edb_organization_collect_persons(EdbContext ec, EdbTupleInfo organ)
{
	EdbCatalogue	ca;

	if(!organ) return NULL;
	ca = edb_catalogue_new();
	edb_organization_collect_titled_persons_func(ec, ca, organ, NULL, ec->max_hierarchy_recurse);
	if(edb_catalogue_size(ca)==0) {
		edb_catalogue_free(ca);
		ca = NULL;
	}
	return ca;
}

EdbCatalogue
edb_organization_collect_titled_persons(EdbContext ec, EdbTupleInfo organ, EdbCatalogue ca_title)
{
	EdbCatalogue	ca;

	if(!organ) return NULL;
	ca = edb_catalogue_new();
	edb_organization_collect_titled_persons_func(ec, ca, organ, ca_title, ec->max_hierarchy_recurse);
	if(edb_catalogue_size(ca)==0) {
		edb_catalogue_free(ca);
		ca = NULL;
	}
	return ca;
}

static
int
edb_organization_person_is_belonged_func(EdbContext ec, EdbTupleInfo organ, EdbTupleInfo person)
{
	EdbTupleColumn	tc;
	EdbDatum	datum;

	if(!organ) return FALSE;
	tc = edb_tuple_seek_column_by_xmlname(ec, organ, (utf8_t *)"@.member");
	if(tc==NULL) return FALSE;
	ForTCDatum(datum, tc) {
		EdbTupleInfo		member;

		if(!DatumEIDisVALID(datum)) continue;
		if(!edb_datum_at_present(ec, datum)) continue;
		member = edb_tuple_get(ec, DatumEID(datum), 0);
		if(TupleIsPerson(member)  && TupleEID(member)==TupleEID(person)) return TRUE;
	}
	return FALSE;
}

int
edb_organization_person_is_belonged(EdbContext ec, EdbTupleInfo organ, EdbTupleInfo person)
{
	EdbTableInfo	organ_ti;
	utf8_t	*condition, *query;
	int	n, belong = FALSE;
	Link	l, l_eid;

	if(organ==NULL || person==NULL) return FALSE;
	if(edb_organization_person_is_belonged_func(ec, organ, person)) return TRUE;
	organ_ti = TupleTable(organ);
	asprintf((char **)&query, "SELECT %s FROM %s WHERE %s",
		EDB_BASE_EID,
		TABLE_SN(organ_ti),
		condition = edb_condition_make(ec, organ_ti, CONDITION_LOGIC_AND,
			ConditionEIDMatchNoExpand("@.upper", TupleEID(organ)),
			ConditionEIDMatch("@.member", TupleEID(person)),
			ConditionTerminate)
		);
	FREE(condition);
	l_eid = edb_pgsql_query_one_field(ec, query, &n);
	FREE(query);
	ForLink(l, l_eid) {
		eid_t	organ_eid = ATOI(l->d);
		EdbTupleInfo	organ_tuple = edb_tuple_get(ec, organ_eid, EDB_TUPLE_GET_template_table, organ_ti, 0);

		if(edb_organization_person_is_belonged_func(ec, organ_tuple, person)) {
			belong = TRUE;
			break;
		}
	}
	EdbLinkFree(l_eid);
	return belong;
}


