/*
$Id: personification.c,v 1.3 2012/04/03 03:51:13 alex Exp alex $
	personification.c--- personification
*/

#ifndef lint
static char * rcsid __attribute__((unused)) = "$Id: personification.c,v 1.3 2012/04/03 03:51:13 alex Exp alex $";
#endif

#include	<stdio.h>
#include	<stdlib.h>
#include	<strings.h>
#include	"edb/edb.h"
#include	"edb/personification.h"

utf8_t *
edb_personification_email(EdbContext ec, EdbTupleInfo tuple)
{

	return edb_person_email(ec, tuple);
}

utf8_t *
edb_personification_certificate(EdbContext ec, EdbTupleInfo tuple)
{
	Link	lc;

	if(tuple==NULL) return NULL;
	if(!TupleIsPersonification(tuple)) return NULL;
	ForTupleColumn(lc, tuple) {
		EdbTupleColumn	tc = (EdbTupleColumn)lc->d;

		if(tc->ci->typeInfo->basicTypeID == EDB_TYPE_CERTIFICATE && tc->datum) {
			EdbDatum	datum = (EdbDatum) tc->datum;

			if(DatumEnglishIsValid(datum)) return (utf8_t *)strdup((char *)DatumEnglish(datum));
		}
	}
	return NULL;
}

#define	ORGAN_COLUMN_CHIEF_XML	"organization.chief"
#define	ORGAN_COLUMN_MEMEBER_XML	"organization.member"
#define	ORGAN_COLUMN_MEMEBER_TITLE_XML	"organization.member.title"

static
void
edb_personification_check_position(EdbContext ec, EdbTupleInfo tuple)
{
	EdbTableInfo	organ_ti;
	utf8_t	*query, *condition, *condition2;
	Link	l, l_eid;
	int	n;

	if(tuple==NULL) return;
	if(!TupleIsPersonification(tuple)) return;

	organ_ti = edb_table_get_by_name(ec, (utf8_t *)EDB_TABLE_ORGANIZATION_XN);
	if(organ_ti==NULL) return;

	condition2 = edb_condition_make(ec, organ_ti, CONDITION_LOGIC_OR,
		ConditionEIDMatchNoExpand("@.member", TupleEID(tuple)),
		ConditionEIDMatchNoExpand("@.chief", TupleEID(tuple)),
		ConditionTerminate);
	asprintf((char **)&query, "SELECT %s FROM %s WHERE %s",
		EDB_BASE_EID,
		TABLE_SN(organ_ti),
		condition = edb_condition_make(ec, organ_ti, CONDITION_LOGIC_AND,
			ConditionEIDMatchNoExpand("@.upper", ORGAN_UNIV_EID),
			ConditionCondition(condition2),
			ConditionTerminate)
		);
	FREE(condition2);
	FREE(condition);
	l_eid = edb_pgsql_query_one_field(ec, query, &n);
	FREE(query);
	LinkAppend(l_eid, link_atom_new_spf("%d", ORGAN_UNIV_EID));
	ForLink(l, l_eid) {
		eid_t	organ_eid = ATOI(l->d);
		EdbTupleInfo	organ_tuple;
		EdbTupleColumn	tc;
		EdbDatum	d;

		if(!EIDisVALID(organ_eid)) continue;
		organ_tuple = edb_tuple_get(ec, organ_eid, 0);
		if(organ_tuple==NULL) continue;
		if(!TupleAvail(organ_tuple)) continue;
		if(!edb_organization_is_inside(ec, organ_tuple, ORGAN_UNIV_EID)) continue;
		tc = edb_tuple_seek_column_by_xmlname(ec, organ_tuple, (utf8_t *)ORGAN_COLUMN_CHIEF_XML);
		if(tc!=NULL) {
			ForTCDatum(d, tc) {
				if(!edb_datum_at_present(ec, d)) continue;
				if(DatumEID(d) == TupleEID(tuple)) tuple->tuple_auth_personnel = TRUE;
				if(tuple->tuple_auth_personnel) break;
			}
		}
		if(tuple->tuple_auth_personnel) break;
		tc = edb_tuple_seek_column_by_xmlname(ec, organ_tuple, (utf8_t *)ORGAN_COLUMN_MEMEBER_XML);
		if(tc!=NULL) {
			ForTCDatum(d, tc) {
				if(!edb_datum_at_present(ec, d)) continue;
				if(DatumEID(d) == TupleEID(tuple)) tuple->tuple_auth_personnel = TRUE;
				if(tuple->tuple_auth_personnel) break;
			}
		}
		if(tuple->tuple_auth_personnel) break;
	}
	EdbLinkFree(l_eid);
}

int
edb_personification_is_member(EdbContext ec, EdbTupleInfo tuple)
{

	if(tuple==NULL || !TupleAvail(tuple)) return FALSE;
	if(!tuple->tuple_auth_level_checked) {
		edb_personification_check_position(ec, tuple);
		tuple->tuple_auth_level_checked = TRUE;
	}
	return tuple->tuple_auth_personnel;
}

