/*
$Id: picture.c,v 1.15 2012/04/03 03:51:13 alex Exp alex $
	picture.c--- picture
	(C) Copyright 2002 by Alex
*/

#ifndef lint
static char * rcsid __attribute__((unused)) = "$Id: picture.c,v 1.15 2012/04/03 03:51:13 alex Exp alex $";
#endif

#include	<stdio.h>
#include	<stdlib.h>
#include	<strings.h>
#include	<setjmp.h>
#include	"edb/edb.h"
#include	"edb/base64.h"
#include	"edb/picture.h"

#define	PICTURE_ERROR	(-1)
#define	PICTURE_UNKNOWN	0
#define	PICTURE_JPEG	1
#define	PICTURE_PNG	2

#if 0
#include "jinclude.h"
#else
#include "jconfig.h"
#endif
#include "jpeglib.h"
#include "jerror.h"
#include	<setjmp.h>
#include "png.h"

static
int
edb_picture_body_type(EdbContext ec, unsigned char *body, int bytes)
{
	int	type = PICTURE_UNKNOWN;

	if(bytes>2
		&& body[0]==0xff
		&& body[1]==0xd8
		&& body[2]==0xff) {
		type = PICTURE_JPEG;
	} else if(bytes>3
		&& body[0]==0x89
		&& body[1]=='P'
		&& body[2]=='N'
		&& body[3]=='G') {
		type = PICTURE_PNG;
	}
	return type;
}

int
edb_picture_body_is_supported(EdbContext ec, unsigned char *body, int bytes)
{

	return (edb_picture_body_type(ec, body, bytes)>0);
}

static
int
edb_picture_type(EdbContext ec, EdbTupleInfo tuple)
{
	EdbTupleColumn	tc;
	int	type = PICTURE_UNKNOWN;
	EdbDatum	datum;

	if(tuple==NULL) return PICTURE_ERROR;
	tc = edb_tuple_seek_column_by_xmlname(ec, tuple, (utf8_t *)"picture.body");
	if(tc==NULL) tc = edb_tuple_seek_picture_column(ec, tuple);
	if(tc==NULL) return PICTURE_ERROR;

	if((datum=tc->datum) && DatumEnglishIsValid(datum)) {
		unsigned char	*body;
		int	len;

		body = base64_decode(DatumEnglish(datum), &len);
		type = edb_picture_body_type(ec, body, len);
		FREE(body);
	}
	return type;
}

char *
edb_picture_mimetype(EdbContext ec, EdbTupleInfo tuple)
{
	int	type;

	type = edb_picture_type(ec, tuple);
	if(type == PICTURE_JPEG) return "image/jpeg";
	if(type == PICTURE_PNG) return "image/png";
	return NULL;
}

struct my_error_mgr {
	struct jpeg_error_mgr	pub;
	jmp_buf				setjmp_buffer;
	EdbContext			ec;
};

typedef struct my_error_mgr *my_error_ptr;

static
void
edb_picture_error_exit(j_common_ptr cinfo)
{
	my_error_ptr myerr;

	myerr = (my_error_ptr) cinfo->err;
	(*cinfo->err->output_message)(cinfo);     /* display error message */
	longjmp(myerr->setjmp_buffer, 1);         /* return from error */
}

static
void
edb_picture_error_output(j_common_ptr cinfo)
{
	my_error_ptr myerr;
	char         buffer[JMSG_LENGTH_MAX];

	myerr = (my_error_ptr) cinfo->err;
	(*cinfo->err->format_message)(cinfo, buffer);

	edb_trace(myerr->ec, "edb_picture_error_output: %s\n", buffer);
}

static
EdbPicture
edb_picture_struct_new(EdbContext ec, unsigned char *body, int bytes)
{
	EdbPicture	picture;

	picture = (EdbPicture)malloc(sizeof(struct EDB_PICTURE));
	if(picture == NULL) return NULL;
	bzero(picture, sizeof(struct EDB_PICTURE));

	picture->body = (char *)malloc(bytes);
	memcpy(picture->body, body, bytes);
	picture->bytes = bytes;
	return picture;
}

void
edb_picture_free(EdbContext ec, EdbPicture picture)
{

	if(picture==NULL) return;
	FREE(picture->body);
	free(picture);
}

static
EdbPicture
edb_picture_make_picture_info_JPEG(EdbContext ec, EdbTupleInfo tuple, unsigned char *body, int bytes)
{
	EdbPicture	picture;
	struct jpeg_decompress_struct    cinfo;
//	struct jpeg_progress_mgr         prog;
	struct my_error_mgr              jerr;
	FILE	*fp;

	picture = edb_picture_struct_new(ec, body, bytes);
	if(picture == NULL) return NULL;

	picture->mimetype = "image/jpeg";

	fp = tmpfile();
	fwrite(body, 1, bytes, fp);
	rewind(fp);

	cinfo.err = jpeg_std_error(&jerr.pub);
	jerr.pub.error_exit     = edb_picture_error_exit;
	jerr.pub.output_message = edb_picture_error_output;
	jerr.ec = ec;

	if(setjmp(jerr.setjmp_buffer)) {
		fclose(fp);
		/* if we're here, it blowed up... */
		jpeg_destroy_decompress(&cinfo);
		if(picture) edb_picture_free(ec, picture);
		return NULL;
	}

	jpeg_create_decompress(&cinfo);

	jpeg_stdio_src(&cinfo, fp);
	(void) jpeg_read_header(&cinfo, TRUE);

	/* do various cleverness regarding decompression parameters & such... */
	jpeg_calc_output_dimensions(&cinfo);
	picture->xsize = cinfo.output_width;
	picture->ysize = cinfo.output_height;

#if 0
	jpeg_start_decompress(&cinfo);
	jpeg_finish_decompress(&cinfo);
#endif
	jpeg_destroy_decompress(&cinfo);
	fclose(fp);
	return picture;
}

static
void
edb_picture_png_error(png_structp png_ptr, png_const_charp message)
{
	EdbContext	ec;

	if(!png_ptr) return;
	ec = (EdbContext)png_get_error_ptr(png_ptr);
	edb_trace(ec, "libpng error: %s\n", message);

//	longjmp(png_ptr->jmpbuf, 1);
	longjmp(png_jmpbuf(png_ptr), 1);
}

static
void
edb_picture_png_warning(png_structp png_ptr, png_const_charp message)
{
	EdbContext	ec;

	if(!png_ptr) return;
	ec = (EdbContext)png_get_error_ptr(png_ptr);
	edb_trace(ec, "libpng warning: %s\n", message);
}

static
EdbPicture
edb_picture_make_picture_info_PNG(EdbContext ec, EdbTupleInfo tuple, unsigned char *body, int bytes)
{
	EdbPicture	picture;
	png_struct	*png_ptr;
	png_info		*info_ptr;
	FILE			*fp;

	picture = edb_picture_struct_new(ec, body, bytes);
	if(picture == NULL) return NULL;

	picture->mimetype = "image/png";

	fp = tmpfile();
	fwrite(body, 1, bytes, fp);
	rewind(fp);

	png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, (png_voidp)ec, edb_picture_png_error, edb_picture_png_warning);
	if(!png_ptr) {
		fclose(fp);
		edb_trace(ec, "malloc failure in LoadPNG\n");
		return NULL;
	}

	info_ptr = png_create_info_struct(png_ptr);

	if(!info_ptr) {
		fclose(fp);
		png_destroy_read_struct(&png_ptr, (png_infopp)NULL, (png_infopp)NULL);
		edb_trace(ec, "malloc failure in LoadPNG\n");
		return NULL;
	}

//	if(setjmp(png_ptr->jmpbuf)) {
	if(setjmp(png_jmpbuf(png_ptr))) {
		fclose(fp);
		png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);
		if(picture) edb_picture_free(ec, picture);
		return NULL;
	}

	png_init_io(png_ptr, fp);
	png_read_info(png_ptr, info_ptr);

//  	picture->xsize = info_ptr->width;
//	picture->ysize = info_ptr->height;
  	picture->xsize = png_get_image_width(png_ptr, info_ptr);
	picture->ysize = png_get_image_height(png_ptr, info_ptr);

	png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);

	fclose(fp);
	return picture;
}

EdbPicture
edb_picture_new(EdbContext ec, EdbTupleInfo tuple)
{
	EdbTupleColumn	tc;
	int			type;
	EdbPicture	picture = NULL;
	EdbDatum		datum;

	if(tuple==NULL) return NULL;

	tc = edb_tuple_seek_column_by_xmlname(ec, tuple, (utf8_t *)"picture.body");
	if(tc==NULL) tc = edb_tuple_seek_picture_column(ec, tuple);
	if(tc==NULL) return NULL;
	if((datum=tc->datum) && DatumEnglishIsValid(datum)) {
		unsigned char	*body;
		int	len;

		body = base64_decode(DatumEnglish(datum), &len);
		type = edb_picture_body_type(ec, body, len);
		if(type==PICTURE_JPEG) {
			picture = edb_picture_make_picture_info_JPEG(ec, tuple, body, len);
		} else if(type==PICTURE_PNG) {
			picture = edb_picture_make_picture_info_PNG(ec, tuple, body, len);
		}
		FREE(body);
	}
	if(picture) {
		picture->dpi = EDB_PICTURE_DPI_DEFAULT;
		tc = edb_tuple_seek_column_by_xmlname(ec, tuple, (utf8_t *)"picture.dpi");
		if(tc) picture->dpi = edb_datum_get_INT(ec, tc->datum);

		picture->height = EDB_PICTURE_HEIGHT_DEFAULT;
		tc = edb_tuple_seek_column_by_xmlname(ec, tuple, (utf8_t *)"picture.height");
		if(tc) picture->height = edb_datum_get_REAL(ec, tc->datum);

		picture->depth = EDB_PICTURE_DEPTH_DEFAULT;
		tc = edb_tuple_seek_column_by_xmlname(ec, tuple, (utf8_t *)"picture.depth");
		if(tc) picture->depth = edb_datum_get_REAL(ec, tc->datum);
	}
	return picture;
}

int
edb_picture_set_attribute(EdbContext ec, EdbPicture picture, char *attribute)
{
	register char	*name, *value;

	if(!STRisVALID(attribute)) return FALSE;
	name = STRDUP(attribute);
	value = strchr(name, '=');
	if(value==NULL) return FALSE;
	*value++ = '\0';
	str_oneline(name);
	str_oneline(value);
	if(!STRisVALID(name) || !STRisVALID(value)) return FALSE;
	if(!strcmp(name, "dpi")) {
		picture->dpi = ATOI(value);
	} else if(!strcmp(name, "height")) {
		picture->height = ATOF(value);
	} else if(!strcmp(name, "depth")) {
		picture->depth = ATOF(value);
	} else {
		return FALSE;
	}
	return TRUE;
}

#define	JPEGtoPNG	"/usr/local/bin/jpegtopnm | /usr/local/bin/pnmtopng"
#define	PNGtoJPEG	"/usr/local/bin/pngtopnm | /usr/local/bin/pnmtojpeg"
#define	JPEGtoEPS	"/usr/local/bin/jpegtopnm | /usr/local/bin/pnmtops -nocenter"
#define	PNGtoEPS	"/usr/local/bin/pngtopnm | /usr/local/bin/pnmtops -nocenter"

static
int
picture2any(EdbContext ec, EdbPicture picture, char *command, FILE *fp)
{
	char	tempfile[BUFSIZ];
	char	*pipename;
	int	fd;
	FILE	*pfp;
//	int	dp;
	char	buf[BUFSIZ];
	int	l;

	strcpy(tempfile, "/tmp/edb-picture.XXXXXXXX");
	fd = mkstemp(tempfile);
	if(fd<0) return FALSE;
	write(fd, picture->body, picture->bytes);
	close(fd);

	asprintf(&pipename, "/bin/cat %s | %s", tempfile, command);
	pfp = popen(pipename, "r");
	FREE(pipename);
	if(pfp==NULL) {
		edb_trace(ec, "picture2any: Netpbm package is required\n");
		unlink(tempfile);
		return FALSE;
	}

	while(!feof(pfp)) {
		l=fread(buf, sizeof(unsigned char), sizeof(buf), pfp);
		if(l>0) fwrite(buf, sizeof(unsigned char), l, fp);
	}
	pclose(pfp);
	unlink(tempfile);
	return TRUE;
}

int
edb_picture_write_JPEG(EdbContext ec, EdbPicture picture, FILE *fp)
{

	if(picture==NULL) return FALSE;
	if(!strcmp(picture->mimetype, "image/jpeg")) {
		fwrite(picture->body, sizeof(unsigned char), picture->bytes, fp);
	} else if(!strcmp(picture->mimetype, "image/png")) {
		return picture2any(ec, picture, PNGtoJPEG, fp);
	} else {
		return FALSE;
	}
	return TRUE;
}

int
edb_picture_write_PNG(EdbContext ec, EdbPicture picture, FILE *fp)
{

	if(picture==NULL) return FALSE;
	if(!strcmp(picture->mimetype, "image/jpeg")) {
		return picture2any(ec, picture, JPEGtoPNG, fp);
	} else if(!strcmp(picture->mimetype, "image/png")) {
		fwrite(picture->body, sizeof(unsigned char), picture->bytes, fp);
	} else {
		return FALSE;
	}
	return TRUE;
}

int
edb_picture_write_EPS(EdbContext ec, EdbPicture picture, FILE *fp)
{

	if(picture==NULL) return FALSE;
	if(!strcmp(picture->mimetype, "image/jpeg")) {
		return picture2any(ec, picture, JPEGtoEPS, fp);
	} else if(!strcmp(picture->mimetype, "image/png")) {
		return picture2any(ec, picture, PNGtoEPS, fp);
	} else {
		return FALSE;
	}
	return TRUE;
}

