/*
$Id: string.c,v 1.17 2012/04/03 03:51:13 alex Exp alex $
	string.c--- misc string functions
	(C) Copyright 1994 by Alex
*/

#ifndef lint
static char *rcsid __attribute__((unused)) = "$Id: string.c,v 1.17 2012/04/03 03:51:13 alex Exp alex $";
#endif

#include	<stdio.h>
#include	<stdlib.h>
#include	<stdarg.h>
#include	<ctype.h>
#include	<string.h>
#include	<strings.h>
#include	<stdlib.h>

#include	"edb/link.h"
#include	"edb/edb.h"

#ifndef	TRUE
#define	TRUE	1
#endif
#ifndef	FALSE
#define	FALSE	0
#endif
#ifndef	ERROR
#define	ERROR	(-1)
#endif

char *
str_alloc(int n)
{

	return (char *)malloc(sizeof(char)*n);
}

/*#strndup--- duplicate string (n-character) */
static
char *
my_strndup(register char *s, register int n)
{
	register char *buf;

	if(s == NULL) return NULL;
	buf = (char *)malloc((n+1) * sizeof(char));
	if(buf == NULL) return NULL;
	strncpy(buf, s, n);
	buf[n] = '\0';
	return buf;
}

/*#str_capitalize--- capitalize string */
void
str_capitalize(register char *dst, register char *src)
{

	while((*dst++ = toupper(*src++))) ;
}

/*#str_capitalize1--- capitalize string */
void
str_capitalize1(register char *s)
{

	for( ; *s; ++s) {
		if(islower(*s)) *s = toupper(*s) ;
	}
}

/*#str_minuscule--- minuscule string */
void
str_minuscule(register char *dst, register char *src)
{

	while((*dst++ = tolower(*src++))) ;
}

/*#str_minuscule1--- minuscule string */
void
str_minuscule1(register char *s)
{

	for( ; *s; ++s) {
		if(isupper(*s)) *s = tolower(*s) ;
	}
}

/*#str_isascii--- is ascii string */
int
str_isascii(register char *s)
{

	while(*s) {
		if(!isascii(*s++)) return FALSE;
	}
	return TRUE;
}

/*#str_iskana--- is kana string */
int
str_iskana(register char *s)
{
	ucs4_t	*u;

	u = UTF8toUCS4((utf8_t *)s);
	return ucs4_str_iskana(u);
}

/*#str_isdigit--- is digit string */
int
str_isdigit(register char *s)
{
	register char	c;

	while((c=*s++)) {
		if(!isdigit(c)) return FALSE;
	}
	return TRUE;
}

#if 0
/*#wstr_isfp--- is floating point string */
int
wstr_isfp(register wchar16_t *s)
{
	register int	pts=0;

	while(*s) {
		if(!wchar16_isdigit(*s)) {
			if(*s == '.') {
				++pts;
			} else {
				return FALSE;
			}
		}
		++s;
	}
	if(pts>1) return FALSE;
	return TRUE;
}
#endif

/*#str_skip_whitespace--- skip white space */
char *
str_skip_whitespace(register char *s)
{

	while((*s ==' ' || *s == '\t' || *s == '\n') && *s!= '\0') ++s;
	return s;
}

#if 0
/*#wstr_skip_whitespace--- skip white space */
wchar16_t *
wstr_skip_whitespace(register wchar16_t *s)
{

	while((*s ==WCHAR_SPACE || *s == WCHAR_TAB || *s == WCHAR_NL) && *s!= WCHAR_NULL) ++s;
	return s;
}
#endif

/*#str_skip_until_whitespace--- skip until white space */
char *
str_skip_until_whitespace(register char *s)
{

	while(!(*s ==' ' || *s == '\t' || *s == '\n') && *s!= '\0') ++s;
	return s;
}

#if 0
/*#wstr_skip_until_whitespace--- skip until white space */
wchar16_t *
wstr_skip_until_whitespace(register wchar16_t *s)
{

	while(!(*s ==WCHAR_SPACE || *s == WCHAR_TAB || *s == WCHAR_NL) && *s!= WCHAR_NULL) ++s;
	return s;
}
#endif

/*#str_convert_to_standard--- convert to standard */
void
str_convert_to_standard(register char *s)
{
	register int i, len;
	register ucs4_t *u;
	register ucs4_t *dst, *src, c;
	register int space;

	/* JWS に変換する．*/
	u = UTF8toUCS4((utf8_t *)s);

	/* 全角の空白文字とTAB，改行文字(CR/NL)を半角の空白に変換する．*/
	for(src = u; *src; ++src) {
		if(*src == UCS4_WSPACE || *src == UCS4_TAB || *src == UCS4_CR) *src = UCS4_SPACE;
	}

	/* 初めの空白文字を削除する．*/
	dst = src = u;
	while( (*src == UCS4_TAB) || (*src == UCS4_SPACE) ) ++src;
	while((*dst++ = *src++)) ;

	/* 末尾の空白文字を削除する．*/
	len = ucs4_strlen(u);
	for(i=len-1; i>=0; --i) {
		c = u[i];
		if(c == UCS4_TAB || c == UCS4_SPACE) {
			u[i] = UCS4_NULL;
		} else {
			break;
		}
	}

	/* 連続した空白を１つにする．*/
	space = FALSE;
	dst = src = u;
	while((c = *dst++ = *src++)) {
		if(space && (c == UCS4_SPACE)) {
			--dst;
		} else {
			space = (c == UCS4_SPACE);
		}
	}

	convUCS4toUTF8(u, (utf8_t *)s);
}

/*#str_standard--- convert to standard */
void
str_standard(register char *s)
{

	str_convert_to_standard(s);
}


/*#str_oneline--- convert to oneline */
void
str_oneline(register char *s)
{
	register char	*dst, *src, c, *l;
	register int	space;

	space = TRUE;
	l = s;
	for(src = dst = s; (c=*src); ++src) {
		/* TAB，NL，CR → 空白 */
		if(c == ' ' || c == '\r' || c == '\t' || c == '\n' ) {
			if(!space) {
				*dst++ = ' ';
				space = TRUE;
			}
		} else {
			*dst++ = c;
			l = dst;
			space = FALSE;
		}
	}
	*l = '\0';
}

#if 0
/*#wstr_oneline--- convert to oneline */
void
wstr_oneline(register wchar16_t *ws)
{
	register wchar16_t	*dst, *src, wc, *wl;
	register int	space;

	space = TRUE;
	wl = ws;
	for(src = dst = ws; (wc=*src); ++src) {
		/* TAB，NL，空白(全角)，CR → 空白(半角) */
		if(wc == WCHAR_SPACE  || wc == WCHAR_WSPACE || wc == WCHAR_CR || wc == WCHAR_TAB || wc == WCHAR_NL ) {
			if(!space) {
				*dst++ = WCHAR_SPACE;
				space = TRUE;
			}
		} else {
			*dst++ = wc;
			wl = dst;
			space = FALSE;
		}
	}
	*wl = WCHAR_NULL;
}

/*#wstr_onephrase--- convert to oneline */
void
wstr_onephrase(register wchar16_t *ws)
{
	register wchar16_t *w, wc, *wd;

	wstr_oneline(ws);

	/* 半角空白文字を消去する．*/
	for(w = wd = ws; *wd = *w; ++w) {
		if(*w != WCHAR_SPACE) ++wd;
	}
}
#endif

/*#str_onephrase--- convert to oneline */
void
str_onephrase(register char *s)
{
	register char *c, *d;

	str_oneline(s);

	/* 半角空白文字を消去する．*/
	for(c = d = s; (*d = *c); ++c) {
		if(*c != ' ') ++d;
	}
}

/*#str_split--- splitting */
Link
str_split(register char *str, register int c)
{
	register Link lstr;
	register char *s, *p;

	s = str;
	lstr = NULL;
	while(*s) {
		p = index(s, c);
		if(p != NULL) {
			LinkAppendTop(link_atom_new(my_strndup(s, p-s), free), lstr);
			s = ++p;
		} else {
			LinkAppendTop(link_atom_new_sdup(s), lstr);
			break;
		}
	}
	return link_reverse(lstr);
}

/*#str_concatenate--- concatenate string */
char *
str_concatenate(register Link lstr, register char c)
{
	register int len;
	register Link l;
	register int n;
	register char *s, *str;
	char buf[2];

	n = link_atoms(lstr);
	if(n <= 0) return strdup("");
	len = n-1;
	for(l=lstr; l; l=l->n) len += strlen(l->d);
	str = s = (char *)malloc(sizeof(char)*(len+1));
	strcpy(s, "");
	buf[0] = c;
	buf[1] = '\0';
	for(l=lstr; l; l=l->n) {
		strcat(s, l->d);
		if(l->n) strcat(s, buf);
		while(*s) ++s;
	}
	return str;
}

/*#str_concatenate0--- concatenate string */
char *
str_concatenate0(register Link lstr)
{
	register int len;
	register Link l;
	register int n;
	register char *s, *str;

	n = link_atoms(lstr);
	if(n <= 0) return strdup("");
	len = 0;
	for(l=lstr; l; l=l->n) len += strlen(l->d);
	str = s = (char *)malloc(sizeof(char)*(len+1));
	strcpy(s, "");
	for(l=lstr; l; l=l->n) {
		strcat(s, l->d);
		while(*s) ++s;
	}
	return str;
}

/*#str_concatenate_str--- concatenate string */
char *
str_concatenate_str(register Link lstr, register char *dstr)
{
	register int len;
	register Link l;
	register int n;
	register char *s, *str;

	n = link_atoms(lstr);
	if(n <= 0) return strdup("");
	len = strlen(dstr)*(n-1);
	for(l=lstr; l; l=l->n) len += strlen(l->d);
	str = s = (char *)malloc(sizeof(char)*(len+1));
	strcpy(s, "");
	for(l=lstr; l; l=l->n) {
		strcat(s, l->d);
		if(l->n) strcat(s, dstr);
		while(*s) ++s;
	}
	return str;
}

#if 0
static	wchar16_t	wstr_null_string[] = {WCHAR_NULL};

/*#wstr_concatenate_wstr--- concatenate string */
wchar16_t *
wstr_concatenate_wstr(register Link lstr, register wchar16_t *w)
{
	register int len;
	register Link l;
	register int n;
	register wchar16_t *s, *str;

	n = link_atoms(lstr);
	if(n <= 0) return wstrdup(wstr_null_string);
	len = (n-1)*wstrlen(w);
	for(l=lstr; l; l=l->n) len += wstrlen((wchar16_t *)l->d);
	str = s = (wchar16_t *)malloc(sizeof(wchar16_t)*(len+1));
	*s = WCHAR_NULL;
	for(l=lstr; l; l=l->n) {
		wstrcat(s, (wchar16_t *)l->d);
		if(l->n) wstrcat(s, w);
		while(*s) ++s;
	}
	return str;
}
#endif

/*#str_duplicate--- duplicate linked string */
Link
str_duplicate(register Link lstr)
{
	register Link	ls, lstr2 = NULL;

	for(ls=link_tail(lstr); ls; ls=ls->p) {
		LinkAppendTop(link_atom_new_sdup((char *)ls->d), lstr2);
	}
	return lstr2;
}

/*#str_match--- check match linked string */
int
str_match(register Link lstr, register char *str)
{
	register Link	ls;

	ForLink(ls, lstr) {
		if(!strcmp(ls->d, str)) return TRUE;
	}
	return FALSE;
}

/*#str_casematch--- check match linked string */
int
str_casematch(register Link lstr, register char *str)
{
	register Link	ls;

	ForLink(ls, lstr) {
		if(!strcasecmp(ls->d, str)) return TRUE;
	}
	return FALSE;
}

#if 0
/*#wstr_duplicate--- duplicate linked wstring */
Link
wstr_duplicate(register Link lwstr)
{
	register Link	lw, lwstr2 = NULL;

	for(lw=link_tail(lwstr); lw; lw=lw->p) {
		LinkAppendTop(link_atom_new(wstrdup((wchar16_t *)lw->d), free), lwstr2);
	}
	return lwstr2;
}

/*#wstr_duplicate_only_ascii--- duplicate linked wstring */
Link
wstr_duplicate_only_ascii(register Link lwstr)
{
	register Link	lw, lwstr2 = NULL;

	for(lw=link_tail(lwstr); lw; lw=lw->p) {
		if(wstr_isascii((wchar16_t *)lw->d)) {
			LinkAppendTop(link_atom_new(wstrdup((wchar16_t *)lw->d), free), lwstr2);
		}
	}
	return lwstr2;
}

/*#wstr_duplicate_only_katakana--- duplicate linked wstring */
Link
wstr_duplicate_only_katakana(register Link lwstr)
{
	register Link	lw, lwstr2 = NULL;

	for(lw=link_tail(lwstr); lw; lw=lw->p) {
		if(wstr_iskatakana((wchar16_t *)lw->d)) {
			LinkAppendTop(link_atom_new(wstrdup((wchar16_t *)lw->d), free), lwstr2);
		}
	}
	return lwstr2;
}

/*#jws_to_jis--- convert jws string to jis string */
char *
jws_to_jis(register wchar16_t *w, register char *j)
{

	convJWStoJIS(w, j);
	return j;
}

/*#jws_to_euc--- convert jws string to euc string */
char *
jws_to_euc(register wchar16_t *w, register char *e)
{

	convJWStoEUC(w, e);
	return e;
}

/*#jis_to_jws--- convert jis string to jws string */
wchar16_t *
jis_to_jws(register char *j, register wchar16_t *w)
{

	convJIStoJWS(j, w);
	return w;
}

/*#euc_to_jws--- convert euc string to jws string */
wchar16_t *
euc_to_jws(register char *e, register wchar16_t *w)
{

	convEUCtoJWS(e, w);
	return w;
}

/*#wstr_encode_escape--- encode escape */
int
wstr_encode_escape(register wchar16_t *src, register wchar16_t *dst)
{
	register int n = 0;

	while(*src) {
		if(*src == '(' || *src == ')' || *src == '\\') {
			if(dst) *dst++ = '\\';
			++n;
		}
		if(dst) *dst++ = *src;
		++src;
		++n;
	}
	if(dst) *dst = WCHAR_NULL;
	return n;
}

/*#wstr_encode_escape_dup--- encode escape */
wchar16_t *
wstr_encode_escape_dup(register wchar16_t *src, register wchar16_t *dst)
{

	wstr_encode_escape(src, dst);
	return dst;
}

/*#wstr_decode_escape--- decode escape */
int
wstr_decode_escape(register wchar16_t *src, register wchar16_t *dst)
{
	register int escape = FALSE;
	register int n = 0;

	while(*src) {
		if(!escape && *src == '\\') {
			escape = TRUE;
			++src;
			continue;
		} else {
			escape = FALSE;
		}
		if(dst) *dst++ = *src;
		++src;
		++n;
	}
	if(dst) *dst = WCHAR_NULL;
	return n;
}

/*#wstr_decode_escape_dup--- encode escape */
wchar16_t *
wstr_decode_escape_dup(register wchar16_t *src, register wchar16_t *dst)
{

	wstr_decode_escape(src, dst);
	return dst;
}
#endif

/*#str_compare--- compare for qsort */
static
int
str_compare(register LinkObj s1, register LinkObj s2)
{

	return strcmp(s1->d, s2->d);
}

/*#str_sort--- sort linked string */
void
str_sort(register Link lstr)
{

	link_sort(lstr, str_compare, NULL, NULL, NULL);
}

#if 0
/*#wstr_compare--- compare for qsort */
static
int
wstr_compare(register LinkObj s1, register LinkObj s2)
{

	return wstrcmp((wchar16_t *)s1->d, (wchar16_t *)s2->d);
}

/*#wstr_sort--- sort linked string */
void
wstr_sort(register Link lstr)
{

	link_sort(lstr, wstr_compare, NULL, NULL, NULL);
}

/*#wstr_rcompare--- compare for qsort */
static
int
wstr_rcompare(register LinkObj s1, register LinkObj s2)
{

	return - wstrcmp((wchar16_t *)s1->d, (wchar16_t *)s2->d);
}

/*#wstr_rsort--- sort linked string */
void
wstr_rsort(register Link lstr)
{

	link_sort(lstr, wstr_rcompare, NULL, NULL, NULL);
}

/*#wstr_casecompare--- compare for qsort */
static
int
wstr_casecompare(register LinkObj s1, register LinkObj s2)
{

	return wstrcasecmp((wchar16_t *)s1->d, (wchar16_t *)s2->d);
}

/*#wstr_casesort--- sort linked string */
void
wstr_casesort(register Link lstr)
{

	link_sort(lstr, wstr_casecompare, NULL, NULL, NULL);
}
#endif

/*#str_uniq--- uniq linked string */
Link
str_uniq(register Link lstr)
{
	register Link	l, lr=NULL;
	register char	*s;

	if(!lstr) return NULL;
	LinkAppend(lr, link_atom_new(lstr->d, lstr->f));
	s = (char *)lstr->d;
	lstr->f = NULL;
	for(l=lstr->n; l; l=l->n) {
		if(strcmp(s, (char *)l->d)) {
			LinkAppend(lr, link_atom_new(l->d, l->f));
			s = (char *)l->d;
			l->f = NULL;
		}
	}
	link_free(lstr);
	return lr;
}

#if 0
/*#wstr_uniq--- uniq linked string */
Link
wstr_uniq(register Link lstr)
{
	register Link	l, lr=NULL;
	register wchar16_t	*w;

	if(!lstr) return NULL;
	LinkAppend(lr, link_atom_new(lstr->d, lstr->f));
	w = (wchar16_t *)lstr->d;
	lstr->f = NULL;
	for(l=lstr->n; l; l=l->n) {
		if(wstrcmp(w, (wchar16_t *)l->d)) {
			LinkAppend(lr, link_atom_new(l->d, l->f));
			w = (wchar16_t *)l->d;
			l->f = NULL;
		}
	}
	link_free(lstr);
	return lr;
}

/*#wstr_caseuniq--- uniq linked string */
Link
wstr_caseuniq(register Link lstr)
{
	register Link	l, lr=NULL;
	register wchar16_t	*w;

	if(!lstr) return NULL;
	LinkAppend(lr, link_atom_new(lstr->d, lstr->f));
	w = (wchar16_t *)lstr->d;
	lstr->f = NULL;
	for(l=lstr->n; l; l=l->n) {
		if(wstrcasecmp(w, (wchar16_t *)l->d)) {
			LinkAppend(lr, link_atom_new(l->d, l->f));
			w = (wchar16_t *)l->d;
			l->f = NULL;
		}
	}
	link_free(lstr);
	return lr;
}

/*#wstr_casesort_and_caseuniq--- sort and uniq linked string */
Link
wstr_casesort_and_caseuniq(register Link lstr)
{

	wstr_casesort(lstr);
	return wstr_caseuniq(lstr);
}

/*#wstr_sort_and_uniq--- sort and uniq linked string */
Link
wstr_sort_and_uniq(register Link lstr)
{

	wstr_sort(lstr);
	return wstr_uniq(lstr);
}
#endif

/*#str_whats_kind_of_code--- 文字列が JIS, EUC, Shift-JISのどれか判断する */
int
str_whats_kind_of_code(register unsigned char *s)
{

	if(index((char *)s, 0x1b)) {
		return STR_CODE_JIS;
	} else if(str_isascii((char *)s)) {
		return STR_CODE_ASCII;
	} else if(utf8string_is_valid(s)) {
		return STR_CODE_UTF8;
	} else {
		register int	second = FALSE;

		while(*s) {
			if((*s)<0x80) {
				second = TRUE;
			} else if(0x80<=(*s) && (*s)<0xa0) {
				return STR_CODE_SJIS;
			} else if(0xa0<=(*s) && (*s)<0xe0) {
				if(!second) return STR_CODE_EUC;
			}
			second = !second;
			++s;
		}
		return STR_CODE_EUC;
	}
}

#if 0
static char	*toascii_table1[128] = {
	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,	/* 0x00 */
	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,	/* 0x08 */
	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,	/* 0x10 */
	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,	NULL,	/* 0x18 */
	"　",	"！",	"“”",	"＃",	"＄",	"％",	"＆",	"’´",	/* 0x20 */
	"（〔【",	"）〕】",	"＊",	"＋",	"，、",	"−",	"．。",	"／",	/* 0x28 */
	"０",	"１",	"２",	"３",	"４",	"５",	"６",	"７",	/* 0x30 */
	"８",	"９",	"：",	"；",	"＜《",	"＝",	"＞》",	"？",	/* 0x38 */
	"＠",	"Ａ",	"Ｂ",	"Ｃ",	"Ｄ",	"Ｅ",	"Ｆ",	"Ｇ",	/* 0x40 */
	"Ｈ",	"Ｉ",	"Ｊ",	"Ｋ",	"Ｌ",	"Ｍ",	"Ｎ",	"Ｏ",	/* 0x48 */
	"Ｐ",	"Ｑ",	"Ｒ",	"Ｓ",	"Ｔ",	"Ｕ",	"Ｖ",	"Ｗ",	/* 0x50 */
	"Ｘ",	"Ｙ",	"Ｚ",	"［「『",	"@",	"］」』",	"＾",	"＿",	/* 0x58 */
	"‘｀",	"ａ",	"ｂ",	"ｃ",	"ｄ",	"ｅ",	"ｆ",	"ｇ",	/* 0x60 */
	"ｈ",	"ｉ",	"ｊ",	"ｋ",	"ｌ",	"ｍ",	"ｎ",	"ｏ",	/* 0x68 */
	"ｐ",	"ｑ",	"ｒ",	"ｓ",	"ｔ",	"ｕ",	"ｖ",	"ｗ",	/* 0x70 */
	"ｘ",	"ｙ",	"ｚ",	"｛",	"｜",	"｝",	"￣",	NULL,	/* 0x78 */
};

static wchar16_t	toascii_table[0x10000];
static int	toascii_table_initialized = FALSE;

/*#wstr_toascii_make_table--- */
static
void
wstr_toascii_make_table()
{
	register int	i;

	if(toascii_table_initialized) return;
	for(i=0; i<Ns(toascii_table); ++i) toascii_table[i] = i;
	for(i=0; i<Ns(toascii_table1); ++i) {
		if(toascii_table1[i]) {
			register wchar16_t	*w = EUCtoJWS(toascii_table1[i]);

			for( ; *w; ++w) {
				toascii_table[((*w)&(Ns(toascii_table)-1))] = i;
			}
		}
	}
	toascii_table_initialized = TRUE;
}

/*#wstr_toascii--- conver wstring to ascii */
void
wstr_toascii(register wchar16_t *wstr)
{
	register wchar16_t	*w;

	wstr_toascii_make_table();
	for(w=wstr; *w; ++w) {
		if((*w)<0xb000 && !wchar16_isascii(*w)) {
			*w = toascii_table[((*w)&(Ns(toascii_table)-1))];
		}
	}
}

/*#wstr_end_char--- get end char */
wchar16_t
wstr_end_char(register wchar16_t *w)
{
	register int	len;

	len = wstrlen(w);
	if(len==0) return WCHAR_NULL;
	return w[len-1];
}
#endif

char *
strdup_null(char *s)
{

	return strdup(s ? s : "");
}

int
strcmp_null(char *s1, char *s2)
{

	if(s1==NULL && s2==NULL) return 0;
	if(s1==NULL) return -1;
	if(s2==NULL) return 1;
	return strcmp(s1, s2);
}

int
str_is_valid(char *s)
{

	if(s==NULL) return FALSE;
	if(strlen(s)==0) return FALSE;
	return TRUE;
}

Link
lstr_append_printf(Link lstr, char *fmt, ...)
{
	char	*buf;
	va_list	ap;

	va_start(ap, fmt);
	vasprintf(&buf, fmt, ap);
	va_end(ap);
	return link_append(lstr, link_atom_new(buf, free));
}

#define	IsWhiteSpace(c) ((c)==' ' || (c)=='\t')

char *
str_get_phrase(char **str)
{
	char	*s, *e;
	int	brace, quote, wquote;
	int	len;

	s = *str;
	/* skip white space */
	while(IsWhiteSpace(*s) && *s) ++s;

	if(*s=='\0') {
		*str = NULL;
		return NULL;
	}
	/* get phrase */
	e = s;
	brace = 0;
	quote = wquote = FALSE;
	for(e=s; *e; ++e) {
		switch(*e) {
		case '"':	wquote = !wquote;	break;
		case '\'':	quote = !quote;	break;
		case '{':	++brace;	break;
		case '}':	if(brace>0) --brace;	break;
		default:	break;
		}
		if(brace==0 && !quote && !wquote && IsWhiteSpace(*e)) break;
	}
	len = (e-s);
	if(len>=0) {
		s = strdup(s);
		s[len] = '\0';
		*str = e;
		return s;
	} else {
		*str = NULL;
		return NULL;
	}
}


