/*
	$Id: table.c,v 1.83 2012/04/03 03:51:13 alex Exp alex $
	table.c--- table info
*/

#ifndef	lint
static char *rcsid __attribute__((unused)) = "$Id: table.c,v 1.83 2012/04/03 03:51:13 alex Exp alex $";
#endif

#include	<stdio.h>
#include	<stdarg.h>
#include	<string.h>
#include	"edb/edb.h"
#include	"edb/xml.h"
#include	"edb/module.h"
#include	"edb/pki.h"

EdbMapLookup
edb_maplookup_new(EdbContext ec, utf8_t *lookup)
{
	EdbMapLookup	map;

	if(!STRisValid(lookup)) return NULL;
	map = (EdbMapLookup)malloc(sizeof(struct EDB_MAPLOOKUP));
	if(map==NULL) return NULL;
	bzero(map, sizeof(struct EDB_MAPLOOKUP));
	map->lookup = (utf8_t *)strdup((char *)lookup);
	map->teid = 0;
	return map;
}

void
edb_maplookup_free(EdbContext ec, EdbMapLookup map)
{

	if(map==NULL) return;
	link_free(map->l_caption_ci);
	FREE(map->lookup);
	FREE(map);
}

/*
	maplookup のフォーマット
	"{xmlname-of-table} CANDIDATE={condition} CAPTION={@.column1 @.column2 ...}"
*/
void
edb_maplookup_parse(EdbContext ec, EdbMapLookup map)
{
	utf8_t	*s, *maplookupstr;
	utf8_t	*xmlname;
	EdbTableInfo	mapto_ti;

	if(map==NULL || map->parsed) return;
	map->parsed = TRUE;
	maplookupstr = (utf8_t *)STRDUP((char *)map->lookup);

	xmlname = (utf8_t *)str_get_phrase((char **)&maplookupstr);
	if(!STRisVALID(xmlname)) return;
	s = (utf8_t *)STRDUP((char *)xmlname);
	FREE(xmlname);
	if(s[0]=='{') ++s;
	xmlname = (utf8_t *)STRDUP((char *)s);
	str_oneline((char *)xmlname);
	if((s = (utf8_t *)strchr((char *)xmlname, '}'))!=NULL) *s = '\0';
	if((s = (utf8_t *)strchr((char *)xmlname, '.'))!=NULL) *s = '\0';

	map->teid = edb_table_xmlname_to_eid(ec, xmlname);
	if(!EIDisVALID(map->teid)) {
		edb_trace(ec, "edb_maplookup_parse: ERROR : illegal table name \"%s\".\n", xmlname);
		return;
	}
	mapto_ti = edb_table_get(ec, map->teid);
	if(mapto_ti) {
		char	*opt;

		while((opt=str_get_phrase((char **)&maplookupstr))!=NULL) {
			char	*p=NULL, *q=NULL;

			s = (utf8_t *)STRDUP(opt);
			FREE(opt);
			opt = (char *)s;
			p = STRDUP(opt);
			if((s = (utf8_t *)strchr(p, '='))!=NULL) {
				*s = '\0';
				q = (char *)(s+1);
			}
			if(p==NULL || q==NULL) continue;

			if(!strcmp(p, "CAPTION")) {
				Link	l, l_columns;

				if(q[0]=='{') {
					++q;
					if((s = (utf8_t *)strchr(q, '}'))!=NULL) *s = '\0';
				}
				l_columns = str_split(q, ' ');
				ForLink(l, l_columns) {
					EdbColumnInfo	ci = edb_table_seek_column_by_xmlname(ec, mapto_ti, l->d);

					if(ci) {
						LinkAppend(map->l_caption_ci, link_atom_new(ci, NULL));
					} else {
						edb_trace(ec, "edb_maplookup_parse: ERROR : can't find caption column (%s) in \"%s\".\n", l->d, TABLE_XN(mapto_ti));
					}
				}
				link_free(l_columns);
			} else if(!strcmp(p, "CANDIDATE")) {
				if(q[0]=='{') {
					++q;
					if((s = (utf8_t *)strchr(q, '}'))!=NULL) *s = '\0';
				}
				LinkAppend(map->l_candidate_condition, link_atom_new_sdup(q));
			} else {
				edb_trace(ec, "edb_maplookup_parse: ERROR : unknown option (%s).\n", opt);
			}
		}
	} else {
		edb_trace(ec, "edb_maplookup_parse: ERROR : can't load mapto table (EID: %d).\n", map->teid);
	}
}

eid_t
edb_maplookup_table_eid(EdbContext ec, EdbMapLookup map)
{

	edb_maplookup_parse(ec, map);
	return  map->teid;
}

static
char *
edb_maplookup_candidate_classify_column(EdbContext ec, EdbMaptoConditionContext cc, char *str)
{
	EdbColumnInfo	ci;
	char	*name = NULL;

	if(cc==NULL) return NULL;
	while(str[0]=='@' && str[1]=='@') {
		cc = cc->outer;
		++str;
	}
	if(cc==NULL || cc->ti==NULL) return NULL;

	if(!strcmp(str, "REF") || !strcmp(str, "@.REF")) {
		name = strdup("REF");
	} else if(!strcmp(str, "PERM") || !strcmp(str, "@.PERM")) {
		name = strdup("PERM");
	} else if(!strcmp(str, "OWN") || !strcmp(str, "@.OWN")) {
		name = strdup("OWN");
	} else if(!strcmp(str, "MAP") || !strcmp(str, "@.MAP")) {
		name = strdup("MAP");
	} else {
		ci = edb_table_seek_column(ec, cc->ti, (utf8_t *)str);
		if(ci==NULL) return NULL;
		name = strdup((char *)ColumnXN(ci));
	}
	return name;
}

static
char *
edb_maplookup_candidate_classify_value(EdbContext ec, EdbMaptoConditionContext cc, char *str)
{
	EdbColumnInfo	ci;
	char	*value = NULL;
	char	*s;

	if(cc==NULL) return NULL;
	if(str[0]=='@') {
		while(str[0]=='@' && str[1]=='@') {
			cc = cc->outer;
			++str;
		}
		if(cc==NULL || cc->tuple==NULL) return NULL;
		if(!strcmp(str, "@")) {
			if(cc->datum) {
				if(DatumEIDisValid(cc->datum)) {
					value = eid2strdup(DatumEID(cc->datum));
				} else if(cc->ci->typeInfo->basicTypeID==EDB_TYPE_INT && DatumEnglishIsValid(cc->datum)) {
					value = strdup((char *)DatumEnglish(cc->datum));
				}
			}
		} else if(!strcmp(str, "@-")) {
			if(cc->sdatum) {
				if(DatumEIDisValid(cc->sdatum)) {
					value = eid2strdup(DatumEID(cc->sdatum));
				} else if(cc->ci->super && cc->ci->super->typeInfo->basicTypeID==EDB_TYPE_INT && DatumEnglishIsValid(cc->sdatum)) {
					value = strdup((char *)DatumEnglish(cc->sdatum));
				}
			}
		} else if(!strcmp(str, "@1")) {
			if(EIDisVALID(TupleEID(cc->tuple))) value = eid2strdup(TupleEID(cc->tuple));
		} else if(!strncmp(str, "@1.", 3)) {
			EdbTupleColumn	tc;
			EdbDatum			datum;

			asprintf(&s, "@.%s", &str[3]);
			tc = edb_tuple_seek_column(ec, cc->tuple, (utf8_t *)s);
			FREE(s);
			if(tc && (datum=tc->datum)) {
				if(DatumEIDisVALID(datum)) {
					asprintf(&value, "%d", DatumEID(datum));
				} else if(tc->ci->typeInfo->basicTypeID==EDB_TYPE_INT && DatumEnglishIsValid(datum)) {
					value = strdup((char *)DatumEnglish(datum));
				}
			}
		}
	} else if(str[0]=='"') {
		value = strdup(&str[1]);
		if((s=strrchr(value, '"'))!=NULL) *s = '\0';
	} else {
		value = strdup(str);
	}
	return value;
}

char *
edb_maplookup_candidate_condition(EdbContext ec, EdbMaptoConditionContext cc, EdbMapLookup map)
{
	char	*cond, *condition;
	Link	l_condition = NULL;

	condition = STRDUP((char *)map->l_candidate_condition->d);
	while((cond = str_get_phrase(&condition))!=NULL) {
		char	*s, *s2, *column = NULL, *value = NULL, op = '=';
		EdbTupleColumn	tc = NULL;

		column = STRDUP(cond);
		FREE(cond);

		if(!strcmp(column, "INSIDE") && TABLE_IsOrganization(cc->ti)) {
			s = (char *)edb_condition_make(ec, cc->ti, CONDITION_LOGIC_OR,
				ConditionEIDMatch("@.EID", ORGAN_TOP_EID),
				ConditionEIDMatch("@.upper", ORGAN_TOP_EID),
				ConditionTerminate);
			LinkAppend(l_condition, link_atom_new_spf("(%s)", s));
			FREE(s);
			continue;
		}

		s=strchr(column, '=');
		s2=strchr(column, '~');
		if(!s) {
			s = s2;
		} else if(s2 && s2<s) {
			s = s2;
		}
		if(s) {
			op = *s;
			*s = '\0';
			value = s+1;
		}
		if(!STRisVALID(column) || !STRisVALID(value)) continue;
		column = edb_maplookup_candidate_classify_column(ec, cc, column);
		value = edb_maplookup_candidate_classify_value(ec, cc, value);
		if(STRisVALID(column) && STRisVALID(value)) {
			EdbColumnInfo	ci = edb_table_seek_column(ec, cc->ti, (utf8_t *)column);
			EdbCatalogue ca = edb_catalogue_new();
			Link	l, lstr = str_split(value, ',');
			ForLink(l, lstr) {
				if(!str_isdigit(l->d)) {
					edb_catalogue_free(ca);
					ca = NULL;
					break;
				}
				EdbTupleInfo tuple = edb_tuple_get(ec, ATOI(l->d), 0);
				if(!tuple) {
					edb_catalogue_free(ca);
					ca = NULL;
					break;
				}
				edb_catalogue_append(ca, TupleEID(tuple), tuple);
			}
			link_free(lstr);
			if(ci && ci->typeInfo->basicTypeID == EDB_TYPE_INT) {
				s = (char *)edb_condition_make(ec, cc->ti, CONDITION_LOGIC_AND,
					ConditionIntegerEQ(column, ATOI(value)),
					ConditionTerminate);
			} else if(str_isdigit(value)) {
				s = (char *)edb_condition_make(ec, cc->ti, CONDITION_LOGIC_AND,
					ConditionEIDMatch(column, ATOI(value)),
					ConditionTerminate);
			} else if(ca) {
				s = (char *)edb_condition_make(ec, cc->ti, CONDITION_LOGIC_AND,
					ConditionCatalogueMatch(column, ca),
					ConditionTerminate);
			} else if(op=='=') {
				s = (char *)edb_condition_make(ec, cc->ti, CONDITION_LOGIC_AND,
					ConditionTextEQ(column, value),
					ConditionTerminate);
			} else {
				s = (char *)edb_condition_make(ec, cc->ti, CONDITION_LOGIC_AND,
					ConditionRegexTextEQ(column, value),
					ConditionTerminate);
			}
			edb_catalogue_free(ca);
			LinkAppend(l_condition, link_atom_new(s, free));
		}
		FREE(column);
		FREE(value);
	}
	condition = NULL;
	if(l_condition) {
		condition = str_concatenate_str(l_condition, " AND ");
		link_free(l_condition);
	}
	return condition;
}

char *
edb_maplookup_candidate_constraint(EdbContext ec, EdbMaptoConditionContext cc, EdbMapLookup map)
{
	char	*cond, *condition;
	Link	l_constraint = NULL;

	condition = STRDUP((char *)map->l_candidate_condition->d);
	while((cond = str_get_phrase(&condition))!=NULL) {
		char	*s, *s2, *column = NULL, *value = NULL;
		char	op = '=';
		EdbTupleColumn	tc = NULL;

		column = STRDUP(cond);
		FREE(cond);

		if(!strcmp(column, "INSIDE") && TABLE_IsOrganization(cc->ti)) {
			LinkAppend(l_constraint, link_atom_new_spf("INSIDE"));
			continue;
		}

		s=strchr(column, '=');
		s2=strchr(column, '~');
		if(!s) {
			s = s2;
		} else if(s2 && s2<s) {
			s = s2;
		}
		if(s) {
			op = *s;
			*s = '\0';
			value = s+1;
		}
		if(!STRisVALID(column) || !STRisVALID(value)) continue;
//		column = edb_maplookup_candidate_classify_column(ec, cc, column);
		value = edb_maplookup_candidate_classify_value(ec, cc, value);
		if(STRisVALID(column) && STRisVALID(value)) {
			LinkAppend(l_constraint, link_atom_new_spf("%s%c%s", column, op, value));
		}
//		FREE(column);
		FREE(value);
	}
	condition = NULL;
	if(l_constraint) {
		condition = str_concatenate(l_constraint, ' ');
		link_free(l_constraint);
	}
	return condition;
}

char *
edb_maplookup_candidate_classify(EdbContext ec, EdbMaptoConditionContext cc, EdbMapLookup map)
{
	char	*cond, *condition;
	Link	l_classify = NULL;

	condition = STRDUP((char *)map->l_candidate_condition->d);
	while((cond = str_get_phrase(&condition))!=NULL) {
		char	*s, *column = NULL, *value = NULL;
		EdbTupleColumn	tc = NULL;

		column = STRDUP(cond);
		FREE(cond);

		if(!strcmp(column, "INSIDE") && TABLE_IsOrganization(cc->ti)) {
			LinkAppend(l_classify, link_atom_new_spf("CLASSIFY=INSIDE"));
			continue;
		}
		if((s=strchr(column, '='))!=NULL) {
			*s = '\0';
			value = s+1;
		}
		if(!STRisVALID(column) || !STRisVALID(value)) continue;
		column = edb_maplookup_candidate_classify_column(ec, cc, column);
		value = edb_maplookup_candidate_classify_value(ec, cc, value);
		if(STRisVALID(column) && STRisVALID(value)) {
			asprintf(&s, "CLASSIFY=%s=%s", column, value);
			LinkAppend(l_classify, link_atom_new(s, free));
		}
		FREE(column);
		FREE(value);
	}
	condition = NULL;
	if(l_classify) {
		condition = str_concatenate(l_classify, '&');
		link_free(l_classify);
	}
	return condition;
}

static
int
edb_table_parse_maplookup_func(EdbContext ec, EdbTableInfo ti, int layer, EdbColumnInfo ci, EdbTableCBData cbd)
{
	Link	lm;

	ForLink(lm, ci->l_maplookup) edb_maplookup_parse(ec, (EdbMapLookup)lm->d);
	return TRUE;
}

void
edb_table_parse_maplookup(EdbContext ec, EdbTableInfo ti)
{

	edb_table_column_callback(ec, ti, NULL, edb_table_parse_maplookup_func, NULL);
}

static
EdbColumnInfo
edb_table_column_new(EdbContext ec, EdbTableInfo ti, EdbColumnInfo super)
{
	EdbColumnInfo	ci;

	ci = (EdbColumnInfo)malloc(sizeof(struct EDB_COLUMN_INFO));
	if(ci==NULL) {
		edb_trace(ec, "edb_table_column_new: can't allocate new column.\n");
		return NULL;
	}
	bzero(ci, sizeof(struct EDB_COLUMN_INFO));
	ci->ti = ti;
	ci->super = super;
	ci->idx = -1;
	return ci;
}

static
void
edb_table_column_free(EdbContext ec, EdbColumnInfo ci)
{

	if(ci==NULL) return;
	EdbLinkFree2(ec, ci->l_child);
	EdbLinkFree2(ec, ci->l_maplookup);
	FREE(ci->reldb_sqlname);
	FREE(ci->xmlname);
	FREE(ci->sqlname);
	FREE(ci->sqlname_variant);
	FREE(ci->sqlname_eid);
	FREE(ci->name.english);
	FREE(ci->name.japanese);
	FREE(ci->name.pronounce);
	FREE(ci->desc.english);
	FREE(ci->desc.japanese);
	FREE(ci->explain.english);
	FREE(ci->explain.japanese);
	FREE(ci->prefix.english);
	FREE(ci->prefix.japanese);
	FREE(ci->suffix.english);
	FREE(ci->suffix.japanese);
	edb_datum_ctype_free(ec, ColumnCType(ci));
	FREE(ci->elasticity);
	free(ci);
}

int
edb_table_reldb_exist(EdbContext ec, utf8_t *table)
{
	utf8_t		*query;
	register int	v;

	asprintf((char **)&query, "SELECT COUNT(*) FROM pg_tables WHERE tablename='%s'", table);
	v = edb_pgsql_query_int(ec, query);
	FREE(query);
	return v;
}

static
EdbColumnInfo
edb_table_column_body_parse_xml(EdbContext ec, EdbTableInfo ti, EdbColumnInfo super, xmlDocPtr doc, xmlNodePtr cur)
{
	EdbColumnInfo	ci;
	Link	ln, l_node;
	char	*type_name;

	ci = edb_table_column_new(ec, ti, super);
	if(ci==NULL) return NULL;

	ci->xmlname = xml_get_element_text(ec, doc, cur, "xmlname", NULL);
	ci->sqlname = xml_get_element_text(ec, doc, cur, "sqlname", NULL);
	if(STRisVALID(ci->sqlname)) {
		asprintf((char **)&ci->reldb_sqlname, "%s_%s", ti->sqlname, &ci->sqlname[2]);
		if(strlen((char *)ci->reldb_sqlname)>25) ci->reldb_sqlname[25] = '\0';		/* Ooops: table name is truncated */
#if 0
		ci->reldb_exist = edb_table_reldb_exist(ec, ci->reldb_sqlname);
#endif

		ci->sqlname_variant = (utf8_t *)strdup((char *)ci->sqlname);
		ci->sqlname_variant[0] = 'v';		/* v_sqlname */
		ci->sqlname_eid = (utf8_t *)strdup((char *)ci->sqlname);
		ci->sqlname_eid[0] = 'e';		/* e_sqlname */
	}
	type_name = (char *)xml_get_element_text(ec, doc, cur, (utf8_t *)"type", NULL);
	if(STRisValid(type_name)) {
		ColumnCType(ci) = edb_datum_ctype_new(ec, ci, type_name);
		ci->typeInfo = ColumnTypeInfo (ci);
		ci->columns = 1;
		ColumnTypeMinimum(ci) = (char *)xml_get_element_text(ec, doc, cur, "minimum", NULL);
		ColumnTypeMaximum(ci) = (char *)xml_get_element_text(ec, doc, cur, "maximum", NULL);
		ColumnTypeDefault(ci) = (char *)xml_get_element_text(ec, doc, cur, "default", NULL);
		ColumnTypeCryptography(ci) = (char *)xml_get_element_text(ec, doc, cur, "cryptography", NULL);
	} else {
		edb_trace(ec, "edb_table_column_body_parse_xml: can't determine the type of column.\n");
	}
	if(!xml_get_element_mtext(ec, &ci->name.english, &ci->name.japanese, &ci->name.pronounce, doc, cur, "name", NULL)) {
		edb_trace(ec, "edb_table_column_body_parse_xml: can't get name of column.\n");
	}
	if(!xml_get_element_mtext(ec, &ci->desc.english, &ci->desc.japanese, NULL, doc, cur, "description", NULL)) {
		edb_trace(ec, "edb_table_column_body_parse_xml: can't get description of column.\n");
	}
	if(!xml_get_element_mtext(ec, &ci->explain.english, &ci->explain.japanese, NULL, doc, cur, "explain", NULL)) {
		edb_trace(ec, "edb_table_column_body_parse_xml: can't get explain of column.\n");
	}
	ci->elements = xml_get_element_text_integer(ec, doc, cur, 1, "elements", NULL);
	ci->elasticity = (char *)xml_get_element_text(ec, doc, cur, "elasticity", NULL);

	if(STRisVALID(ci->elasticity)) {
		sscanf(ci->elasticity, "%u+%u", &ci->elements_base, &ci->elements_delta);
	}
	if(ci->elements_base<1) ci->elements_base = 1;
	if(ci->elements_delta<1) ci->elements_delta = 1;

	l_node = xml_find_element_node_list(ec, doc, cur, "attribute", NULL);
	ForLink(ln, l_node) {
		utf8_t	*s;

		s = xml_get_node_text(ec, doc, (xmlNodePtr)ln->d);
		if(!strcasecmp((char *)s, "authorize")) {
			ci->authorize = TRUE;
		} else if(!strcasecmp((char *)s, "exclusive")) {
			ci->exclusive = TRUE;
		} else if(!strcasecmp((char *)s, "hierarchyparent")) {
			if(ci->ti->hierarchical) ci->hierarchyparent = TRUE;
		} else if(!strcasecmp((char *)s, "multilingualcaption")) {
			ci->multilingualcaption = TRUE;
		} else if(!strcasecmp((char *)s, "personname")) {
			ci->personname = TRUE;
		} else if(!strcasecmp((char *)s, "student")) {
			ci->sometimesStudent = TRUE;
		} else if(!strcasecmp((char *)s, "pagenumber")) {
			ci->pagenumber = TRUE;
		} else if(!strcasecmp((char *)s, "patentnumber")) {
			ci->patentnumber = TRUE;
		} else if(!strcasecmp((char *)s, "nothing")) {
			ci->canbeNothing = TRUE;
		} else if(!strcmp((char *)s, "CLASSIFY")) {
			ci->classify = CLASSIFY_PRIMARY;
		} else if(!strcmp((char *)s, "classify")) {
			ci->classify = CLASSIFY_SECONDARY;
		} else if(!strcmp((char *)s, "alias")) {
			ci->alias = TRUE;
		} else if(!strcmp((char *)s, "chronological")) {
			ci->chronological = TRUE;
		} else if(!strcmp((char *)s, "conductor")) {
			ci->conductor = TRUE;
			LinkAppend(ci->ti->l_conductors, link_atom_new(ci, NULL));
		} else if(!strcmp((char *)s, "required")) {
			ci->usage = USAGE_REQUIRED;
		} else if(!strcmp((char *)s, "preferred")) {
			ci->usage = USAGE_PREFERRED;
		} else if(!strcmp((char *)s, "optional")) {
			ci->usage = USAGE_OPTIONAL;
		} else if(!strcmp((char *)s, "unused")) {
			ci->usage = USAGE_UNUSED;
		} else if(!strcmp((char *)s, "obsoleted")) {
			ci->obsoleted = TRUE;
		} else {
			edb_trace(ec, "edb_table_column_body_parse_xml: unknown attribute (%s)\n", s);
		}
		FREE(s);
	}

	l_node = xml_find_element_node_list(ec, doc, cur, "conductor", NULL);
	ForLink(ln, l_node) {
		utf8_t	*s = xml_get_node_text(ec, doc, (xmlNodePtr)ln->d);
		if(!strcmp((char *)s, "required")) {
			ci->conductor_usage = USAGE_REQUIRED;
		} else if(!strcmp((char *)s, "preferred")) {
			ci->conductor_usage = USAGE_PREFERRED;
		} else if(!strcmp((char *)s, "optional")) {
			ci->conductor_usage = USAGE_OPTIONAL;
		} else if(!strcmp((char *)s, "unused")) {
			ci->conductor_usage = USAGE_UNUSED;
		}
		FREE(s);
	}

	l_node = xml_find_element_node_list(ec, doc, cur, "maplookup", NULL);
	ForLink(ln, l_node) {
		utf8_t	*s = xml_get_node_text(ec, doc, (xmlNodePtr)ln->d);
		if(s) LinkAppend(ci->l_maplookup, link_atom_new(edb_maplookup_new(ec, s), edb_maplookup_free));
	}

	if(!xml_get_element_permission(ec, &ci->perm_tuple, doc, cur, "permtuple", NULL)) {
		edb_trace(ec, "edb_table_column_body_parse_xml: can't get permission for tuple.\n");
	}
	ci->caption = xml_get_element_text_integer(ec, doc, cur, 0, "caption", NULL);
	xml_get_element_mtext(ec, &ci->prefix.english, &ci->prefix.japanese, NULL, doc, cur, "prefix", NULL);
	xml_get_element_mtext(ec, &ci->suffix.english, &ci->suffix.japanese, NULL, doc, cur, "suffix", NULL);
	ci->order = xml_get_element_text_integer(ec, doc, cur, 0, "order", NULL);
	ci->size = xml_get_element_text_integer(ec, doc, cur, 0, "size", NULL);

	return ci;
}

static
utf8_t *
edb_table_column_body_make_xml(EdbContext ec, EdbTableInfo ti, EdbColumnInfo ci)
{
	register Link	l, l_xml = NULL;
	utf8_t	*s, *xml;

	xml = xml_make_element_text(ec, (utf8_t *)"xmlname", ci->xmlname);
	if(STRisVALID(xml)) LinkAppend(l_xml, link_atom_new(xml, free));

	if(STRisVALID(ci->sqlname)) {
		xml = xml_make_element_text(ec, (utf8_t *)"sqlname", ci->sqlname);
		if(STRisVALID(xml)) LinkAppend(l_xml, link_atom_new(xml, free));
	}

	LinkAppend(l_xml, link_atom_new_sdup("<" EDB_XMLNS ":name>"));
	xml = xml_make_element_mtext(ec, ci->name.english, ci->name.japanese, ci->name.pronounce);
	if(STRisVALID(xml)) LinkAppend(l_xml, link_atom_new(xml, free));
	LinkAppend(l_xml, link_atom_new_sdup("</" EDB_XMLNS ":name>"));

	LinkAppend(l_xml, link_atom_new_sdup("<" EDB_XMLNS ":description>"));
	xml = xml_make_element_mtext(ec, ci->desc.english, ci->desc.japanese, NULL);
	if(STRisVALID(xml)) LinkAppend(l_xml, link_atom_new(xml, free));
	LinkAppend(l_xml, link_atom_new_sdup("</" EDB_XMLNS ":description>"));

	LinkAppend(l_xml, link_atom_new_sdup("<" EDB_XMLNS ":explain>"));
	xml = xml_make_element_mtext(ec, ci->explain.english, ci->explain.japanese, NULL);
	if(STRisVALID(xml)) LinkAppend(l_xml, link_atom_new(xml, free));
	LinkAppend(l_xml, link_atom_new_sdup("</" EDB_XMLNS ":explain>"));

	if(ci->typeInfo) {
		xml = xml_make_element_text(ec, (utf8_t *)"type", (utf8_t *)ci->typeInfo->name);
		if(STRisVALID(xml)) LinkAppend(l_xml, link_atom_new(xml, free));
	}
	if(STRisValid(ColumnTypeMinimum(ci))) {
		xml = xml_make_element_text(ec, (utf8_t *)"minimum", (utf8_t *)ColumnTypeMinimum(ci));
		if(STRisValid(xml)) LinkAppend(l_xml, link_atom_new(xml, free));
	}
	if(STRisValid(ColumnTypeMaximum(ci))) {
		xml = xml_make_element_text(ec, (utf8_t *)"maximum", (utf8_t *)ColumnTypeMaximum(ci));
		if(STRisValid(xml)) LinkAppend(l_xml, link_atom_new(xml, free));
	}
	if(STRisValid(ColumnTypeDefault(ci))) {
		xml = xml_make_element_text(ec, (utf8_t *)"default", (utf8_t *)ColumnTypeDefault(ci));
		if(STRisValid(xml)) LinkAppend(l_xml, link_atom_new(xml, free));
	}
	if(STRisValid(ColumnTypeCryptography(ci))) {
		xml = xml_make_element_text(ec, (utf8_t *)"cryptography", (utf8_t *)ColumnTypeCryptography(ci));
		if(STRisValid(xml)) LinkAppend(l_xml, link_atom_new(xml, free));
	}

	if(ci->elements==0) {
		xml = xml_make_element_text(ec, (utf8_t *)"elements", (utf8_t *)"*");
	} else {
		asprintf((char **)&s, "%d", ci->elements);
		xml = xml_make_element_text(ec, (utf8_t *)"elements", (utf8_t *)s);
		FREE(s);
	}
	if(STRisVALID(xml)) LinkAppend(l_xml, link_atom_new(xml, free));

	if(ci->elements==0 && STRisVALID(ci->elasticity)) {
		xml = xml_make_element_text(ec, (utf8_t *)"elasticity", (utf8_t *)ci->elasticity);
		if(STRisVALID(xml)) LinkAppend(l_xml, link_atom_new(xml, free));
	}

	xml = xml_make_element_permission(ec, &ci->perm_tuple, FALSE);
	if(STRisVALID(xml)) {
		LinkAppend(l_xml, link_atom_new_sdup("<" EDB_XMLNS ":permtuple>"));
		LinkAppend(l_xml, link_atom_new(xml, free));
		LinkAppend(l_xml, link_atom_new_sdup("</" EDB_XMLNS ":permtuple>"));
	}

	if(ci->authorize) {
		xml = xml_make_element_text(ec, (utf8_t *)"attribute", (utf8_t *)"authorize");
		if(STRisVALID(xml)) LinkAppend(l_xml, link_atom_new(xml, free));
	}
	if(ci->exclusive) {
		xml = xml_make_element_text(ec, (utf8_t *)"attribute", (utf8_t *)"exclusive");
		if(STRisVALID(xml)) LinkAppend(l_xml, link_atom_new(xml, free));
	}
	if(ci->hierarchyparent) {
		xml = xml_make_element_text(ec, (utf8_t *)"attribute", (utf8_t *)"hierarchyparent");
		if(STRisVALID(xml)) LinkAppend(l_xml, link_atom_new(xml, free));
	}
	if(ci->multilingualcaption) {
		xml = xml_make_element_text(ec, (utf8_t *)"attribute", (utf8_t *)"multilingualcaption");
		if(STRisVALID(xml)) LinkAppend(l_xml, link_atom_new(xml, free));
	}
	if(ci->personname) {
		xml = xml_make_element_text(ec, (utf8_t *)"attribute", (utf8_t *)"personname");
		if(STRisVALID(xml)) LinkAppend(l_xml, link_atom_new(xml, free));
	}
	if(ci->sometimesStudent) {
		xml = xml_make_element_text(ec, (utf8_t *)"attribute", (utf8_t *)"student");
		if(STRisVALID(xml)) LinkAppend(l_xml, link_atom_new(xml, free));
	}
	if(ci->pagenumber) {
		xml = xml_make_element_text(ec, (utf8_t *)"attribute", (utf8_t *)"pagenumber");
		if(STRisVALID(xml)) LinkAppend(l_xml, link_atom_new(xml, free));
	}
	if(ci->patentnumber) {
		xml = xml_make_element_text(ec, (utf8_t *)"attribute", (utf8_t *)"patentnumber");
		if(STRisVALID(xml)) LinkAppend(l_xml, link_atom_new(xml, free));
	}
	if(ci->canbeNothing) {
		xml = xml_make_element_text(ec, (utf8_t *)"attribute", (utf8_t *)"nothing");
		if(STRisVALID(xml)) LinkAppend(l_xml, link_atom_new(xml, free));
	}
	if(ci->classify==CLASSIFY_PRIMARY) {
		xml = xml_make_element_text(ec, (utf8_t *)"attribute", (utf8_t *)"CLASSIFY");
		if(STRisVALID(xml)) LinkAppend(l_xml, link_atom_new(xml, free));
	} else if(ci->classify==CLASSIFY_SECONDARY) {
		xml = xml_make_element_text(ec, (utf8_t *)"attribute", (utf8_t *)"classify");
		if(STRisVALID(xml)) LinkAppend(l_xml, link_atom_new(xml, free));
	}
	if(ci->alias) {
		xml = xml_make_element_text(ec, (utf8_t *)"attribute", (utf8_t *)"alias");
		if(STRisVALID(xml)) LinkAppend(l_xml, link_atom_new(xml, free));
	}
	if(ci->chronological) {
		xml = xml_make_element_text(ec, (utf8_t *)"attribute", (utf8_t *)"chronological");
		if(STRisVALID(xml)) LinkAppend(l_xml, link_atom_new(xml, free));
	}
	if(ci->conductor) {
		xml = xml_make_element_text(ec, (utf8_t *)"attribute", (utf8_t *)"conductor");
		if(STRisVALID(xml)) LinkAppend(l_xml, link_atom_new(xml, free));
	}
	if(ci->usage!=0) {
		if(ci->usage==USAGE_REQUIRED) {
			xml = xml_make_element_text(ec, (utf8_t *)"attribute", (utf8_t *)"required");
		} else if(ci->usage==USAGE_PREFERRED) {
			xml = xml_make_element_text(ec, (utf8_t *)"attribute", (utf8_t *)"preferred");
		} else if(ci->usage==USAGE_OPTIONAL) {
			xml = xml_make_element_text(ec, (utf8_t *)"attribute", (utf8_t *)"optional");
		} else if(ci->usage==USAGE_UNUSED) {
			xml = xml_make_element_text(ec, (utf8_t *)"attribute", (utf8_t *)"unused");
		}
		if(STRisVALID(xml)) LinkAppend(l_xml, link_atom_new(xml, free));
	}
	if(ci->obsoleted) {
		xml = xml_make_element_text(ec, (utf8_t *)"attribute", (utf8_t *)"obsoleted");
		if(STRisVALID(xml)) LinkAppend(l_xml, link_atom_new(xml, free));
	}

	if(ci->conductor_usage!=0) {
		if(ci->conductor_usage==USAGE_REQUIRED) {
			xml = xml_make_element_text(ec, (utf8_t *)"conductor", (utf8_t *)"required");
		} else if(ci->conductor_usage==USAGE_PREFERRED) {
			xml = xml_make_element_text(ec, (utf8_t *)"conductor", (utf8_t *)"preferred");
		} else if(ci->conductor_usage==USAGE_OPTIONAL) {
			xml = xml_make_element_text(ec, (utf8_t *)"conductor", (utf8_t *)"optional");
		} else if(ci->conductor_usage==USAGE_UNUSED) {
			xml = xml_make_element_text(ec, (utf8_t *)"conductor", (utf8_t *)"unused");
		}
		if(STRisVALID(xml)) LinkAppend(l_xml, link_atom_new(xml, free));
	}

	ForLink(l, ci->l_maplookup) {
		EdbMapLookup	map = (EdbMapLookup)l->d;

		xml = xml_make_element_text(ec, (utf8_t *)"maplookup", (utf8_t *)map->lookup);
		if(STRisVALID(xml)) LinkAppend(l_xml, link_atom_new(xml, free));
	}


	xml = xml_make_element_text_integer(ec, (utf8_t *)"caption", ci->caption);
	if(STRisVALID(xml)) LinkAppend(l_xml, link_atom_new(xml, free));

	/* prefix */
	xml = xml_make_element_mtext(ec, ci->prefix.english, ci->prefix.japanese, NULL);
	if(STRisVALID(xml)) {
		LinkAppend(l_xml, link_atom_new_sdup("<" EDB_XMLNS ":prefix>"));
		LinkAppend(l_xml, link_atom_new(xml, free));
		LinkAppend(l_xml, link_atom_new_sdup("</" EDB_XMLNS ":prefix>"));
	}

	/* suffix */
	xml = xml_make_element_mtext(ec, ci->suffix.english, ci->suffix.japanese, NULL);
	if(STRisVALID(xml)) {
		LinkAppend(l_xml, link_atom_new_sdup("<" EDB_XMLNS ":suffix>"));
		LinkAppend(l_xml, link_atom_new(xml, free));
		LinkAppend(l_xml, link_atom_new_sdup("</" EDB_XMLNS ":suffix>"));
	}

	xml = xml_make_element_text_integer(ec, (utf8_t *)"order", ci->order);
	if(STRisVALID(xml)) LinkAppend(l_xml, link_atom_new(xml, free));

	xml = xml_make_element_text_integer(ec, (utf8_t *)"size", ci->size);
	if(STRisVALID(xml)) LinkAppend(l_xml, link_atom_new(xml, free));

	xml = (utf8_t *)str_concatenate0(l_xml);
	link_free(l_xml);
	return xml;
}


static
Link
edb_table_column_parse_xml(EdbContext ec, EdbTableInfo ti, EdbColumnInfo super, xmlDocPtr doc, xmlNodePtr cur)
{
	Link	l_column = NULL;

	while(cur != NULL) {
		if(!xmlStrcmp((xmlChar *)cur->name, (xmlChar *)"column")) {
			EdbColumnInfo	ci;
			Link	l;

			ci = edb_table_column_body_parse_xml(ec, ti, super, doc, cur->xmlChildrenNode);
			if(ci) {
				LinkAppend(l_column, link_atom_new(ci, edb_table_column_free));
			} else {
				edb_trace(ec, "edb_table_column_parse_xml: can't make column.\n");
			}
			ci ->l_child = edb_table_column_parse_xml(ec, ti, ci, doc, cur->xmlChildrenNode);
			ForLink(l, ci->l_child) ci->columns += ((EdbColumnInfo)l->d)->columns;
		}
		cur = cur->next;
	}
	return l_column;
}

static
utf8_t *
edb_table_column_make_xml(EdbContext ec, EdbTableInfo ti, Link l_column)
{
	register Link	l, l_xml;
	register utf8_t	*xml;

	l_xml = NULL;
	ForLink(l, l_column) {
		register EdbColumnInfo	ci = (EdbColumnInfo)l->d;

		if(ci==NULL) continue;
		LinkAppend(l_xml, link_atom_new_sdup("<" EDB_XMLNS ":column>"));
		xml = edb_table_column_body_make_xml(ec, ti, ci);
		LinkAppend(l_xml, link_atom_new(xml, free));
		if(ci ->l_child) {
			xml = edb_table_column_make_xml(ec, ti, ci->l_child);
			if(STRisVALID(xml)) LinkAppend(l_xml, link_atom_new(xml, free));
		}
		LinkAppend(l_xml, link_atom_new_sdup("</" EDB_XMLNS ":column>"));
	}
	xml = (utf8_t *)str_concatenate0(l_xml);
	link_free(l_xml);
	return xml;
}

static
void
edb_table_column_set(EdbContext ec, Link l_column, int *idx)
{
	Link	l;

	ForLink(l, l_column) {
		EdbColumnInfo	ci = (EdbColumnInfo)l->d;

		if(ci->typeInfo) {
			ci->ti->column[*idx] = ci;
			ci->idx = (*idx)++;
		} else {
			ci->idx = -1;
		}
		edb_table_column_set(ec, ci->l_child, idx);
	}
}

EdbTableInfo
edb_table_parse_xml(EdbContext ec, utf8_t *xmlstr)
{
	xmlDocPtr	doc;
	xmlNodePtr	root, base;
	EdbTableInfo	ti;
	Link	l, l_node;
	int	idx;
	utf8_t	*s;

	doc = xmlParseMemory((char *)xmlstr, xmlStrlen(xmlstr));
	if(doc==NULL) {
		edb_trace(ec, "edb_table_parse_xml: can't parse xml.\n");
		return NULL;
	}

	root = xmlDocGetRootElement(doc);
	if(root==NULL) {
		edb_trace(ec, "edb_table_parse_xml: can't get root element.\n");
		return NULL;
	}

	ti = (EdbTableInfo)malloc(sizeof(struct EDB_TABLE_INFO));
	if(ti == NULL) return NULL;
	bzero(ti, sizeof(struct EDB_TABLE_INFO));
	edb_base_init(ec, TABLE_Base(ti));

	base = xml_find_element_node(ec, doc, root, "table", XML_EN_BASE, NULL);
	edb_base_parse_xml(ec, TABLE_Base(ti), doc, base);
	ti->xmlname = xml_get_element_text(ec, doc, root, (utf8_t *)"table", (utf8_t *)"xmlname", NULL);
	ti->sqlname = xml_get_element_text(ec, doc, root, (utf8_t *)"table", (utf8_t *)"sqlname", NULL);
	if(!xml_get_element_mtext(ec, &ti->name.english, &ti->name.japanese, &ti->name.pronounce, doc, root, "table", "name", NULL)) {
		edb_trace(ec, "edb_table_parse_xml: can't get name of table.\n");
	}
	if(!xml_get_element_mtext(ec, &ti->desc.english, &ti->desc.japanese, NULL, doc, root, "table", "description", NULL)) {
		edb_trace(ec, "edb_table_parse_xml: can't get descriptoin of table.\n");
	}
	if(!xml_get_element_mtext(ec, &ti->explain.english, &ti->explain.japanese, NULL, doc, root, "table", "explain", NULL)) {
		edb_trace(ec, "edb_table_parse_xml: can't get explain of table.\n");
	}
	if(!xml_get_element_permission(ec, &ti->perm_tuple, doc, root, "table", "permtuple", NULL)) {
		edb_trace(ec, "edb_table_parse_xml: can't get permission for tuple.\n");
	}
	if(!xml_get_element_permission(ec, &ti->perm_column, doc, root, "table", "permcolumn", NULL)) {
		edb_trace(ec, "edb_table_parse_xml: can't get permission for column.\n");
	}
	l_node = xml_find_element_node_list(ec, doc, root, "table", "attribute", NULL);
	ForLink(l, l_node) {
		s = xml_get_node_text(ec, doc, (xmlNodePtr)l->d);
		if(s==NULL) continue;
		if(!xmlStrcasecmp((xmlChar *)s, (xmlChar *)"auxiliary")) {
			ti->auxiliary = TRUE;
		} else if(!xmlStrcasecmp((xmlChar *)s, (xmlChar *)"hierarchical")) {
			ti->hierarchical = TRUE;
		} else if(!xmlStrcasecmp((xmlChar *)s, (xmlChar *)"censorship")) {
			ti->censorship = TRUE;
		} else if(!xmlStrcasecmp((xmlChar *)s, (xmlChar *)"chronological")) {
			ti->chronological = TRUE;
		} else {
			edb_trace(ec, "edb_table_parse_xml: unknown attribute \"%s\".\n", s);
		}
		xmlFree(s);
	}
	link_free(l_node);
	ti->module = (char *)xml_get_element_text(ec, doc, root, (utf8_t *)"table", (utf8_t *)"module", NULL);

	l_node = xml_find_element_node_list(ec, doc, root, (utf8_t *)"table", (utf8_t *)"maplookup", NULL);
	ForLink(l, l_node) {
		s = xml_get_node_text(ec, doc, (xmlNodePtr)l->d);
		if(s) LinkAppend(ti->l_maplookup, link_atom_new(edb_maplookup_new(ec, s), edb_maplookup_free));
	}
	link_free(l_node);

	/* get column definition */
	ti->l_column = edb_table_column_parse_xml(ec, ti, NULL, doc, root->xmlChildrenNode);

	xmlFreeDoc(doc);

	ti->ncol = 0;
	ForLink(l, ti->l_column) ti->ncol += ((EdbColumnInfo)l->d)->columns;
	ti->column = (EdbColumnInfo *)malloc(sizeof(EdbColumnInfo)*ti->ncol);
	idx = 0;
	edb_table_column_set(ec, ti->l_column, &idx);
	return ti;
}

utf8_t *
edb_table_make_xml(EdbContext ec, EdbTableInfo ti)
{
	utf8_t	*base_xml, *body_xml, *xml;
	register Link	l, l_xml = NULL;

	base_xml = edb_base_make_xml(ec, TABLE_Base(ti), TRUE);

	xml = xml_make_element_text(ec, (utf8_t *)"xmlname", ti->xmlname);
	if(STRisVALID(xml)) LinkAppend(l_xml, link_atom_new(xml, free));

	xml = xml_make_element_text(ec, (utf8_t *)"sqlname", ti->sqlname);
	if(STRisVALID(xml)) LinkAppend(l_xml, link_atom_new(xml, free));

	xml = xml_make_element_mtext(ec, ti->name.english, ti->name.japanese, ti->name.pronounce);
	LinkAppend(l_xml, link_atom_new_sdup("<" EDB_XMLNS ":name>"));
	if(STRisVALID(xml)) LinkAppend(l_xml, link_atom_new(xml, free));
	LinkAppend(l_xml, link_atom_new_sdup("</" EDB_XMLNS ":name>"));

	LinkAppend(l_xml, link_atom_new_sdup("<" EDB_XMLNS ":description>"));
	xml = xml_make_element_mtext(ec, ti->desc.english, ti->desc.japanese, NULL);
	if(STRisVALID(xml)) LinkAppend(l_xml, link_atom_new(xml, free));
	LinkAppend(l_xml, link_atom_new_sdup("</" EDB_XMLNS ":description>"));

	LinkAppend(l_xml, link_atom_new_sdup("<" EDB_XMLNS ":explain>"));
	xml = xml_make_element_mtext(ec, ti->explain.english, ti->explain.japanese, NULL);
	if(STRisVALID(xml)) LinkAppend(l_xml, link_atom_new(xml, free));
	LinkAppend(l_xml, link_atom_new_sdup("</" EDB_XMLNS ":explain>"));

	xml = xml_make_element_permission(ec, &ti->perm_tuple, TRUE);
	if(STRisVALID(xml)) {
		LinkAppend(l_xml, link_atom_new_sdup("<" EDB_XMLNS ":permtuple>"));
		LinkAppend(l_xml, link_atom_new(xml, free));
		LinkAppend(l_xml, link_atom_new_sdup("</" EDB_XMLNS ":permtuple>"));
	}
	xml = xml_make_element_permission(ec, &ti->perm_column, TRUE);
	if(STRisVALID(xml)) {
		LinkAppend(l_xml, link_atom_new_sdup("<" EDB_XMLNS ":permcolumn>"));
		LinkAppend(l_xml, link_atom_new(xml, free));
		LinkAppend(l_xml, link_atom_new_sdup("</" EDB_XMLNS ":permcolumn>"));
	}

	if(ti->auxiliary) {
		xml = xml_make_element_text(ec, (utf8_t *)"attribute", (utf8_t *)"auxiliary");
		if(STRisVALID(xml)) LinkAppend(l_xml, link_atom_new(xml, free));
	}
	if(ti->hierarchical) {
		xml = xml_make_element_text(ec, (utf8_t *)"attribute", (utf8_t *)"hierarchical");
		if(STRisVALID(xml)) LinkAppend(l_xml, link_atom_new(xml, free));
	}
	if(ti->censorship) {
		xml = xml_make_element_text(ec, (utf8_t *)"attribute", (utf8_t *)"censorship");
		if(STRisVALID(xml)) LinkAppend(l_xml, link_atom_new(xml, free));
	}
	if(ti->chronological) {
		xml = xml_make_element_text(ec, (utf8_t *)"attribute", (utf8_t *)"chronological");
		if(STRisVALID(xml)) LinkAppend(l_xml, link_atom_new(xml, free));
	}

	if(STRisVALID(ti->module)) {
		xml = xml_make_element_text(ec, (utf8_t *)"module", (utf8_t *)ti->module);
		if(STRisVALID(xml)) LinkAppend(l_xml, link_atom_new(xml, free));
	}

	ForLink(l, ti->l_maplookup) {
		EdbMapLookup	map = (EdbMapLookup)l->d;

		xml = xml_make_element_text(ec, (utf8_t *)"maplookup", (utf8_t *)map->lookup);
		if(STRisVALID(xml)) LinkAppend(l_xml, link_atom_new(xml, free));
	}

	xml = edb_table_column_make_xml(ec, ti, ti->l_column);
	if(STRisVALID(xml)) LinkAppend(l_xml, link_atom_new(xml, free));

	body_xml = (utf8_t *)str_concatenate0(l_xml);
	link_free(l_xml);
	asprintf((char **)&xml, "<?xml version=\"%s\" encoding=\"utf-8\"?><!DOCTYPE %s:%s SYSTEM \"%s%s.dtd\"><%s:%s xmlns:%s=\"%s\">%s%s</%s:%s>",
		XML_DEFAULT_VERSION,
		EDB_XMLNS, "table", EDB_DTD_URI, "table",
		EDB_XMLNS, "table", EDB_XMLNS, EDB_XMLNS_URI,
		base_xml, body_xml,
		EDB_XMLNS, "table");
	FREE(base_xml);
	FREE(body_xml);
	return xml;
}

/*
	edb_table_column_reldb_elements() return the number of elements in specified column.
	If super_ci==NULL, then returns the nubmer of elements for entire nestings,
	else returns the number of elements until reaching to spuer_ci.
	If super_ci and ci are identical, this function assume the own elemement ci->elements is 1.
*/
int
edb_table_column_reldb_elements(EdbContext ec, EdbTableInfo ti, EdbColumnInfo super_ci, EdbColumnInfo ci)
{
	register int	elements;
	register EdbColumnInfo	sci;
	register Link	lm;

	if(ci->hierarchyparent) return 0;

#if 0
	elements = 1;
	for(sci=ci; elements!=0 && sci && sci!=super_ci; sci=sci->super) {
		elements *= sci->elements;
		if(elements==0) return 0;
	}
#else
	elements = 1;
	for(sci=ci; elements!=0 && sci && sci!=super_ci; sci=sci->super) {
		elements *= sci->elements;
		if(sci->chronological) elements = 0;
		if(elements==0) return 0;
	}
#endif

	ForLink(lm, ci->l_maplookup) {
		EdbMapLookup	map = (EdbMapLookup)lm->d;
		eid_t			meid;
		EdbTableInfo	mti;

		meid = edb_maplookup_table_eid(ec, map);
		if(!EIDisVALID(meid)) {
			edb_trace(ec, "edb_table_column_reldb_elements: can't get table eid.\n");
			continue;
		}
		mti = edb_table_get(ec, meid);
		if(mti==NULL) {
			edb_trace(ec, "edb_table_column_reldb_elements: can't get table.\n");
			continue;
		}
		if(mti->hierarchical) elements = 0;
		if(elements==0) break;
	}
	return elements;
}

int
edb_table_column_xml_elements(EdbContext ec, EdbTableInfo ti, EdbColumnInfo super_ci, EdbColumnInfo ci)
{
	register int	elements;
	register EdbColumnInfo	sci;

	elements = 1;
	for(sci=ci; elements!=0 && sci && sci!=super_ci; sci=sci->super) {
		elements *= sci->elements;
		if(sci->chronological) elements = 0;
		if(elements==0) return 0;
	}

	return elements;
}

#define	EDB_TABLE_XMLNAME		"edb_xmlname"
#define	EDB_TABLE_SQLNAME		"edb_sqlname"
#define	EDB_TABLE_NAME			"edb_name"
#define	EDB_TABLE_DESCRIPTION	"edb_desc"
#define	EDB_TABLE_EXPLAIN		"edb_ex"
#define	EDB_TABLE_ATTR			"edb_attr"
#define	EDB_TABLE_RDL			"edb_rd"
#define	EDB_TABLE_WRL			"edb_wr"

/*
	edb_table_new--- get info
	edb_table_get--- get info
	edb_table_put--- put info
	edb_talbe_dup--- duplicate info
	edb_table_free--- free info
*/
EdbTableInfo
edb_table_new(EdbContext ec)
{
	EdbTableInfo	ti;

	ti = (EdbTableInfo)malloc(sizeof(struct EDB_TABLE_INFO));
	if(ti == NULL) return NULL;
	bzero(ti, sizeof(struct EDB_TABLE_INFO));
	edb_base_init(ec, TABLE_Base(ti));
	TABLE_EID(ti) = 0;
	TABLE_Operator(ti) = ec->masterUID;
	TABLE_Read(ti) = EDB_AUTH_PUBLIC;
	TABLE_Write(ti) = EDB_AUTH_OWNER;
	TABLE_Delete(ti) = EDB_AUTH_OWNER;
	TABLE_Owner(ti) = ec->masterUID;
	TABLE_TupleRead(ti)	= EDB_AUTH_PUBLIC;
	TABLE_TupleCreate(ti)	= EDB_AUTH_USER;
	TABLE_TupleWrite(ti)	= EDB_AUTH_LIMITED;
	TABLE_TupleDelete(ti)	= EDB_AUTH_LIMITED;
	TABLE_ColumnRead(ti)	= EDB_AUTH_PUBLIC;
	TABLE_ColumnCreate(ti)	= EDB_AUTH_LIMITED;
	TABLE_ColumnWrite(ti)	= EDB_AUTH_LIMITED;
	TABLE_ColumnDelete(ti)	= EDB_AUTH_LIMITED;
	ti->xmlname = (utf8_t *)strdup("xmlname");
	ti->sqlname = (utf8_t *)strdup("t_00000");
	ti->name.english = (utf8_t *)strdup("EnglishName");
	ti->name.japanese = (utf8_t *)strdup("テーブルの名前");
	ti->name.pronounce = (utf8_t *)strdup("てーぶるのなまえ");
	ti->desc.english = (utf8_t *)strdup("description of table");
	ti->desc.japanese = (utf8_t *)strdup("テーブルの説明");
	ti->explain.english = (utf8_t *)strdup("explain of table");
	ti->explain.japanese = (utf8_t *)strdup("テーブルの説明");
	return ti;
}

EdbTableInfo
edb_table_get_XMLDB(EdbContext ec, eid_t eid, eoid_t eoid)
{
	utf8_t *query, *select;
	utf8_t *latest;
	EdbTableInfo	ti;
	utf8_t	*xml;

	select = (utf8_t *)"SELECT " EDB_XMLDB_COLUMN_XML " FROM " EDB_XMLDB_TABLE " AS t1 WHERE ";
	if(eoid) {
		asprintf((char **)&query, "%s t1.eoid = %d", select, eoid);
	} else {
		latest = edb_create_latest_phrase(ec, (utf8_t *)EDB_XMLDB_TABLE, (utf8_t *)"t1", (utf8_t *)"t2", NULL, eid);
		asprintf((char **)&query, "%s t1.%s = %d AND %s", select, EDB_BASE_EID, eid, latest);
		FREE(latest);
	}
	xml = edb_pgsql_query_text(ec, query);
	if(xml==NULL) {
		edb_trace(ec, "edb_table_get_XMLDB: can't get table information (%d, %d). query (%s).\n", eid, eoid, query);
		FREE(query);
		return NULL;
	}
	FREE(query);
	ti = edb_table_parse_xml(ec, xml);
	FREE(xml);

	return ti;
}

struct TABLE_CHECK {
	int	shortcaptions;
	int	status;
};

static
void
edb_table_column_check(EdbContext ec, utf8_t *xmlsuper, Link l_column, struct TABLE_CHECK *table_check)
{
	register Link	l;

	ForLink(l, l_column) {
		EdbColumnInfo	ci = (EdbColumnInfo)l->d;
		register Link	ls;

		ForLink(ls, l->n) {
			EdbColumnInfo	cis = (EdbColumnInfo)ls->d;

			if(!strcasecmp((char *)ci->xmlname, (char *)cis->xmlname)) {
				edb_trace(ec, "edb_table_column_check: xmlname were not unique.\n");
				table_check->status = FALSE;
			}
			if(STRisVALID(ci->sqlname) && STRisVALID(cis->sqlname) && !strcasecmp((char *)ci->sqlname, (char *)cis->sqlname)) {
				edb_trace(ec, "edb_table_column_check: sqlname were not unique.\n");
				table_check->status = FALSE;
			}
		}
	}

	ForLink(l, l_column) {
		EdbColumnInfo	ci = (EdbColumnInfo)l->d;
		int	len;

		if(ci->caption==1) ++table_check->shortcaptions;
		if(!STRisVALID(ci->xmlname)) {
			edb_trace(ec, "edb_table_column_check: invalid xmlname.\n");
			table_check->status = FALSE;
			continue;
		}
		len = strlen((char *)xmlsuper);
		if(strncmp((char *)ci->xmlname, (char *)xmlsuper, len) || ci->xmlname[len]!='.' || ci->xmlname[len+1]=='\0') {
			edb_trace(ec, "edb_table_column_check: xmlname is not in hierarchy.\n");
			table_check->status = FALSE;
			continue;
		}
		if(!STRisVALID(ci->name.english) || !str_isascii((char *)ci->name.english)) {
			edb_trace(ec, "edb_table_column_check: %s : invalid name.english.\n", ci->xmlname);
			table_check->status = FALSE;
		}
		if(!STRisVALID(ci->name.japanese)) {
			edb_trace(ec, "edb_table_column_check: %s : name.japanese is not set.\n", ci->xmlname);
			table_check->status = FALSE;
		}
		if(!STRisVALID(ci->name.pronounce) || !str_iskana((char *)ci->name.pronounce)) {
			edb_trace(ec, "edb_table_column_check: %s : invalid name.pronounce.\n", ci->xmlname);
			table_check->status = FALSE;
		}
		if(!STRisVALID(ci->desc.english) || !str_isascii((char *)ci->desc.english)) {
			edb_trace(ec, "edb_table_column_check: %s : invalid desc.english.\n", ci->xmlname);
			table_check->status = FALSE;
		}
		if(!STRisVALID(ci->desc.japanese)) {
			edb_trace(ec, "edb_table_column_check: %s : desc.japanese is not set.\n", ci->xmlname);
			table_check->status = FALSE;
		}
		if(!STRisVALID(ci->explain.english) || !str_isascii((char *)ci->explain.english)) {
			edb_trace(ec, "edb_table_column_check: %s : invalid explain.english.\n", ci->xmlname);
			table_check->status = FALSE;
		}
		if(!STRisVALID(ci->explain.japanese)) {
			edb_trace(ec, "edb_table_column_check: %s : explain.japanese is not set.\n", ci->xmlname);
			table_check->status = FALSE;
		}
		if(ci->l_child==NULL) {
			if(!STRisVALID(ci->sqlname)) {
				edb_trace(ec, "edb_table_column_check: column (%s) sqlname was not set (%s).\n", ci->xmlname, ci->sqlname);
				table_check->status = FALSE;
			} else if(strncmp((char *)ci->sqlname, "c_", 2)) {
				edb_trace(ec, "edb_table_column_check: WARNING : column (%s) sqlname was not start with \"c_\" (%s).\n", ci->xmlname, ci->sqlname);
#if 0
				table_check->status = FALSE;
#endif
			}
			if(!STRisVALID(ColumnTypeName(ci))) {
				edb_trace(ec, "edb_table_column_check: column (%s) type was not set (%s).\n", ci->xmlname, ci->sqlname);
				table_check->status = FALSE;
			}
#if 0
変更: サブカラムの要素数は任意にする．
			if(ci->super && ci->elements==0) {
				edb_trace(ec, "edb_table_column_check: column.column (%s) elements == 0.\n", ci->xmlname);
				table_check->status = FALSE;
			}
#endif
		}
		if(ci->classify) {
			switch(ci->typeInfo->typeID) {
			case EDB_TYPE_EID:
			case EDB_TYPE_NAME:
			case EDB_TYPE_SENTENCE:
			case EDB_TYPE_YEAR:
			case EDB_TYPE_DATE:
			case EDB_TYPE_DATE2:
				break;
			default:
				edb_trace(ec, "edb_table_column_check: column.column (%s)'s classify is bad.\n", ci->xmlname);
				table_check->status = FALSE;
				break;
			}
		}
		if(ci->l_child) edb_table_column_check(ec, ci->xmlname, ci->l_child, table_check);
	}
}

int
edb_table_check(EdbContext ec, EdbTableInfo ti)
{
	register int	i;
	struct TABLE_CHECK	table_check;

	table_check.status = TRUE;
	table_check.shortcaptions = 0;

	if(!STRisVALID(ti->xmlname)) {
		edb_trace(ec, "edb_table_check: invalid xmlname.\n");
		table_check.status = FALSE;
	}
	if(!STRisVALID(ti->sqlname)) {
		edb_trace(ec, "edb_table_check: invalid sqlname.\n");
		table_check.status = FALSE;
	}
	if(!STRisVALID(ti->name.english) || !str_isascii((char *)ti->name.english)) {
		edb_trace(ec, "edb_table_check: invalid name.english.\n");
		table_check.status = FALSE;
	}
	if(!STRisVALID(ti->name.japanese)) {
		edb_trace(ec, "edb_table_check: name.japanese is not set.\n");
		table_check.status = FALSE;
	}
	if(!STRisVALID(ti->name.pronounce) || !str_iskana((char *)ti->name.pronounce)) {
		edb_trace(ec, "edb_table_check: invalid name.pronounce.\n");
		table_check.status = FALSE;
	}
	if(!STRisVALID(ti->desc.english) || !str_isascii((char *)ti->desc.english)) {
		edb_trace(ec, "edb_table_check: desc.english is not set.\n");
		table_check.status = FALSE;
	}
	if(!STRisVALID(ti->desc.japanese)) {
		edb_trace(ec, "edb_table_check: desc.japanese is not set.\n");
		table_check.status = FALSE;
	}
#if 0
	if(!STRisVALID(ti->explain.english) || !str_isascii((char *)ti->explain.english)) {
		edb_trace(ec, "edb_table_check: invalid explain.english.\n");
		table_check.status = FALSE;
	}
	if(!STRisVALID(ti->explain.japanese)) {
		edb_trace(ec, "edb_table_check: explain.japanese is not set.\n");
		table_check.status = FALSE;
	}
#endif
	edb_table_column_check(ec, ti->xmlname, ti->l_column, &table_check);
	for(i=0; i<ti->ncol; ++i) {
		register int	j;

		for(j=i+1; j<ti->ncol; ++j) {
			if(!strcasecmp((char *)ti->column[i]->xmlname, (char *)ti->column[j]->xmlname)) {
				edb_trace(ec, "edb_table_check: %s : xmlname (%s) were duplicated.\n", ti->xmlname, ti->column[i]->xmlname);
				table_check.status = FALSE;
			}
			if(!strcasecmp((char *)ti->column[i]->sqlname, (char *)ti->column[j]->sqlname)) {
				edb_trace(ec, "edb_table_check: %s : sqlname (%s) were duplicated.\n", ti->xmlname, ti->column[i]->sqlname);
				table_check.status = FALSE;
			}
			if(!strcasecmp((char *)ti->column[i]->name.english, (char *)ti->column[j]->name.english) || !strcasecmp((char *)ti->column[i]->name.english, (char *)ti->column[j]->name.japanese)) {
				edb_trace(ec, "edb_table_check: %s : name.english (%s) were duplicated.\n", ti->xmlname, ti->column[i]->name.english);
				table_check.status = FALSE;
			}
			if(!strcasecmp((char *)ti->column[i]->name.japanese, (char *)ti->column[j]->name.english) || !strcasecmp((char *)ti->column[i]->name.japanese, (char *)ti->column[j]->name.japanese)) {
				edb_trace(ec, "edb_table_check: %s : name.japanese (%s) were duplicated.\n", ti->xmlname, ti->column[i]->name.japanese);
				table_check.status = FALSE;
			}
		}
	}
	if(table_check.shortcaptions==0) {
		edb_trace(ec, "edb_table_check: %s : number of shortcaption == 0\n", ti->xmlname);
		table_check.status = FALSE;
	}

	return table_check.status;
}

EdbTableInfo
edb_table_get(EdbContext ec, eid_t eid)
{
	EdbTableInfo	ti;

	ti = edb_eid_list_restore(ec, eid);
	if(ti) return ti;

	ti = edb_table_get_XMLDB(ec, eid, 0);

	if(ti) ti = edb_eid_list_store(ec, eid, EDB_EID_IS_TABLE, ti);
	return ti;
}

EdbTableInfo
edb_table_dup(EdbContext ec, EdbTableInfo src)
{
	EdbTableInfo	ti;

	if(src == NULL) return NULL;
	ti = (EdbTableInfo)malloc(sizeof(struct EDB_TABLE_INFO));
	if(ti == NULL) return NULL;
	bcopy(src, ti, sizeof(struct EDB_TABLE_INFO));
	edb_base_init(ec, TABLE_Base(ti));

	ti->sqlname		= (utf8_t *)strdup_null((char *)ti->sqlname);
	ti->name.english	= (utf8_t *)strdup_null((char *)ti->name.english);
	ti->name.japanese	= (utf8_t *)strdup_null((char *)ti->name.japanese);
	ti->name.pronounce	= (utf8_t *)strdup_null((char *)ti->name.pronounce);
	ti->desc.english		= (utf8_t *)strdup_null((char *)ti->desc.english);
	ti->desc.japanese	= (utf8_t *)strdup_null((char *)ti->desc.japanese);
	ti->explain.english	= (utf8_t *)strdup_null((char *)ti->explain.english);
	ti->explain.japanese	= (utf8_t *)strdup_null((char *)ti->explain.japanese);

#if 0
	ti->column = (EdbColumnInfo *)malloc(sizeof(EdbColumnInfo)*ti->ncol);
	for(i=0; i<ti->ncol; ++i) ti->column[i] = edb_column_dup(ec, src->column[i]);
#endif
	return ti;
}

void
edb_table_free(EdbContext ec, EdbTableInfo ti)
{

	if(ti == NULL) return;
	FREE(ti->xmlname);
	FREE(ti->sqlname);
	FREE(ti->name.english);
	FREE(ti->name.japanese);
	FREE(ti->name.pronounce);
	FREE(ti->desc.english);
	FREE(ti->desc.japanese);
	FREE(ti->explain.english);
	FREE(ti->explain.japanese);
	EdbLinkFree(ti->l_conductors);
	EdbLinkFree2(ec, ti->l_maplookup);
	EdbLinkFree2(ec, ti->l_column);
	FREE(ti->column);
	free(ti);
}

EdbTableInfo
edb_table_get_by_name(EdbContext ec, utf8_t *name)
{
	register eid_t	eid;
	register EdbTableInfo	ti;

	if(ec->ca_all_tables) {
		register Link	l;

		ForCatalogueLink(l, ec->ca_all_tables) {
			ti = CatalogueObject(l);
			if(!strcmp((char *)TABLE_XN(ti), (char *)name)) return ti;
		}
		return NULL;
	} else if(ec->ca_tables) {
		register Link	l;

		ForCatalogueLink(l, ec->ca_tables) {
			ti = CatalogueObject(l);
			if(!strcmp((char *)TABLE_XN(ti), (char *)name)) return ti;
		}
	}
	eid = edb_table_xmlname_to_eid(ec, name);
	if(!EIDisVALID(eid)) return NULL;
	ti = edb_table_get(ec, eid);
	if(ti) {
		if(ec->ca_tables==NULL) ec->ca_tables = edb_catalogue_new();
		edb_catalogue_append(ec->ca_tables, TABLE_EID(ti), ti);
	}
	return ti;
}

/*
	edb_talbe_create -- new table
*/
struct NAME2EID {
	eid_t	eid;
	char	*xn;
	char	*sn;
	char	*en;
	char	*ja;
};

static
void
edb_table_get_names(EdbContext ec)
{

	if(!ec->ca_table_name2eid) {
		register struct	NAME2EID *n2e;
		utf8_t	*query;
		int	nt, nf;
		Link	l1, l2, lres;

		ec->ca_table_name2eid = edb_catalogue_new();
		asprintf((char **)&query, "SELECT %s,%s,%s,english(%s),japanese(%s) FROM %s ORDER BY %s",
			EDB_BASE_EID,
			EDB_TABLE_XMLNAME,
			EDB_TABLE_SQLNAME,
			EDB_TABLE_NAME,
			EDB_TABLE_NAME,
			EDB_TABLE_TABLE,
			EDB_BASE_EID);
		lres = edb_pgsql_query(ec, query, &nt, &nf);
		FREE(query);
		if(nf==5) {
			ForLink(l1, lres) {
				l2 = l1->d;
				if(link_atoms(l2)==5) {
					n2e = (struct NAME2EID *)malloc(sizeof(struct NAME2EID));
					n2e->eid = ATOI(l2->d);
					n2e->xn = strdup(l2->n->d);
					n2e->sn = strdup(l2->n->n->d);
					n2e->en = strdup(l2->n->n->n->d);
					n2e->ja = strdup(l2->n->n->n->n->d);
					edb_catalogue_append(ec->ca_table_name2eid, n2e->eid, n2e);
				}
			}
		}
		link_free(lres);
	}
}

eid_t
edb_table_name_to_eid(EdbContext ec, utf8_t *tablename)
{
#if 0
	char	*query;
	eid_t	eid;

	asprintf(&query, "SELECT %s FROM %s WHERE english(%s) = '%s' OR japanese(%s) = '%s'",
		EDB_BASE_EID,
		EDB_TABLE_TABLE,
		EDB_TABLE_NAME, tablename,
		EDB_TABLE_NAME, tablename);
	eid = edb_pgsql_query_int(ec, query);
	FREE(query);
	return eid;
#else
	Link	l;

	edb_table_get_names(ec);
	ForCatalogueLink(l, ec->ca_table_name2eid) {
		register struct	NAME2EID *n2e = CatalogueObject(l);

		if(!strcmp(n2e->en, (char *)tablename) || !strcmp(n2e->ja, (char *)tablename)) return n2e->eid;
	}
	return 0;
#endif
}

eid_t
edb_table_xmlname_to_eid(EdbContext ec, utf8_t *tablename)
{
#if 0
	char	*query;
	eid_t	eid;

	asprintf(&query, "SELECT %s FROM %s WHERE %s='%s'",
		EDB_BASE_EID,
		EDB_TABLE_TABLE,
		EDB_TABLE_XMLNAME, tablename);
	eid = edb_pgsql_query_int(ec, query);
	FREE(query);
	return eid;
#else
	Link	l;

	edb_table_get_names(ec);
	ForCatalogueLink(l, ec->ca_table_name2eid) {
		register struct	NAME2EID *n2e = CatalogueObject(l);

		if(!strcmp(n2e->xn, (char *)tablename)) return n2e->eid;
	}
	return 0;
#endif
}

eid_t
edb_table_sqlname_to_eid(EdbContext ec, utf8_t *sqlname)
{
#if 0
	return edb_EID_get(ec, EDB_TABLE_TABLE, "%s='%s'", EDB_TABLE_SQLNAME, sqlname);
#else
	Link	l;

	edb_table_get_names(ec);
	ForCatalogueLink(l, ec->ca_table_name2eid) {
		register struct	NAME2EID *n2e = CatalogueObject(l);

		if(!strcmp(n2e->sn, (char *)sqlname)) return n2e->eid;
	}
	return 0;
#endif
}


utf8_t *
edb_table_make_order(EdbContext ec, EdbTableInfo ti, utf8_t *as)
{
	register int	i, o;
	register Link	l_order = NULL;
	char	*buf;
	utf8_t	*order;

	for(o=1; o<=EDB_ORDER_MAX; ++o) {
		for(i=0; i<ti->ncol; ++i) {
			EdbColumnInfo	ci = ti->column[i];

			if(ci->order==o) {
				int	elements = edb_table_column_reldb_elements(ec, ti, NULL, ci);

				if(elements==1) {
					if(ci->typeInfo->typeID==EDB_TYPE_IPv4CIDR || ci->typeInfo->typeID==EDB_TYPE_IPv4ADDR) {
						asprintf(&buf, "%s.%s::inet", as ? as : ti->sqlname, ci->sqlname);
						LinkAppend(l_order, link_atom_new(buf, free));
					} else {
						asprintf(&buf, "%s.%s", as ? as : ti->sqlname, ci->sqlname);
						LinkAppend(l_order, link_atom_new(buf, free));
					}
				} else {
					asprintf(&buf, "%s.%s[1]", as ? as : ti->sqlname, ci->sqlname);
					LinkAppend(l_order, link_atom_new(buf, free));
				}
			}
		}
	}
	asprintf(&buf, "%s.%s", as ? as : ti->sqlname, EDB_BASE_EID);
	LinkAppend(l_order, link_atom_new(buf, free));

	order = (utf8_t *)str_concatenate(l_order, ',');
	link_free(l_order);
	return order;
}

Link
edb_table_get_eid_of_tuples(EdbContext ec, EdbTableInfo ti, char *fmt, ...)
{
	utf8_t *order;
	utf8_t *query;
	utf8_t *target;
	Link	l_eid;
	int	n;
	va_list	ap;

	order = edb_table_make_order(ec, ti, (utf8_t *)"t1");
	if(fmt) {
		va_start(ap, fmt);
		vasprintf((char **)&target, fmt, ap);
		va_end(ap);
		asprintf((char **)&query,
			"SELECT %s FROM %s AS t1 WHERE %s %s %s",
			EDB_BASE_EID,
			ti->sqlname,
			target,
			STRisVALID(order) ? "ORDER BY" : "", order);
		FREE(target);
	} else {
		asprintf((char **)&query,
			"SELECT %s FROM %s AS t1 %s %s",
			EDB_BASE_EID,
			ti->sqlname,
			STRisVALID(order) ? "ORDER BY" : "", order);
	}
	FREE(order);
	l_eid = edb_pgsql_query_one_field(ec, query, &n);
	FREE(query);
	return l_eid;
}

Link
edb_table_get_eid_of_available_tuples(EdbContext ec, EdbTableInfo ti, char *fmt, ...)
{
	utf8_t *order;
	utf8_t *query;
	utf8_t *target;
	Link	l_eid;
	int	n;
	va_list	ap;

	order = edb_table_make_order(ec, ti, (utf8_t *)"t1");
	if(fmt) {
		va_start(ap, fmt);
		vasprintf((char **)&target, fmt, ap);
		va_end(ap);
		asprintf((char **)&query,
			"SELECT %s FROM %s AS t1 WHERE t1.av AND %s %s %s",
			EDB_BASE_EID,
			ti->sqlname,
			target,
			STRisVALID(order) ? "ORDER BY" : "", order);
		FREE(target);
	} else {
		asprintf((char **)&query,
			"SELECT %s FROM %s AS t1 WHERE t1.av %s %s",
			EDB_BASE_EID,
			ti->sqlname,
			STRisVALID(order) ? "ORDER BY" : "", order);
	}
	FREE(order);
	l_eid = edb_pgsql_query_one_field(ec, query, &n);
	FREE(query);
	return l_eid;
}

void
edb_table_prefetch_tuples(EdbContext ec, EdbTableInfo ti)
{
	Link	l, lt, l_eid;
	EdbTupleInfo	tuple;

	if(ti->ca_all_tuples) return;
	lt = edb_tuple_get_XMLDBs(ec, ti);
	ForLink(l, lt) {
		tuple = (EdbTupleInfo)l->d;

		tuple = edb_eid_list_store(ec, TupleEID(tuple), EDB_EID_IS_TUPLE, tuple);
		l->d = NULL;
	}
	EdbLinkFree2(ec, lt);
	ti->ca_all_tuples = edb_catalogue_new();
	l_eid = edb_table_get_eid_of_tuples(ec, ti, NULL);
	ForLink(l, l_eid) {
		tuple = edb_tuple_get(ec, ATOI(l->d), EDB_TUPLE_GET_template_table, ti, 0);
		if(tuple) edb_catalogue_append(ti->ca_all_tuples, TupleEID(tuple), tuple);
	}
	link_free(l_eid);
}

Link
edb_table_get_tuples(EdbContext ec, EdbTableInfo ti, char *fmt, ...)
{
	utf8_t *order;
	utf8_t *query;
	utf8_t *target;
	Link	l_eid, l_tuple, l;
	int	n;
	va_list	ap;

	order = edb_table_make_order(ec, ti, (utf8_t *)"t1");
	if(fmt) {
		va_start(ap, fmt);
		vasprintf((char **)&target, fmt, ap);
		va_end(ap);
		asprintf((char **)&query,
			"SELECT %s FROM %s AS t1 WHERE %s %s %s",
			EDB_BASE_EID,
			ti->sqlname,
			target,
			STRisVALID(order) ? "ORDER BY" : "", order);
		FREE(target);
	} else {
		asprintf((char **)&query,
			"SELECT %s FROM %s AS t1 %s %s",
			EDB_BASE_EID,
			ti->sqlname,
			STRisVALID(order) ? "ORDER BY" : "", order);
	}
	FREE(order);
	l_eid = edb_pgsql_query_one_field(ec, query, &n);
	FREE(query);
	l_tuple = NULL;
	ForLink(l, l_eid) {
		LinkAppend(l_tuple, link_atom_new(edb_tuple_get(ec, ATOI((char *)l->d), EDB_TUPLE_GET_template_table, ti, 0), NULL));
	}
	link_free(l_eid);
	return l_tuple;
}

Link
edb_table_get_ordered_tuples(EdbContext ec, EdbTableInfo ti, Link l_order, char *fmt, ...)
{
	utf8_t *order;
	utf8_t *query;
	utf8_t *target;
	Link	l_eid, l_tuple, l;
	int	n;
	va_list	ap;

	order = edb_order_make(ec, ti, (utf8_t *)"t1", l_order);
	if(fmt) {
		va_start(ap, fmt);
		vasprintf((char **)&target, fmt, ap);
		va_end(ap);
		asprintf((char **)&query,
			"SELECT %s FROM %s AS t1 WHERE %s %s %s",
			EDB_BASE_EID,
			ti->sqlname,
			target,
			STRisVALID(order) ? "ORDER BY" : "", order);
		FREE(target);
	} else {
		asprintf((char **)&query,
			"SELECT %s FROM %s AS t1 %s %s",
			EDB_BASE_EID,
			ti->sqlname,
			STRisVALID(order) ? "ORDER BY" : "", order);
	}
	FREE(order);
	l_eid = edb_pgsql_query_one_field(ec, query, &n);
	FREE(query);
	l_tuple = NULL;
	ForLink(l, l_eid) {
		LinkAppend(l_tuple, link_atom_new(edb_tuple_get(ec, ATOI((char *)l->d), EDB_TUPLE_GET_template_table, ti, 0), NULL));
	}
	link_free(l_eid);
	return l_tuple;
}

Link
edb_table_get_tuples_for_look(EdbContext ec, EdbTableInfo ti, Link l_order, char *fmt, ...)
{
	utf8_t *order, *query, *target = NULL;
	Link	l_eid, l_tuple, l;
	int	n;
	va_list	ap;

	if(l_order) {
		order = edb_order_make(ec, ti, (utf8_t *)"t1", l_order);
	} else {
		order = edb_table_make_order(ec, ti, (utf8_t *)"t1");
	}
	if(fmt) {
		va_start(ap, fmt);
		vasprintf((char **)&target, fmt, ap);
		va_end(ap);
	}
	asprintf((char **)&query,
		"SELECT t1.%s FROM %s AS t1, %s AS t2 WHERE t1.eoid=t2.eoid %s%s %s%s",
		EDB_BASE_EID,
		ti->sqlname,
		EDB_INDEX_TABLE,
		STRisVALID(target) ? " AND " : "",
		STRisVALID(target) ? target : (utf8_t *)"",
		STRisVALID(order) ? " ORDER BY " : "",
		STRisVALID(order) ? order : (utf8_t *)"");
	FREE(target);
	FREE(order);
	l_eid = edb_pgsql_query_one_field(ec, query, &n);
	FREE(query);
	l_tuple = NULL;
	ForLink(l, l_eid) {
		LinkAppend(l_tuple, link_atom_new(edb_tuple_get(ec, ATOI((char *)l->d), EDB_TUPLE_GET_template_table, ti, 0), NULL));
	}
	link_free(l_eid);
	return l_tuple;
}

#if 0
Link
edb_table_get_tuples_use_index(EdbContext ec, EdbTableInfo ti, char *fmt, ...)
{
	utf8_t *order;
	utf8_t *query;
	utf8_t *latest;
	utf8_t *target;
	Link	l_eid, l_tuple, l;
	int	n;
	va_list	ap;

	order = edb_table_make_order(ec, ti, (utf8_t *)"t1");
	latest = strdup("t1.oid = (SELECT t2.edb_oid FROM " EDB_INDEX_TABLE " AS t2 WHERE t2.eid=t1.eid)");
	if(fmt) {
		va_start(ap, fmt);
		vasprintf((char **)&target, fmt, ap);
		va_end(ap);
		asprintf((char **)&query,
			"SELECT %s FROM %s AS t1 WHERE %s AND %s %s %s",
			EDB_BASE_EID,
			ti->sqlname,
			latest,
			target,
			STRisVALID(order) ? "ORDER BY" : "", order);
		FREE(target);
	} else {
		asprintf((char **)&query,
			"SELECT %s FROM %s AS t1 WHERE %s %s %s",
			EDB_BASE_EID,
			ti->sqlname,
			latest,
			STRisVALID(order) ? "ORDER BY" : "", order);
	}
	FREE(latest);
	FREE(order);
	l_eid = edb_pgsql_query_one_field(ec, query, &n);
	FREE(query);
	l_tuple = NULL;
	ForLink(l, l_eid) {
		LinkAppend(l_tuple, link_atom_new(edb_tuple_get(ec, ATOI((char *)l->d), EDB_TUPLE_GET_template_table, ti, 0), NULL));
	}
	link_free(l_eid);
	return l_tuple;
}
#endif

#if 0
/*
	edb_table_create_column--- create new column
	edb_table_delete_column--- create new column
*/

int
edb_table_put_column(EdbContext ec, EdbTableInfo ti, EdbColumnInfo ci)
{
	char	sqlname[BUFSIZ];
	char	*command;
	char	*type;

	/* 同じ名前のカラムがあればエラー */
	if(edb_table_seek_column(ec, ti, ci->name.english) || edb_table_seek_column(ec, ti, ci->name.japanese)) return FALSE;
	/* 登録 */
	type = edb_datum_type_to_sql_type(ec, ci->ctype.name);
	sprintf(sqlname, EDB_COLUMN_FMT, ti->ncol+1);
	asprintf(&command, "ALTER TABLE %s ADD COLUMN %s %s%s", ti->sqlname, sqlname, type, ci->elements==1 ? "" : "[]");
	if(!edb_pgsql_command_exec(ec, command)) {
		edb_trace(ec, "ERROR %s\n", command);
	}
#if 0
	edb_column_put(ec, ci, 0);
#endif
	FREE(command);
	return TRUE;
}


int
edb_table_create_column(EdbContext ec, EdbTableInfo ti, int ac, char *av[])
{
#if 0
	char	sqlname[BUFSIZ];
	char	*name_e, *name_j, *name_p;
	EdbColumnInfo	ci;
	int	v;
	Link	l_attr = NULL;
	eid_t	mapto;

	if(ac<15) return 0;
	/*チェック */
	if(!str_isdigit(av[1])) return 0;
	if(!str_isdigit(av[3])) return 0;
	if(!str_isdigit(av[5])) return 0;
	if(!str_isdigit(av[6])) return 0;
	if(!str_isdigit(av[8])) return 0;
	if(!str_isdigit(av[9])) return 0;
	if(!str_isdigit(av[10])) return 0;
	/* 属性のチェック */ 
	l_attr = NULL;
	if(av[4] && strcmp(av[4], ".")) {
		Link	l, lstr;

		lstr = str_split(av[4], ',');
		ForLink(l, lstr) {
			eid_t	teid;

			teid = edb_table_sqlname_to_eid(ec, (char *)l->d);
			if(teid==0) {
				fprintf(stderr, "table: %s is not found\n", (char *)l->d);
				return;
			} else {
				LinkAppend(l_attr, link_atom_new(eid2refstrdup(teid), free));
			}
		}
		link_free(lstr);
	}
	/* 作成 */
	ci = edb_column_new(ec);
	ColumnEID(ci) = 0;
	ci->teid = TABLE_EID(ti);
	ci->col = atoi(av[1]);
	ci->type = strdup(av[2]);
	ci->typeInfo = edb_datum_type_to(ec, ci->type);
	ci->elements = atoi(av[3]);
	if(l_attr) {
		ci->attr = str_concatenate(l_attr, ',');
		link_free(l_attr);
	} else {
		ci->attr = strdup("");
	}
	ColumnTupleRead(ci) = atoi(av[5]);
	ColumnTupleWrite(ci) = atoi(av[6]);
	ci->check = strcmp(av[7], ".") ? strdup(av[7]) : strdup("");

	ci->caption = atoi(av[8]);
	ci->order = atoi(av[9]);
	ci->size = atoi(av[10]);
	sscanf(av[11], "%x", &v);
	ci->flag = v;

	edb_mtext_split(ec, &mapto, &name_e, &name_j, &name_p, av[12]);
	ci->name.english = strdup(name_e);
	ci->name.japanese = strdup(name_j);
	ci->name.pronounce = strdup(name_p);
	FREE(name_e);
	FREE(name_j);
	FREE(name_p);

	edb_mtext_split(ec, &mapto, &name_e, &name_j, &name_p, av[13]);
	ci->desc.english = strdup(name_e);
	ci->desc.japanese = strdup(name_j);
	FREE(name_e);
	FREE(name_j);
	FREE(name_p);

	edb_mtext_split(ec, &mapto, &name_e, &name_j, &name_p, av[14]);
	ci->explain.english = strdup(name_e);
	ci->explain.japanese = strdup(name_j);
	FREE(name_e);
	FREE(name_j);
	FREE(name_p);

	/* 同じ名前のカラムがあればエラー */
	if(edb_table_seek_column(ec, ti, ci->name.english) || edb_table_seek_column(ec, ti, ci->name.japanese)) return FALSE;
	return edb_table_put_column(ec, ti, ci);
#endif
	return TRUE;
}
#endif

/*
	seek column
*/
EdbColumnInfo
edb_table_seek_column(EdbContext ec, EdbTableInfo ti, utf8_t *name)
{
	register EdbColumnInfo	ci;

	if((ci=edb_table_seek_column_by_xmlname(ec, ti, name))) return ci;
	if((ci=edb_table_seek_column_by_sqlname(ec, ti, name))) return ci;
	if((ci=edb_table_seek_column_by_name(ec, ti, name))) return ci;
	return NULL;
}

EdbColumnInfo
edb_table_seek_column_by_name(EdbContext ec, EdbTableInfo ti, utf8_t *name)
{
	int	i;

	for(i=0; i<ti->ncol; ++i) {
		if(!strcasecmp((char *)ti->column[i]->name.english, (char *)name)
			|| !strcasecmp((char *)ti->column[i]->name.japanese, (char *)name)) {
			return ti->column[i];
		}
	}
	return NULL;
}

EdbColumnInfo
edb_table_seek_column_by_xmlname(EdbContext ec, EdbTableInfo ti, utf8_t *xmlname)
{
	int	i;
	EdbColumnInfo	ci = NULL;

	if(xmlname[0]=='@' && xmlname[1]=='.') {
		char	*s;

		asprintf(&s, "%s.%s", ti->xmlname, &xmlname[2]);
		xmlname = (utf8_t *)STRDUP(s);
		FREE(s);
	}
	for(i=0; i<ti->ncol; ++i) {
		if(!strcasecmp((char *)ti->column[i]->xmlname, (char *)xmlname)) {
			ci = ti->column[i];
			break;
		}
	}
	return ci;
}

EdbColumnInfo
edb_table_seek_column_by_sqlname(EdbContext ec, EdbTableInfo ti, utf8_t *sqlname)
{
	int	i;

	for(i=0; i<ti->ncol; ++i) {
		if(!strcasecmp((char *)ti->column[i]->sqlname, (char *)sqlname)) return ti->column[i];
	}
	return NULL;
}

EdbColumnInfo
edb_table_seek_picture_column(EdbContext ec, EdbTableInfo ti)
{
	Link l;
	EdbColumnInfo	ci = NULL;
	ForLink(l, ti->l_column) {
		if(((EdbColumnInfo)l->d)->typeInfo->typeID==EDB_TYPE_PICTURE) { ci = l->d; break; }
	}
	return ci;
}

#if 0
int
edb_table_seek_column_index_by_sqlname(EdbContext ec, EdbTableInfo ti, utf8_t *sqlname)
{
	int	i;

	for(i=0; i<ti->ncol; ++i) {
		if(!strcasecmp((char *)ti->column[i]->sqlname, (char *)sqlname)) return i;
	}
	return NULL;
}
#endif

/*
	edb_table_make_caption	-- make caption
*/
Caption
edb_table_make_caption(EdbContext ec, EdbTableInfo ti, int mode)
{
	Caption	cap;

	cap = edb_caption_new(ec);
	if(cap == NULL) return NULL;
	if(ti) {
		cap->c_main = (utf8_t *)strdup((char *)(ec->japanese ? ti->name.japanese : ti->name.english));
		cap->c_main_english = (utf8_t *)strdup((char *)ti->name.english);
		cap->c_main_japanese = (utf8_t *)strdup((char *)ti->name.japanese);
		cap->c_main_pronounce = (utf8_t *)strdup((char *)ti->name.pronounce);
		cap->c_sub = (utf8_t *)strdup((char *)(ec->japanese ? ti->desc.japanese : ti->desc.english));
	} else {
		cap->c_main = (utf8_t *)strdup("(no caption)");
		cap->c_main_english = (utf8_t *)strdup("(no caption)");
		cap->c_main_japanese = (utf8_t *)strdup("(見出しなし)");
		cap->c_main_pronounce = (utf8_t *)strdup("みだしなし");
		cap->c_sub = (utf8_t *)strdup("(no caption)");
	}
	return cap;
}

/*
	edb_table_get_eid_of_tables
*/
Link
edb_table_get_eid_of_tables(EdbContext ec)
{
	utf8_t	*query;
	utf8_t	*latest;
	Link	l_eid;
	int	n;

	latest = edb_create_latest_phrase(ec, (utf8_t *)EDB_TABLE_TABLE, (utf8_t *)"t1", (utf8_t *)"t2", NULL, 0);
	asprintf((char **)&query,
		"SELECT %s FROM %s AS t1 WHERE %s ORDER BY %s",
		EDB_BASE_EID,
		EDB_TABLE_TABLE,
		latest,
		EDB_TABLE_NAME);
	FREE(latest);
	l_eid = edb_pgsql_query_one_field(ec, query, &n);
	FREE(query);
	return l_eid;
}

Link
edb_table_get_tables(EdbContext ec)
{
	Link	l, l_eid, l_table;

	l_eid = edb_table_get_eid_of_tables(ec);
	l_table = NULL;
	ForLink(l, l_eid) {
		EdbTableInfo	ti;

		ti = edb_table_get(ec, ATOI((char *)l->d));
		LinkAppend(l_table, link_atom_new(ti, NULL));
	}
	link_free(l_eid);
	return l_table;
}

/*
	edb_table_search_tuples
*/
#if 0
Link
edb_table_search_tuples_byText(EdbContext ec, EdbTableInfo ti, utf8_t *target, int lang)
{
	utf8_t *order;
	utf8_t *query;
	utf8_t *match;
	Link	l_eid, l_tuple, l, l_target, l_str;
	int	n;
	int	c;
	utf8_t	*s, *constraint, langstr;

	l_str = NULL;
	for(c=0; c<ti->ncol; ++c) {
		EdbColumnInfo	ci = ti->column[c];
		int	e, elements = edb_table_column_reldb_elements(ec, ti, NULL, ci);
		s = NULL;

		if(ci->caption==1) {
			if(ci->typeInfo->basicTypeID==EDB_TYPE_NAME) {
				switch(lang) {
				case 0:	asprintf(&s, "english(t1.%s%s)", ColumnSN(ci), elements==1 ? "":"[1]");		break;
				case 1:	asprintf(&s, "japanese(t1.%s%s)", ColumnSN(ci), elements==1 ? "":"[1]");		break;
				case 2:	asprintf(&s, "pronounce(t1.%s%s)", ColumnSN(ci), elements==1 ? "":"[1]");	break;
				}
			} else if(ci->typeInfo->basicTypeID==EDB_TYPE_TEXT
				|| ci->typeInfo->basicTypeID==EDB_TYPE_INETNAME
				|| ci->typeInfo->basicTypeID==EDB_TYPE_INETFQDN) {
				asprintf(&s, "t1.%s%s", ColumnSN(ci), elements==1 ? "":"[1]");
			}
		}
		if(STRisVALID(s)) LinkAppend(l_str, link_atom_new(s, free));
	}
	s = (utf8_t *)str_concatenate_str(l_str, "||");
	link_free(l_str);
	match = (utf8_t *)STRDUP((char *)s);
	FREE(s);
	if(!STRisVALID(match)) return NULL;

	l_target = str_split(target, '%');
	l_str = NULL;
	ForLink(l, l_target) {
		char *s;

		asprintf(&s, "(%s)~~*'%%%s%%'", match, l->d);
		LinkAppend(l_str, link_atom_new(s, free));
	}
	link_free(l_target);
	constraint = (utf8_t *)str_concatenate_str(l_str, " AND ");
	link_free(l_str);
	if(!STRisVALID(constraint)) return NULL;

	order = edb_table_make_order(ec, ti, (utf8_t *)"t1");
	asprintf((char **)&query,
		"SELECT %s FROM %s AS t1 WHERE %s AND %s %s %s",
		EDB_BASE_EID,
		TABLE_SN(ti),
		EDB_BASE_AVAILABLE,
		constraint,
		STRisVALID(order) ? "ORDER BY" : "", order);
	FREE(order);
	FREE(constraint);
	l_eid = edb_pgsql_query_one_field(ec, query, &n);
	FREE(query);
	l_tuple = NULL;
	ForLink(l, l_eid) {
		LinkAppend(l_tuple, link_atom_new(edb_tuple_get(ec, ATOI((char *)l->d), EDB_TUPLE_GET_template_table, ti, 0), NULL));
	}
	link_free(l_eid);
	return l_tuple;
}
#else
Link
edb_table_search_tuples_byText(EdbContext ec, EdbTableInfo ti, utf8_t *target, int lang)
{
	utf8_t *order;
	utf8_t *query;
	utf8_t *match;
	Link	l_eid, l_tuple, l, l_target, l_str, l_tname;
	int	n, tno;
	int	c;
	utf8_t	*s, *constraint, *tnamestr, *as;

	tno = 1;
	l_tname = NULL;
	asprintf((char **)&s, "%s AS t%d", TABLE_SN(ti), tno++);
	LinkAppend(l_tname, link_atom_new(s, free));

	l_str = NULL;
	for(c=0; c<ti->ncol; ++c) {
		EdbColumnInfo	ci = ti->column[c];
		int	elements = edb_table_column_reldb_elements(ec, ti, NULL, ci);
		s = NULL;

		if(elements==1) {
			if(ci->caption==1) {
#if	defined(TEXT_VARIANT_CONVERSION)
				if(ci->typeInfo->basicTypeID==EDB_TYPE_NAME) {
					switch(lang) {
					case 0:	asprintf((char **)&s, "english(t1.%s%s)", ColumnSN_VARIANT(ci), elements==1 ? "":"[1]");		break;
					case 1:	asprintf((char **)&s, "japanese(t1.%s%s)", ColumnSN_VARIANT(ci), elements==1 ? "":"[1]");		break;
					case 2:	asprintf((char **)&s, "pronounce(t1.%s%s)", ColumnSN_VARIANT(ci), elements==1 ? "":"[1]");	break;
					}
				} else if(ci->typeInfo->basicTypeID==EDB_TYPE_TEXT
					|| ci->typeInfo->basicTypeID==EDB_TYPE_INETNAME
					|| ci->typeInfo->basicTypeID==EDB_TYPE_INETFQDN) {
					asprintf((char **)&s, "t1.%s%s", ColumnSN_VARIANT(ci), elements==1 ? "":"[1]");
				}
#else
				if(ci->typeInfo->basicTypeID==EDB_TYPE_NAME) {
					switch(lang) {
					case 0:	asprintf((char **)&s, "english(t1.%s%s)", ColumnSN(ci), elements==1 ? "":"[1]");		break;
					case 1:	asprintf((char **)&s, "japanese(t1.%s%s)", ColumnSN(ci), elements==1 ? "":"[1]");		break;
					case 2:	asprintf((char **)&s, "pronounce(t1.%s%s)", ColumnSN(ci), elements==1 ? "":"[1]");	break;
					}
				} else if(ci->typeInfo->basicTypeID==EDB_TYPE_TEXT
					|| ci->typeInfo->basicTypeID==EDB_TYPE_INETNAME
					|| ci->typeInfo->basicTypeID==EDB_TYPE_INETFQDN) {
					asprintf((char **)&s, "t1.%s%s", ColumnSN(ci), elements==1 ? "":"[1]");
				}
#endif	/* TEXT_VARIANT_CONVERSION */
				if(STRisVALID(s)) LinkAppend(l_str, link_atom_new(s, free));
			}
		} else {
			if(ci->caption==1) {
				asprintf((char **)&as, "t%d", tno);
				asprintf((char **)&s, "%s AS t%d", ColumnRdbSN(ci), tno++);
				LinkAppend(l_tname, link_atom_new(s, free));
				s = NULL;
				elements = edb_table_column_reldb_elements(ec, ti, ci, ci);
#if	defined(TEXT_VARIANT_CONVERSION)
				if(ci->typeInfo->basicTypeID==EDB_TYPE_NAME) {
					switch(lang) {
					case 0:	asprintf((char **)&s, "english(%s.%s%s)", as, ColumnSN_VARIANT(ci), elements==1 ? "":"[1]");		break;
					case 1:	asprintf((char **)&s, "japanese(%s.%s%s)", as, ColumnSN_VARIANT(ci), elements==1 ? "":"[1]");	break;
					case 2:	asprintf((char **)&s, "pronounce(%s.%s%s)", as, ColumnSN_VARIANT(ci), elements==1 ? "":"[1]");	break;
					}
				} else if(ci->typeInfo->basicTypeID==EDB_TYPE_TEXT
					|| ci->typeInfo->basicTypeID==EDB_TYPE_INETNAME
					|| ci->typeInfo->basicTypeID==EDB_TYPE_INETFQDN) {
					asprintf((char **)&s, "%s.%s%s", as, ColumnSN_VARIANT(ci), elements==1 ? "":"[1]");
				}
#else
				if(ci->typeInfo->basicTypeID==EDB_TYPE_NAME) {
					switch(lang) {
					case 0:	asprintf((char **)&s, "english(%s.%s%s)", as, ColumnSN(ci), elements==1 ? "":"[1]");		break;
					case 1:	asprintf((char **)&s, "japanese(%s.%s%s)", as, ColumnSN(ci), elements==1 ? "":"[1]");	break;
					case 2:	asprintf((char **)&s, "pronounce(%s.%s%s)", as, ColumnSN(ci), elements==1 ? "":"[1]");	break;
					}
				} else if(ci->typeInfo->basicTypeID==EDB_TYPE_TEXT
					|| ci->typeInfo->basicTypeID==EDB_TYPE_INETNAME
					|| ci->typeInfo->basicTypeID==EDB_TYPE_INETFQDN) {
					asprintf((char **)&s, "%s.%s%s", as, ColumnSN(ci), elements==1 ? "":"[1]");
				}
#endif	/* TEXT_VARIANT_CONVERSION */
				if(STRisVALID(s)) LinkAppend(l_str, link_atom_new(s, free));
				FREE(as);
			}
		}
	}
	s = (utf8_t *)str_concatenate_str(l_str, "||");
	link_free(l_str);
	match = (utf8_t *)STRDUP((char *)s);
	FREE(s);
	if(!STRisVALID(match)) return NULL;

	l_str = NULL;
	for(n=2; n<tno; ++n) {
		asprintf((char **)&s, "t1." EDB_BASE_EID "=t%d." EDB_BASE_EID, n);
		LinkAppend(l_str, link_atom_new(s, free));
	}
	l_target = str_split((char *)target, '%');
#if	defined(TEXT_VARIANT_CONVERSION)
	ForLink(l, l_target) {
		utf8_t *s, *v = edb_text_variant_conversion(l->d, FALSE);
		for(s=v; *s; ++s) if(*s=='\'') *s = '.';

		asprintf((char **)(char **)&s, "(%s)~~*'%%%s%%'", match, v);
		FREE(v);
		LinkAppend(l_str, link_atom_new(s, free));
	}
#else
	ForLink(l, l_target) {
		char *s;

		asprintf((char **)&s, "(%s)~~*'%%%s%%'", match, l->d);
		LinkAppend(l_str, link_atom_new(s, free));
	}
#endif	/* TEXT_VARIANT_CONVERSION */
	link_free(l_target);
	constraint = (utf8_t *)str_concatenate_str(l_str, " AND ");
	link_free(l_str);
	if(!STRisVALID(constraint)) return NULL;

	order = edb_table_make_order(ec, ti, (utf8_t *)"t1");
	tnamestr = (utf8_t *)str_concatenate(l_tname, ',');
	link_free(l_tname);
	asprintf((char **)&query,
		"SELECT t1.%s"
		" FROM %s"
		" WHERE t1.%s AND %s"
		" GROUP BY %s"
		" %s %s",
		EDB_BASE_EID,
		tnamestr,
		EDB_BASE_AVAILABLE,
		constraint,
		order,
		STRisVALID(order) ? "ORDER BY" : "", order);
	FREE(tnamestr);
	FREE(order);
	FREE(constraint);
	l_eid = edb_pgsql_query_one_field(ec, query, &n);
	FREE(query);
	l_tuple = NULL;
	ForLink(l, l_eid) {
		LinkAppend(l_tuple, link_atom_new(edb_tuple_get(ec, ATOI((char *)l->d), EDB_TUPLE_GET_template_table, ti, 0), NULL));
	}
	link_free(l_eid);
	return l_tuple;
}
#endif

/*
	edb_table_lookup_tuples
*/
Link
edb_table_lookup_tuples_byText(EdbContext ec, EdbTableInfo ti, EdbColumnInfo ci, utf8_t *target)
{
	utf8_t *order;
	utf8_t *query;
	utf8_t *match = NULL;
	Link	l_eid, l_tuple, l;
	int	n;
	int	elements = edb_table_column_reldb_elements(ec, ti, NULL, ci);

	if(elements==1) {
		if(ci->typeInfo->basicTypeID==EDB_TYPE_NAME) {
			asprintf((char **)&match, "mtext_texticregexeq(%s, '.*%s.*')", ci->sqlname, target);
		} else {
			asprintf((char **)&match, "%s ~* '.*%s.*'", ci->sqlname, target);
		}
	} else {
		if(ci->typeInfo->basicTypeID==EDB_TYPE_NAME) {
			asprintf((char **)&match, "mtext_array_texticregexeq(%s, '.*%s.*')", ci->sqlname, target);
		} else {
			asprintf((char **)&match, "%s *~ '.*%s.*'", ci->sqlname, target);
		}
	}
	order = edb_table_make_order(ec, ti, (utf8_t *)"t1");
	asprintf((char **)&query,
		"SELECT %s FROM %s AS t1 WHERE %s AND %s %s %s",
		EDB_BASE_EID,
		ti->sqlname,
		match,
		EDB_BASE_AVAILABLE,
		STRisVALID(order) ? "ORDER BY" : "", order);
	FREE(order);
	FREE(match);
	l_eid = edb_pgsql_query_one_field(ec, query, &n);
	FREE(query);
	l_tuple = NULL;
	ForLink(l, l_eid) {
		LinkAppend(l_tuple, link_atom_new(edb_tuple_get(ec, atoi((char *)l->d), EDB_TUPLE_GET_template_table, ti, 0), NULL));
	}
	link_free(l_eid);
	return l_tuple;
}

static
int
edb_table_column_callback_sub(
	EdbContext				ec,
	EdbTableInfo				ti,
	int						layer,
	Link						l_column,
	TableColumn_callback_func_t	callback,
	EdbTableCBData			cbd)
{
	register Link	l;
	register int cont = TRUE;

	ForLink(l, l_column) {
		register EdbColumnInfo	ci = (EdbColumnInfo)l->d;

		if(callback) {
			cont = (*callback)(ec, ti, layer, ci, cbd);
			if(!cont) break;
		}
		if(ci->l_child) {
			cont = edb_table_column_callback_sub(ec, ti, layer+1, ci->l_child, callback, cbd);
			if(!cont) break;
		}
	}
	return cont;
}

void
edb_table_column_callback(
	EdbContext				ec,
	EdbTableInfo				ti,
	EdbColumnInfo				ci,
	TableColumn_callback_func_t	callback,
	void						*data)
{
	struct EDB_TABLE_CB_DATA	cb_data;
	EdbTableCBData	cbd = &cb_data;

	if(ti==NULL) return;
	bzero(cbd, sizeof(struct EDB_TABLE_CB_DATA));
	cbd->user_data = data;
	edb_table_column_callback_sub(ec, ti, 0, ci ? ci->l_child : ti->l_column, callback, cbd);
}

int
edb_table_is_hierarchical(EdbContext ec, EdbTableInfo ti)
{

	return (ti->hierarchical)!=0;
}

EdbColumnInfo
edb_table_hierarchy_column(EdbContext ec, EdbTableInfo ti)
{
	Link	l;

	ForLink(l, ti->l_column) {
		EdbColumnInfo	ci = (EdbColumnInfo)l->d;

		if(ci->hierarchyparent) return ci;
	}
	return NULL;
}

int
edb_table_is_chronological(EdbContext ec, EdbTableInfo ti)
{

	return (ti->chronological)!=0;
}

int
edb_table_has_conductor(EdbContext ec, EdbTableInfo ti)
{

	return (ti->l_conductors)!=NULL;
}

Link
edb_table_get_conductors(EdbContext ec, EdbTableInfo ti)
{

	return ti->l_conductors;
}

utf8_t *
edb_table_get_column_minimum(EdbContext ec, EdbTableInfo ti, EdbColumnInfo ci, utf8_t *constraint)
{
	utf8_t	*query;
	utf8_t	*value;

	asprintf((char **)&query, "SELECT MIN(%s) FROM %s%s%s",
		ci->sqlname,
		ti->sqlname,
		STRisVALID(constraint) ? " WHERE " : "",
		STRisVALID(constraint) ? constraint : (utf8_t *)"");
	value = edb_pgsql_query_text(ec, query);
	FREE(query);
	return value;
}

utf8_t *
edb_table_get_column_maximum(EdbContext ec, EdbTableInfo ti, EdbColumnInfo ci, utf8_t *constraint)
{
	utf8_t	*query;
	utf8_t	*value;

	asprintf((char **)&query, "SELECT MAX(%s) FROM %s%s%s",
		ci->sqlname,
		ti->sqlname,
		STRisVALID(constraint) ? " WHERE " : "",
		STRisVALID(constraint) ? constraint : (utf8_t *)"");
	value = edb_pgsql_query_text(ec, query);
	FREE(query);
	return value;
}

utf8_t *
edb_table_get_date2_minimum(EdbContext ec, EdbTableInfo ti, EdbColumnInfo ci, utf8_t *constraint)
{
	utf8_t	*query;
	utf8_t	*value;

	asprintf((char **)&query, "SELECT MIN(%s/int8(100000000)) FROM %s%s%s",
		ci->sqlname,
		ti->sqlname,
		STRisVALID(constraint) ? " WHERE " : "",
		STRisVALID(constraint) ? constraint : (utf8_t *)"");
	value = edb_pgsql_query_text(ec, query);
	FREE(query);
	return value;
}

utf8_t *
edb_table_get_date2_maximum(EdbContext ec, EdbTableInfo ti, EdbColumnInfo ci, utf8_t *constraint)
{
	utf8_t	*query;
	utf8_t	*value;

	asprintf((char **)&query, "SELECT MAX(%s%%int8(100000000)) FROM %s%s%s",
		ci->sqlname,
		ti->sqlname,
		STRisVALID(constraint) ? " WHERE " : "",
		STRisVALID(constraint) ? constraint : (utf8_t *)"");
	value = edb_pgsql_query_text(ec, query);
	FREE(query);
	return value;
}

int
edb_table_get_number_of_tuples(EdbContext ec, EdbTableInfo ti, char *fmt, ...)
{
	utf8_t	*query, *target;
	int	v;
	va_list	ap;

	target = (utf8_t *)strdup("");
	if(fmt) {
		char	*s;

		va_start(ap, fmt);
		vasprintf(&s, fmt, ap);
		va_end(ap);
		if(STRisVALID(s)) {
			FREE(target);
			asprintf((char **)&target, " WHERE %s", s);
		}
		FREE(s);
	}
	asprintf((char **)&query, "SELECT COUNT(*) FROM %s AS t1 %s", ti->sqlname, target);
	FREE(target);
	v = edb_pgsql_query_int(ec, query);
	FREE(query);
	return v;
}

int
edb_table_get_number_of_available_tuples(EdbContext ec, EdbTableInfo ti, char *fmt, ...)
{
	utf8_t	*query, *target;
	int	v;
	va_list	ap;

	target = (utf8_t *)strdup(" WHERE " EDB_BASE_AVAILABLE);
	if(fmt) {
		char	*s;

		va_start(ap, fmt);
		vasprintf(&s, fmt, ap);
		va_end(ap);
		if(STRisVALID(s)) {
			FREE(target);
			asprintf((char **)&target, " WHERE %s AND %s", EDB_BASE_AVAILABLE, s);
		}
		FREE(s);
	}
	asprintf((char **)&query, "SELECT COUNT(*) FROM %s AS t1 %s", ti->sqlname, target);
	FREE(target);
	v = edb_pgsql_query_int(ec, query);
	FREE(query);
	return v;
}

Link
edb_table_get_all_tables(EdbContext ec)
	/*
		Must be call from outer libedb, becase it use EDB_TABLE_TABLE.
	*/
{
	Link	l, l_tables;
//	Link	l_eid;
	EdbTableInfo	ti;

	if(ec->ca_all_tables) {
		l_tables = NULL;
		ForCatalogueLink(l, ec->ca_all_tables) {
			ti = CatalogueObject(l);
			LinkAppend(l_tables, link_atom_new(ti, NULL));
		}
#if 1
	} else {
		utf8_t	*query;
		int	n;
		Link	l_xml;

		asprintf((char **)&query,
			"SELECT %s FROM %s AS t1, %s AS t2 WHERE t1.%s=t2.%s ORDER BY t2.%s",
			EDB_XMLDB_COLUMN_XML,
			EDB_XMLDB_TABLE,
			EDB_TABLE_TABLE,
			EDB_XMLDB_COLUMN_EOID, EDB_BASE_EOID,
			EDB_TABLE_NAME);
		l_xml = edb_pgsql_query_one_field(ec, query, &n);
		FREE(query);
		l_tables = NULL;
		ec->ca_all_tables = edb_catalogue_new();
		ForLink(l, l_xml) {
			utf8_t	*xml = l->d;

			ti = edb_table_parse_xml(ec, xml);
			if(ti) {
				ti = edb_eid_list_store(ec, TABLE_EID(ti), EDB_EID_IS_TABLE, ti);
				LinkAppend(l_tables, link_atom_new(ti, NULL));
				edb_catalogue_append(ec->ca_all_tables, TABLE_EID(ti), ti);
			}
		}
		link_free(l_xml);
#else
	} else {
		utf8_t	*query;
		utf8_t	*latest;
		int	n;

		latest = edb_create_latest_phrase(ec, (utf8_t *)EDB_XMLDB_TABLE, (utf8_t *)"t1", (utf8_t *)"t2", NULL, 0);
		asprintf((char **)&query,
			"SELECT %s FROM %s AS t1 WHERE t1.%s=%d AND %s ORDER BY t1.%s",
			EDB_XMLDB_COLUMN_EID,
			EDB_XMLDB_TABLE,
			EDB_XMLDB_COLUMN_TEID, 0,
			latest,
			EDB_BASE_EOID);
		FREE(latest);
		l_eid = edb_pgsql_query_one_field(ec, query, &n);
		FREE(query);
		l_tables = NULL;
		ForLink(l, l_eid) {
			ti = edb_table_get(ec, ATOI(l->d));
			if(ti) LinkAppend(l_tables, link_atom_new(ti, NULL));
		}
		link_free(l_eid);
		l_eid = edb_table_get_eid_of_tables(ec);
		ec->ca_all_tables = edb_catalogue_new();
		ForLink(l, l_eid) {
			ti = edb_table_get(ec, ATOI(l->d));
			if(ti) edb_catalogue_append(ec->ca_all_tables, TABLE_EID(ti), ti);
		}
		link_free(l_eid);
#endif
	}
	return l_tables;
}

int
edb_table_module_load(EdbContext ec, EdbTableInfo ti)
{
	char	*s, *module_name, *info_name;
	edb_module_info_t	*mi;
	edb_module_method_t	*mm;

#if defined(__DARWIN__)
	return FALSE;
#endif
	if(ti==NULL || !STRisVALID(ti->module)) return FALSE;
	if(ti->module_checked) return (ti->module_handle!=NULL);
	ti->module_checked = TRUE;
	asprintf(&s, EDB_MODULE_DIR "/%s", ti->module);
	module_name = STRDUP(s);
	FREE(s);
	ti->module_handle = dlopen(module_name, RTLD_NOW);
	if(ti->module_handle==NULL) {
		edb_trace(ec, "edb_table_module_load: %s\n", dlerror());
		return FALSE;
	}

	asprintf(&s, EDB_MODULE_INFO_PREFIX "%s", TABLE_XN(ti));
	info_name = STRDUP(s);
	FREE(s);
	mi = dlsym(ti->module_handle, info_name);
	if(mi == NULL) {
		edb_trace(ec, "edb_table_module_load: %s\n", dlerror());
		return FALSE;
	}
	if(strcmp(mi->xn, (char *)TABLE_XN(ti))!=0) {
		edb_trace(ec, "edb_table_module_load: mismatch xmlname '%s'<->'%s'.\n", mi->xn, TABLE_XN(ti));
		return FALSE;
	}
	if(mi->methods) {
		for(mm=mi->methods; mm->method; ++mm) {
			if(mm->method==&edb_tuple_make_caption_func) {
				ti->tuple_make_caption_func = mm->instance;
			} else if(mm->method==&edb_tuple_check_updatable) {
				ti->tuple_check_updatable = mm->instance;
			} else if(mm->method==&edb_tuple_check_writable_by) {
				ti->tuple_check_writable_by = mm->instance;
			} else if(mm->method==&edb_tuple_print_additional_information) {
				ti->tuple_print_additional_information = mm->instance;
#if defined(EDB_MANAGE_PKI)
			} else if(mm->method==&edb_pki_tuple_certificate_request_info) {
				ti->tuple_certificate_request_info = mm->instance;
#endif
			}
		}
	}
	return TRUE;
}

static
int
edb_table_tuple_caption_size_func(EdbContext ec, EdbTableInfo ti, int mode, int recurse)
{
	int	c, size = 0;

	if(--recurse<=0) return 0;
	for(c=0; c<ti->ncol; ++c) {
		EdbColumnInfo	ci = ti->column[c];

		if(ci->obsoleted) continue;
		if(ci->caption==1 || ((mode&CAP_long) && ci->caption==2)) {
			int	siz = 0;
			Link	lm;

			ForLink(lm, ci->l_maplookup) {
				EdbMapLookup	map = (EdbMapLookup)lm->d;
				eid_t			meid;
				EdbTableInfo	mti;
				int			s;

				meid = edb_maplookup_table_eid(ec, map);
				if(!EIDisVALID(meid)) {
					edb_trace(ec, "edb_table_column_reldb_elements: can't get table eid.\n");
					continue;
				}
				mti = edb_table_get(ec, meid);
				if(mti==NULL) {
					edb_trace(ec, "edb_table_column_reldb_elements: can't get table.\n");
					continue;
				}
				s = edb_table_tuple_caption_size_func(ec, mti, mode&(~CAP_long), recurse);
				siz = max(s, siz);
			}
			siz = max(siz, ci->size);
			size += siz*ci->elements_base;
		}
	}
	return size;
}

int
edb_table_tuple_caption_size(EdbContext ec, EdbTableInfo ti, int mode)
{

	return edb_table_tuple_caption_size_func(ec, ti, mode, ec->max_mapto_recurse);
}

int
edb_table_column_is_studentcode(EdbContext ec, EdbColumnInfo ci)
{
	int	st = FALSE;

	if(ci->typeInfo->typeID==EDB_TYPE_EXTID && ci->super && ci->super->sometimesStudent) {
		char *cxn;
		asprintf(&cxn, "%s.%s", ColumnXN(ci->super), CXN_STUDENT);
		if(strcmp(cxn, (char *)ColumnXN(ci))==0) st = TRUE;
		FREE(cxn);
	}
	return st;
}

#define	STUDENT_COLUMNs	"c_id,c_sid,c_code,c_name_ja,c_name_en,c_name_pr,c_enter,c_leave,c_faculty,c_department,c_division,c_title"
static
StudentInfo
edb_table_get_student_info_parse(Link l_fields, int rdate)
{

	if(link_atoms(l_fields)<12) return NULL;
	StudentInfo	info = (StudentInfo)malloc(sizeof(struct STUDENT_INFO));
	bzero(info, sizeof(struct STUDENT_INFO));
	int	no = 0;
	info->v_id = (utf8_t *)strdup((char *)link_get_obj(l_fields, no++));
	info->v_sid = (utf8_t *)strdup((char *)link_get_obj(l_fields, no++));
	info->v_code = (utf8_t *)strdup((char *)link_get_obj(l_fields, no++));
	info->v_name_ja = (utf8_t *)strdup((char *)link_get_obj(l_fields, no++));
	info->v_name_en = (utf8_t *)strdup((char *)link_get_obj(l_fields, no++));
	info->v_name_pr = (utf8_t *)strdup((char *)link_get_obj(l_fields, no++));
	info->v_enter = atoi((char *)link_get_obj(l_fields, no++));
	info->v_leave = atoi((char *)link_get_obj(l_fields, no++));
	info->v_faculty = (utf8_t *)strdup((char *)link_get_obj(l_fields, no++));
	info->v_department = (utf8_t *)strdup((char *)link_get_obj(l_fields, no++));
	info->v_division = (utf8_t *)strdup((char *)link_get_obj(l_fields, no++));
	info->v_title = (utf8_t *)strdup((char *)link_get_obj(l_fields, no++));

	info->ref_date = rdate;
	return info;
}

Link
edb_table_get_student_info(EdbContext ec, utf8_t *code) {
	utf8_t	*query;
	int		nrows;
	Link		l_students = NULL;

	utf8_t *text = edb_text_for_pgsql_literal(ec, code);
	asprintf((char **)&query, "SELECT " STUDENT_COLUMNs " FROM " TN_STUDENT " WHERE c_code=%s", text);
	FREE(text);
	Link	l_results = edb_pgsql_query(ec, query, &nrows, NULL);
	FREE(query);
	if(nrows>0) {
		Link	l1;
		ForLink(l1, l_results) {
			StudentInfo info = edb_table_get_student_info_parse(l1->d, 0);
			if(info) LinkAppendTop(link_atom_new(info, edb_table_free_student_info), l_students);
		}
		LinkReverse(l_students);
	}
	link_free(l_results);
	return l_students;
}

void
edb_table_free_student_info(StudentInfo info) {
	FREE(info->v_id);
	FREE(info->v_sid);
	FREE(info->v_code);
	FREE(info->v_name_ja);
	FREE(info->v_name_en);
	FREE(info->v_name_pr);
	FREE(info->v_faculty);
	FREE(info->v_department);
	FREE(info->v_division);
	FREE(info->v_title);
	FREE(info);
}

char *
edb_table_create_student_dict(EdbContext ec, StudentInfo info) {
	if(info==NULL) return NULL;

	Doc xml = doc_sopen(DocML_XML);
	doc_puts(xml,		"			<utlf:dict>\n");
	doc_text_puts3(xml,	"				<utlf:key>ID</utlf:key><utlf:string>", (char *)info->v_id, "</utlf:string>\n");
	doc_text_puts3(xml,	"				<utlf:key>SID</utlf:key><utlf:string>", (char *)info->v_sid, "</utlf:string>\n");
	doc_text_puts3(xml,	"				<utlf:key>学籍番号</utlf:key><utlf:string>", (char *)info->v_code, "</utlf:string>\n");
	doc_puts(xml,		"				<utlf:key>個人</utlf:key>\n");
	doc_puts(xml,		"				<utlf:dict>\n");
	doc_text_puts3(xml,	"					<utlf:key>氏名(英)</utlf:key><utlf:string>", (char *)info->v_name_en, "</utlf:string>\n");
	doc_text_puts3(xml,	"					<utlf:key>氏名(日)</utlf:key><utlf:string>", (char *)info->v_name_ja, "</utlf:string>\n");
	doc_text_puts3(xml,	"					<utlf:key>氏名(読)</utlf:key><utlf:string>", (char *)info->v_name_pr, "</utlf:string>\n");
	doc_puts(xml,		"				</utlf:dict>\n");
	doc_puts(xml,		"				<utlf:key>所属</utlf:key>\n");
	doc_puts(xml,		"				<utlf:dict>\n");
	doc_text_puts3(xml,	"					<utlf:key>FACULTY</utlf:key><utlf:string>", (char *)info->v_faculty, "</utlf:string>\n");
	doc_text_puts3(xml,	"					<utlf:key>DEPARTMENT</utlf:key><utlf:string>", (char *)info->v_department, "</utlf:string>\n");
	doc_text_puts3(xml,	"					<utlf:key>DIVISION</utlf:key><utlf:string>", (char *)info->v_division, "</utlf:string>\n");
	doc_text_puts3(xml, "					<utlf:key>学生等区分</utlf:key><utlf:string>", (char *)info->v_title, "</utlf:string>\n");
	doc_puts(xml,		"				</utlf:dict>\n");
	if(info->v_enter!=0) {
		doc_puts(xml,	"				<utlf:key>入学</utlf:key>\n");
		doc_puts(xml,	"				<utlf:dict>\n");
		doc_printf(xml, "					<utlf:key>年月日</utlf:key><utlf:date>%04d-%02d-%02d</utlf:date>\n", DATE_Y(info->v_enter), DATE_M(info->v_enter), DATE_D(info->v_enter));
		doc_puts(xml,	"				</utlf:dict>\n");
	}
	if(info->v_leave!=0) {
		doc_puts(xml,	"				<utlf:key>出学</utlf:key>\n");
		doc_puts(xml,	"				<utlf:dict>\n");
		doc_printf(xml, "					<utlf:key>年月日</utlf:key><utlf:date>%04d-%02d-%02d</utlf:date>\n", DATE_Y(info->v_leave), DATE_M(info->v_leave), DATE_D(info->v_leave));
		doc_puts(xml,	"				</utlf:dict>\n");
	}
	doc_puts(xml,		"			</utlf:dict>");
	char *s = doc_sgets(xml);
	doc_close(xml);
	return s;
}

#if 0
static
int
compare_student(LinkObj o1, LinkObj o2)
{
	StudentInfo s1 = o1->d;
	StudentInfo s2 = o2->d;
	return strcmp((char *)s1->v_id, (char *)s2->v_id);
}
#else
static
int
compare_student(LinkObj o1, LinkObj o2)
{
	StudentInfo s1 = o1->d;
	StudentInfo s2 = o2->d;
	int	v;
	if(s1->v_leave!=s2->v_leave) {			// 出学
		if(s1->v_leave==0) return -1;		// 在学している方を優先
		if(s2->v_leave==0) return 1;
		if(s1->ref_date!=0) {				// 基準日が有る場合，出学が基準日に近い方を優先
			int leave1 = abs(s1->v_leave - s1->ref_date);
			int leave2 = abs(s2->v_leave - s2->ref_date);
			if((v = leave1 - leave2)!=0) return v;
		} else {						// 基準日が無い場合，出学が現在に近い方を優先
			if((v = s2->v_leave - s1->v_leave)!=0) return v;
		}
	}
	v = s2->v_enter - s1->v_enter;			// 入学が現在に近い方を優先
	if(v!=0) return v;
	return strcmp((char *)s1->v_id, (char *)s2->v_id);
}
#endif


Link
edb_table_search_students_byText(EdbContext ec, utf8_t *target, int lang, int date)
{
	utf8_t *query;
	utf8_t *match;
	Link	l_eid, l_tuple, l, l_target, l_str;
	utf8_t	*s, *constraint, *as;

	l_str = NULL;
	s = NULL;

#if	defined(TEXT_VARIANT_CONVERSION)
	switch(lang) {
	case 0:	asprintf((char **)&s, "%s", "v_name_en");		break;
	case 1:	asprintf((char **)&s, "%s", "v_name_ja");		break;
	case 2:	asprintf((char **)&s, "%s", "v_name_pr");		break;
	}
#else
	switch(lang) {
	case 0:	asprintf((char **)&s, "%s", "c_name_en");		break;
	case 1:	asprintf((char **)&s, "%s", "c_name_ja");		break;
	case 2:	asprintf((char **)&s, "%s", "c_name_pr");		break;
	}
#endif	/* TEXT_VARIANT_CONVERSION */
	if(!STRisVALID(s)) return NULL;
	match = (utf8_t *)STRDUP((char *)s);
	FREE(s);
	if(!STRisVALID(match)) return NULL;

	l_target = str_split((char *)target, '%');
#if	defined(TEXT_VARIANT_CONVERSION)
	ForLink(l, l_target) {
		utf8_t *v = edb_text_variant_conversion(l->d, FALSE);
		for(s=v; *s; ++s) if(*s=='\'') *s = '.';

		asprintf((char **)(char **)&s, "(%s)~~*'%%%s%%'", match, v);
		FREE(v);
		LinkAppend(l_str, link_atom_new(s, free));
	}
#else
	ForLink(l, l_target) {
		asprintf((char **)&s, "(%s)~~*'%%%s%%'", match, l->d);
		LinkAppend(l_str, link_atom_new(s, free));
	}
#endif	/* TEXT_VARIANT_CONVERSION */
	link_free(l_target);
	constraint = (utf8_t *)str_concatenate_str(l_str, " AND ");
	link_free(l_str);
	if(!STRisVALID(constraint)) return NULL;

	asprintf((char **)&query, "SELECT " STUDENT_COLUMNs " FROM " TN_STUDENT "  WHERE %s", constraint);
	FREE(constraint);

	int nrows = 0;
	Link	l_results = edb_pgsql_query(ec, query, &nrows, NULL);
	FREE(query);

	Link	l_students = NULL;
	if(nrows>0) {
		Link	l1;
		ForLink(l1, l_results) {
			StudentInfo info = edb_table_get_student_info_parse(l1->d, date);
			if(info) {
#define	STUDENT_PREV_ENTER	10000	// 入学1年前まで
#define	STUDENT_AFTER_LEAVE	100000	// 出学10年後まで
				if(date!=0 &&(
					(info->v_enter!=0 && date < info->v_enter-STUDENT_PREV_ENTER)
					|| (info->v_leave!=0 && info->v_leave+STUDENT_AFTER_LEAVE < date)
					)) {
					edb_table_free_student_info(info);
					continue;
				}
				LinkAppendTop(link_atom_new(info, edb_table_free_student_info), l_students);
			}
		}
		LinkReverse(l_students);
	}
	link_free(l_results);
	if(l_students) l_students = link_sort(l_students, compare_student, NULL, NULL, NULL);
	return l_students;
}

static
Link
edb_table_search_student_information(EdbContext ec, Link l_found, char *str, int lang, int date)
{
	Link	l, l_info;
	char	*target = NULL;
	ucs4_t	*u, *ustr;

	if(str==NULL || strlen(str)==0) return FALSE;
	ustr = UTF8toUCS4((utf8_t *)str);
	for(u=ustr; *u; ++u) {
		if(ucs4_isascii(*u) && !ucs4_isalnum(*u)) *u = ' ';
	}
	ucs4_str_oneline(ustr);
	if(!UCS4_STRisVALID(ustr)) return FALSE;
	for(u=ustr; *u; ++u) {
		if(*u==' ') *u = '%';
	}
	str = (char *)UCS4toUTF8(ustr);
	asprintf(&target, "%s", str);
	l_info = edb_table_search_students_byText(ec, (utf8_t *)target, lang, date);
	FREE(target);
	ForLink(l, l_info) {
		StudentInfo info = l->d;
		int	found = FALSE;
		Link	l2;
		ForLink(l2, l_found) {
			if(strcmp((char *)info->v_code, (char *)((StudentInfo)l2->d)->v_code)==0) { found = TRUE; break; }
		}
		if(found) continue;
		LinkAppend(l_found, link_atom_new(info, l->f));
		l->d = NULL;
	}
	link_free(l_info);
	return l_found;
}

Link
edb_table_search_students_byDatum(EdbContext ec, EdbDatum datum, int date)
{
	if(datum==NULL) return NULL;

	Link	l_found = NULL;
	if(DatumEnglishIsValid(datum)) l_found = edb_table_search_student_information(ec, l_found, (char *)DatumEnglish(datum), 0, date);
	if(DatumJapaneseIsValid(datum)) l_found = edb_table_search_student_information(ec, l_found, (char *)DatumJapanese(datum), 1, date);
	if(DatumPronounceIsValid(datum)) l_found = edb_table_search_student_information(ec, l_found, (char *)DatumPronounce(datum), 2, date);
	if(l_found) l_found = link_sort(l_found, compare_student, NULL, NULL, NULL);
	return l_found;
}

static
int
edb_table_get_date_column_callback(EdbContext ec, EdbTableInfo ti, int layer, EdbColumnInfo ci, EdbTableCBData cbd)
{
	EdbColumnInfo	*cip = (EdbColumnInfo *)cbd->user_data;

	if(ci->typeInfo->typeID == EDB_TYPE_DATE) {
		*cip = ci;
		return FALSE;
	}
	return TRUE;
}

EdbColumnInfo
edb_table_get_date_column(EdbContext ec, EdbTableInfo ti)
{
	EdbColumnInfo ci = NULL;

	edb_table_column_callback(ec, ti, NULL, edb_table_get_date_column_callback, &ci);
	return ci;
}




