/*
	$Id: xml.c,v 1.10 2008/08/09 09:21:37 alex Exp $
	xml.c--- XML misc command
*/

#ifndef	lint
static char *rcsid __attribute__((unused)) = "$Id: xml.c,v 1.10 2008/08/09 09:21:37 alex Exp $";
#endif

#include	<stdio.h>
#include	<string.h>
#include	<stdlib.h>
#include	<stdarg.h>
#include	"edb/edb.h"
#include	"edb/xml.h"

utf8_t *
xml_get_node_text(EdbContext ec, xmlDocPtr doc, xmlNodePtr node)
{

	return  xmlNodeListGetString(doc, node->xmlChildrenNode, 1);
}

xmlNodePtr
xml_find_element_node(EdbContext ec, xmlDocPtr doc, xmlNodePtr cur, ...)
{
	va_list	ap;
	xmlChar	*target;

	va_start(ap, cur);
	target = va_arg(ap, xmlChar *);
	while(target && cur) {
		while(cur) {
			if(!xmlStrcmp(cur->name, target)) break;
			cur = cur->next;
		}
		if(cur==NULL) break;
		target = va_arg(ap, xmlChar *);
		if(target==NULL) break;
		cur = cur->xmlChildrenNode;
	}
	va_end(ap);

	if(target==NULL && cur) return cur;
	return NULL;
}

Link
xml_find_element_node_list(EdbContext ec, xmlDocPtr doc, xmlNodePtr node, ...)
{
	va_list	ap;
	xmlChar	*target;
	Link	l_node = NULL, l_found=NULL;

	va_start(ap, node);
	target = va_arg(ap, xmlChar *);
	if(target==NULL) {
		va_end(ap);
		return NULL;
	}
	l_node = link_atom_new(node, NULL);
	while(target) {
		Link	l;

		l_found = NULL;
		ForLink(l, l_node) {
			xmlNodePtr	cur = (xmlNodePtr)l->d;

			while(cur) {
				if(!xmlStrcmp(cur->name, target)) {
					LinkAppend(l_found, link_atom_new(cur, NULL));
				}
				cur=cur->next;
			}
		}
		EdbLinkFree(l_node);
		l_node = l_found;
		if(l_node==NULL) break;
		target = va_arg(ap, xmlChar *);
		if(target==NULL) break;
		l_found = NULL;
		ForLink(l, l_node) {
			LinkAppend(l_found, link_atom_new(((xmlNodePtr)l->d)->xmlChildrenNode, NULL));
		}
		EdbLinkFree(l_node);
		l_node = l_found;
	}
	va_end(ap);

	if(target) {
		EdbLinkFree(l_node);
		return NULL;
	}
	return l_node;
}

xmlChar *
xml_get_element_text(EdbContext ec, xmlDocPtr doc, xmlNodePtr cur, ...)
{
	va_list	ap;
	xmlChar	*target;

	va_start(ap, cur);
	target = va_arg(ap, xmlChar *);
	while(target && cur) {
		while(cur) {
			if(!xmlStrcmp(cur->name, target)) break;
			cur = cur->next;
		}
		if(cur==NULL) break;
		target = va_arg(ap, xmlChar *);
		if(target==NULL) break;
		cur = cur->xmlChildrenNode;
	}
	va_end(ap);

	if(target==NULL && cur) {
		return  xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
	}
	return NULL;
}

int
xml_get_element_mtext(EdbContext ec, utf8_t **en, utf8_t **ja, utf8_t **pr, xmlDocPtr doc, xmlNodePtr cur, ...)
{
	va_list	ap;
	xmlChar	*target;

	if(en) *en = NULL;
	if(ja) *ja = NULL;
	if(pr) *pr = NULL;
	va_start(ap, cur);
	target = va_arg(ap, xmlChar *);
	while(target && cur) {
		while(cur) {
			if(!xmlStrcmp(cur->name, target)) break;
			cur = cur->next;
		}
		if(cur==NULL) break;
		target = va_arg(ap, xmlChar *);
		if(target==NULL) break;
		cur = cur->xmlChildrenNode;
	}
	va_end(ap);

	if(target==NULL && cur) {
		if(en) *en = xml_get_element_text(ec, doc, cur, cur->name, XML_EN_ENGLISH, NULL);
		if(ja) *ja = xml_get_element_text(ec, doc, cur, cur->name, XML_EN_JAPANESE, NULL);
		if(pr) *pr = xml_get_element_text(ec, doc, cur, cur->name, XML_EN_PRONOUNCE, NULL);
		return TRUE;
	}
	return FALSE;
}

int
xml_get_element_text_integer(EdbContext ec, xmlDocPtr doc, xmlNodePtr cur, int default_value, ...)
{
	va_list	ap;
	xmlChar	*target;
	int	v = default_value;

	va_start(ap, default_value);
	target = va_arg(ap, xmlChar *);
	while(target && cur) {
		while(cur) {
			if(!xmlStrcmp(cur->name, target)) break;
			cur = cur->next;
		}
		if(cur==NULL) break;
		target = va_arg(ap, xmlChar *);
		if(target==NULL) break;
		cur = cur->xmlChildrenNode;
	}
	va_end(ap);

	if(target==NULL && cur) {
		xmlChar	*s;

		s = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
		if(STRisVALID(s)) v = atoi((char *)s);
		xmlFree(s);
	}
	return v;
}

int
xml_get_element_text_permission(EdbContext ec, xmlDocPtr doc, xmlNodePtr cur, int default_value, ...)
{
	va_list	ap;
	xmlChar	*target;
	int	v = default_value;

	va_start(ap, default_value);
	target = va_arg(ap, xmlChar *);
	while(target && cur) {
		while(cur) {
			if(!xmlStrcmp(cur->name, target)) break;
			cur = cur->next;
		}
		if(cur==NULL) break;
		target = va_arg(ap, xmlChar *);
		if(target==NULL) break;
		cur = cur->xmlChildrenNode;
	}
	va_end(ap);

	if(target==NULL && cur) {
		xmlChar	*s;

		s = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
		if(s) {
			v = edb_auth_xmlname_to_permission(s, v);
			xmlFree(s);
		} else {
		edb_trace(ec, "xml_get_element_text_permission: not found attr.\n");
		}
	} else {
		edb_trace(ec, "xml_get_element_text_permission: not found.\n");
	}
	return v;
}

int
xml_get_element_permission(EdbContext ec, void *p, xmlDocPtr doc, xmlNodePtr cur, ...)
{
	va_list	ap;
	xmlChar	*target;
	struct EDB_PERMISSION *perm = p;

	va_start(ap, cur);
	target = va_arg(ap, xmlChar *);
	while(target && cur) {
		while(cur) {
			if(!xmlStrcmp(cur->name, target)) break;
			cur = cur->next;
		}
		if(cur==NULL) break;
		target = va_arg(ap, xmlChar *);
		if(target==NULL) break;
		cur = cur->xmlChildrenNode;
	}
	va_end(ap);

	if(target==NULL && cur) {
		perm->p_read	= xml_get_element_text_permission(ec, doc, cur, EDB_AUTH_INHERIT, cur->name, "permread", NULL);
		perm->p_create	= xml_get_element_text_permission(ec, doc, cur, EDB_AUTH_INHERIT, cur->name, "permcreate", NULL);
		perm->p_write	= xml_get_element_text_permission(ec, doc, cur, EDB_AUTH_INHERIT, cur->name, "permwrite", NULL);
		perm->p_delete	= xml_get_element_text_permission(ec, doc, cur, EDB_AUTH_INHERIT, cur->name, "permdelete", NULL);
		return TRUE;
	} else {
		return FALSE;
	}
}

xmlChar *
xml_get_element_attr(EdbContext ec, xmlDocPtr doc, xmlNodePtr cur, xmlChar *attr, ...)
{
	va_list	ap;
	xmlChar	*target;

	va_start(ap, attr);
	target = va_arg(ap, xmlChar *);
	while(target && cur) {
		while(cur) {
			if(!xmlStrcmp(cur->name, target)) break;
			cur = cur->next;
		}
		if(cur==NULL) break;
		target = va_arg(ap, xmlChar *);
		if(target==NULL) break;
		cur = cur->xmlChildrenNode;
	}
	va_end(ap);

	if(target==NULL && cur) {
		return xmlGetProp(cur, attr);
	}
	return NULL;
}

int
xml_get_element_attr_integer(EdbContext ec, xmlDocPtr doc, xmlNodePtr cur, xmlChar *attr, int default_value, ...)
{
	va_list	ap;
	xmlChar	*target;
	int	v = default_value;

	va_start(ap, default_value);
	target = va_arg(ap, xmlChar *);
	while(target && cur) {
		while(cur) {
			if(!xmlStrcmp(cur->name, target)) break;
			cur = cur->next;
		}
		if(cur==NULL) break;
		target = va_arg(ap, xmlChar *);
		if(target==NULL) break;
		cur = cur->xmlChildrenNode;
	}
	va_end(ap);

	if(target==NULL && cur) {
		register xmlChar	*s;

		s = xmlGetProp(cur, attr);
		if(STRisVALID(s)) v = atoi((char *)s);
		xmlFree(s);
	}
	return v;
}

int
xml_get_element_attr_bool(EdbContext ec, xmlDocPtr doc, xmlNodePtr cur, xmlChar *attr, int default_value, ...)
{
	va_list	ap;
	xmlChar	*target;
	int	v = default_value;

	va_start(ap, default_value);
	target = va_arg(ap, xmlChar *);
	while(target && cur) {
		while(cur) {
			if(!xmlStrcmp(cur->name, target)) break;
			cur = cur->next;
		}
		if(cur==NULL) break;
		target = va_arg(ap, xmlChar *);
		if(target==NULL) break;
		cur = cur->xmlChildrenNode;
	}
	va_end(ap);

	if(target==NULL && cur) {
		register xmlChar	*s;

		s = xmlGetProp(cur, attr);
		if(STRisVALID(s) && !xmlStrcasecmp(s, (xmlChar *)"false")) v = FALSE;
		xmlFree(s);
	}
	return v;
}

int
xml_get_element_attr_permission(EdbContext ec, xmlDocPtr doc, xmlNodePtr cur, xmlChar *attr, int default_value, ...)
{
	va_list	ap;
	xmlChar	*target;
	int	v = default_value;

	va_start(ap, default_value);
	target = va_arg(ap, xmlChar *);
	while(target && cur) {
		while(cur) {
			if(!xmlStrcmp(cur->name, target)) break;
			cur = cur->next;
		}
		if(cur==NULL) break;
		target = va_arg(ap, xmlChar *);
		if(target==NULL) break;
		cur = cur->xmlChildrenNode;
	}
	va_end(ap);

	if(target==NULL && cur) {
		register xmlChar	*s;

		s = xmlGetProp(cur, attr);
		if(STRisVALID(s)) v = edb_auth_xmlname_to_permission(s, v);
		xmlFree(s);
	} else {
		edb_trace(ec, "xml_get_element_attr_permission: not found.\n");
	}
	return v;
}

utf8_t *
xml_text_encode(register EdbContext ec, register utf8_t *s)
{
	register utf8_t *d, *dst, c;

	dst = d = (utf8_t *)alloca(sizeof(utf8_t)*strlen((char *)s)*10);
	while((c=*s++)) {
		switch(c) {
		case	'<':	strcpy((char *)d, "&lt;");	d += 4;	break;
		case	'>':	strcpy((char *)d, "&gt;");	d += 4;	break;
		case	'&':	strcpy((char *)d, "&amp;");	d += 5;	break;
		case	'\'':	strcpy((char *)d, "&apos;");	d += 6;	break;
		case	'"':	strcpy((char *)d, "&quot;");	d += 6;	break;
		default:	*d++ = c;			break;
		}
	}
	*d = '\0';
	return (utf8_t *)strdup((char *)dst);
}

utf8_t *
xml_make_element_text(
	register EdbContext	ec,
	register utf8_t	*xmlname,
	register utf8_t	*text)
{
	register utf8_t	*s;
	utf8_t	*xml;

	asprintf((char **)&xml, "<%s:%s>%s</%s:%s>", EDB_XMLNS, xmlname, s=xml_text_encode(ec, text), EDB_XMLNS, xmlname);
	FREE(s);
	return xml;
}

utf8_t *
xml_make_element_text_with_attr(
	register EdbContext	ec,
	register utf8_t	*xmlname,
	register utf8_t	*text,
	register utf8_t	*attr)
{
	register utf8_t	*s;
	utf8_t	*xml;

	asprintf((char **)&xml, "<%s:%s %s>%s</%s:%s>", EDB_XMLNS, xmlname, attr, s=xml_text_encode(ec, text), EDB_XMLNS, xmlname);
	FREE(s);
	return xml;
}

utf8_t *
xml_make_element_text_integer(
	register EdbContext	ec,
	register utf8_t	*xmlname,
	register int	v)
{
	utf8_t	*xml;

	asprintf((char **)&xml, "<%s:%s>%d</%s:%s>", EDB_XMLNS, xmlname, v, EDB_XMLNS, xmlname);
	return xml;
}

utf8_t *
xml_make_element_mtext(
	register EdbContext	ec,
	register utf8_t	*en,
	register utf8_t	*ja,
	register utf8_t	*pr)
{
	utf8_t	*xml;

	if(en) {
		en = xml_make_element_text(ec, (utf8_t *)XML_EN_ENGLISH, en);
	} else {
		en = (utf8_t *)strdup("");
	}
	if(ja) {
		ja = xml_make_element_text(ec, (utf8_t *)XML_EN_JAPANESE, ja);
	} else {
		ja = (utf8_t *)strdup("");
	}
	if(pr) {
		pr = xml_make_element_text(ec, (utf8_t *)XML_EN_PRONOUNCE, pr);
	} else {
		pr = (utf8_t *)strdup("");
	}

	asprintf((char **)&xml, "%s%s%s", en, ja, pr);
	FREE(en);
	FREE(ja);
	FREE(pr);
	return xml;
}

utf8_t *
xml_make_element_permission(EdbContext ec, void *p, int last)
{
	utf8_t	*xml, *r, *c, *w, *d;
	register struct EDB_PERMISSION *perm = p;

	r = xml_make_element_text(ec, (utf8_t *)"permread", edb_auth_xmlstring(ec, perm->p_read, last));
	c = xml_make_element_text(ec, (utf8_t *)"permcreate", edb_auth_xmlstring(ec, perm->p_create, last));
	w = xml_make_element_text(ec, (utf8_t *)"permwrite", edb_auth_xmlstring(ec, perm->p_write, last));
	d = xml_make_element_text(ec, (utf8_t *)"permdelete", edb_auth_xmlstring(ec, perm->p_delete, last));
	asprintf((char **)&xml, "%s%s%s%s", r, c, w, d);
	FREE(r);
	FREE(c);
	FREE(w);
	FREE(d);
	return xml;
}


